/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "api.h"
#include "global.h"
#include "anim-channel.h"
#include "animatable.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <math.h>

//
// static Animatable tokenTable declaration
//
Pad_HashTable  Pad_Animatable::tokenTable(PAD_STRING_TABLE); // maps tokens to animatables
int            Pad_Animatable::tokenIndex=0;            // Current token index
//
// static Animatable tokenTable declaration
//
//Pad_HashTable  Pad_AnimChannel::tokenTable(PAD_STRING_TABLE); // maps tokens to animatables
//int            Pad_AnimChannel::tokenIndex=0;            // Current token index


static void  Animate(void*);  // Declaration of
                            // toplevel function that controls the
                            // timer event for playing a channel
                              
Pad_AnimChannel::Pad_AnimChannel(Pad_OptionType, Pad_AnimPath* path, int resolution)
{
    _animOption       = Pad_NoOption;
    _path             = path;
    _timeResolution   = resolution;// in milliseconds
    _warp             = 0;
    animTimer         = NULL;
    _lastRenderTime   = 0.0;
}

Pad_AnimChannel::~Pad_AnimChannel()
{
}

// Call this to begin an animation
Pad_Bool
Pad_AnimChannel::Start_animation()
{
    time.Update();
    _startSec      = time.Get_sec();
    _startMicroSec = time.Get_usec();

    _interrupt     = FALSE;
    Animate(this); // begin animation
    Get_path()->Set_run(TRUE);// Set to TRUE to use "optimizaton"
    return TRUE;
}

// Get realTime/ if warp is applied, apply time transformation/
// Get interpolated value from Pad_AnimPath/ 
// Based upon the _type of _object,  
// send interpolated value to stuffer (option changer)
static void
Animate(void* clientData)
{
    float                 currentTimeSec;
    Pad_AnimChannel      *channel = (Pad_AnimChannel *)clientData;

    channel->time.Update();
          // Calculate time in seconds
    currentTimeSec = float(1000000 * (channel->time.Get_sec() - channel->Get_start_sec()) + 
			   (channel->time.Get_usec() - channel->Get_start_micro_sec())) / 1000000.0;

    if (!(channel->Get_interrupt())) {
	if ((currentTimeSec<=channel->Get_end_time())) {
	        // Use time to set option value
	    channel->Apply(currentTimeSec);
	    // Set timer to callback this function in timeResolution miliseconds
	    channel->animTimer = Pad_CreateTimerHandler(channel->Get_time_resolution(), Animate, channel);
	
	} else {
	    // Set option to end value.
	    channel->Apply(channel->Get_end_time());
	    Pad_DeleteTimerHandler(channel->animTimer);
	    channel->Get_path()->Set_run(FALSE);
	} 
    } else {
	Pad_DeleteTimerHandler(channel->animTimer);
    }
	    
}

////////////////////////////
//   Setters
///////////////////////////
Pad_Bool 
Pad_AnimChannel::Set_path(Pad_AnimPath *path) 
{
    _path = path;
    
    Get_path()->Set_run(TRUE);
    if (_path) {
	return TRUE;
    } else {
	return FALSE;
    }
}

Pad_Bool
Pad_AnimChannel::Set_option(Pad_OptionType opt)
{
    _animOption = opt;

    if (_animOption) {
	return TRUE;
    } else {
	return FALSE;
    }
}

Pad_Bool
Pad_AnimChannel::Set_begin_time(float beginT)
{
    _path->Set_begt(beginT);
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_end_time(float endT)
{
    _path->Set_endt(endT);
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Apply(float time)
{
    Pad_AnimPath::datum3  interpResult;

    if (_warp) { // if there is a warp such as siso, apply proper tansform to time
	time = Calc_transformed_time(time);
    }

    if (Has_path()) {
	_path->Interpolate(time, interpResult);

	// decide which data type to send to the Pad_ObjHandle::Set_option()
	// method, based upon the option type to set
	switch(_animOption) {
	case Pad_PositionOption:
	case Pad_RPositionOption:
	case Pad_ViewOption:
	case Pad_AngleCtrOption:
	    _objHandle.Set_option(_animOption,
				  (double)interpResult[0],
				  (double)interpResult[1],
				  (double)interpResult[2]);
	    break;
	case Pad_PenOption:
	case Pad_FillOption:
	    _objHandle.Set_option(_animOption,
				  (int)interpResult[0],
				  (int)interpResult[1],
				  (int)interpResult[2]);
	    break;
	default:
	    _objHandle.Set_option(_animOption, (double)interpResult[0]);// temp change from (double *) interpResult
	    break;
	}
    }

    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_time_resolution(int res)
{
    _timeResolution = res;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_interrupt(Pad_Bool intpt)
{
    _interrupt = intpt;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_warp(int warp)
{
    _warp = warp;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_object_tag_or_id(Pad_String objTagOrId)
{
    _objTagOrId = objTagOrId;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_object_handle(Pad_String surfaceName, Pad_String tagOrId)
{
    Pad_SurfaceHandle surface;
       
    _objHandle.Attach(surfaceName.Get(), tagOrId.Get());
    _objTagOrId = tagOrId;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_object_handle(Pad_ObjHandle &objHandle)
{
    _objHandle.Attach(objHandle);
    if (objHandle._Get_tag()) {
	_objTagOrId = objHandle._Get_tag();
    } else {
	_objTagOrId = objHandle.Get_id();
    }
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_token(Pad_String token)
{
    _token = token;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_start_sec(long sec) 
{
    _startSec = sec;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_start_micro_sec(long micsec)
{
    _startMicroSec = micsec;
    return TRUE;
}

Pad_Bool
Pad_AnimChannel::Set_last_render_time(float lastT)
{
    _lastRenderTime = lastT;
    return TRUE;
}

////////////////////////
// Getters
////////////////////////
Pad_AnimPath*
Pad_AnimChannel::Get_path() const
{
	return _path;
}

Pad_Bool
Pad_AnimChannel::Get_object_handle(Pad_ObjHandle &objHandle) 
{
    if (_objHandle._Get_first_object()) {
	objHandle = _objHandle;
	return TRUE;
    } else {
	return FALSE;
    }
}
 
Pad_OptionType
Pad_AnimChannel::Get_option() const 
{
    return _animOption;
}
 
int
Pad_AnimChannel::Get_warp() const 
{
    return _warp;
}

float
Pad_AnimChannel::Get_begin_time() const
{
    if (Has_path()) {
	return _path->Get_begt();
    } else {
	return 0.0;
    }
}

Pad_AnimPath::PRE_POST 
Pad_AnimChannel::Get_pre()  const
{
    if (Has_path()) {
	return _path->Get_pre();
    } else {
	return (Pad_AnimPath::PRE_POST)0;
    }
}

Pad_AnimPath::PRE_POST 
Pad_AnimChannel::Get_post() const
{
    if (Has_path()) {
	return _path->Get_post();
    } else {
	return (Pad_AnimPath::PRE_POST)0;
    }
}

float
Pad_AnimChannel::Get_end_time() const
{
    if (Has_path()) {
	return _path->Get_endt();
    } else {
	return 0.0;
    }
}

Pad_Bool
Pad_AnimChannel::Get_option_val_from_time(float time, Pad_AnimPath::datum3& interpVal) const
{
    _path->Interpolate(time, interpVal); 

    if (interpVal[0]) {
	return TRUE;
    } else {
	return FALSE;
    }
}

int
Pad_AnimChannel::Get_time_resolution() const
{
    return _timeResolution;
}

Pad_Bool
Pad_AnimChannel::Get_interrupt() const
{
    return _interrupt;
}

Pad_String
Pad_AnimChannel::Get_object_tag_or_id() const
{
    return _objTagOrId;
}

Pad_String
Pad_AnimChannel::Get_token() const
{
    return _token;
}

long
Pad_AnimChannel::Get_start_sec() const
{
    return _startSec;
}

long
Pad_AnimChannel::Get_start_micro_sec()  const
{ 
    return _startMicroSec;
}

float
Pad_AnimChannel::Get_last_render_time()  const
{ 
    return _lastRenderTime;
}

///////////////////////////////////////////////////////
// Utility functions
////////////////////////////////////////////////

Pad_Bool 
Pad_AnimChannel::Has_path() const
{
    if (_path) {
	return TRUE;
    } else {
	return FALSE;
    }
}
    
Pad_Bool 
Pad_AnimChannel::Has_object_handle() const
{
    if (_objHandle.Get_id()) {
	return TRUE;
    } else {
	return FALSE;
    }
}    

Pad_Bool 
Pad_AnimChannel::Has_option() const
{
    if (_animOption) {
	return TRUE;
    } else {
	return FALSE;
    }
}    

Pad_Bool 
Pad_AnimChannel::Has_token() const
{
    if (!strcmp(_token.Get(), "")) {
	return TRUE;
    } else {
	return FALSE;
    }
}    

