/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef ANIM_CHANNEL_H
#define ANIM_CHANNEL_H

#include "defs.h"
#include "global.h"
#include "anim-path.h"
#include "misc.h"
#include "api.h"
#include "pad.h"
#include "object.h"
#include "animatable.h"

class Pad_Handle;

class Pad_AnimChannel : public Pad_Animatable
{

  public:

    Pad_TimerToken animTimer;        // Event timer token, primarily used to 
                                     // Turn off event timer at end of animation
    Pad_Time       time;             // The timer that tracks real time 
                                     //_startSec and _startMicroSec are in the channel

    //    static   Pad_HashTable  tokenTable;
    //    static   int            tokenIndex;

    float          Calc_transformed_time(float){return 0.0;} // for future warp implementation

    Pad_Bool       Set_object_handle(Pad_String surface, Pad_String tadOrId); // 
    Pad_Bool       Set_object_handle(Pad_ObjHandle &objHandle); // 
    Pad_Bool       Set_path(Pad_AnimPath *path);         // Set pointer to animation path to use for animation
    Pad_Bool       Set_option(Pad_OptionType opt);       // Set enum integer that specifies animOption
    Pad_Bool       Set_object_tag_or_id(Pad_String objTagOrId);// Set tag or id that specifies the animation object
    Pad_Bool       Set_begin_time(float beginT);         // Time is stored in _path (Pad_AnimPath)
    Pad_Bool       Set_end_time(float endT);             // This time is stored in _path (Pad_AnimPath)
    Pad_Bool       Set_time_resolution(int res);         // Resolution is miliseconds between time events
    Pad_Bool       Set_interrupt(Pad_Bool intpt);        // For stoping an animation
    Pad_Bool       Set_warp(int warp);                   // For future warp functions
    Pad_Bool       Set_token(Pad_String token);          // The channel's token
    Pad_Bool       Set_start_sec(long sec);              // Set second resolution of start time _startSec
    Pad_Bool       Set_start_micro_sec(long micsec);     // Set microsecond resolution of start time _startMicroSec
    
    virtual  Pad_Bool       Apply(float time);                 // Given the time, this method
                                                               // will set the option value to 
                                                               // the proper interopolated option
                                                               // value based upon the path

    virtual  Pad_Bool       Set_last_render_time(float lastT); // Most recent time that channel option 
                                                               //  value was set (in seconds)

    virtual  float          Get_begin_time()       const;      // this time is stored in _path (Pad_AnimPath)
    virtual  float          Get_end_time()         const;      // this time is stored in _path (Pad_AnimPath)
    virtual  int            Get_time_resolution()  const;
    virtual  float          Get_last_render_time() const;      // In float seconds

    virtual  Pad_AnimPath::PRE_POST Get_pre()  const;
    virtual  Pad_AnimPath::PRE_POST Get_post() const;

    Pad_AnimPath*  Get_path()             const;
    Pad_Bool       Get_object_handle(Pad_ObjHandle &objHandle);
    Pad_String     Get_object_tag_or_id() const;
    
    virtual Pad_OptionType Get_option()           const;

    int            Get_warp()             const;
    Pad_Bool       Get_interrupt()        const;
    Pad_String     Get_token()            const;
    long           Get_start_sec()        const;
    long           Get_start_micro_sec()  const;

    Pad_Bool       Get_option_val_from_time(float time, Pad_AnimPath::datum3& interpVal) const;

    Pad_Bool       Start_animation();          // Call this to begin the aimation. 
                                               // This method initializes the animation process.
                                               // This method calls the callback function:
                                               // Animate(void*) (below), which has a timer
                                               // event function that calls itself until the
                                               // animation time has passed
    Pad_Bool       Has_path()          const; 
    Pad_Bool       Has_object_handle() const; 
    Pad_Bool       Has_option()        const;
    Pad_Bool       Has_token()         const;

    Pad_AnimChannel(Pad_OptionType opt=Pad_NoOption, Pad_AnimPath* path=NULL, int resolution=100);
    virtual ~Pad_AnimChannel();

  private:
    friend class Pad_AnimChannelHandle;
    friend class Pad_Animatable;

    Pad_Handle*    _handle;         // Pointer to first handle to this channel, or NULL if none

    Pad_AnimPath*  _path;           // The user supplied path
    Pad_Type       _type;           // Type of _object
    Pad_OptionType _animOption;     // Option of object to be animated
    Pad_String     _objTagOrId;     // The tag or id referring to the object/s
    Pad_ObjHandle  _objHandle;      // The handle to animation object/s
    Pad_Bool       _interrupt;      // Used to stop animation early
    int            _timeResolution; // Update interval (miliseconds) 
                                    // to cover time span
    int            _warp;           // a flag to set the type of warping
    Pad_String     _token;          // tcl token for accessing channel

    long           _startSec;       // Time in seconds resolution since start of this channel
    long           _startMicroSec;  // The micro second resolution of start time to add to seconds resolution
    float          _lastRenderTime; // Time (in ms) of current render so far

};

#endif    // Include ANIM_CHANNEL_H
