/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "pad.h"
#include "object.h"
#include "global.h"
#include "misc.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <math.h>

#include "anim-path.h"

Pad_HashTable Pad_AnimPath::tokenTable(PAD_STRING_TABLE);  // Table to index Pad_AnimPath's by Pad_AnimPathToken's
int           Pad_AnimPath::tokenIndex = 0;      // Current token index

Pad_HashTable Pad_AnimPath::handleTokenTable(PAD_STRING_TABLE);  // Table to index Pad_AnimPath's by Pad_AnimPathToken's
int           Pad_AnimPath::handleTokenIndex = 0;      // Current token index

/////////////////////////////////////////////////////
// begin new methods to handle new data structure
/////////////////////////////////////////////////////
// Constructor
Pad_AnimPath::Pad_AnimPath(int num, anim_data*, enum PRE_POST post, enum PRE_POST pre, float bt, float et) 
{
    _begt = bt; 
    _endt = et;
    _dataLen = num;
    _pre  = pre;
    _post = post;
    _run = 0; 
    _beginInterval = 0;
    _data = new anim_data;
    _data->order = 0;
    _data->timePath = FALSE;
    _siso = FALSE;
}

Pad_AnimPath::~Pad_AnimPath()
{
    int _ord;

    for (_ord=0; _ord<_data->order; _ord++){
	delete [] _data->data[_ord];
    }
    if (_data->timePath) {
	delete [] _data->times;
    }
    delete _data;
    

    delete [] _normDist;
    _normDist = NULL;
}

void
Pad_AnimPath::_Init_data()
{
    _normDist = new float [_dataLen];
    _Calc_path_len();
    _Calc_norm_dist();
} 

////////////////////////////////////////////////////////////
//   Setters
///////////////////////////////////////////////////////////
Pad_Bool
Pad_AnimPath::Set_data(int num, anim_data *data, int pathOrder, Pad_Bool timeData)
{
    int orderIter;
    int dataIter;

    // delete old data if it exists
     if (Has_data()) {
 	for (orderIter=0; orderIter<_data->order; orderIter++){
	    delete [] _data->data[orderIter];
 	}
 	if (_data->timePath) {
 	    delete [] _data->times;
 	}
     }

    _data = new anim_data;
    _dataLen = num;
    _data->order = pathOrder;
    _data->timePath = timeData;

    // Allocate enough memory to each of the float array pointers to hold
    // the data to be copied
    if (_data->timePath) {
	_data->times = new datum [_dataLen];
    }
    for (orderIter=0; orderIter<_data->order; orderIter++){
	_data->data[orderIter] = new datum [_dataLen];
    }

    // Copy the data from the user "data" to  "_data"
    for (dataIter=0; dataIter<_dataLen; dataIter++){
	for (orderIter=0; orderIter<_data->order; orderIter++){
	    _data->data[orderIter][dataIter] = data->data[orderIter][dataIter];
	}
	// If a time path, set time values
	if (_data->timePath) {
	    _data->times[dataIter] = data->times[dataIter];
	}
    }
    // Use last time in -timepath + _begt data as the
    // default end time for the path. This is enough
    // time for the entire animation to play after the 
    // the _begt time delay
    if (_data->timePath) {
	_endt = _data->times[_dataLen-1] + _begt;// Use final time in -timepath as default end time
	_data->firstTime = _data->times[0];
	_data->lastTime  = _data->times[_dataLen-1];
    }

    _inTime  = _begt;
    _outTime = _endt;

    _Init_data();
    
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_begt(float time)
{
    _begt   = time;
    // By default, _inTime is the same as _begt
    // so it one changes _begt, set _inTime to the same value
    _inTime = time;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_endt(float time)
{
    _endt    = time;
    // By default, _outTime is the same as _endt
    // so it one changes _endt, set _outTime to the same value
    _outTime = time;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_post(enum PRE_POST post)
{
    _post = post;
    return TRUE;
}

Pad_Bool   
Pad_AnimPath::Set_pre(enum PRE_POST pre)
{
    _pre = pre;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_token(Pad_String token)
{
    _token = token;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_run(Pad_Bool on)
{
    _run = on;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_time_path_flag(Pad_Bool flag) 
{
    _data->timePath = flag;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_siso(Pad_Bool set)
{
    _siso = set;
    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_in_time(float inTime) 
{
    if ((inTime >= _begt) && (inTime <= _endt)) {
	_inTime = inTime;
    } 

    return TRUE;
}

Pad_Bool
Pad_AnimPath::Set_out_time(float outTime) 
{
    if ((outTime >= _begt) && (outTime <= _endt)) {
	_outTime = outTime;
    } 

    return TRUE;
}
///////////////////////////////////////////
//     Getters
/////////////////////////////////////////
Pad_AnimPath::anim_data* 
Pad_AnimPath::Get_data(void) const
{
    if (_data){
	return(_data);
    } else {
	return 0;
    }
} 

int
Pad_AnimPath::Get_order(void) const
{
    if (_data) {
	return _data->order;
    } else {
	return 0;
    }
}

//
// Returns # of elements of data
//
int
Pad_AnimPath::Get_data_len(void) const
{
    if (_data){
	return _dataLen;
    } else {
	return 0;
    }
}

//
// Length of "polyline" connecting data
//
float
Pad_AnimPath::Get_path_len(void) const
{
    if (_data){
	return _pathLen;
    } else {
	return 0.0;
    }
}

float
Pad_AnimPath::Get_begt() const
{
    return(_begt);
}


float
Pad_AnimPath::Get_endt() const
{
    return(_endt);
}

Pad_AnimPath::PRE_POST  
Pad_AnimPath::Get_post() const
{
    return  _post;
}

Pad_AnimPath::PRE_POST  
Pad_AnimPath::Get_pre() const
{
    return  _pre;
}

Pad_String
Pad_AnimPath::Get_token() const
{
    return _token;
}

Pad_Bool
Pad_AnimPath::Get_time_path_flag() const
{
    return _data->timePath;
}

Pad_Bool
Pad_AnimPath::Get_siso() const
{
    return _siso;
}

float
Pad_AnimPath::Get_in_time() const
{
    return _inTime;
}

float
Pad_AnimPath::Get_out_time() const
{
    return _outTime;
}

////////////////////////////////////////////
// End Getters
////////////////////////////////////////////

// _Calc_path_len calculates the sum of 
// the line segments from data ponit to
// data point in the path data to get
// a total length of that path
float
Pad_AnimPath::_Calc_path_len()
{
    int     _dataIter, _valIter; // for loop iterators of datum and values
    float   len[3];              // lengths along individual dimensions
    float   _valChange  = 0.0;   // store change in dimension over single segment
    float   _sumSquares = 0.0;   // store sum of squares for paths of order > 1

    _pathLen = 0;
    len[0] = 0.0;
    len[1] = 0.0;
    len[2] = 0.0;

    if (_data->order == 1) {  // Calculate length if path order == 1 
	for (_dataIter=0; _dataIter<(_dataLen-1); _dataIter++){
	    _pathLen += ABS(_data->data[0][_dataIter + 1]  - _data->data[0][_dataIter]);
	}
    } else { // Calculate length if path order greater than 1
	for (_dataIter=0; _dataIter<(_dataLen-1); _dataIter++){
	    for (_valIter=0; _valIter<(_data->order); _valIter++){ // get length squared of individual segments
		_valChange = (_data->data[_valIter][_dataIter+1] - _data->data[_valIter][_dataIter]);

		_sumSquares += (_valChange * _valChange);

		len[_valIter] += ABS(_data->data[_valIter][_dataIter+1] - _data->data[_valIter][_dataIter]);
	    }
		
	    _pathLen += sqrt(_sumSquares);
	    _sumSquares = 0.0;

	}
    }

    return _pathLen;
}

// _Calc_norm_dist creates an array (normDist) that
// stores the normalized distance to each data point
// from the beginning of the path
void
Pad_AnimPath::_Calc_norm_dist()
{
    int    _dataIter, _valIter; // iterator for datum and within the datum
    float  _accumLen;           // to accumulate the distance along the path
    float  _sumSquares = 0.0;
    float  _valChange  = 0.0;   // store change in dimension over single segment

    _accumLen    = 0.0;
    _normDist[0] = 0;

    if (_data->order == 1) {  // Calculate length if path order == 1 
	for (_dataIter=0; _dataIter<(_dataLen-1); _dataIter++){
	    _accumLen += ABS(_data->data[0][_dataIter + 1] - _data->data[0][_dataIter]);
	    _normDist[_dataIter + 1] = _accumLen/_pathLen;
	}
    } else {
	for (_dataIter=0; _dataIter<(_dataLen-1); _dataIter++){
	    for (_valIter=0; _valIter<(_data->order); _valIter++){ // get length squared of individual segments
		_valChange = (_data->data[_valIter][_dataIter+1] - _data->data[_valIter][_dataIter]);

		_sumSquares += (_valChange * _valChange);

	    }
		
	    _accumLen += sqrt(_sumSquares);
	    if (_pathLen != 0.0){
		_normDist[_dataIter + 1] = _accumLen/_pathLen;
	    }
	    _sumSquares = 0.0;
	}
    }
}

// Interpolate calculates that value of the data given the
// time. It stores the interpolated value in a datum3 variable
// passed by the api. 
// It first normalizes that time by deviding by the distance
// that time is from the beginning of the animation, then finds the
// normalized interval (the indicies of the normData between which
// the interpolated value lies), then uses linear interpolation to
// determing the data value that cooresponds to the given time.
void
Pad_AnimPath::Interpolate(float curTime, datum3 &interpVal)
{
    float      animTime;
    float      normTime;
    float      outNorm;
    float      appliedTime;          // Used in time paths to apply the pre/post conditions
    int        _dataIter, _valIter;
    float      ratio;

    if (_data->timePath) {
	animTime = _data->lastTime - _inTime;
    } else {
	animTime = _endt - _begt;
    }

        // Calculate the normalized time
    if (animTime != 0.0) {// check for division by zero
	normTime = (curTime - _begt)/(animTime);// Normalized time since beginning of animation
    } else {
	normTime = _inTime;
    }

        // Handle times before begin time using the proper 
        // pre specification
    if (normTime<0.0){
	normTime = _Pre(normTime);
    }

        // Calculate the normalized _outTime (outNorm) to apply
        // the _outTime to the animation
    if (_data->lastTime != 0.0) {
	if (_data->timePath && (_outTime !=_data->lastTime) && (_outTime > _inTime)) {
	    outNorm = (_outTime - _inTime)/(animTime);
	} else if (!_data->timePath && (_outTime != _endt) && (_outTime > _inTime)) {
	    outNorm = _outTime/animTime;
	} else {
	    outNorm = 1.0;
	}
    } else {
	outNorm = 1.0;
    }

        // Adjust normTime to account for _inTime if path is not a timePath
    if (!_data->timePath && (animTime != 0.0) && (_inTime!=_begt)) {
	normTime = normTime + ((_inTime - _begt)/animTime);
    }

        // Handle times beyond end time using the proper 
        // post specification
    if (normTime>outNorm){
        normTime = _Post(normTime, outNorm);
    }
    
        // Handle warp siso
    if (_siso && !(_data->timePath)) {
	normTime = Pad_Siso_lerp(normTime);
	    //_Siso(normTime);
    }
    
        // First guess for the index of the datum beginning 
        // the interval that contains the desired value.
    if (_run && _beginInterval<_dataLen) {
	_dataIter = _beginInterval;
    } else {
	_dataIter = int(floor(normTime*_dataLen));
    }

        // Interpolate if the path is not a timePath
    if (_data && !(_data->timePath)) {
	if (_endt == _begt || _pathLen == 0.0) { // special case to jump instantly to last value
	    for (_valIter=0; _valIter<_data->order; _valIter++){ 
		interpVal[_valIter] = _data->data[_valIter][_dataLen-1];
	    }
	} else {
	    // Find the two data vals that surround the value cooresponding to the normTime
	    // normTime values must be between 0.0 and 1.0
	    while (!(_normDist[_dataIter]<=normTime && _normDist[_dataIter+1]>=normTime) &&
		   normTime < 1.0 && normTime > 0.0) {
	    
		if (_normDist[_dataIter]>=normTime){
		    _dataIter--;
		} else if (_normDist[_dataIter+1]<=normTime){
		    _dataIter++;
		} else if (normTime>1){
		    _dataIter = _dataLen - 1;
		}
	    }
	
	    if (_run && _dataIter<_dataLen) { // save the present interval for first guess during run mode
		_beginInterval = _dataIter;
	    }

	    if ((normTime < 1.0) && (normTime > 0.0)){
		ratio = (normTime - _normDist[_dataIter])/(_normDist[_dataIter+1] - _normDist[_dataIter]);
		for (_valIter=0; _valIter<_data->order; _valIter++){ 
		    interpVal[_valIter] = LERP(ratio, _data->data[_valIter][_dataIter], 
					       _data->data[_valIter][_dataIter+1]);
		}
	    } else if (normTime >= 1.0) {
		for (_valIter=0; _valIter<_data->order; _valIter++){ 
		    interpVal[_valIter] = _data->data[_valIter][_dataLen-1];
		}
	    } else if (normTime <= 0.0){
		for (_valIter=0; _valIter<_data->order; _valIter++){ 
		    interpVal[_valIter] = _data->data[_valIter][0];
		}
	    } 
	}
	// The path is a timePath
    } else if (_data && (_data->timePath)) {
	// To get proper time after applying the pre/post conditions
	appliedTime = normTime*animTime + _inTime;

	// "if" takes care of special case. If the appliedTime is greater than 0 and less than the
	//  first time value in _data->times (_data->firstTime), set _normTime to value less than 0 (-1) so that
	//  the first data value in _data->data is used at this time.
	if (appliedTime<_data->firstTime) {
	    _dataIter =  0;
	    normTime = -1;
	} else {
	    // Find the time val that begins the interval containing appliedTime
	    // This "while" loop differs from the above "while" loop in that it checks
	    // appliedTime (not normalized) against the time values in _data->times[],
	    //  as opposed to checking a normalized time against the normalized data.
	    // When the time interval is found that contains applied time, this interval
	    // is used in the linear interpolation (lerp) to find that proper data value
	    // for that time.
	    while (!(_data->times[_dataIter]<=appliedTime && _data->times[_dataIter+1]>=appliedTime) &&
		   normTime < 1.0 && normTime > 0.0) {
	    
		if ((_data->times[_dataIter])>=appliedTime){
		    _dataIter--;
		} else if ((_data->times[_dataIter])<=appliedTime){
		    _dataIter++;
		} else if (appliedTime>=(_data->lastTime)){
		    _dataIter = _dataLen - 1;
		}

	    }
	}
	// save the present interval for first guess when _run is TRUE
	if (_run && _dataIter<_dataLen) { 
	    _beginInterval = _dataIter;
	}
	
	// Calculate the lerp ratio from the appliedTime and the interval begin and end times,
	// then calculate the data interpolated value and stuff it into interpVal
	if ((normTime < 1.0) && (normTime > 0.0)){
	    ratio = (appliedTime - (_data->times[_dataIter]))/((_data->times[_dataIter+1]) - 
							                 (_data->times[_dataIter]));
	    for (_valIter=0; _valIter<_data->order; _valIter++){ 
		interpVal[_valIter] = LERP(ratio, _data->data[_valIter][_dataIter], 
					   _data->data[_valIter][_dataIter+1]);
	    }
	} else if ((normTime >= 1.0) || (appliedTime >= (_data->lastTime))) {
	    for (_valIter=0; _valIter<_data->order; _valIter++){ 
		interpVal[_valIter] = _data->data[_valIter][_dataLen-1];
	    }
	} else if (normTime <= 0.0){
	    for (_valIter=0; _valIter<_data->order; _valIter++){ 
		interpVal[_valIter] = _data->data[_valIter][0];
	    }
	} 
    } else{ // if no data return zero's for the answer
	for (_valIter=0; _valIter<3; _valIter++){ 
	    interpVal[_valIter] = 0.0;
	}
    }

} 

// 
// Chech to see of data has been set for path
//
Pad_Bool
Pad_AnimPath::Has_data()
{
    Pad_Bool  hasPath = FALSE;

    if (_data->order != 0) {
	hasPath = TRUE;
    }
    return hasPath;
}

//
// Slow in Slow out warp for -path animations
// (ie. not for -timepath animations)
//
float
Pad_AnimPath::_Siso(float normTime)
{
    float rampTime  = 0.2;
    float slowSlope = 0.5;
    float fastSlope = 1.33333;

    if (normTime<rampTime) {
	normTime = slowSlope*normTime;
    } else if (normTime >= rampTime && normTime <= (1-rampTime)) {
	normTime = fastSlope*(normTime-rampTime) + slowSlope*rampTime;

    } else if (normTime > 1-rampTime) {
	normTime = slowSlope*(normTime-(1-rampTime))
			      + slowSlope*rampTime + fastSlope*(1-2*rampTime);

    }
    return normTime;
}

//
// Process the post conditions
//
float
Pad_AnimPath::_Post(float normTime, float outNorm)
{
    if (normTime>outNorm){
	if(_post==CONSTANT){
	    normTime = outNorm;
	} else if (_post==CYCLE){
	    if (fmod(normTime, (outNorm))==0.0){
		normTime = outNorm;
	    } else {
		normTime = fmod(normTime, (outNorm));
	    }
        } else if (_post==OSCILLATE){
	    if ((int(floor(normTime)) % 2)){ 
		normTime = outNorm - fmod(normTime, (outNorm));
	    } else {
		normTime = fmod(normTime, (outNorm));
	    }
	}
    }
    return normTime;
}

// _Pre manipulates the time in the interpolation
// calculation to manage the requested PRE condition
// for the animation. The PRE condition states what
// should occure if the animation time is earlier 
// than the animation begin time
float
Pad_AnimPath::_Pre(float normTime)
{
    if (normTime<1.0){
	if(_pre==CONSTANT){
	    normTime = 0.0;
	} else if (_pre==CYCLE){
	    if (fmod(normTime, 1.0f)==0.0){
		normTime = 1.0;
	    } else {
		normTime = 1.0 + fmod(normTime, 1.0f);
	    }
        } else if (_pre==OSCILLATE){
	    if ((int(ceil(normTime)) % 2)){ // if floor normTime is odd neg number 
		normTime = 1.0 + fmod(normTime, 1.0f);
	    } else {                      // if floor normTime even neg number
		normTime = -fmod(normTime, 1.0f);
	    }
	}
    }
    return normTime;
}

