/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/
#ifndef ANIM_PATH_H
#define ANIM_PATH_H

// Anim_Path objects are paths generater from the user 
// supplied data. They are of order n; where the n
// denotes the dimentions of the parmeter being varied 

#include "defs.h"
#include "pad.h"

class Pad_Handle;

// Animation path base class
class Pad_AnimPath
{
  public:
    static  Pad_HashTable tokenTable; // hash table for path tokens to path pointers
    static  int           tokenIndex;

    static  Pad_HashTable handleTokenTable; // hash table for path tokens to path handles
    static  int           handleTokenIndex;

    typedef float datum; 
    typedef float datum3[3];
    
    struct anim_data { // To hold the data that defines that path
	datum*    data[3];
	datum*    times;
	int       order;
	Pad_Bool  timePath;   // Flag to indicate if path includes times
	datum     firstTime;  // If -timepath, this is the first time
	datum     lastTime;   // If -timepath, this is the last time
    };

    enum INDEX {FSTI, SECI, TRDI};
    enum PRE_POST  {CONSTANT, CYCLE, OSCILLATE, UNDEFINED}; // "UNDEFINED" is used for error checking.
                                                            // These same values are used for the pre
                                                            // condition. 
    
            Pad_Bool Set_pre(enum PRE_POST pre);          // Sets the pre condition
            Pad_Bool Set_data(int num, anim_data* data, int pathOrder, Pad_Bool timeData=FALSE);  // For setting _data
            Pad_Bool Set_begt(float time);                // Set the begin time of animation
            Pad_Bool Set_endt(float time);                // Set the ent time of animation
            Pad_Bool Set_post(enum PRE_POST post);        // Sets the post condition
            Pad_Bool Set_token(Pad_String token);         // Set token name for this path
            Pad_Bool Set_run(Pad_Bool on);                // Used to optimize interpolation
            Pad_Bool Set_time_path_flag(Pad_Bool flag);   // Set flag that tells if time is included in data
            Pad_Bool Set_siso(Pad_Bool set);              // Set flag that indicates whether to use a siso 
                                                          // effect ( only applied when _data->timePath is FALSE)
            Pad_Bool Set_in_time(float inTime);           // Set time within animation to begin play
            Pad_Bool Set_out_time(float outTime);         // Set time within animation to end play
    

            float      Get_path_len()       const;        // Length of "polyline" connecting data
            int        Get_data_len(void)   const;        // Returns # of elements in data 
            int        Get_order(void)      const;        // Returns order of path
            float      Get_begt(void)       const;        // Return the begin time of animation
            float      Get_endt(void)       const;        // Return the end time of animation
    enum    PRE_POST   Get_post()           const;        // Returns the post condition 
    enum    PRE_POST   Get_pre()            const;        // Returns the pre condition 
            Pad_String Get_token()          const;        // Return token name 
            Pad_Bool   Get_time_path_flag() const;        // Return state of time data flag
            Pad_Bool   Get_siso()           const;        // Return state of _siso flag
            float      Get_in_time()        const;        // Get time within animation to begin play
            float      Get_out_time()       const;        // Get time within animation to end play
            anim_data* Get_data()           const;        // Returns a pointer to the _data 

    void             Interpolate(float curTime, datum3 &result); // Interpolates along the path to
                                                                 // to the value cooresponding to the 
                                                                 // the time "t".
    Pad_Bool         Has_data();

    Pad_AnimPath(int num=0, anim_data* data = NULL, 
		  enum PRE_POST post=CONSTANT,  enum PRE_POST pre=CONSTANT,
		  float bt=0.0, float et=1.0); 
    
    ~Pad_AnimPath();
 

  private:
    friend class  Pad_AnimPathHandle;

    enum PRE_POST _post;        // Post condition. 
    enum PRE_POST _pre;         // Pre condition
    int           _dataLen;     // Number of verticies defining path
    float         _pathLen;     // Sum of segment lengths between verticies
    float         _begt, _endt; // Animation begin and end times
    float         _time;        // Time passed since tbeg
    float        *_normDist;	// Array of normalized distances to
				// each data value from the beginning
				// of the path
    Pad_Bool      _siso;        // Flag to determine whether to apply slow in slow out or not
    Pad_Bool      _run;         // If in run mode (_run=1), when interpolating use the
                                // the most recently accessed interval as guess for next
                                // value. This is an optimization for finding the proper interval

    int           _beginInterval; // If in run mode (ie. running an animation),
                                     // this int holds the index of the vertex at
                                     // the beginning of the interval where the
                                     // last interp val was found. It is used as
                                     // the first guess for interval of the present interpolation

    Pad_Handle *  _handle;      // Pointer to first handle to this path, or NULL if none
    anim_data*    _data;        // Pointer to sturcture that holds the 
                                // data array that defines the user defined path
                                // This is a tricky allocation.
                                // The datum1 for this array is allocated to slot usrDat1, in 
                                // pad-tcl-cmd.cpp, in method _Animate_create_path. 
                                // The number of datum1 in the "data" array depends on the 
                                // number of datum1 supplied by the user."usrDat1" is given
                                // to "data" by passing the pointer to the array to the 
                                // constructor of Pad_AnimPath1 (data), and "data" is set equal to 
                                // data in Pad_AnimPath1::Init_data(datum1 *data).
                                // This memory is released in ~Pad_AnimPath1

    float        _inTime;       // The time within a defined animation, to start the playing of
                                // that animation. 
    float        _outTime;      // The time within a defined animation, to stop the playing of
                                // that animation. 

    Pad_String   _token;            // Tcl token name for this path

    void    _Init_data();           // Calls Calc_* routines to set up for interpolation
    float   _Calc_path_len(void);   // Calculates the sum of the lengths of the 
                                    // segments comprising a path
    void    _Calc_norm_dist(void);  // Makes an array that holds the normalized distance 
                                    // along the path to each vertex devining the path
    float   _Pre( float normTime);  // Calculates pre condition,
                                    // which tells the interpolating function
                                    // what to do when the time it is passed
                                    // is less than begt

    float   _Post(float normTime, float outNorm);  // Calculates post condition,
                                    // which tells the interpolating function
                                    // what to do when it is passed a time
                                    // that is greater than endt
    float   _Siso(float normTime);  // Applies a slow in slow out transform to normTime
                                    // and returns the transformed value.

};

#endif    // Include ANIM_PATH_H

