/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "anim.h"
#include "animatable.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <math.h>

Pad_HashTable Pad_Animation::tokenTable(PAD_STRING_TABLE);  // Table to index Pad_Animations's by token anim#
int           Pad_Animation::tokenIndex = 0;                // Current token index

static void  Play(void*);    // Declaration of
                            // toplevel function that controls the
                            // timer event for playing an animatable

//
// Animatable destructor.  Can't define in header file
// due to bug with gcc in mklinux.
//
Pad_Animatable::~Pad_Animatable()
{
};

Pad_Animation::Pad_Animation()
{
    _interrupt          = FALSE;
    _update             = TRUE;
    _endTimeAutoSet     = TRUE;
    _beginTime          = 0.0;
    _animatableList     = new Pad_List;
    _animTimeResolution = 100;
    _endTime            = 1.0;
    _started            = FALSE;
    _speedFactor        = 1.0;
}

Pad_Animation::~Pad_Animation()
{
    delete _animatableList;
}

// Call this to play an animation
Pad_Bool
Pad_Animation::Start()
{
    time.Update();
    _startSec        = time.Get_sec();
    _startMicroSec   = time.Get_usec();
    _interrupt       = FALSE;
    //    _beginTime       = 0.0;
    _update          = TRUE;
    Play(this); // begin animation
    return TRUE;
}

//
// Play the animation
//
static void
Play(void* clientData)
{
    float                 currentTimeSec;
    Pad_Animation        *animation = (Pad_Animation *)clientData;
    float                 endTime;

    endTime      = animation->Get_end_time();

    animation->time.Update();
        // Calculate time in seconds
    currentTimeSec = float(1000000.0 * float(animation->time.Get_sec() - animation->Get_start_sec()) + 
		   float(animation->time.Get_usec() - animation->Get_start_micro_sec())) / 1000000.0;

    currentTimeSec = (animation->Get_speed_factor())*currentTimeSec;
    if (!animation->Get_interrupt()) {

	animation->Do_animatables(currentTimeSec);

	if ((currentTimeSec<=endTime)) {
	        // Set timer to callback this function in timeResolution miliseconds
	    animation->animTimer = Pad_CreateTimerHandler(animation->Get_time_resolution(), Play, animation);
	    if (animation->Get_update()) {
		animation->Set_update(FALSE);
	    }
	    
	} else {
	    animation->Do_animatables(animation->Get_end_time());
	    Pad_DeleteTimerHandler(animation->animTimer);
	}
			    
    } else {
	animation->Interrupt_members();
	Pad_DeleteTimerHandler(animation->animTimer);
	animation->Set_interrupt(FALSE);
    }

}

Pad_Bool 
Pad_Animation::Interrupt_members()
{
    Pad_Iterator     animatablei;
    Pad_Animatable  *animatable;

    DOLIST(animatablei, *_animatableList, Pad_Animatable, animatable) {
	if (animatable->Get_type()== Pad_Animatable::ANIM) {
	    animatable->Set_interrupt(TRUE);
	    //	    ((Pad_Animation*)animatable)->Set_started(FALSE);
	}
    }
    return TRUE;
}

Pad_Bool
Pad_Animation::Do_animatables(float currentTimeSec)
{
    Pad_Animatable       *animatable;
    Pad_Iterator          animatablei;
    float                 checkTime;
    float                 animatableEndTime;
    float                 animatableBeginTime;    

    currentTimeSec = currentTimeSec;
    DOLIST(animatablei, *_animatableList, Pad_Animatable, animatable) {
	
	if (animatable) {
	    animatableBeginTime = animatable->Get_begin_time();
	    animatableEndTime = animatable->Get_end_time();
	    // Update the animation time resolution and end time on when the animation
	    // begins incase to reflect changes in member's (ie. animatable) resolution and
	    // end times
	    if (_update ) {

		if (_endTimeAutoSet) {
		    Update_end_time_and_resolution(animatable);
		}		    
		
		if (animatable->Get_type()==Pad_Animatable::CHAN) {
		    animatable->Apply(animatableBeginTime);
		    animatable->Set_last_render_time(currentTimeSec);

 		} else if (animatable->Get_type()== Pad_Animatable::ANIM) {
		    // Set channels in member animation to initial value at begining of animation
		    ((Pad_Animation*)animatable)->Reset();
		    // if parent animation speed had been altered, alter the animation
		    // speeds of its children
		    if (_speedFactor != 1.0) {
			((Pad_Animation*)animatable)->Set_speed_factor(_speedFactor);
		    }
 		}
	    }

	    
	    if ((animatable && currentTimeSec<=_endTime)) {
		if ((animatable->Get_type())==(Pad_Animatable::CHAN)) {
		    checkTime = animatable->Get_last_render_time() + float(animatable->Get_time_resolution()/1000);
   		    // checkTime is used to determine if a enough time has passed since last rendering,
		    // relative to this animatable's timeResolution, to do an update of the animatable
		    if (currentTimeSec >= checkTime) {
			if (animatableBeginTime == animatableEndTime){// Special case
		            // Set option value corresponding to current time		    
			    animatable->Apply(currentTimeSec);
			    animatable->Set_last_render_time(currentTimeSec);

			} else if (currentTimeSec >= animatableBeginTime &&
				   currentTimeSec <= animatableEndTime) {
			    
			    animatable->Apply(currentTimeSec);
			    animatable->Set_last_render_time(currentTimeSec);
		    
			    // if before begin time and pre condition is not constant (0)
			} else if (currentTimeSec < animatableBeginTime && 
				   animatable->Get_pre() != 0) {

			    animatable->Apply(currentTimeSec);
			    animatable->Set_last_render_time(currentTimeSec);

			    // if after channel end time and post condition not constant (0)
			} else if (currentTimeSec > animatableEndTime && 
				   animatable->Get_post() != 0) {

			    animatable->Apply(currentTimeSec);
			    animatable->Set_last_render_time(currentTimeSec);
			    
			} else if (currentTimeSec > animatableEndTime && 
				   animatable->Get_post() == 0) {

			    animatable->Apply(animatableEndTime);
			    animatable->Set_last_render_time(currentTimeSec);
			}
			
		    }
		    // start animaion members of this animation, if they have not been started
		} else if (animatable && (animatable->Get_type())==(Pad_Animatable::ANIM)) {
		    if ((animatable!=this) && (currentTimeSec>=animatableBeginTime)
			&& (currentTimeSec<_endTime) && !(animatable->Get_started())) {
		    
			animatable->Apply(0.0);
			animatable->Set_started(TRUE);
		    }
		}
	    } else if (animatable) {
	        // Interrupt the animatable if it is not finnished.
		animatable->Set_interrupt(TRUE);

		// Set option to end value.
		//animatable->Apply(_endTime);
		animatable->Set_last_render_time(0.0);
		
		// Set animation "started" flags to 0
 		if ((animatable->Get_type()) == ANIM) {
 		    animatable->Set_started(FALSE);
 		}
	    }	
	}
    }

     return TRUE;
}

// 
Pad_Bool
Pad_Animation::Apply(float)
{
    return Start();
}

//
// For setting the value of each channel in an animation
// to its initial value. This recurses into the animations
// within an animation.
//
Pad_Bool
Pad_Animation::Reset() 
{
    Pad_Iterator          animatablei;
    Pad_Animatable       *animatable;
    float                 animatableBeginTime;

    DOLIST(animatablei, *_animatableList, Pad_Animatable, animatable) {
	if (animatable->Get_type()==Pad_Animatable::CHAN) {
	    animatable->Apply(animatableBeginTime);
	    //animatable->Set_last_render_time(currentTimeSec);
	} else if (animatable->Get_type()==Pad_Animatable::ANIM) {
	    ((Pad_Animation*)animatable)->Reset();
	}
    }

    return TRUE;    
}
    



// Checks the present _endTime and  _animTimeResolution against animatable
// and decides if the values need to be updated
Pad_Bool
Pad_Animation::Update_end_time_and_resolution(Pad_Animatable *animatable)
{
    int                   tempRes;
    float                 tempEndTime;

    tempRes = animatable->Get_time_resolution();
    if (tempRes < _animTimeResolution && tempRes > 10) {
	_animTimeResolution = tempRes;
    } else if (tempRes < 10) {
	_animTimeResolution = 10;
    }

    if (_endTimeAutoSet) {
	if (animatable->Get_type() == ANIM) {
	    tempEndTime = animatable->Get_end_time() + animatable->Get_begin_time() + .1;
	} else if (animatable->Get_type() == CHAN) {
	    tempEndTime = animatable->Get_end_time(); 
	}

	if (tempEndTime > _endTime) {
	    _endTime = tempEndTime;
	}
    }
    return TRUE;
}

////////////////////////
// Setters
////////////////////////
Pad_Bool
Pad_Animation::Set_start_sec(long sec)
{
    _startSec = sec;
    return TRUE;
}

Pad_Bool
Pad_Animation::Set_token(Pad_String animToken)
{
    _token = animToken;
    return TRUE;
}

Pad_Bool
Pad_Animation::Set_start_micro_sec(long microSec)
{
    _startMicroSec = microSec;
    return TRUE;
}

Pad_Bool 
Pad_Animation::Set_started(Pad_Bool started)
{
    _started = started;
    return TRUE;
}

Pad_Bool
Pad_Animation::Set_end_time(float endt)
{
    _endTime = endt;
       // see explanation of _endTimeAutoSet in anim.h
    if (_endTime > 0) {
	_endTimeAutoSet = FALSE;
    } else {
	_endTimeAutoSet = TRUE;
    }

    return TRUE;
}

Pad_Bool
Pad_Animation::Set_begin_time(float begint)
{
    _beginTime = begint;
    return TRUE;
}

Pad_Bool
Pad_Animation::Set_interrupt(Pad_Bool interpt)
{
    _interrupt = interpt;

    if (_interrupt == TRUE) {
	Set_started(FALSE);
    }

    return TRUE;
}
	
Pad_Bool
Pad_Animation::Set_update(Pad_Bool update)
{
    _update = update;
    return TRUE;
}

Pad_Bool 
Pad_Animation::Set_speed_factor(float speedF) 
{
    _speedFactor = speedF;
    return TRUE;
}

Pad_Bool
Pad_Animation::Set_time_resolution(int res)
{
    _animTimeResolution = res;
    return TRUE;
}

//////////////////////
// Getters
//////////////////////
long
Pad_Animation::Get_start_sec() const
{
    return _startSec;
}

long
Pad_Animation::Get_start_micro_sec() const
{
    return _startMicroSec;
}

Pad_Bool 
Pad_Animation::Get_started() const
{
    return _started;
}

Pad_String
Pad_Animation::Get_token() const
{
    return _token;
}

Pad_List*
Pad_Animation::Get_animatable_list() const
{
    return _animatableList;
}

float
Pad_Animation::Get_end_time() const
{
    return _endTime;
}

float
Pad_Animation::Get_begin_time() const
{
    return _beginTime;
}

float
Pad_Animation::Get_speed_factor() const 
{
    return _speedFactor;
}

Pad_Bool
Pad_Animation::Get_interrupt() const
{
    return _interrupt;
}

int
Pad_Animation::Get_time_resolution() const
{
    return _animTimeResolution;
}

Pad_Bool
Pad_Animation::Get_update() const
{
    return _update;
}

Pad_Bool
Pad_Animation::Get_end_time_auto_set() const
{
    return _endTimeAutoSet;
}

Pad_Bool
Pad_Animation::Get_animatable_tokens(Pad_String &tokens)
{
    Pad_Iterator     atomi;
    Pad_List        *animatables;
    Pad_Animatable      *animatable;
    char*            spacer = " ";

    animatables = _animatableList;

    DOLIST(atomi, *animatables, Pad_Animatable, animatable) {    
	tokens.Append(animatable->Get_token());
	tokens.Append(spacer);
    }
    return TRUE;
}

/////////////////////////////////////////

Pad_Bool
Pad_Animation::Add_animatable(Pad_Animatable *animatable)
{
    _animatableList->Push_last(animatable);
    animatable->parents.Push_last(this);
    // Set the event time resolution for playing the animation
    // and _endTime
    Update_end_time_and_resolution(animatable);

    return TRUE;
}

Pad_Bool
Pad_Animation::Remove_animatable(Pad_Animatable *animatable)
{
    _animatableList->Remove(animatable);
    return TRUE;
}


