/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef ANIMATION_H
#define ANIMATION_H

#include "defs.h"
#include "pad.h"
#include "global.h"
#include "misc.h"
#include "api.h"
#include "object.h"
#include "list.h"
#include "animatable.h"

class Pad_Animation : public Pad_Animatable
{
  public:

    Pad_TimerToken animTimer;        // Event timer token, primarily used to 
                                     // Turn off event timer at end of animation
    Pad_Time       time;             // The timer that tracks real time 
                                     //_startSec and _startMicroSec are in the anaton

    static  Pad_HashTable tokenTable;
    static  int           tokenIndex;

    Pad_Bool       Reset();
    Pad_Bool       Add_animatable(Pad_Animatable *animatable);    // Add "animatable" to "_animatableList"  
    Pad_Bool       Remove_animatable(Pad_Animatable *animatable); // Remove "animatable" from "_animatableList"
    Pad_Bool       Update_end_time_and_resolution(Pad_Animatable *animatable);         
    Pad_Bool       Interrupt_members();
    Pad_Bool       Do_animatables(float currentTimeSec);
    Pad_Bool       Start();                                  // Begin the animation of the animatable

    Pad_Bool       Set_token(Pad_String animToken);
    Pad_Bool       Set_start_sec(long sec);
    Pad_Bool       Set_start_micro_sec(long microSec);
    Pad_Bool       Set_begin_time(float begint);
    Pad_Bool       Set_end_time(float endt);
    Pad_Bool       Set_speed_factor(float speedF);
    Pad_Bool       Set_interrupt(Pad_Bool interpt);
    Pad_Bool       Set_update(Pad_Bool update);
    Pad_Bool       Set_time_resolution(int res);

        // Virtual methods that allow animations to be animatable
    virtual     Pad_Bool      Apply(float);
    virtual     Pad_Bool      Set_started(Pad_Bool started);

    virtual     float         Get_begin_time()       const;
    virtual     float         Get_end_time()         const;
    virtual     Pad_String    Get_token()            const;
    virtual     Pad_Bool      Get_started()          const;
        // end virtual
    float          Get_speed_factor()      const;
    Pad_Bool       Get_interrupt()         const;
    int            Get_time_resolution()   const;
    Pad_Bool       Get_update()            const;
    Pad_Bool       Get_end_time_auto_set() const;
    Pad_Bool       Get_animatable_tokens(Pad_String &tokens);

        // These three are not to be api supported. They are not private, in order to make them 
        // easier to access by the Player function. This may also be accomplished by making
        // Play a friend
    long           Get_start_sec()         const; 
    long           Get_start_micro_sec()   const;
    Pad_List*      Get_animatable_list()   const;


    Pad_Animation();
    ~Pad_Animation();

  private:
    friend  class Pad_AnimationHandle;
    friend  class Pad_Animatable;
    
    Pad_Handle*      _handle;
    int              _animTimeResolution; // The smallest of the timeResolutions of the animatables or 10
                                          // whichever is greater
    float            _speedFactor;    // Default 1 for normal play. Use to speed up or slow down 
                                      // play. (i.e. if 2, double speed of play
    Pad_String       _token;
    Pad_Bool         _interrupt;
    Pad_Bool         _update;
    Pad_Bool         _endTimeAutoSet; // This is a flag to determine if the _endTime is automatically set to 
                                      // the largest end time of the animatable or if it has been manually set
                                      // set to some time. If one sets explicitly sets the endTime to a positive value,
                                      // the auto set is turned off. To turn auto set back on, one explicitly sets the
                                      // endTime to -1 (or any negative value);
    Pad_List        *_animatableList;
    long             _startSec;
    long             _startMicroSec; 
    float            _beginTime;
    float            _endTime;
    Pad_Bool         _started;        // Flag to let Play know that this animation has already been fired

};

#endif    // Include ANIMATION_H







