/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef ANIMATABLE
#define ANIMATABLE

#include "anim-path.h"
#include "api.h"

class Pad_Animatable // a unit of animtion
{

  public:

    enum ANIMATABLE_TYPE {CHAN, ANIM};

    static               Pad_HashTable  tokenTable; // maps tokens to animatables (initialized in anim-channle.cpp
    static int           tokenIndex;            // Current token index

    enum    ANIMATABLE_TYPE Get_type() const {return _type;}
    Pad_Bool                Set_type(enum ANIMATABLE_TYPE type){_type = type; return TRUE;}

    virtual   Pad_Bool   Apply(float) {return TRUE;}   // Given he time, this method will
                                           //  o set the option value for the given time (if member is channel)
                                           //  o will run a script (if member is script)

    virtual   Pad_Bool   Set_last_render_time(float) {return TRUE;}    // Most recent time that member 
                                                           //option value was set (in seconds)
    virtual   Pad_Bool   Set_started(Pad_Bool) {return FALSE;}
    virtual   Pad_Bool   Set_interrupt(Pad_Bool) {return TRUE;}

    virtual   Pad_Bool   Get_started()          const {return FALSE;}
    virtual   float      Get_begin_time()       const =0;  // this time is stored in _path (Pad_AnimPath)
    virtual   float      Get_end_time()         const =0;  // this time is stored in _path (Pad_AnimPath)
    virtual   int        Get_time_resolution()  const {return 0;}     // applies only to channel members
    virtual   float      Get_last_render_time() const {return 0.0;}   // In seconds
    virtual   Pad_String Get_token()            const =0;  // Return animatable token

    virtual  Pad_AnimPath::PRE_POST Get_pre()  const {return (Pad_AnimPath::PRE_POST)3;}
    virtual  Pad_AnimPath::PRE_POST Get_post() const {return (enum Pad_AnimPath::PRE_POST)3;}
    virtual  Pad_OptionType Get_option() const {return (Pad_OptionType)0;}

    virtual   ~Pad_Animatable();

    //static   Pad_Bool Exists(Pad_String token); 


    Pad_List             parents;        // A list of animations that are immediate parrents of this channel

  private:
    enum ANIMATABLE_TYPE _type;
    Pad_Bool             _firstEnd; // Optimization flag to detect when animation ends
};

/*
Pad_Bool 
Pad_Animatable::Exists(Pad_String token) {
    Pad_Animatable  *animatable;
    Pad_Bool         exists = FALSE;
    
    if (token.Get()) {
	animatable = (Pad_Animatable *)Pad_Animatable::tokenTable.Get(token.Get());
	if (animatable) {
	    exists = TRUE;
	}
    }
    return exists;
}
*/
#endif    // Include ANIMATABLE


