/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef PADCACHE_H
#define PADCACHE_H 1

#include "object.h"
#include "hashtab.h"

#define MEGABYTES 1024*1000

class Pad_Image;
struct Pad_CacheInfo;

class Pad_CacheMgr {
public:
   Pad_CacheMgr(float cacheSize=2, 
		char *cacheDir=NULL,
		Pad_Bool disable=FALSE,
		long delayTime=5000);
   virtual ~Pad_CacheMgr();
                                            
   virtual Pad_Bool CacheOut(Pad_Object*);  // Indicates object can be cached out.
                                            // The cache manager will decide what to do.

   virtual Pad_Bool CacheIn(Pad_Object*);   // Ensures object is fully in memory.
                                            // Its data is reconstructed if cached out.

   void   Forget(Pad_Object*);              // Rremove any info the cache manager
                                            // may for the object.

	   Pad_Bool UpdateCache(Pad_Object*, Pad_CacheInfo*);

   // Methods for configuring the cache manager

   void     Cache(Pad_Bool);       // Enable/disable caching
   void     Size(float);           // Set the total cache size
   float    Size(void);
   Pad_Bool Dir(char*);            // Set the cache directory
   char*    Dir(void);
   void     ViewMultiple(float);   // Set the view bbox multiple the
   float    ViewMultiple(void);    // cache manager uses for visibility criteria
   void     UpdateDelay(long);
   long     UpdateDelay(void);
   Pad_Bool RmDir(char*);

protected:
   virtual Pad_Bool Reconstruct(Pad_Object*, Pad_CacheInfo*);
   virtual long     CacheSize(Pad_Object*);

           Pad_Bool CacheIn(Pad_Image*, Pad_CacheInfo*);
	   Pad_Bool CacheOut(Pad_Image*, Pad_CacheInfo*);
	   long     CacheSize(Pad_Image*, Pad_CacheInfo*);

   virtual Pad_CacheInfo **SelectObjs(Pad_Object*, int&);
           void            CacheFile(Pad_Object*, Pad_String&);
	   Pad_CacheInfo  *Allocate_Info(Pad_Object*, int);
	   void            ViewBBox(Pad_BBox&);

private:

   Pad_String    _cacheDir;
   long          _cacheSize;
   long          _usedBytes;
   Pad_HashTable _cacheInfoTable;
   Pad_CacheInfo **_infoArr;
   int           _infoArrSize, _infoArrLen;
   Pad_Bool      _noCaching;
   float         _viewMultiple;
   Pad           *_pad;
   long          _updateDelay;

public:
   Pad_TimerToken _updateTimer;
};

#endif
