/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef CALLBACK_H
#define CALLBACK_H 1

#include "bind.h"
#include "defs.h"
#include "pad-string.h"

class Pad_TkWin;
class Pad_Language;
class Pad_Event;
class Pad_Object;
class Pad_Callback;

typedef void (Pad_InitProc)      (Pad_Language *language, int argc, char **argv);
typedef int  (Pad_CommandProc)   (ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
typedef int  (Pad_CompleteProc)  (char *cmd);
typedef void (Pad_CreateProc)    (Pad_Language *language, char *pathname, Pad_TkWin *tkwin);
typedef void (Pad_PromptProc)    (Pad_Language *language, int partial, Pad_String &prompt);
typedef int  (Pad_EvalProc)      (Pad_Language *language, char *cmd);
typedef int  (Pad_CallbackProc)  (ClientData clientData);

//typedef int (Pad_EventCallbackProc) (Pad_Object*, ClientData, Pad_Event*);

// Sun's C++ compiler for Solaris can't handle typedef'ed class methods.
// So we use a macro instead:(

#define Pad_EventCallbackProc(Event) int Event (Pad_Object *, ClientData, Pad_Event *)

typedef Pad_Language *Pad_LanguagePtr;

//
// A Pad_Language represents a scripting language for Pad++.
// There is one of these for each language, and it contains callback
// functions for initializing the language interpreter, creating
// commands for pad widgets, for determining if a command is complete,
// and for creating the prompt.
//
class Pad_Language
{
  public:
    Pad_String       name;	     // Name of this language
    Pad_String       cmdName;	     // Name of command for accessing this language
    Tcl_Interp       *interp;	     // Tcl interpreter, or NULL for other languages
    Pad_InitProc     *initProc;	     // Callback to initialize interpreter
    Pad_CommandProc  *commandProc;   // Callback to process pad command
    Pad_CompleteProc *completeProc;  // Callback specifies if command is complete and should be evaluated
    Pad_CreateProc   *createProc;    // Callback when a pad widget is created
    Pad_PromptProc   *promptProc;    // Callback to generate a prompt
    Pad_EvalProc     *evalProc;      // Callback to evaluate a string
    
				     // General form for scripting languages
    Pad_Language(Tcl_Interp *interp, char *name, char *command, Pad_InitProc *initProc, 
		 Pad_CreateProc *createProc, Pad_CommandProc *commandProc, 
		 Pad_PromptProc *promptProc, Pad_CompleteProc *completeProc, 
		 Pad_EvalProc *evalProc);
};

//
// A Pad_Callback contains a function in any of the available scripting languages.
// It stores the function, the language, and contains methods for modification,
// access, and evaluation.
//
class Pad_Callback
{
  private:
    Pad_Language * language;	         // Language of callback
    Pad_String *   script;	         // Text of script
    Pad_CallbackProc * func;	         // C++ function pointer
    Pad_EventCallbackProc((*eventFunc)); // C++ function pointer
    ClientData     clientData;           // Data for function callback
    int            refCount;	         // Number of time this callback used within an event binding

  public:
    void     Append(char *newScript);   // Add on to script
    int      Eval(void);	        // Evaluates script
    int      Eval(Pad_Object *obj, Pad_Event *padEvent);   // Evaluates script, and passes obj and event pointers
    char *   Get(void);		  // Pointer to script
    Pad_Bool Is_script(void);     // Return TRUE if callback is a script (as opposed to a C call)
    Pad_Language *Language(void); // Script language is to be evaluated in
    int      Length(void);	  // Length of script in chars
    void     Set(Pad_Callback *newCallBack);                         // Replace callback with new one
    void     Set(char *newScript, Pad_Language *newLanguage);        // Replace callback with new script
    void     Set(Pad_String &newScript, Pad_Language *newLanguage);  // Replace callback with new script
    void     Increment_refcount(void);
    void     Decrement_refcount(void);
    int      Get_refcount(void);

    Pad_Callback(char *newScript, Pad_Language *newLanguage);
    Pad_Callback(Pad_String &newScript, Pad_Language *newLanguage);
    Pad_Callback(Pad_CallbackProc *newFunc, ClientData clientData=NULL);
    Pad_Callback(Pad_EventCallbackProc((*newEventFunc)), ClientData clientData=NULL);
    Pad_Callback(Pad_Callback &callback);
    ~Pad_Callback();

  private:
    void _Init(char *newScript, Pad_Language *newLanguage);
    void _Guess_language(void);
};

///////////////////////////////////////////////////////////////////////
//
// Function Prototypes
//
///////////////////////////////////////////////////////////////////////

int      Pad_GlobalEval(Pad_Language *language, char *script);
Pad_Bool Pad_GetLanguage(char *languageStr, Pad_Language **language);

#endif
