/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

/*
   MI -- I've strived to keep the original source code and logic as much
   as possible.  Most of the Tk_ routines used (e.g. Tk_Width()) have
   been changed to their Pad equivalent (defined in this file).
   Also coordinate and distance information is changed from int to float.
   Code that wasn't needed or relevant is mostly commented (or ifdef'ed).
*/
/* 
 * tkGrid.c --
 *
 *	Grid based geometry manager.
 *

 * Copyright (c) 1997 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *
 * SCCS: @(#) tkGrid.c 1.21 96/02/21 10:50:58
 */

#include "defs.h"
#include "tkInt.h"
#include "tclInt.h"
#include "grid.h"
#include "pad-tcl.h"
#include "win.h"
#include "global.h"

#include <stdlib.h>
/*
 * LayoutInfo structure.  We shouldn't be using hard-wired limits!
 */

// MI -- Instead of harcoding MAXGRIDSIZE I've put it in the
// GridBag structure and it is dynamically updated for each
// grid as slaves are positioned in the grid.

// #define MAXGRIDSIZE 128

#ifndef MAXINT
#  define MAXINT 0x7fff
#endif
#define MINWEIGHT	0.0001		/* weight totals < this are considered to be zero */

/*
 * Special characters to support relative layouts
 */

#define REL_SKIP	'x'	/* skip this column */
#define REL_HORIZ	'-'	/* extend previous widget horizontally */
#define REL_VERT	'^'	/* extend previous widget verticallly */

/*
 *  structure to hold collected constraints temporarily:
 *  needs to use a "Constrain" thingy
 */

// MI -- have to turn this into a C++ structure so its arrays can be
// dynamically allocated with the right size for individual grids.

struct LayoutInfo {
    LayoutInfo(int MAXGRIDSIZE=128);
    ~LayoutInfo();

    int width, height;	/* number of cells horizontally, vertically */
    int lastRow;	/* last cell with a window in it */
    float *minWidth;	/* largest minWidth in each column */
    float *minHeight;	/* largest minHeight in each row */
    double *weightX;	/* largest weight in each column */
    double *weightY;	/* largest weight in each row */
};

LayoutInfo::LayoutInfo(int MAXGRIDSIZE)
{
  width = height = lastRow = 0;
  minWidth = new float[MAXGRIDSIZE];
  minHeight = new float[MAXGRIDSIZE];
  weightX = new double[MAXGRIDSIZE];
  weightY = new double[MAXGRIDSIZE];
  for (int i=0; i < MAXGRIDSIZE; i++) {
    minWidth[i] = minHeight[i] = weightX[i] = weightY[i] = 0;
  }
}

LayoutInfo::~LayoutInfo()
{
  delete[] minWidth;
  delete[] minHeight;
  delete[] weightX;
  delete[] weightY;
}

/* structure for holding row and column constraints */

typedef struct {
    int used;		/* maximum element used */
    int max;		/* maximum element allocated */
    float *minsize;		/* array of minimum column/row sizes */
    double *weight;	/* array of column/row weights */
} Constrain;

/* For each window that the gridbag cares about (either because
 * the window is managed by the gridbag or because the window
 * has slaves that are managed by the gridbag), there is a
 * structure of the following type:
 */

typedef struct GridBag {
    Pad_Object* tkwin;		/* Tk token for window.  NULL means that
				 * the window has been deleted, but the
				 * packet hasn't had a chance to clean up
				 * yet because the structure is still in
				 * use. */
    struct GridBag *masterPtr;	/* Master window within which this window
				 * is managed (NULL means this window
				 * isn't managed by the gridbag). */
    struct GridBag *nextPtr;	/* Next window managed within same
				 * parent.  List is priority-ordered:
				 * first on list gets layed out first. */
    struct GridBag *slavePtr;	/* First in list of slaves managed
				 * inside this window (NULL means
				 * no gridbag slaves). */

    int gridColumn, gridRow;
    int gridWidth, gridHeight;

    int tempX, tempY;
    int tempWidth, tempHeight;

    double weightX, weightY;
    float minWidth, minHeight;

    float padX, padY;		/* Total additional pixels to leave around the
				 * window (half of this space is left on each
				 * side).  This is space *outside* the window:
				 * we'll allocate extra space in frame but
				 * won't enlarge window). */
    float iPadX, iPadY;		/* Total extra pixels to allocate inside the
				 * window (half this amount will appear on
				 * each side). */
    float startx, starty;		/* starting location of layout */

    float doubleBw;		/* Twice the window's last known border
				 * width.  If this changes, the window
				 * must be re-arranged within its parent. */
    int *abortPtr;		/* If non-NULL, it means that there is a nested
				 * call to ArrangeGrid already working on
				 * this window.  *abortPtr may be set to 1 to
				 * abort that nested call.  This happens, for
				 * example, if tkwin or any of its slaves
				 * is deleted. */
    int flags;			/* Miscellaneous flags;  see below
				 * for definitions. */

    Constrain row, column;		/* column and row constraints */

    int valid;
    LayoutInfo *layoutCache;
    float pad_x, pad_y;   // saved coord in parent group

    int MAXGRIDSIZE;
} GridBag;

/*
 * Flag values for GridBag structures:
 *
 * REQUESTED_RELAYOUT:		1 means a Tk_DoWhenIdle request
 *				has already been made to re-arrange
 *				all the slaves of this window.
 * STICK_NORTH  		1 means this window sticks to the edgth of its
 * STICK_EAST			cavity
 * STICK_SOUTH
 * STICK_WEST
 *
 * DONT_PROPAGATE:		1 means don't set this window's requested
 *				size.  0 means if this window is a master
 *				then Tk will set its requested size to fit
 *				the needs of its slaves.
 */

#define STICK_NORTH		1
#define STICK_EAST		2
#define STICK_SOUTH		4
#define STICK_WEST		8
#define STICK_ALL		(STICK_NORTH|STICK_EAST|STICK_SOUTH|STICK_WEST)

#define REQUESTED_RELAYOUT	16
#define DONT_PROPAGATE		32

/*
 * Hash table used to map from Tk_Window tokens to corresponding
 * GridBag structures:
 */

static Tcl_HashTable gridBagHashTable;

/*
 * Have statics in this module been initialized?
 */

static int initialized = 0;

/*
 * Prototypes for procedures used only in this file:
 */

static void		ArrangeGrid (ClientData clientData);
static int		ConfigureSlaves (Tcl_Interp *interp, Pad_Win *win,
			    Pad_Grid* tkwin, int argc, char *argv[]);
static void		DestroyGridBag (char *memPtr);
static void		GetCachedLayoutInfo (GridBag *masterPtr);
static GridBag *	GetGridBag (Pad_Object* tkwin);
static void		GetLayoutInfo (GridBag *masterPtr,
			    LayoutInfo *r);
static void		GetMinSize (GridBag *masterPtr,
			    LayoutInfo *info, float *minw, float *minh);

#ifdef notdef
static void		GridBagStructureProc(ClientData clientData, 
					     XEvent *eventPtr);
static void		GridLostSlaveProc(ClientData clientData,
					  Tk_Window tkwin);
static void		GridReqProc (ClientData clientData,
				     Tk_Window tkwin);
static void		GridBagStructureProc (
					      ClientData clientData, XEvent *eventPtr);
#endif

static void		StickyToString (int flags, char *result);
static int		StringToSticky (char *string);
static void		Unlink (GridBag *gridPtr);

#ifdef notdef
static Tk_GeomMgr gridMgrType = {
    "grid",			/* name */
    GridReqProc,		/* requestProc */
    GridLostSlaveProc,		/* lostSlaveProc */
};
#endif

int Pad_GridCmd(ClientData, Tcl_Interp*, Pad_Win*, int, char**);

//
// Routine for handling pad tcl grid command
//
int Padtcl_grid(ClientData clientdata, Tcl_Interp *interp, int argc, char **argv) {
  Pad_Win *win = (Pad_Win *)clientdata;
  return Pad_GridCmd(NULL, interp, win, argc, argv);
}

//
// Pad_Table constructor is different from most other pad objects
// since attribute configuration is done by the underlying grid code.
//
Pad_Grid::Pad_Grid(Pad *pad)
    : Pad_Group(pad) 
{
  _type = PAD_GRID;
  grid = GetGridBag(this);
}

//
// Handle configure options on the create grid command line.
//
int
Pad_Grid::Create_obj_args(int argc, char **argv)
{
  // call grid routines to create the grid structures 
  char **hack_argv = new char*[argc+1];
  hack_argv[0] = "grid";
  for (int i = 0; i < argc; i++) hack_argv[i+1] = argv[i];
  Pad_Win *win = pad->view->win;
  Tcl_Interp *interp = pad->view->win->interp;
  Pad_GridCmd(this, interp, win, argc+1, hack_argv);
  return argc;
}

//
// The grid destructor deletes all its C structures and cancels
// any relayout requests.
//
Pad_Grid::~Pad_Grid() {

  GridBag *gridPtr = (GridBag*) grid;
  if (gridPtr->flags & REQUESTED_RELAYOUT) {
    Tk_CancelIdleCall(ArrangeGrid, (ClientData) gridPtr);
  }

  GridBag *freeSlave;

  for (GridBag *slavePtr = gridPtr->slavePtr; slavePtr != NULL; /*see body*/) {
    freeSlave = slavePtr;
    freeSlave->tkwin->flags &= ~PAD_MASK_GRID_SLAVE;
    slavePtr = slavePtr->nextPtr;
             // free slave's gridbag now for non-grid slaves 
             // grid slaves gridbag will be deallocated at the end here when
             // slave destructor is called.
    if (freeSlave->tkwin->Type() != PAD_GRID) {
      DestroyGridBag((char*)freeSlave);
    }
  }
  DestroyGridBag((char*)grid);
}

//
// Convineince get & set methods for the C grid structure of the
// grid object.
//
void*
Pad_Grid::Get_Grid() {
  return grid;
}

void
Pad_Grid::Set_Grid(void *g) {
  grid = g;
}


//
// Have the grid compute its bounding box after all its slaves
// are positioned (and not while a relayout is in progress).
//
void
Pad_Grid::Compute_bounding_box() {
  GridBag *gridPtr = (GridBag*) Get_Grid();
  if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
    // recompute bbox but keep the anchor point (group doesn't do this).
    Set_bounding_box_to_enclose_list(members);
    Pad_Object::Compute_bounding_box();
    Pad_Group::Update_global_bounding_box();
  }
}

void
Pad_Grid::Update_global_bounding_box(void)
{
  GridBag *gridPtr = (GridBag*) Get_Grid();
  if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
    Pad_Group::Update_global_bounding_box();
  }
}


//
// Remove a slave object from the grid.  The internal structures for
// the slave are deallocated and it is removed from the grid group.
//
Pad_Bool
Pad_Grid::Remove(Pad_Object *obj, Pad_Bool transform)
{
  Pad_Bool rc = TRUE;

  obj->flags &= ~PAD_MASK_GRID_SLAVE;
  GridBag *slavePtr = GetGridBag(obj);
  Unlink(slavePtr);
  if (obj->_type != PAD_GRID) {
    DestroyGridBag((char*)slavePtr);
  }

  rc = Pad_Group::Remove(obj, transform);
  return(rc);
}

//
// Deleting a slave object from the grid is similar to removing it from
// the grid except that it is deleted from the grid group.
//
Pad_Bool
Pad_Grid::Delete(Pad_Object *obj)
{
  Pad_Bool rc = TRUE;

  obj->flags &= ~PAD_MASK_GRID_SLAVE;
  GridBag *slavePtr = GetGridBag(obj);
  if (slavePtr->masterPtr != NULL) {
    if (!(slavePtr->masterPtr->flags & REQUESTED_RELAYOUT)) {
      slavePtr->masterPtr->flags |= REQUESTED_RELAYOUT;
      Tk_DoWhenIdle(ArrangeGrid, (ClientData) slavePtr->masterPtr);
    }
  }
  Unlink(slavePtr);
  DestroyGridBag((char*)slavePtr);

  rc = Pad_Group::Delete(obj);
  return(rc);
}

//
// Compute eclosing bbox of all the slaves.
// Take slaves x & x padding into account.
//
void 
Pad_Grid::Set_bounding_box_to_enclose_list(Pad_List &list)
{
    Pad_BBox bb;
    Pad_BBox obbox;
    Pad_Object *f;
    Pad_Iterator oi;
    Pad_Object *o;
    
    f = (Pad_Object *)list.First();
    
    if (f) {
	f->Get_bbox(bb);
	GridBag *gridPtr = GetGridBag(f);
	bb.Set_xmin(bb.Xmin() - gridPtr->padX/2);
	bb.Set_xmax(bb.Xmax() + gridPtr->padX/2);
	bb.Set_ymin(bb.Ymin() - gridPtr->padY/2);
	bb.Set_ymax(bb.Ymax() + gridPtr->padY/2);
	obbox.Set(bb);
	f->transform.Apply(obbox);
    } else {
	obbox.Set_zero();
    }
    
    DOLIST(oi, list, Pad_Object, o) { 
	o->Get_bbox(bb);
	GridBag *gridPtr = GetGridBag(o);
	bb.Set_xmin(bb.Xmin() - gridPtr->padX/2);
	bb.Set_xmax(bb.Xmax() + gridPtr->padX/2);
	bb.Set_ymin(bb.Ymin() - gridPtr->padY/2);
	bb.Set_ymax(bb.Ymax() + gridPtr->padY/2);
	o->transform.Apply(bb);
	obbox.Union(bb);
    }

    if (!(optionFlags & PAD_WIDTH_SET)) {
	//bbox.Set_x(obbox.Xmin(), obbox.Xmax());
	Set_bbox_xmin(obbox.Xmin());
        Set_bbox_xmax(obbox.Xmax());
    }
    if (!(optionFlags & PAD_HEIGHT_SET)) {
	//bbox.Set_y(obbox.Ymin(), obbox.Ymax());
	Set_bbox_ymin(obbox.Ymin());
	Set_bbox_ymax(obbox.Ymax());
    }
}

//
// The write methods for Pad_Grid are different from normal pad objects
// since slave attributes are stored in underlying C structures.
// We have to create the grid object first, and after writing
// each slave we can then do its grid configuration.
//

void
Pad_Grid::Write_obj(char *padname, Pad_String &str)
{
  Pad_Iterator oi;
  Pad_Option *option;
  Pad_Type *type = (Pad_Type *)pad->typeTable->Get(Pad_allUid);
  Tcl_Interp *interp = pad->view->win->interp;
  Tk_Uid typeUid;

  // code for creating the grid
  str += "set Pad_Grid";
  str += id;
  str += " [";
  str += padname;
  str += " create ";
  typeUid = Tk_GetUid(Type_name());
  str += typeUid;
  Write_obj_args(padname, str);
  str += "]\n";

  // code for grid group configuration
  str += padname;
  str += " ic $Pad_Grid";
  str += id;
  str += " ";
  
  // First, extract generic options
  DOLIST(oi, type->options, Pad_Option, option) {
    Pad_AppendOption(interp, this, option, str, TRUE, FALSE);
  }

  // Then, extract options for this object's type
  type = (Pad_Type *)pad->typeTable->Get(typeUid);
  DOLIST(oi, type->options, Pad_Option, option) {
    Pad_AppendOption(interp, this, option, str, TRUE, FALSE);
  }
  str += "\n";

  // write the slaves
  Write_obj_pre(padname, str);
  Write_obj_post(padname, str);

  str += "set Pad_ID $Pad_Grid";
  str += id;
  str += "\n";
  str += "set Pad_NewID(";
  str += id;
  str += ") $Pad_ID\n";
}

//
// Writes all grid slaves and their grid configuration.
// It is assumed the grid object is already created.
//
Pad_Bool
Pad_Grid::Write_obj_pre(char *padname, Pad_String &str) {
  Pad_Bool rc;
  Pad_Object *obj;
  Pad_String cb_str;
  Pad_StreamData sd;
  
  rc = TRUE;
  char string[200];
  // GridBag *masterPtr = (GridBag*)Get_Grid();  doesn't seem to work
  GridBag *masterPtr = GetGridBag(this), *slavePtr = NULL;

  // write out each slave
  for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
       slavePtr = slavePtr->nextPtr) {
    
    // write object
    obj = slavePtr->tkwin;
    cb_str = "";
    if (obj->Write_obj_callback(cb_str)) {
      obj->Write_obj_wrapper(PAD_SCRIPT_FORMAT, padname, str, sd, TRUE, FALSE);

      // write obj's grid configuration 
      str += padname;
      str += " grid $Pad_ID -in $Pad_Grid";
      str += id;
      
      sprintf(string, " -column %d -row %d -columnspan %d -rowspan %d",
	      slavePtr->gridColumn, slavePtr->gridRow,
	      slavePtr->gridWidth, slavePtr->gridHeight);
      str += string;
      // NOTE: not writing internal padding since we are not handling them
      sprintf(string, /*" -ipadx %f -ipady %f*/ " -padx %f -pady %f",
	      /*slavePtr->iPadX/2, slavePtr->iPadY/2,*/ slavePtr->padX/2,
	      slavePtr->padY/2);
      str += string;
      str += " -sticky ";
      StickyToString(slavePtr->flags, string);
      str += string;
     str += "\n";
    }      
    
    str += cb_str;
  }

  return rc;
}

void
Pad_Grid::Write_obj_args(char *, Pad_String &)
{
}

void
Pad_Grid::Write_obj_post(char *, Pad_String &)
{
}

//
// Private method for getting all the slaves grid configurations.
//
void
Pad_Grid::Write_slaves_config(char *padname, Pad_String &str)
{
  Pad_Object *obj;
  char string[200];
  GridBag *masterPtr = GetGridBag(this), *slavePtr = NULL;

  for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
       slavePtr = slavePtr->nextPtr) {
    
    // write slave configuration 
    obj = slavePtr->tkwin;
    str += padname;
    //str += " grid $Pad_ID -in $Pad_Grid";
    sprintf(string," grid %d -in %d", obj->id, id);
    str += string;
      
    sprintf(string, " -column %d -row %d -columnspan %d -rowspan %d",
	    slavePtr->gridColumn, slavePtr->gridRow,
	    slavePtr->gridWidth, slavePtr->gridHeight);
    str += string;
    // NOTE: not writing internal padding since we are not handling them
    sprintf(string, /*" -ipadx %f -ipady %f*/ " -padx %f -pady %f",
	    /*slavePtr->iPadX/2, slavePtr->iPadY/2,*/ slavePtr->padX/2, slavePtr->padY/2);
    str += string;
    str += " -sticky ";
    StickyToString(slavePtr->flags, string);
    str += string;
    str += "\n";
  }      
}

//
// Add the grid and all the members to the specified layer.
// This is different from Pad_Group::Add_to_layer since the
// we don't want the grid Remove() method to get called (will
// lose slaves configurations).  We just have to make sure the 
// group Remove() method is called.
//
void
Pad_Grid::Add_to_layer(char *name)
{
    Pad_Object *obj;
    Pad_Iterator oi;
    Pad_List copyOfMembers;
    Pad_Layer *layer;
    Tk_Uid nameUid;

                                // If already on the same layer, don't do anything
    layer = pad->Get_layer_from_id(layerId);
    nameUid = Tk_GetUid(name);
    if (layer->name == nameUid) {
	return;
    }

    copyOfMembers = members;

				// First remove all members from group
    DOLIST(oi, copyOfMembers, Pad_Object, obj) {
        Pad_Group::Remove(obj, FALSE);  // HACK: call group remove method
    }
				// Change group to new layer
    Pad_Component::Add_to_layer(name);

				// Then add members back (which will put them on the new layer)
    copyOfMembers.Reverse();
    DOLIST(oi, copyOfMembers, Pad_Object, obj) {
        Pad_Group::Add(obj, FALSE);  // HACK: call group add method
    }
}

//
// Binary format for grid objects:
// - stream for options flags, divisible and number of slaves
// - zero or more streams slaves and the tcl scripts that adds
// each object to the grid.
//
Pad_Bool
Pad_Grid::Write_obj(char *padname, Pad_StreamData &sd) {

    Pad_Bool rc = Pad_Object::Write_obj(padname, sd);

    // option flags
    sd << groupFlags;
    sd << divisible;

    // members
    sd << members.Length();
    sd << id;
    sd.Flush();  // since we want members in seperate streams

    GridBag *masterPtr = (GridBag*)Get_Grid();
    Pad_Object *obj;
    Pad_String str;
    char string[200];

    // write out each slave
    for (GridBag *slavePtr = masterPtr->slavePtr; slavePtr != NULL;
	 slavePtr = slavePtr->nextPtr) {
        obj = slavePtr->tkwin;
        obj->Write_obj_wrapper(sd.Get_rep(), padname, str, sd, TRUE, FALSE);

	// write obj's grid configuration 
	str << padname << " grid $Pad_NewID(" << obj->id << ")"
	    << " -in $Pad_Grid" << id;
	
	GridBag *slavePtr = GetGridBag(obj);
	sprintf(string, " -column %d -row %d -columnspan %d -rowspan %d",
		slavePtr->gridColumn, slavePtr->gridRow,
		slavePtr->gridWidth, slavePtr->gridHeight);
	str += string;
	// NOTE: not writing internal padding since we are not handling them
	sprintf(string, /*" -ipadx %f -ipady %f */ " -padx %f -pady %f",
		/*slavePtr->iPadX/2, slavePtr->iPadY/2,*/ slavePtr->padX/2,
		slavePtr->padY/2);
	str += string;
	str += " -sticky ";
	StickyToString(slavePtr->flags, string);
	str += string;
	str += "\n";
	
	sd << str.Get();
	sd.Flush();
    }
    
    return rc;
}

Pad_Bool 
Pad_Grid::Read_obj(Pad_StreamData &sd) {

   Pad_Bool rc = Pad_Object::Read_obj(sd);
   int n=0, tag, oldid=0;

   // option flags 
   rc &= !sd.Reconstruct(groupFlags);
   rc &= !sd.Reconstruct(divisible);
   rc &= !sd.Reconstruct(n);
   rc &= !sd.Reconstruct(oldid);
   sd.Clear();

   // members
   Pad_Object *obj=NULL;
   Tcl_Interp *interp=pad->view->win->interp;

   Pad_String str;
   str << "set Pad_Grid" << oldid << " " << id;
   Tcl_GlobalEval(interp, str.Get());

   for (int i=0; i < n; i++) {
      sd.Fill();
      if (!sd.Reconstruct(tag)) {
	// create object
	obj = pad->Create_object(tag);
	if (obj) {
	  // get rest of the object
	  rc &= obj->Read_obj(sd);
	  Add(obj, FALSE);
	  // get object script
	  Read_script(NULL, sd, interp);

	} else if (tag == PAD_OBJ_SCRIPT) {
	  // get script
	  Read_script(NULL, sd, interp);
	} else if (tag == PAD_USER_TYPE) {
	  Pad_errorString << "Ignoring PAD_USER_TYPE stream\n";
	  rc &= FALSE;
	} else {
	  Pad_errorString << "ERR: Unknown stream tag: " << tag << " -- SKIPPING STREAM\n";
	  rc &= FALSE;
	}

	// get script to place in grid
	sd.Fill();
	Read_script(NULL, sd, interp);

      } else {
	rc &= FALSE;
      }

      sd.Clear();
   }

   if (!rc) {
     Pad_errorString << "ERR: problems reading grid object: " << id << "\n";
   }

   this->Damage();
   return rc;
}


//
// Pad replacements for Tk routines that the original grid C code uses.
//

static float
Pad_Width(Pad_Object* obj) {
  float width;
  Pad_BBox *bb;
  
  bb = obj->Get_global_bbox();
  width = bb->Width();

  return(width); 
}

static float
Pad_Height(Pad_Object* obj) {
  float height;
  Pad_BBox *bb;
  
  bb = obj->Get_global_bbox();
  height = bb->Height();
  
  return(height);
}

//
// For pad, requested width and height are the same
// as stored width.
//
static float
Pad_ReqWidth(Pad_Object* obj) {
    return Pad_Width(obj);
}

static float
Pad_ReqHeight(Pad_Object* obj) {
    return Pad_Height(obj);
}

static float
Pad_BorderWidth(Pad_Object*) {
  return 0;  // since we use bounding box
}

static float
Pad_InternalBorderWidth(Pad_Object*) {
    return 0;
}

//
// The Pad_X and Pad_Y() routines are supposed to return the x & y
// of the object.  We use the bbox to do this.
//
/*  Not used
static float
Pad_X(Pad_Object* obj) {
  Pad_BBox box;
  obj->Get_bbox(box);
  return box.Xmin();
}

static float
Pad_Y(Pad_Object* obj) {
  Pad_BBox box;
  obj->Get_bbox(box);
  return box.Ymin();
}
*/

static char*
Pad_PathName( Pad_Object* obj) {
  static char buf[16];
  sprintf(buf,"%d", obj->id);
  return buf;
}

static Pad_Object*
Pad_NameToWindow(Tcl_Interp*, char *tagorid, Pad_Object* anyobj) {
  Pad_List objs;
  anyobj->pad->Find_with_tagorid(tagorid, objs, TRUE);
  if (objs.Is_empty())
    return NULL;
  else
    return (Pad_Object*)objs.First();
}

/*  Not used
static int 
Pad_GeometryRequest(Pad_Object* obj, float w, float h) {
  obj->Set_width(w);
  obj->Set_height(h);
  return TCL_OK;
}
*/

//
// Assume MoveResizeWindow() is only called for grid slaves and 
// that x & y are relative to the grid.
//
static int
Pad_MoveResizeWindow(Pad_Object* slave, float x, float y, float /*w*/, float /*h*/) {
  // get the grid and transform x & y.
  Pad_Grid *grid = (Pad_Grid*) slave->group;
  Pad_BBox *bb;
  bb = grid->Get_global_bbox();
  float transf_x = bb->Xmin() + x;
  float transf_y = bb->Ymax() - y;  // trickery to get rows from top to bottom:).
              
  //fprintf(stderr, "%s at %f,%f is %fx%f\n", 
  //  Pad_PathName(slave), transf_x, transf_y, w, h);
	
  // find upper right corner of object and slide the object
  Pad_BBox *slavebb;
  //slave->Get_bbox(slavebb);
  slavebb = slave->Get_global_bbox();
  //slave->Transform(slavebb);
  slave->Slide(transf_x - slavebb->Xmin(),
	       transf_y - slavebb->Ymax(), TRUE);
/*
  if (w != Pad_Width(slave)) {
    float invw = w;
    slave->Invert(&invw);
    slave->Set_width(invw);
  }
  if (h != Pad_Height(slave)) {
    float invh = h;
    slave->Invert(&invh);
    slave->Set_height(invh);
  }
*/

  return TCL_OK;
}

//
// This routine is suppose to compute a pixel value from a string.
// We're not doing any conversion for now...
//
static int
Pad_GetPixels(Tcl_Interp*, Pad_Object*, char *string, float *ptr) {
  *ptr = atof(string); // hack...
  return TCL_OK;
}

static int
Pad_IsMapped(Pad_Object* obj) {
    if (obj->Get_transparency() == 0.0) {
	return(0);
    } else {
	return(1);
    }
}

static int
Pad_MapWindow(Pad_Object* obj) {
  obj->Set_transparency(1);
  return TCL_OK;
}

static int
Pad_UnmapWindow(Pad_Object* obj) {
  obj->Set_transparency(0);
  return TCL_OK;
}

/* Not used
static Pad_Object*
Pad_Parent( Pad_Object* obj) {
  if (obj->group)
    return obj->group;
  else
    //return obj->pad;
    return NULL;
}
*/

static int
Pad_IsTopLevel(Pad_Object* obj) {
  return (obj->id == 1); // hack...
}

//
// Event handler for grid slaves.  If a slave is mofdified then
// its grid is scheduled for relayout.  If a slave is deleted then
// it is removed from the grid and its C structures are deleted.
//
void
Pad_Grid::SlaveEventProc(
    Tcl_Interp *,
    Pad_Object* slave,
    int event_type)
{
    GridBag *slavePtr = GetGridBag(slave);
    GridBag *gridPtr = slavePtr->masterPtr;

    if (gridPtr) {
      if (event_type == Pad_ModifyNotify) {
	gridPtr->valid = 0;
	if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
	  gridPtr->flags |= REQUESTED_RELAYOUT;
	  Tk_DoWhenIdle(ArrangeGrid, (ClientData) gridPtr);
	}
      }
    } else if (event_type == Pad_DeleteNotify) {
      slave->flags &= ~PAD_MASK_GRID_SLAVE;
      Unlink(slavePtr);
      gridPtr->valid = 0;
      if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
	gridPtr->flags |= REQUESTED_RELAYOUT;
	Tk_DoWhenIdle(ArrangeGrid, (ClientData) gridPtr);
      }
    }
}


/*
 *--------------------------------------------------------------
 *
 * Tk_GridCmd --
 *
 *	This procedure is invoked to process the "grid" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */
int
Pad_GridCmd(
    ClientData clientData,	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp,		/* Current interpreter. */
    Pad_Win *win,
    int argc,			/* Number of arguments. */
    char **argv) 		/* Argument strings. */
{
    size_t length;
    char c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option arg ?arg ...?\"\n", 
			 "option must be bbox, columnconfigure, configure, forget, info, location\n",
			 "propagate, rowconfigure, size, or slaves", (char *) NULL);
	return TCL_ERROR;
    }

    //if ((argc >= 2) && (argv[1][0] == '.')) {
    Pad_Object *obj = Pad_NameToWindow(interp, argv[1], win->view);
    if ((argc >= 2) && (obj != NULL)) {
        if (clientData != NULL)
	    return ConfigureSlaves(interp, win, (Pad_Grid*)clientData, argc-1, argv+1);
	else {
	    return ConfigureSlaves(interp, win, NULL, argc-1, argv+1);
	}
    }

    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option arg ?arg ...?\"\n", 
			 "option must be bbox, columnconfigure, configure, forget, info, location\n",
			 "propagate, rowconfigure, size, or slaves", (char *) NULL);
	return TCL_ERROR;
    }

    Pad_Win *tkwin = win;
    c = argv[1][0];
    length = strlen(argv[1]);
  
    if ((c == 'a') && (strncmp(argv[1], "arrange", length) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "Wrong number of arguments: ",
		    "must be \"",argv[0],
		    " arrange <master>\"", (char *) NULL);
	    return TCL_ERROR;
	}
	Pad_Object *master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL || master->_type != PAD_GRID) {
	    return TCL_ERROR;
	}
	GridBag *masterPtr = GetGridBag(master);
	Tk_CancelIdleCall(ArrangeGrid, (ClientData) masterPtr);
	ArrangeGrid(masterPtr);

    } else if ((c == 'b') && (strncmp(argv[1], "bbox", length) == 0)) {
	Pad_Object* master;
	GridBag *masterPtr;
	int row, column;
	int i; float x, y;
	float prevX, prevY;
	float width, height;
	double weight;
	float diff;

	if (argc != 5) {
	    Tcl_AppendResult(interp, "Wrong number of arguments: ",
		    "must be \"",argv[0],
		    " bbox <master> <column> <row>\"", (char *) NULL);
	    return TCL_ERROR;
	}
        
	master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL) {
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[3], &column) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[4], &row) != TCL_OK) {
	    return TCL_ERROR;
	}
	masterPtr = GetGridBag(master);

	/* make sure the grid is up to snuff */

	while ((masterPtr->flags & REQUESTED_RELAYOUT)) {
	    Tk_CancelIdleCall(ArrangeGrid, (ClientData) masterPtr);
	    ArrangeGrid((ClientData) masterPtr);
	}
	GetCachedLayoutInfo(masterPtr);

	if (row < 0 || column < 0) {
	    *interp->result = '\0';
	    return TCL_OK;
	}
	if (column >= masterPtr->layoutCache->width ||
		row >= masterPtr->layoutCache->height) {
	    *interp->result = '\0';
	    return TCL_OK;
	}
	x = masterPtr->startx;
	y = masterPtr->starty;
	GetMinSize(masterPtr, masterPtr->layoutCache, &width, &height);

	diff = Pad_Width(masterPtr->tkwin) - (width + masterPtr->iPadX);
	for (weight=0.0, i=0; i<masterPtr->layoutCache->width; i++)
	    weight += masterPtr->layoutCache->weightX[i];

	prevX = 0;			/* Needed to prevent gcc warning. */
	for (i=0; i<=column; i++) {
	    float dx = 0;
	    if (weight > MINWEIGHT) {
		dx = ((((double)diff) * masterPtr->layoutCache->weightX[i])
			/ weight);
	    }
	    prevX = x;
	    x += masterPtr->layoutCache->minWidth[i] + dx;
	}
	diff = Pad_Height(masterPtr->tkwin) - (height + masterPtr->iPadY);
	for (weight=0.0, i=0; i<masterPtr->layoutCache->width; i++) {
	    weight += masterPtr->layoutCache->weightY[i];
	}
	prevY = 0;			/* Needed to prevent gcc warning. */
	for (i=0; i<=row; i++) {
	    float dy = 0;
	    if (weight > MINWEIGHT) {
		dy = ((((double)diff) * masterPtr->layoutCache->weightY[i])
			/ weight);
	    }
	    prevY = y;
	    y += masterPtr->layoutCache->minHeight[i] + dy;
	}
	sprintf(interp->result,"%f %f %f %f",prevX,prevY,x - prevX,y - prevY);
    } else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	//if (argv[2][0] != '.') {
	Pad_Object *obj = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (obj == NULL) {
	    Tcl_AppendResult(interp, "bad argument \"", argv[2],
		    "\": must be name of window", (char *) NULL);
	    return TCL_ERROR;
	}
	return ConfigureSlaves(interp, win, NULL, argc-2, argv+2);
    } else if ((c == 'f') && (strncmp(argv[1], "forget", length) == 0)) {
	Pad_Object* slave;
	GridBag *slavePtr;
	int i;
    
	for (i = 2; i < argc; i++) {
	    slave = Pad_NameToWindow(interp, argv[i], tkwin->view);
	    if (slave == NULL) {
		return TCL_ERROR;
	    }
	    slavePtr = GetGridBag(slave);
	    if (slavePtr->masterPtr != NULL) {
	        Pad_Grid *grid = (Pad_Grid*) slavePtr->masterPtr->tkwin;
		grid->Remove(slave, TRUE);
		//Pad_ManageGeometry(slave, (Tk_GeomMgr *) NULL,
		//	(ClientData) NULL);
		//Unlink(slavePtr);
		//Tk_UnmapWindow(slavePtr->tkwin);
		//DestroyGridBag((char*)slavePtr);
	    }
	}
    } else if ((c == 'i') && (strncmp(argv[1], "info", length) == 0)) {
	register GridBag *slavePtr;
	Pad_Object* slave;
	char buffer[64];
    
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " info window\"", (char *) NULL);
	    return TCL_ERROR;
	}
	slave = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (slave == NULL) {
	    return TCL_ERROR;
	}
	slavePtr = GetGridBag(slave);
	if (slavePtr->masterPtr == NULL) {
	    interp->result[0] = '\0';
	    return TCL_OK;
	}
    
	Tcl_AppendElement(interp, "-in");
	Tcl_AppendElement(interp, Pad_PathName(slavePtr->masterPtr->tkwin));
	sprintf(buffer, " -column %d -row %d -columnspan %d -rowspan %d",
		slavePtr->gridColumn, slavePtr->gridRow,
		slavePtr->gridWidth, slavePtr->gridHeight);
	Tcl_AppendResult(interp, buffer, (char *) NULL);
	//sprintf(buffer, " -ipadx %f -ipady %f -padx %f -pady %f",
		//slavePtr->iPadX/2, slavePtr->iPadY/2, slavePtr->padX/2,
	sprintf(buffer, " -padx %f -pady %f",
		slavePtr->padX/2,
		slavePtr->padY/2);
	Tcl_AppendResult(interp, buffer, (char *) NULL);
	StickyToString(slavePtr->flags,buffer);
	Tcl_AppendResult(interp, " -sticky ", buffer, (char *) NULL);
/*
	sprintf(buffer, " -weightx %.2f -weighty %.2f",
		slavePtr->weightX, slavePtr->weightY);
	Tcl_AppendResult(interp, buffer, (char *) NULL);
*/
    } else if ((c == 'p') && (strncmp(argv[1], "propagate", length) == 0)) {
	Pad_Object* master;
	GridBag *masterPtr;
	int propagate;
    
	if (argc > 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " propagate window ?boolean?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL) {
	    return TCL_ERROR;
	}
	masterPtr = GetGridBag(master);
	if (argc == 3) {
	    interp->result = (char *)((masterPtr->flags & DONT_PROPAGATE) ? "0" : "1");
	    return TCL_OK;
	}
	if (Tcl_GetBoolean(interp, argv[3], &propagate) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (propagate) {
	    masterPtr->flags &= ~DONT_PROPAGATE;
      
	    /*
	     * Re-arrange the master to allow new geometry information to
	     * propagate upwards to the master\'s master.
	     */
      
	    if (masterPtr->abortPtr != NULL) {
		*masterPtr->abortPtr = 1;
	    }
	    masterPtr->valid = 0;
	    if (!(masterPtr->flags & REQUESTED_RELAYOUT)) {
		masterPtr->flags |= REQUESTED_RELAYOUT;
		Tk_DoWhenIdle(ArrangeGrid, (ClientData) masterPtr);
	    }
	} else {
	    masterPtr->flags |= DONT_PROPAGATE;
	}
    } else if ((c == 's') && (strncmp(argv[1], "size", length) == 0)) {
	Pad_Object* master;
	GridBag *masterPtr;

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " size window\"", (char *) NULL);
	    return TCL_ERROR;
	}
	master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL)
	    return TCL_ERROR;
	masterPtr = GetGridBag(master);
	GetCachedLayoutInfo(masterPtr);

	sprintf(interp->result, "%d %d", masterPtr->layoutCache->width,
		masterPtr->layoutCache->height);
    } else if ((c == 's') && (strncmp(argv[1], "slaves", length) == 0)) {
	Pad_Object* master;
	GridBag *masterPtr, *slavePtr;
	int i, value;
	int row = -1, column = -1;
 
	if (argc < 3 || argc%2 ==0) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " slaves window ?-option value...?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}

	for (i=3; i<argc; i+=2) {
	    if (*argv[i] != '-' || (length = strlen(argv[i])) < 2) {
		Tcl_AppendResult(interp, "Invalid args: should be \"",
			argv[0], " slaves window ?-option value...?\"",
			(char *) NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[i+1], &value) != TCL_OK) {
		return TCL_ERROR;
	    }
	    if (value < 0) {
		Tcl_AppendResult(interp, argv[i],
			" is an invalid value: should NOT be < 0",
			(char *) NULL);
		return TCL_ERROR;
	    }
	    if (strncmp(argv[i], "-column", length) == 0) {
		column = value;
	    } else if (strncmp(argv[i], "-row", length) == 0) {
		row = value;
	    } else {
		Tcl_AppendResult(interp, argv[i],
			" is an invalid option: should be \"",
			"-row, -column\"",
			(char *) NULL);
		return TCL_ERROR;
	    }
	}
	master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL) {
	    return TCL_ERROR;
	}
	masterPtr = GetGridBag(master);

	for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
					     slavePtr = slavePtr->nextPtr) {
	    if (column>=0 && (slavePtr->gridColumn > column
		    || slavePtr->gridColumn+slavePtr->gridWidth-1 < column)) {
		continue;
	    }
	    if (row>=0 && (slavePtr->gridRow > row ||
		    slavePtr->gridRow+slavePtr->gridHeight-1 < row)) {
		continue;
	    }
	    Tcl_AppendElement(interp, Pad_PathName(slavePtr->tkwin));
	}

    /*
     * grid columnconfigure <master> <index> -option
     * grid columnconfigure <master> <index> -option value -option value
     * grid rowconfigure <master> <index> -option
     * grid rowconfigure <master> <index> -option value -option value
     */
   
    } else if(((c=='c') && (strncmp(argv[1], "columnconfigure", length) == 0)) ||
            ((c=='r') && (strncmp(argv[1], "rowconfigure", length) == 0))) {
	Pad_Object* master;
	GridBag *masterPtr;
	Constrain *con;
	int index, i; float size;
	double weight;

	if (argc != 5 && (argc < 5 || argc%2 == 1)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " ", argv[1], " master index ?-option value...?\"",
		    (char *)NULL);
	    return TCL_ERROR;
	}

	master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL) {
	    return TCL_ERROR;
	}
	masterPtr = GetGridBag(master);
	con = (c=='c') ? &(masterPtr->column) : &(masterPtr->row);

	if (Tcl_GetInt(interp, argv[3], &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (index < 0 || index >= masterPtr->MAXGRIDSIZE) {
	    Tcl_AppendResult(interp, argv[3], " is out of range",
		    (char *)NULL);
	    return TCL_ERROR;
	}

	/*
	 *  make sure the row/column constraint array is allocated.  This
	 *  Should be changed to avoid hard-wired limits.  We'll wimp out
	 *  for now.
	 */

	if (con->max == 0) {
	    unsigned int size;
	    con->max = masterPtr->MAXGRIDSIZE;
	    con->used = 0;

	    size = masterPtr->MAXGRIDSIZE * sizeof(con->minsize[0]);
	    con->minsize = (float *) ckalloc(size);
	    memset(con->minsize, 0, size);

	    size = masterPtr->MAXGRIDSIZE * sizeof(con->weight[0]);
	    con->weight = (double *) ckalloc(size);
	    memset(con->weight, 0, size);
	}

	for (i=4; i<argc; i+=2) {
	    if (*argv[i] != '-' || (length = strlen(argv[i])) < 2) {
		Tcl_AppendResult(interp, "Invalid arg: \"",
			argv[0], "\" expecting -minsize or -weight",
			(char *) NULL);
		return TCL_ERROR;
	    }
	    if (strncmp(argv[i], "-minsize", length) == 0) {
		if (argc == 5) {
		    float size = con->used <= index ?  0 : con->minsize[index];
		    sprintf(interp->result, "%f", size);
		} else if (Pad_GetPixels(interp, master, argv[i+1], &size)
			!= TCL_OK) {
		    return TCL_ERROR;
		} else {
		    con->minsize[index] = size;
		    if (size > 0 && index >= con->used) 
			con->used = index+1;
		    else if (size == 0 && index+1 == con->used) {
			while (index >= 0  && (con->minsize[index]==0) &&
				(con->weight[index] == 0.0)) {
			    index--;
			}
			con->used = index + 1;
		    }
		}
	    } else if (strncmp(argv[i], "-weight", length) == 0) {
		if (argc == 5) {
		    weight = con->used <= index ?  0 : con->weight[index];
		    sprintf(interp->result, "%.2f", weight);
		} else if (Tcl_GetDouble(interp, argv[i+1], &weight) != TCL_OK) {
		    return TCL_ERROR;
		} else {
		    con->weight[index] = weight;
		    if (weight > MINWEIGHT && index >= con->used) 
			con->used = index+1;
		    else if (weight == 0.0 && index+1 == con->used) {
			while (index >= 0 && (con->minsize[index]==0) &&
				(con->weight[index] == 0.0)) {
			    index--;
			}
			con->used = index + 1;
		    }
		}
	    } else {
		Tcl_AppendResult(interp, argv[i],
			" is an invalid option: should be \"",
			"-minsize, -weight\"",
			(char *) NULL);
		return TCL_ERROR;
	    }
	}

	/* if we changed a property, re-arrange the table */

	if (argc != 5) {
	    if (masterPtr->abortPtr != NULL) {
		*masterPtr->abortPtr = 1;
	    }
	    masterPtr->valid = 0;
	    if (!(masterPtr->flags & REQUESTED_RELAYOUT)) {
		masterPtr->flags |= REQUESTED_RELAYOUT;
		Tk_DoWhenIdle(ArrangeGrid, (ClientData) masterPtr);
	    }
	}
    } else if((c == 'l') && (strncmp(argv[1], "location", length) == 0)) {
	Pad_Object* master;
	GridBag *masterPtr;
	float x, y; int i, j; float w, h;
	float width, height;
	double weight;
	float diff;

	if (argc != 5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " location master x y\"", (char *)NULL);
	    return TCL_ERROR;
	}

	master = Pad_NameToWindow(interp, argv[2], tkwin->view);
	if (master == NULL) {
	    return TCL_ERROR;
	}
	masterPtr = GetGridBag(master);

	if (Pad_GetPixels(interp, master, argv[3], &x) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (Pad_GetPixels(interp, master, argv[4], &y) != TCL_OK) {
	    return TCL_ERROR;
	}

	/* make sure the grid is up to snuff */

	while ((masterPtr->flags & REQUESTED_RELAYOUT)) {
	    Tk_CancelIdleCall(ArrangeGrid, (ClientData) masterPtr);
	    ArrangeGrid((ClientData) masterPtr);
	}
	GetCachedLayoutInfo(masterPtr);
	GetMinSize(masterPtr, masterPtr->layoutCache, &width, &height);

	diff = Pad_Width(masterPtr->tkwin) - (width + masterPtr->iPadX);
	for (weight=0.0, i=0; i<masterPtr->layoutCache->width; i++) {
	    weight += masterPtr->layoutCache->weightX[i];
	}
	w = masterPtr->startx;
	if (w > x) {
	    i = -1;
	} else {
	    for (i=0; i<masterPtr->layoutCache->width; i++) {
		float dx = 0;
		if (weight > MINWEIGHT) {
		    dx = ((((double)diff) * masterPtr->layoutCache->weightX[i])
			    / weight);
		    }
		w += masterPtr->layoutCache->minWidth[i] + dx;
		if (w > x) {
		    break;
		}
	    }
	}

	diff = Pad_Height(masterPtr->tkwin) - (height + masterPtr->iPadY);
	for (weight=0.0, j = 0; j < masterPtr->layoutCache->height; j++)
	    weight += masterPtr->layoutCache->weightY[j];
	h = masterPtr->starty;
	if (h > y) {
	    j = -1;
	} else {
	    for (j=0; j<masterPtr->layoutCache->height; j++) {
		float dy = 0;
		if (weight > MINWEIGHT) {
		    dy = ((((double)diff) * masterPtr->layoutCache->weightY[j])
			    / weight);
		}
		h += masterPtr->layoutCache->minHeight[j] + dy;
		if (h > y) {
		    break;
		}
	    }
	}
	sprintf(interp->result, "%d %d", i, j);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be bbox, columnconfigure, configure, forget, info, ",
		"location, propagate, rowconfigure, size, or slaves",
		(char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

#ifdef notdef
/*
 *--------------------------------------------------------------
 *
 * GridReqProc --
 *
 *	This procedure is invoked by Tk_GeometryRequest for
 *	windows managed by the gridbag.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Arranges for tkwin, and all its managed siblings, to
 *	be re-arranged at the next idle point.
 *
 *--------------------------------------------------------------
 */

/* ARGSUSED */
static void
GridReqProc(clientData, tkwin)
    ClientData clientData;	/* GridBag's information about
				 * window that got new preferred
				 * geometry.  */
    Tk_Window tkwin;		/* Other Tk-related information
				 * about the window. */
{
    register GridBag *gridPtr = (GridBag *) clientData;

    gridPtr = gridPtr->masterPtr;
    gridPtr->valid = 0;
    if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
	gridPtr->flags |= REQUESTED_RELAYOUT;
	Tk_DoWhenIdle(ArrangeGrid, (ClientData) gridPtr);
    }
}


/*
 *--------------------------------------------------------------
 *
 * GridLostSlaveProc --
 *
 *	This procedure is invoked by Tk whenever some other geometry
 *	claims control over a slave that used to be managed by us.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Forgets all grid-related information about the slave.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static void
GridLostSlaveProc(clientData, tkwin)
    ClientData clientData;	/* GridBag structure for slave window that
				 * was stolen away. */
    Tk_Window tkwin;		/* Tk's handle for the slave window. */
{
    register GridBag *slavePtr = (GridBag *) clientData;

    if (slavePtr->masterPtr->tkwin != Tk_Parent(slavePtr->tkwin)) {
	Tk_UnmaintainGeometry(slavePtr->tkwin, slavePtr->masterPtr->tkwin);
    }
    Unlink(slavePtr);
    Tk_UnmapWindow(slavePtr->tkwin);
}
#endif

/*
 * Fill in an instance of the above structure for the current set
 * of managed children.  This requires two passes through the
 * set of children, first to figure out what cells they occupy
 * and how many rows and columns there are, and then to distribute
 * the weights and min sizes amoung the rows/columns.
 *
 * This also caches the minsizes for all the children when they are
 * first encountered.
 */

static void
GetLayoutInfo(
    GridBag *masterPtr,
    LayoutInfo *r)
{
    register GridBag *slavePtr;
    int i, k; int px, py; float pixels_diff; int nextSize;
    double weight_diff, weight;
    register int curX, curY, curWidth, curHeight, curRow, curCol;
    //float xMax[masterPtr->MAXGRIDSIZE];
    //float yMax[masterPtr->MAXGRIDSIZE];
    float *xMax = new float[masterPtr->MAXGRIDSIZE];
    float *yMax = new float[masterPtr->MAXGRIDSIZE];

    /*
     * Pass #1
     *
     * Figure out the dimensions of the layout grid.
     */

    r->width = r->height = 0;
    curRow = curCol = -1;
    memset(xMax, 0, sizeof(float) * masterPtr->MAXGRIDSIZE);
    memset(yMax, 0, sizeof(float) * masterPtr->MAXGRIDSIZE);

    for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
					 slavePtr = slavePtr->nextPtr) {

	curX = slavePtr->gridColumn;
	curY = slavePtr->gridRow;
	curWidth = slavePtr->gridWidth;
	curHeight = slavePtr->gridHeight;

	/* Adjust the grid width and height */
	for (px = curX + curWidth; r->width < px; r->width++) {
	    /* Null body. */
	}
	for (py = curY + curHeight; r->height < py; r->height++) {
	    /* Null body. */
	}

	/* Adjust the xMax and yMax arrays */
	for (i = curX; i < (curX + curWidth); i++) {
	    yMax[i] = py;
	}
	for (i = curY; i < (curY + curHeight); i++) {
	    xMax[i] = px;
	}

	/* Cache the current slave's size. */
	slavePtr->minWidth = Pad_ReqWidth(slavePtr->tkwin) + slavePtr->doubleBw;
	slavePtr->minHeight = Pad_ReqHeight(slavePtr->tkwin) + slavePtr->doubleBw;
    }

    /*
     * Apply minimum row/column dimensions
     */ 
    if (r->width < masterPtr->column.used) {
	r->width = masterPtr->column.used;
    }
    r->lastRow = r->height;
    if (r->height < masterPtr->row.used) {
	r->height = masterPtr->row.used;
    }

    /*
     * Pass #2
     */

    curRow = curCol = -1;
    memset(xMax, 0, sizeof(float) * masterPtr->MAXGRIDSIZE);
    memset(yMax, 0, sizeof(float) * masterPtr->MAXGRIDSIZE);

    for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
					 slavePtr = slavePtr->nextPtr) {
	curX = slavePtr->gridColumn;
	curY = slavePtr->gridRow;
	curWidth = slavePtr->gridWidth;
	curHeight = slavePtr->gridHeight;

	px = curX + curWidth;
	py = curY + curHeight;

	for (i = curX; i < (curX + curWidth); i++) {
	    yMax[i] = py;
	}
	for (i = curY; i < (curY + curHeight); i++) {
	    xMax[i] = px;
	}

	/* Assign the new values to the gridbag slave */
	slavePtr->tempX = curX;
	slavePtr->tempY = curY;
	slavePtr->tempWidth = curWidth;
	slavePtr->tempHeight = curHeight;
    }

    /*
     * Pass #3
     *
     * Distribute the minimun widths and weights:
     */

    /* Initialize arrays to zero */
    memset(r->minWidth, 0, r->width * sizeof(float));
    memset(r->minHeight, 0, r->height * sizeof(float));
    memset(r->weightX, 0, r->width * sizeof(double));
    memset(r->weightY, 0, r->height * sizeof(double));
    nextSize = MAXINT;

    for (i = 1; i != MAXINT; i = nextSize, nextSize = MAXINT) {
	for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
					     slavePtr = slavePtr->nextPtr) {

	    if (slavePtr->tempWidth == i) {
		px = slavePtr->tempX + slavePtr->tempWidth; /* right column */

		/* 
		 * Figure out if we should use this slave\'s weight.  If the weight
		 * is less than the total weight spanned by the width of the cell,
		 * then discard the weight.  Otherwise split it the difference
		 * according to the existing weights.
		 */

		weight_diff = slavePtr->weightX;
		for (k = slavePtr->tempX; k < px; k++)
		    weight_diff -= r->weightX[k];
		if (weight_diff > 0.0) {
		    weight = 0.0;
		    for (k = slavePtr->tempX; k < px; k++)
			weight += r->weightX[k];
		    for (k = slavePtr->tempX; weight > MINWEIGHT; k++) {
			double wt = r->weightX[k];
			double dx = (wt * weight_diff) / weight;
			r->weightX[k] += dx;
			weight_diff -= dx;
			weight -= wt;
		    }
		    /* Assign the remainder to the rightmost cell */
		    r->weightX[px-1] += weight_diff;
		}

		/*
		 * Calculate the minWidth array values.
		 * First, figure out how wide the current slave needs to be.
		 * Then, see if it will fit within the current minWidth values.
		 * If it won\'t fit, add the difference according to the weightX array.
		 */

		pixels_diff = slavePtr->minWidth + slavePtr->padX + slavePtr->iPadX;
		for (k = slavePtr->tempX; k < px; k++)
		    pixels_diff -= r->minWidth[k];
		if (pixels_diff > 0) {
		    weight = 0.0;
		    for (k = slavePtr->tempX; k < px; k++)
			weight += r->weightX[k];
		    for (k = slavePtr->tempX; weight > MINWEIGHT; k++) {
			double wt = r->weightX[k];
			int dx = (int)((wt * ((double)pixels_diff)) / weight);
			r->minWidth[k] += dx;
			pixels_diff -= dx;
			weight -= wt;
		    }
		    /* Any leftovers go into the rightmost cell */
		    r->minWidth[px-1] += pixels_diff;
		}
	    }
	    else if (slavePtr->tempWidth > i && slavePtr->tempWidth < nextSize)
		nextSize = slavePtr->tempWidth;


	    if (slavePtr->tempHeight == i) {
		py = slavePtr->tempY + slavePtr->tempHeight; /* bottom row */

		/* 
		 * Figure out if we should use this slave\'s weight.  If the weight
		 * is less than the total weight spanned by the height of the cell,
		 * then discard the weight.  Otherwise split it the difference
		 * according to the existing weights.
		 */

		weight_diff = slavePtr->weightY;
		for (k = slavePtr->tempY; k < py; k++)
		    weight_diff -= r->weightY[k];
		if (weight_diff > 0.0) {
		    weight = 0.0;
		    for (k = slavePtr->tempY; k < py; k++)
			weight += r->weightY[k];
		    for (k = slavePtr->tempY; weight > MINWEIGHT; k++) {
			double wt = r->weightY[k];
			double dy = (wt * weight_diff) / weight;
			r->weightY[k] += dy;
			weight_diff -= dy;
			weight -= wt;
		    }
		    /* Assign the remainder to the bottom cell */
		    r->weightY[py-1] += weight_diff;
		}

		/*
		 * Calculate the minHeight array values.
		 * First, figure out how tall the current slave needs to be.
		 * Then, see if it will fit within the current minHeight values.
		 * If it won\'t fit, add the difference according to the weightY array.
		 */

		pixels_diff = slavePtr->minHeight + slavePtr->padY + slavePtr->iPadY;
		for (k = slavePtr->tempY; k < py; k++)
		    pixels_diff -= r->minHeight[k];
		if (pixels_diff > 0) {
		    weight = 0.0;
		    for (k = slavePtr->tempY; k < py; k++)
			weight += r->weightY[k];
		    for (k = slavePtr->tempY; weight > MINWEIGHT; k++) {
			double wt = r->weightY[k];
			int dy = (int)((wt * ((double)pixels_diff)) / weight);
			r->minHeight[k] += dy;
			pixels_diff -= dy;
			weight -= wt;
		    }
		    /* Any leftovers go into the bottom cell */
		    r->minHeight[py-1] += pixels_diff;
		}
	    }
	    else if (slavePtr->tempHeight > i && slavePtr->tempHeight < nextSize)
		nextSize = slavePtr->tempHeight;
	}
    }

    /*
     * Apply minimum row/column dimensions
     */
    for (i=0; i<masterPtr->column.used; i++) {
	if (r->minWidth[i] < masterPtr->column.minsize[i])
	    r->minWidth[i] = masterPtr->column.minsize[i];
	if (r->weightX[i] < masterPtr->column.weight[i])
	    r->weightX[i] = masterPtr->column.weight[i];
    }
    for (i=0; i<masterPtr->row.used; i++) {
	if (r->minHeight[i] < masterPtr->row.minsize[i])
	    r->minHeight[i] = masterPtr->row.minsize[i];
	if (r->weightY[i] < masterPtr->row.weight[i])
	    r->weightY[i] = masterPtr->row.weight[i];
    }

    delete[] xMax;
    delete[] yMax;
}

/*
 * Cache the layout info after it is calculated.
 */
static void
GetCachedLayoutInfo(
    GridBag *masterPtr)
{
    if (masterPtr->valid == 0) {
        if (!masterPtr->layoutCache) {
	    // MI -- have to use C++ new to allocate the layout struct
	    // so its arrays are alloacted with the appropriate size
	    // for this grid.

	    //masterPtr->layoutCache = (LayoutInfo *)ckalloc(sizeof(LayoutInfo));
	    masterPtr->layoutCache = new LayoutInfo(masterPtr->MAXGRIDSIZE);
	} else {
	    delete  masterPtr->layoutCache;
	    masterPtr->layoutCache = new LayoutInfo(masterPtr->MAXGRIDSIZE);
	}

	GetLayoutInfo(masterPtr, masterPtr->layoutCache);
	masterPtr->valid = 1;
    }
}

/*
 * Adjusts the x, y, width, and height fields to the correct
 * values depending on the constraint geometry and pads.
 */

static void
AdjustForGravity(
    GridBag *gridPtr,
    float *x,
    float *y,
    float *width,
    float *height)
{
    float diffx=0, diffy=0;
    int sticky = gridPtr->flags&STICK_ALL;

    *x += gridPtr->padX/2;
    *width -= gridPtr->padX;
    *y += gridPtr->padY/2;
    *height -= gridPtr->padY;

    if (*width > (gridPtr->minWidth + gridPtr->iPadX)) {
	diffx = *width - (gridPtr->minWidth + gridPtr->iPadX);
	*width = gridPtr->minWidth + gridPtr->iPadX;
    }

    if (*height > (gridPtr->minHeight + gridPtr->iPadY)) {
	diffy = *height - (gridPtr->minHeight + gridPtr->iPadY);
	*height = gridPtr->minHeight + gridPtr->iPadY;
    }

    if (sticky&STICK_EAST && sticky&STICK_WEST)
	*width += diffx;
    if (sticky&STICK_NORTH && sticky&STICK_SOUTH)
	*height += diffy;
    if (!(sticky&STICK_WEST)) {
	if (sticky&STICK_EAST)
	    *x += diffx;
	else
	    *x += diffx/2;
    }
    if (!(sticky&STICK_NORTH)) {
	if (sticky&STICK_SOUTH)
	    *y += diffy;
	else
	    *y += diffy/2;
    }
}

/*
 * Figure out the minimum size (not counting the X border) of the
 * master based on the information from GetLayoutInfo()
 */

static void
GetMinSize(
    GridBag *masterPtr,
    LayoutInfo *info,
    float *minw,
    float *minh)
{
    int i; float t;
    float intBWidth;	/* Width of internal border in parent window,
			 * if any. */

    intBWidth = Pad_InternalBorderWidth(masterPtr->tkwin);

    t = 0;
    for(i = 0; i < info->width; i++)
	t += info->minWidth[i];
    *minw = t + 2*intBWidth;

    t = 0;
    for(i = 0; i < info->height; i++)
	t += info->minHeight[i];
    *minh = t + 2*intBWidth;
}

/*
 *--------------------------------------------------------------
 *
 * ArrangeGrid --
 *
 *	This procedure is invoked (using the Tk_DoWhenIdle
 *	mechanism) to re-layout a set of windows managed by
 *	the gridbag.  It is invoked at idle time so that a
 *	series of gridbag requests can be merged into a single
 *	layout operation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The slaves of masterPtr may get resized or moved.
 *
 *--------------------------------------------------------------
 */

static void
ArrangeGrid(
    ClientData clientData)	/* Structure describing parent whose slaves
				 * are to be re-layed out. */
{
    register GridBag *masterPtr = (GridBag *) clientData;
    register GridBag *slavePtr;	
    int abort;
    int i; float x, y, width, height;
    float diffw=0, diffh=0;
    double weight;
    //Pad_Object *parent, *ancestor;
    LayoutInfo info(masterPtr->MAXGRIDSIZE);
    float intBWidth;	/* Width of internal border in parent window,
			 * if any. */
    float iPadX, iPadY;

    //masterPtr->flags &= ~REQUESTED_RELAYOUT;

    /*
     * If the parent has no slaves anymore, then don't do anything
     * at all:  just leave the parent's size as-is.
     * Even if row and column constraints have been set!
     */

    if (masterPtr->slavePtr == NULL) {
	masterPtr->flags &= ~REQUESTED_RELAYOUT;
	return;
    }

    /*
     * Abort any nested call to ArrangeGrid for this window, since
     * we'll do everything necessary here, and set up so this call
     * can be aborted if necessary.  
     */

    if (masterPtr->abortPtr != NULL) {
	*masterPtr->abortPtr = 1;
    }
    masterPtr->abortPtr = &abort;
    abort = 0;
    Tk_Preserve((ClientData) masterPtr);

    masterPtr->flags |= REQUESTED_RELAYOUT; // hack for pad

    /*
     * Pass #1: scan all the slaves to figure out the total amount
     * of space needed.
     */

    GetLayoutInfo(masterPtr, &info);
    GetMinSize(masterPtr, &info, &width, &height);

#ifdef notdef
    //MI -- following code does not apply to pad.  If grid's width & height
    //are not set then they will be right when the group
    //computes its bounding box after slaves have been position.

    if (((width != Pad_ReqWidth(masterPtr->tkwin))
	 || (height != Pad_ReqHeight(masterPtr->tkwin)))
	&& !(masterPtr->flags & DONT_PROPAGATE)) {
	Pad_GeometryRequest(masterPtr->tkwin, width, height);
	masterPtr->flags |= REQUESTED_RELAYOUT;
	masterPtr->valid = 0;
	Tk_DoWhenIdle(ArrangeGrid, (ClientData) masterPtr);
	//goto done;
        masterPtr->abortPtr = NULL;
	Tk_Release((ClientData) masterPtr);
	return;
    }
#endif

    /*
     * If the parent isn't mapped then don't do anything more:  wait
     * until it gets mapped again.  Need to get at least to here to
     * reflect size needs up the window hierarchy, but there's no
     * point in actually mapping the slaves.
     */

    if (!Pad_IsMapped(masterPtr->tkwin)) {
	//goto done;
        masterPtr->abortPtr = NULL;
	Tk_Release((ClientData) masterPtr);
	return;
    }

    /*
     * If the current dimensions of the window don't match the desired
     * dimensions, then adjust the minWidth and minHeight arrays
     * according to the weights.
     */

if (masterPtr->tkwin->optionFlags & PAD_WIDTH_SET &&
    masterPtr->tkwin->optionFlags & PAD_WIDTH_SET) {

    diffw = Pad_Width(masterPtr->tkwin) - (width + masterPtr->iPadX);
    if (diffw != 0) {
	weight = 0.0;
	for (i = 0; i < info.width; i++)
	    weight += info.weightX[i];
	if (weight > MINWEIGHT) {
	    for (i = 0; i < info.width; i++) {
		float dx = (( ((double)diffw) * info.weightX[i]) / weight);
		info.minWidth[i] += dx;
		width += dx;
		if (info.minWidth[i] < 0) {
		    width -= info.minWidth[i];
		    info.minWidth[i] = 0;
		}
	    }
	}
	diffw = Pad_Width(masterPtr->tkwin) - (width + masterPtr->iPadX);
    }
    else {
	diffw = 0;
    }

    diffh = Pad_Height(masterPtr->tkwin) - (height + masterPtr->iPadY);
    if (diffh != 0) {
	weight = 0.0;
	for (i = 0; i < info.height; i++)
	    weight += info.weightY[i];
	if (weight > MINWEIGHT) {
	    for (i = 0; i < info.height; i++) {
		float dy = (( ((double)diffh) * info.weightY[i]) / weight);
		info.minHeight[i] += dy;
		height += dy;
		if (info.minHeight[i] < 0) {
		    height -= info.minHeight[i];
		    info.minHeight[i] = 0;
		}
	    }
	}
	diffh = Pad_Height(masterPtr->tkwin) - (height + masterPtr->iPadY);
    }
    else {
	diffh = 0;
    }
}

    /*
     * Now do the actual layout of the slaves using the layout information
     * that has been collected.
     */

    iPadX = masterPtr->iPadX/2;
    iPadY = masterPtr->iPadY/2;
    intBWidth = Pad_InternalBorderWidth(masterPtr->tkwin);

    for (slavePtr = masterPtr->slavePtr; slavePtr != NULL;
					 slavePtr = slavePtr->nextPtr) {

	masterPtr->startx = x = diffw/2 + intBWidth + iPadX;
	for(i = 0; i < slavePtr->tempX; i++)
	    x += info.minWidth[i];

	masterPtr->starty = y = diffh/2 + intBWidth + iPadY;
	for(i = 0; i < slavePtr->tempY; i++)
	    y += info.minHeight[i];

	width = 0;
	for(i = slavePtr->tempX; i < (slavePtr->tempX + slavePtr->tempWidth); i++)
	    width += info.minWidth[i];

	height = 0;
	for(i = slavePtr->tempY; i < (slavePtr->tempY + slavePtr->tempHeight); i++)
	    height += info.minHeight[i];

	AdjustForGravity(slavePtr, &x, &y, &width, &height);

#ifdef notdef
	/*
	 * If the window in which slavePtr is managed is not its
	 * parent in the window hierarchy, translate the coordinates
	 * to the coordinate system of the real X parent.
	 */
	parent = Pad_Parent(slavePtr->tkwin);
	for (ancestor = masterPtr->tkwin; ancestor != parent;
					  ancestor = Pad_Parent(ancestor)) {
	    x += Pad_X(ancestor) + Pad_Changes(ancestor)->border_width;
	    y += Pad_Y(ancestor) + Pad_Changes(ancestor)->border_width;
	}
# endif

	/*
	 * If the window is too small to be interesting then
	 * unmap it.  Otherwise configure it and then make sure
	 * it's mapped.
	 */

	if ((width <= 0) || (height <= 0)) {
	    Pad_UnmapWindow(slavePtr->tkwin);
	} else {
	    // NOTE: original grid code avoided unnecessary positioning
	    //   and resizing.  I can't quiet get the same efficiency in pad...
	    /*
	    if ((x != slavePtr->pad_x)
		    || (y != slavePtr->pad_y)
		    || (width != Pad_Width(slavePtr->tkwin))
		    || (height != Pad_Height(slavePtr->tkwin))) {
            }
            */

	    Pad_MoveResizeWindow(slavePtr->tkwin, x, y, width, height);
	    slavePtr->pad_x = x;
	    slavePtr->pad_y = y;
	    if (abort) {
	        //goto done;
	        masterPtr->flags &= ~REQUESTED_RELAYOUT;
		masterPtr->tkwin->Update_and_compute_anchor();
		masterPtr->abortPtr = NULL;
		Tk_Release((ClientData) masterPtr);
		return;
	    }
	    Pad_MapWindow(slavePtr->tkwin);
	}

	/*
	 * Changes to the window's structure could cause almost anything
	 * to happen, including deleting the parent or child.  If this
	 * happens, we'll be told to abort.
	 */

	if (abort) {
	   //goto done;
	   masterPtr->flags &= ~REQUESTED_RELAYOUT;
	   masterPtr->tkwin->Update();  // recompute bbox (keep anchor point!)
	   masterPtr->abortPtr = NULL;
	   masterPtr->tkwin->Generate_event(Pad_ModifyNotify, NULL, "geometry");
	   Tk_Release((ClientData) masterPtr);
           return;
	}
    }

    //done:
    masterPtr->flags &= ~REQUESTED_RELAYOUT;
    masterPtr->tkwin->Update();  // recompute bbox (keep anchor point!)
    masterPtr->abortPtr = NULL;
    Tk_Release((ClientData) masterPtr);
    // generate a modify event for the grid (atleast changing grids
    // withing grids cause master grids to relayout).
    masterPtr->tkwin->Generate_event(Pad_ModifyNotify, NULL, "geometry");
}



/*
 *--------------------------------------------------------------
 *
 * GetGridBag --
 *
 *	This internal procedure is used to locate a GridBag
 *	structure for a given window, creating one if one
 *	doesn't exist already.
 *
 * Results:
 *	The return value is a pointer to the GridBag structure
 *	corresponding to tkwin.
 *
 * Side effects:
 *	A new gridbag structure may be created.  If so, then
 *	a callback is set up to clean things up when the
 *	window is deleted.
 *
 *--------------------------------------------------------------
 */

static GridBag *
GetGridBag(
    Pad_Object* tkwin)		/* Token for window for which
				 * gridbag structure is desired. */
{
    register GridBag *gridPtr;
    Tcl_HashEntry *hPtr;
    int newflag;

    if (!initialized) {
	initialized = 1;
	Tcl_InitHashTable(&gridBagHashTable, TCL_ONE_WORD_KEYS);
    }

    /*
     * See if there's already gridbag for this window.  If not,
     * then create a new one.
     */

    hPtr = Tcl_CreateHashEntry(&gridBagHashTable, (char *) tkwin, &newflag);
    if (!newflag) {
	return (GridBag *) Tcl_GetHashValue(hPtr);
    }
    gridPtr = (GridBag *) ckalloc(sizeof(GridBag));
    //gridPtr = new GridBag;
    gridPtr->tkwin = tkwin;
    gridPtr->masterPtr = NULL;
    gridPtr->nextPtr = NULL;
    gridPtr->slavePtr = NULL;

    gridPtr->gridColumn = gridPtr->gridRow = -1;
    gridPtr->gridWidth = gridPtr->gridHeight = 1;
    gridPtr->weightX = gridPtr->weightY = 0.0;
    gridPtr->minWidth = gridPtr->minHeight = 0;

    gridPtr->padX = gridPtr->padY = 0;
    gridPtr->iPadX = gridPtr->iPadY = 0;
    gridPtr->startx = gridPtr->starty = 0;
    //gridPtr->doubleBw = 2*Tk_Changes(tkwin)->border_width;
    gridPtr->doubleBw = 2*Pad_BorderWidth(tkwin);
    gridPtr->abortPtr = NULL;
    gridPtr->flags = 0;

    gridPtr->column.max = 0;
    gridPtr->row.max = 0;
    gridPtr->column.used = 0;
    gridPtr->row.used = 0;

    gridPtr->valid = 0;
    gridPtr->layoutCache = NULL;

    gridPtr->MAXGRIDSIZE = 128;

    gridPtr->pad_x = gridPtr->pad_y = 0;

    Tcl_SetHashValue(hPtr, gridPtr);
#ifdef notdef
    Tk_CreateEventHandler(tkwin, StructureNotifyMask,
	    GridBagStructureProc, (ClientData) gridPtr);
# endif
    return gridPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Unlink --
 *
 *	Remove a gridbag from its parent's list of slaves.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The parent will be scheduled for re-arranging.
 *
 *----------------------------------------------------------------------
 */

static void
Unlink(
    register GridBag *gridPtr)		/* Window to unlink. */
{
    register GridBag *masterPtr, *gridPtr2;

    masterPtr = gridPtr->masterPtr;
    if (masterPtr == NULL) {
	return;
    }
    if (masterPtr->slavePtr == gridPtr) {
	masterPtr->slavePtr = gridPtr->nextPtr;
    }
    else {
	for (gridPtr2 = masterPtr->slavePtr; ; gridPtr2 = gridPtr2->nextPtr) {
	    if (gridPtr2 == NULL) {
		printf("Unlink couldn't find previous window\n");
		exit(1);
	    }
	    if (gridPtr2->nextPtr == gridPtr) {
		gridPtr2->nextPtr = gridPtr->nextPtr;
		break;
	    }
	}
    }
    masterPtr->valid = 0;
    if (!(masterPtr->flags & REQUESTED_RELAYOUT)) {
	masterPtr->flags |= REQUESTED_RELAYOUT;
	Tk_DoWhenIdle(ArrangeGrid, (ClientData) masterPtr);
    }
    if (masterPtr->abortPtr != NULL) {
	*masterPtr->abortPtr = 1;
    }

    gridPtr->masterPtr = NULL;
}



/*
 *----------------------------------------------------------------------
 *
 * DestroyGridBag --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a gridbag at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the gridbag is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyGridBag(
    char *memPtr)		/* Info about window that is now dead. */
{
    register GridBag *gridPtr = (GridBag *) memPtr;

    if (gridPtr->column.max) {
	ckfree((char *) gridPtr->column.minsize);
	ckfree((char *) gridPtr->column.weight);
	//delete gridPtr->column.minsize;
	//delete gridPtr->column.weight;
    }
    if (gridPtr->row.max) {
	ckfree((char *) gridPtr->row.minsize);
	ckfree((char *) gridPtr->row.weight);
	//delete gridPtr->row.minsize;
	//delete gridPtr->row.weight;
    }
    if (gridPtr->layoutCache)
	//ckfree((char *) gridPtr->layoutCache);
	delete gridPtr->layoutCache;

    // MI -- also remove from the gridBagHashTable
    Tcl_DeleteHashEntry(Tcl_FindHashEntry(&gridBagHashTable,
					  (char *) gridPtr->tkwin));

    ckfree((char *) gridPtr);
    //delete gridPtr;
}

#ifdef notdef
/*
 *----------------------------------------------------------------------
 *
 * GridBagStructureProc --
 *
 *	This procedure is invoked by the Tk event dispatcher in response
 *	to StructureNotify events.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If a window was just deleted, clean up all its gridbag-related
 *	information.  If it was just resized, re-configure its slaves, if
 *	any.
 *
 *----------------------------------------------------------------------
 */

static void
GridBagStructureProc(clientData, eventPtr)
    ClientData clientData;		/* Our information about window
					 * referred to by eventPtr. */
    XEvent *eventPtr;			/* Describes what just happened. */
{
    register GridBag *gridPtr = (GridBag *) clientData;

    if (eventPtr->type == ConfigureNotify) {
	gridPtr->valid = 0;
	if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
	    gridPtr->flags |= REQUESTED_RELAYOUT;
	    Tk_DoWhenIdle(ArrangeGrid, (ClientData) gridPtr);
	}
	if (gridPtr->doubleBw != 2*Tk_Changes(gridPtr->tkwin)->border_width) {
	    if ((gridPtr->masterPtr != NULL) &&
		    !(gridPtr->masterPtr->flags & REQUESTED_RELAYOUT)) {
		gridPtr->doubleBw = 2*Tk_Changes(gridPtr->tkwin)->border_width;
		gridPtr->masterPtr->flags |= REQUESTED_RELAYOUT;
		Tk_DoWhenIdle(ArrangeGrid, (ClientData) gridPtr->masterPtr);
	    }
	}
    }
    else if (eventPtr->type == DestroyNotify) {
	register GridBag *gridPtr2, *nextPtr;

	if (gridPtr->masterPtr != NULL) {
	    Unlink(gridPtr);
	}
	for (gridPtr2 = gridPtr->slavePtr; gridPtr2 != NULL;
					   gridPtr2 = nextPtr) {
	    //Pad_UnmapWindow(gridPtr2->tkwin);
	    gridPtr2->masterPtr = NULL;
	    nextPtr = gridPtr2->nextPtr;
	    gridPtr2->nextPtr = NULL;
	}
	Tcl_DeleteHashEntry(Tcl_FindHashEntry(&gridBagHashTable,
		(char *) gridPtr->tkwin));
	if (gridPtr->flags & REQUESTED_RELAYOUT) {
	    Tk_CancelIdleCall(ArrangeGrid, (ClientData) gridPtr);
	}
	gridPtr->tkwin = NULL;
	Tk_EventuallyFree((ClientData) gridPtr, DestroyGridBag);
    }
    else if (eventPtr->type == MapNotify) {
	gridPtr->valid = 0;
	if (!(gridPtr->flags & REQUESTED_RELAYOUT)) {
	    gridPtr->flags |= REQUESTED_RELAYOUT;
	    Tk_DoWhenIdle(ArrangeGrid, (ClientData) gridPtr);
	}
    }
    else if (eventPtr->type == UnmapNotify) {
	register GridBag *gridPtr2;

	for (gridPtr2 = gridPtr->slavePtr; gridPtr2 != NULL;
					   gridPtr2 = gridPtr2->nextPtr) {
	    Tk_UnmapWindow(gridPtr2->tkwin);
	}
    }
}
#endif


/*
 *----------------------------------------------------------------------
 *
 * ConfigureSlaves --
 *
 *	This implements the guts of the "grid configure" command.  Given
 *	a list of slaves and configuration options, it arranges for the
 *	gridbag to manage the slaves and sets the specified options.
 *
 * Results:
 *	TCL_OK is returned if all went well.  Otherwise, TCL_ERROR is
 *	returned and interp->result is set to contain an error message.
 *
 * Side effects:
 *	Slave windows get taken over by the gridbag.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureSlaves(
    Tcl_Interp *interp,	/* Interpreter for error reporting. */
    Pad_Win *win,
    Pad_Grid* tkwin,		/* Any window in application containing
				 * slaves.  Used to look up slave names. */
    int argc,			/* Numb = 0er of elements in argv. */
    char *argv[])		/* Argument strings:  contains one or more
				 * window names followed by any number
				 * of "option value" pairs.  Caller must
				 * make sure that there is at least one
				 * window name. */
{
    GridBag *masterPtr, *slavePtr, *prevPtr;
    Pad_Object* other, *slave;//Pad_Object *parent, *ancestor;
    int i, j, numWindows, c, length, inttmp, positionGiven; float tmp;
    int currentColumn=0, numColumns=1;
    int gotLayout = 0;
    int gotWidth = 0;
    int width;
    int MAXGRIDSIZE=0;

    /*
     * Find out how many windows are specified. (shouldn't use harwired symbols)
     */
#ifdef notdef
    for (numWindows = 0; numWindows < argc; numWindows++) {
	if (argv[numWindows][0] != '.'
		 && strcmp(argv[numWindows],"-")!=0
		 && strcmp(argv[numWindows],"^")!=0
		 && strcmp(argv[numWindows],"x")!=0) {
	    break;
	}
    }
#endif

    for (numWindows = 0; numWindows < argc; numWindows++) {
	if (Pad_NameToWindow(interp, argv[numWindows], win->view) == NULL
	    && strcmp(argv[numWindows],"-")!=0
	    && strcmp(argv[numWindows],"^")!=0
	    && strcmp(argv[numWindows],"x")!=0) {
	   break;
	}
    }

    slave = NULL;

    /*
     * Iterate over all of the slave windows, parsing the configuration
     * options for each slave.  It's a bit wasteful to re-parse the
     * options for each slave, but things get too messy if we try to
     * parse the arguments just once at the beginning.  For example,
     * if a slave already is managed we want to just change a few
     * existing values without resetting everything.  If there are
     * multiple windows, the -in option only gets processed for the
     * first window.
     */

    //masterPtr = NULL;
    if (tkwin)
        masterPtr = GetGridBag(tkwin);
    else
        masterPtr = NULL;
    prevPtr = NULL;
    positionGiven = 0;
    for (j = 0; j < numWindows; j++) {

	/* adjust default widget location for non-widgets */
        //if (*argv[j] != '.') {
        if (Pad_NameToWindow(interp, argv[j], win->view) == NULL) {
	    switch (*argv[j]) {
		case '^':	/* extend the widget in the previous row 
				 * Since we don't know who the master is yet,
				 * handle these in a separate pass at the end
				 */
		    /* no break */
		case REL_SKIP:	/* skip over the next column */
		    currentColumn++;
		    break;
		case REL_HORIZ:	/* increase the span, already dealt with */
		    /* not quite right */
		    if (j>0 && (*argv[j-1] == REL_SKIP || *argv[j-1] == '^')) {
			Tcl_AppendResult(interp, "Invalid grid combination:",
				" \"-\" can't follow \"", argv[j-1], "\"",NULL);
			return TCL_ERROR;
		    }
		    break;
		default:
		    printf("Invalid grid position indicator");
		    exit(1);
	    }
	    continue;
	}

	for (numColumns=1; j+numColumns < numWindows && *argv[j+numColumns] == REL_HORIZ;
		numColumns++) {
	    /* null body */
	}
	slave = Pad_NameToWindow(interp, argv[j], win->view);
	if (slave == NULL) {
	    return TCL_ERROR;
	}
	if (Pad_IsTopLevel(slave)) {
	    Tcl_AppendResult(interp, "can't manage \"", argv[j],
		    "\": it's a top-level window", (char *) NULL);
	    return TCL_ERROR;
	}
	slavePtr = GetGridBag(slave);

	/*
	 * The following statement is taken from tkPack.c:
	 *
	 * "If the slave isn't currently managed, reset all of its
	 * configuration information to default values (there could
	 * be old values left from a previous packer)."
	 *
	 * I disagree with this statement.  If a slave is disabled (using
	 * "forget") and then re-enabled, I submit that 90% of the time the
	 * programmer will want it to retain its old configuration information.
	 * If the programmer doesn't want this behavior, then she can reset the
	 * defaults for herself, but she will never have to worry about keeping
	 * track of the old state. 
	 */

	for (i = numWindows; i < argc; i+=2) {
	    if ((i+2) > argc) {
		Tcl_AppendResult(interp, "extra option \"", argv[i],
			"\" (option with no value?)", (char *) NULL);
		return TCL_ERROR;
	    }
	    length = strlen(argv[i]);
	    if (length < 2) {
		goto badOption;
	    }
	    c = argv[i][1];
	    if ((c == 'i') && (strcmp(argv[i], "-in") == 0)) {
		if (j == 0) {
		    other = Pad_NameToWindow(interp, argv[i+1], win->view);
		    if (other == NULL) {
			return TCL_ERROR;
		    }
		    if (other == slave) {
			sprintf(interp->result,"Window can't be managed in itself");
			return TCL_ERROR;
		    }
		    masterPtr = GetGridBag(other);
		    prevPtr = masterPtr->slavePtr;
		    if (prevPtr != NULL) {
			while (prevPtr->nextPtr != NULL) {
			    prevPtr = prevPtr->nextPtr;
			}
		    }
		    positionGiven = 1;
		}
/*
	    } else if ((c == 'i') && (strcmp(argv[i], "-ipadx") == 0)) {
		if ((Pad_GetPixels(interp, slave, argv[i+1], &tmp) != TCL_OK)
			|| (tmp < 0)) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad ipadx value \"", argv[i+1],
			    "\": must be positive screen distance",
			    (char *) NULL);
		    return TCL_ERROR;
		}
		slavePtr->iPadX = tmp*2;
	    } else if ((c == 'i') && (strcmp(argv[i], "-ipady") == 0)) {
		if ((Pad_GetPixels(interp, slave, argv[i+1], &tmp) != TCL_OK)
			|| (tmp< 0)) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad ipady value \"", argv[i+1],
			    "\": must be positive screen distance",
			    (char *) NULL);
		    return TCL_ERROR;
		}
		slavePtr->iPadY = tmp*2;
*/
	    } else if ((c == 'p') && (strcmp(argv[i], "-padx") == 0)) {
		if ((Pad_GetPixels(interp, slave, argv[i+1], &tmp) != TCL_OK)
			|| (tmp< 0)) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad padx value \"", argv[i+1],
			    "\": must be positive screen distance",
			    (char *) NULL);
		    return TCL_ERROR;
		}
		slavePtr->padX = tmp*2;
	    } else if ((c == 'p') && (strcmp(argv[i], "-pady") == 0)) {
		if ((Pad_GetPixels(interp, slave, argv[i+1], &tmp) != TCL_OK)
			|| (tmp< 0)) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad pady value \"", argv[i+1],
			    "\": must be positive screen distance",
			    (char *) NULL);
		    return TCL_ERROR;
		}
		slavePtr->padY = tmp*2;
	    } else if ((c == 'c') && (strcmp(argv[i], "-column") == 0)) {
		if (Tcl_GetInt(interp, argv[i+1], &inttmp) != TCL_OK || inttmp<0) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad column value \"", argv[i+1],
			    "\": must be a non-negative integer", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->gridColumn = inttmp;
		
		// MI -- keep track of max grid size
		if (slavePtr->gridColumn >= MAXGRIDSIZE) {
		  MAXGRIDSIZE = slavePtr->gridColumn;
		}

	    } else if ((c == 'r') && (strcmp(argv[i], "-row") == 0)) {
		if (Tcl_GetInt(interp, argv[i+1], &inttmp) != TCL_OK || inttmp<0) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad grid value \"", argv[i+1],
			    "\": must be a non-negative integer", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->gridRow = inttmp;
		// MI -- keep track of max grid size
		if (slavePtr->gridRow >= MAXGRIDSIZE) {
		  MAXGRIDSIZE = slavePtr->gridRow;
		}

	    } else if ((c == 'c') && (strcmp(argv[i], "-columnspan") == 0)) {
		if (Tcl_GetInt(interp, argv[i+1], &inttmp) != TCL_OK || inttmp <= 0) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad columnspan value \"", argv[i+1],
			    "\": must be a positive integer", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->gridWidth = inttmp;
		gotWidth++;
	    } else if ((c == 'r') && (strcmp(argv[i], "-rowspan") == 0)) {
		if (Tcl_GetInt(interp, argv[i+1], &inttmp) != TCL_OK) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad rowspan value \"", argv[i+1],
			    "\": must be a positive integer", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->gridHeight = inttmp;
/*
	    } else if ((c == 'w') &&
		    (!strcmp(argv[i], "-weightx") || !strcmp(argv[i], "-wx"))) {
		if (Tcl_GetDouble(interp, argv[i+1], &tmp_dbl) != TCL_OK) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad weight value \"", argv[i+1],
			    "\": must be a double", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->weightX = tmp_dbl;
	    }
	    else if ((c == 'w') &&
		    (!strcmp(argv[i], "-weighty") || !strcmp(argv[i], "-wy"))) {
		if (Tcl_GetDouble(interp, argv[i+1], &tmp_dbl) != TCL_OK) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "bad weight value \"", argv[i+1],
			    "\": must be a double", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->weightY = tmp_dbl;
*/
	    } else if ((c == 's') && strcmp(argv[i], "-sticky") == 0) {
		int sticky = StringToSticky(argv[i+1]);
		if (sticky == -1) {
		    Tcl_AppendResult(interp, "bad stickyness value \"", argv[i+1],
			    "\": must be a string containing n, e, s, and/or w", (char *)NULL);
		    return TCL_ERROR;
		}
		slavePtr->flags = sticky | (slavePtr->flags & ~STICK_ALL);
	    } else {
		badOption:
		Tcl_AppendResult(interp, "unknown or ambiguous option \"",
			argv[i], "\": must be -in, -sticky, ",
			"-row, -column, -rowspan, -columnspan, ",
			//"-ipadx, -ipady, -padx or -pady.",
			"-padx or -pady.",
			(char *) NULL);
		return TCL_ERROR;
	    }
	}

	/*
	 * If no position in a gridbag list was specified and the slave
	 * is already managed, then leave it in its current location in
	 * its current gridbag list.
	 */

	if (!positionGiven && (slavePtr->masterPtr != NULL)) {
	    masterPtr = slavePtr->masterPtr;
	    goto scheduleLayout;
	}

	if (masterPtr == NULL && slavePtr->masterPtr == NULL) {
	    Tcl_AppendResult(interp, "no grid for object \"",
		   argv[j],"\": must be in a grid", (char*)NULL);
	    return TCL_ERROR;
	}

	/*
	 * If the slave is going to be put back after itself then
	 * skip the whole operation, since it won't work anyway.
	 */

	if (prevPtr == slavePtr) {
	    masterPtr = slavePtr->masterPtr;
	    goto scheduleLayout;
	}
    
	/*
	 * If the "-in" option has not been specified, arrange for the
	 * slave to go at the end of the order for its parent.
	 */
    
	if (!positionGiven) {
	    //masterPtr = GetGridBag(Pad_Parent(slave));
	    if (masterPtr != NULL)
	        prevPtr = masterPtr->slavePtr;
	    else
	        prevPtr = slavePtr->masterPtr->slavePtr;
	    if (prevPtr != NULL) {
		while (prevPtr->nextPtr != NULL) {
		    prevPtr = prevPtr->nextPtr;
		}
	    }
	}

#ifdef notdef    
	/*
	 * Make sure that the slave's parent is either the master or
	 * an ancestor of the master.
	 */
	parent = Pad_Parent(slave);
	for (ancestor = masterPtr->tkwin; ; ancestor = Pad_Parent(ancestor)) {
	    if (ancestor == parent) {
		break;
	    }
	    if (Pad_IsTopLevel(ancestor)) {
		Tcl_AppendResult(interp, "can't put ", argv[j],
			" inside ", Pad_PathName(masterPtr->tkwin),
			(char *) NULL);
		return TCL_ERROR;
	    }
	}
#endif
	// add the slave to grid group and hack to get pad grid
        // to position itself around the first slave.
	if (slavePtr->masterPtr == NULL) {
	    Pad_Grid *padGrid = (Pad_Grid*)masterPtr->tkwin;
	    if (!(masterPtr->flags & REQUESTED_RELAYOUT)) {
	        masterPtr->flags |= REQUESTED_RELAYOUT;
		Tk_DoWhenIdle(ArrangeGrid, (ClientData) masterPtr);
	    }
	    if (padGrid->members.Is_empty()) {
	        masterPtr->flags &= ~REQUESTED_RELAYOUT;
	        padGrid->Add(slave, FALSE);
		masterPtr->flags |= REQUESTED_RELAYOUT;
	    } else {
	      // add but don't move slave yet (will be done in ArrangeGrid)
	      padGrid->Add(slave, TRUE);
	    }
            slave->flags |= PAD_MASK_GRID_SLAVE;
	}

	/*
	 * Unlink the slave if it's currently managed, then position it
	 * after prevPtr.
	 */
	if (slavePtr->masterPtr != NULL) {
	    Unlink(slavePtr);
	}
	slavePtr->masterPtr = masterPtr;
	if (prevPtr == NULL) {
	    slavePtr->nextPtr = masterPtr->slavePtr;
	    masterPtr->slavePtr = slavePtr;
	} else {
	    slavePtr->nextPtr = prevPtr->nextPtr;
	    prevPtr->nextPtr = slavePtr;
	}
	//Tk_ManageGeometry(slave, &gridMgrType, (ClientData) slavePtr);
	prevPtr = slavePtr;

	/* assign default row and column */

	if (slavePtr->gridColumn == -1) {
	    slavePtr->gridColumn = currentColumn;
	    // MI -- keep track of max grid size
	    if (slavePtr->gridColumn >= MAXGRIDSIZE) {
	       MAXGRIDSIZE = slavePtr->gridColumn;
	    }
	}
	slavePtr->gridWidth += numColumns - 1;
	if (slavePtr->gridRow == -1) {

            // MI -- keep track of max grid size
	    if (masterPtr->layoutCache == NULL) {
	       slavePtr->gridRow = 0;
            } else {
	       slavePtr->gridRow = masterPtr->layoutCache->lastRow;
            }

            if (masterPtr->layoutCache && 
	        (masterPtr->layoutCache->lastRow >= masterPtr->MAXGRIDSIZE-1)) {
	       masterPtr->MAXGRIDSIZE *= 2;
	       masterPtr->valid = 0;
	       GetCachedLayoutInfo(masterPtr);
            }

	    if (!gotLayout++) GetCachedLayoutInfo(masterPtr);
	    //slavePtr->gridRow = masterPtr->layoutCache->lastRow;
	    slavePtr->gridRow = masterPtr->layoutCache->lastRow-1;

	    if (slavePtr->gridRow >= MAXGRIDSIZE) {
	       MAXGRIDSIZE = slavePtr->gridRow;
	    }
	}

	/*
	 * Arrange for the parent to be re-arranged at the first
	 * idle moment.
	 */

	scheduleLayout:
	if (masterPtr->abortPtr != NULL) {
	    *masterPtr->abortPtr = 1;
	}

	masterPtr->valid = 0;

	// MI -- make sure grid max size is set before relayout calculations
	if (masterPtr->MAXGRIDSIZE <= MAXGRIDSIZE) {
	  masterPtr->MAXGRIDSIZE = MAXGRIDSIZE*2;
	  GetCachedLayoutInfo(masterPtr);
	}

	if (!(masterPtr->flags & REQUESTED_RELAYOUT)) {
	    masterPtr->flags |= REQUESTED_RELAYOUT;
	    Tk_DoWhenIdle(ArrangeGrid, (ClientData) masterPtr);
	}
	currentColumn += slavePtr->gridWidth;
	numColumns = 1;
    }

    /* now look for all the "^"'s */

    for (j = 0; j < numWindows; j++) {
	struct GridBag *otherPtr;
    	char *lastWindow;	/* use this window to base current row/col on */
	int match;		/* found a match for the ^ */

    	//if (*argv[j] == '.') {
	if (Pad_NameToWindow(interp, argv[j], win->view) != NULL) {
	    lastWindow = argv[j];
	}
	if (*argv[j] != '^') {
	    continue;
	}
	for (width=1; width+j < numWindows && *argv[j+width] == '^'; width++) {
	    /* Null Body */
	}
	other = Pad_NameToWindow(interp, lastWindow, win->view);
	otherPtr = GetGridBag(other);
	if (!gotLayout++) GetCachedLayoutInfo(masterPtr);

	for (match=0, slavePtr = masterPtr->slavePtr; slavePtr != NULL;
					 slavePtr = slavePtr->nextPtr) {

	    if (slavePtr->gridWidth == width
		    && slavePtr->gridColumn == otherPtr->gridColumn + otherPtr->gridWidth
		    && slavePtr->gridRow + slavePtr->gridHeight == otherPtr->gridRow) {
		slavePtr->gridHeight++;
		match++;
	    }
	    lastWindow = Pad_PathName(slavePtr->tkwin);
	}
	if (!match) {
	    Tcl_AppendResult(interp, "can't find slave to extend with \"^\"",
		    " after ",lastWindow,
		    (char *) NULL);
	    return TCL_ERROR;
	}
	j += width - 1;
    }
    return TCL_OK;
}

/* convert "Sticky" bits into a string */

static void
StickyToString(
    int flags,		/* the sticky flags */
    char *result)	/* where to put the result */
{
    int count = 0;
    if (flags&STICK_NORTH) result[count++] = 'n';
    if (flags&STICK_EAST) result[count++] = 'e';
    if (flags&STICK_SOUTH) result[count++] = 's';
    if (flags&STICK_WEST) result[count++] = 'w';
    if (count) {
	result[count] = '\0';
    } else {
	sprintf(result,"{}");
    }
}

/* convert sticky string to flags */

static int
StringToSticky(
    char *string)
{
    int sticky = 0;
    char c;

    while ((c = *string++) != '\0') {
	switch (c) {
	    case 'n': case 'N': sticky |= STICK_NORTH; break;
	    case 'e': case 'E': sticky |= STICK_EAST;  break;
	    case 's': case 'S': sticky |= STICK_SOUTH; break;
	    case 'w': case 'W': sticky |= STICK_WEST;  break;
	    case ' ': case ',': case '\t': case '\r': case '\n': break;
	    default: return -1;
	}
    }
    return sticky;
}		
