/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

/*
   Wrapper for grid geometry management in Pad.
     
   Other than having grid as a pad command the Tcl syntax for grids 
   is the same.  For example:

    # create a grid
    % set tbl = [.pad create grid]

    # place objects in grid
    % .pad grid $obj1 $obj2 -in $tbl
    % .pad grid $obj3 -in $tbl -row 1 -column 0
    % .pad grid configure $obj1 $obj2 $obj3 -padx 10 -pady 10

   A derived class of Pad_Group (Pad_Grid) is defined for master grid
   objects in pad and it interfaces to the original C routines and
   structures.  This provides a convinient mechanism for interface to
   Pad and for event handling.  Relevant Tk_ routines that were used
   in the original code are replaced by equivalent Pad_ routines.

   Configuration of slave attributes is done as in the original code 
   (they are stored in underlying C grid structures).

   A PAD_MASK_GRID_SLAVE bit is added to Pad_Object flags and the
   Pad_Object::Generate_event() methods check  this and call
   SlaveEventProc() when necessary.

   Differences and limitations:
     - if the grid is not positioned then it places itself around 
     its first slave.  Once all the slaves have been positioned
     the grid bbox will be computed to enclose them all.
     - added the "arrange" command for forcing grid arrangement.
     - internal pading for slaves is not supported but external padding 
     (padx & pady) works as expected.
     - slaves that are removed from grids are not unmapped.
     - propogation is not done.
     - grid slaves must be (and will be) in the grid group (they can not
     be in some other group).
*/


#ifndef GRID_H
#define GRID_H 1

#include "defs.h"
#include "object.h"
#include "group.h"
#include "line.h"



class Pad_Grid : public Pad_Group
{
public:
    Pad_Grid(Pad *pad);
    virtual ~Pad_Grid();

    void     Compute_bounding_box();
    void     Update_global_bounding_box(void);
    void     Set_bounding_box_to_enclose_list(Pad_List &list);
    int      Create_obj_args(int argc, char **argv);
    void     Add_to_layer(char *name);
    void     Write_obj(char *padname, Pad_String &str);
    Pad_Bool Write_obj_pre(char *padname, Pad_String &str);
    void     Write_obj_args(char *padname, Pad_String &str);
    void     Write_obj_post(char *padname, Pad_String &str);
    Pad_Bool Write_obj(char *padname, Pad_StreamData &ds);
    Pad_Bool Read_obj(Pad_StreamData &sd);
    Pad_Bool Delete(Pad_Object *obj); // Remove object from group (don't put on surface)
    Pad_Bool Remove(Pad_Object *obj, Pad_Bool transform);  // Move object from group onto surface

    static void SlaveEventProc(Tcl_Interp*, Pad_Object*, int);
    void*       Get_Grid();
    void        Set_Grid(void*);
    PAD_TYPE("grid");

protected:
    void  Write_slaves_config(char *padname, Pad_String &str);
    void *grid;
};

#endif




