/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

//
// JM - Wrapper on Tcl Tables
//

#include "defs.h"
#include "hashtab.h"

void
Pad_HashTable::Init()
{
    Tcl_InitHashTable(&table, type);
    len = 0;
}

Pad_HashTable::Pad_HashTable(int t)
{
    // Setup type according to t, then call Init

    if (t == PAD_STRING_TABLE) {
	type = TCL_STRING_KEYS;
    } else if (t == PAD_VOID_TABLE) {
	type = TCL_ONE_WORD_KEYS;
    } else {
	type = t;
    }

    Init();
}

Pad_HashTable::Pad_HashTable()
{
    // Setup the type, then call Init
    type = TCL_ONE_WORD_KEYS;
    Init();
}

Pad_HashTable::~Pad_HashTable()
{
    Tcl_DeleteHashTable(&table);
}

void
Pad_HashTable::Make_empty()
{
    // Not the most efficient way to empty a table!
    Tcl_DeleteHashTable(&table);
    Init();
}

int
Pad_HashTable::Remove(void *key)
{
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_FindHashEntry(&table, (char*)key);
    if (entryPtr) {
	Tcl_DeleteHashEntry(entryPtr);
	len--;
    }
    return (entryPtr ? 1 : 0);
}

int
Pad_HashTable::Set(void *key, void *val)
{
    Tcl_HashEntry *entryPtr;
    int isnew;
    entryPtr = Tcl_CreateHashEntry(&table, (char*)key, &isnew);
    Tcl_SetHashValue(entryPtr, (ClientData)val);
    if (isnew) len++;
    return isnew;
}

int
Pad_HashTable::Length() const
{
    return len;
}

int
Pad_HashTable::Is_empty() const
{
    return (len == 0);
}


void *
Pad_HashTable::Get(void *key)
{
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_FindHashEntry(&table, (char*)key);

    return entryPtr ? (void*)Tcl_GetHashValue(entryPtr) : NULL;
}

// Used by Pad_HashTableIterator
void *
Pad_HashTable::Init(void *sPtr, voidPtr &key)
{
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_FirstHashEntry(&table, (Tcl_HashSearch*)sPtr);
    key =  entryPtr ? ((void*)Tcl_GetHashKey(&table, entryPtr)) : NULL;
    return entryPtr ? (void*)Tcl_GetHashValue(entryPtr) : NULL;
}

void *
Pad_HashTable::Next(void *sPtr, voidPtr &key)
{
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_NextHashEntry((Tcl_HashSearch*)sPtr);
    key =  entryPtr ? ((void*)Tcl_GetHashKey(&table, entryPtr)) : NULL;
    return entryPtr ? (void*)Tcl_GetHashValue(entryPtr) : NULL;
}

Pad_HashTableIterator::Pad_HashTableIterator()
{
    key = NULL;
    table = NULL;
}

void *
Pad_HashTableIterator::Init(Pad_HashTable &t)
{
    table = &t;
    return table->Init((void*)&ptr, key);
}

void *
Pad_HashTableIterator::Next()
{
    return table ? table->Next((void*)&ptr, key) : NULL;
}
