/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef HTML_H
#define HTML_H 1


#include "defs.h"
#include "line.h"
#include "list.h"
#include "object.h"
#include "font.h"
#include "pad-tcl.h"
#include "pad-string.h"
#include "renderer.h"
#include "view.h"
#include "group.h"
#include <stdio.h>


#define HTML_BORDER_SET       (1<<0)  // Set if -border       has been configured
#define HTML_BORDERWIDTH_SET  (1<<1)  // Set if -borderwidth  has been configured
#define HTML_DONESCRIPT_SET   (1<<2)  // Set if -donescript   has been configured
#define HTML_ERRORSCRIPT_SET  (1<<3)  // Set if -errorscript  has been configured
#define HTML_FILL_SET         (1<<4)  // Set if -fill         has been configured
#define HTML_FONT_SET         (1<<5)  // Set if -font         has been configured
#define HTML_UPDATESCRIPT_SET (1<<6)  // Set if -updatescript has been configured
#define HTML_URL_SET          (1<<7)  // Set if -url          has been configured

#define HTML_LEVELS 2             // Maximum refinement level for HTML

#define HTML_DIPLAYITEM_UNDEF   0 // HTML_Display::itemType types
#define HTML_DIPLAYITEM_TEXT    1
#define HTML_DIPLAYITEM_ANCHOR  2
#define HTML_DIPLAYITEM_POLYGON 3
#define HTML_DIPLAYITEM_IMAGE   4

				// HTML_Anchor states
#define HTML_ANCHOR_UNVISITED 0
#define HTML_ANCHOR_ACTIVE    1
#define HTML_ANCHOR_VISITED   2
#define HTML_ANCHOR_NOTLOADED 3

#define H1_SIZE 2.0             // The scaling factors for header tags
#define H2_SIZE 1.75
#define H3_SIZE 1.5
#define H4_SIZE 1.25
#define H5_SIZE 1.15
#define H6_SIZE 1.10

#define HTML_XOFFSET 3          // Offset of items within page
#define HTML_YOFFSET 3

#define HTML_DEFAULT_WIDTH 75

enum HtmlEntryTagType {
  TAG_HTML,
  TAG_BODY,
  TAG_ADDRESS,
  TAG_H1,
  TAG_H2,
  TAG_H3,
  TAG_H4,
  TAG_H5,
  TAG_H6,
  TAG_P,
  TAG_A,
  TAG_IMG,
  TAG_UL,
  TAG_LI,
  TAG_DL,
  TAG_DT,
  TAG_DE,
  TAG_DD,
  TAG_HR,
  TAG_BR,
  TAG_TITLE,
  TAG_HEAD,
  TAG_EM,
  TAG_I,
  TAG_B,
  TAG_CENTER,
  TAG_MENU,
  TAG_CITE,
  TAG_FONT,
  TAG_PRE,
  TAG_KBD,
  TAG_STRONG,
  TAG_BASE,
  TAG_CODE,
  TAG_OL,
  TAG_COMMENT1,
  TAG_COMMENT2,
  TAG_META,
  TAG_TEXTAREA,
  TAG_FORM,
  TAG_INPUT,
  TAG_TABLE,
  TAG_TR,
  TAG_TD,
  TAG_TH,
  TAG_BLOCKQUOTE,
  TAG_BLINK,
  TAG_LINK,
  TAG_CAPTION,
  TAG_HEADER,
  TAG_UNKNOWN};

enum KeyType {
   KEY_HREF,
   KEY_SRC,
   KEY_ALT,
   KEY_NAME,
   KEY_ALIGN,
   KEY_BORDER,
   KEY_WIDTH,
   KEY_HEIGHT,
   KEY_ISMAP,
   KEY_BGCOLOR,
   KEY_TEXT,
   KEY_LINK,
   KEY_ALINK,
   KEY_VLINK,
   KEY_PAD_SCALE_BEGIN,         // The <META PAD_TAG>'s, for MSML
   KEY_PAD_SCALE_END,         
   KEY_PAD_TCL,                 // The pad_tcl meta tag
   KEY_UNKNOWN};

enum HtmlEntryType {
  HTML_EOF = 0,
  HTML_ENTRY_UNDEF,
  HTML_ENTRY_TEXT,
  HTML_ENTRY_START,
  HTML_ENTRY_END,
  HTML_ENTRY_IMAGE,
  HTML_ENTRY_ANCHOR};

class HTML_Key 
{
public:		
    ~HTML_Key();
    HTML_Key();
    HTML_Key(HTML_Key *);
    void Parse_script(HTML *html, int &index);
    void Set_type();

    int        key;      
    Pad_String value;    
};

class HTML_Anchor;
class HTML_Display;
class Pad_Callback;


///////////////////////////////////////////////////////////////////////
//
//   Entry-related classes
//
///////////////////////////////////////////////////////////////////////

class HTML_Entry 
{
  friend class HTML;            // FIXME why do I need this?
public:
  virtual Pad_Bool    Contains_text(char *);
  virtual Pad_Bool    Get_text(Pad_String &string);
  virtual void        Layout(HTML *) = 0;
  virtual Pad_Bool    Remove_obj (Pad_Object *obj);
  virtual void        Remove_display_item(void);
          Pad_String *ParseStartTag (int tag);
          Pad_String *ParseEndTag (int tag);

  virtual ~HTML_Entry();
  HTML_Entry();
};

class HTML_EntryPadObject : public HTML_Entry 
{
public:
            void     Layout(HTML *html);
    virtual Pad_Bool Remove_obj (Pad_Object *obj);
    
    virtual ~HTML_EntryPadObject();
    HTML_EntryPadObject();
    HTML_EntryPadObject(Pad_Object *obj);  
    
private:
    Pad_Object *_obj;
    float       _origSize;
    float       _scaledSize;
};

class HTML_EntryText : public HTML_Entry 
{
public:
  virtual Pad_Bool Contains_text(char *);
  virtual Pad_Bool Get_text(Pad_String &string);
          void     Layout(HTML *html);
          void     Set_text (Pad_String *);

  virtual ~HTML_EntryText();
  HTML_EntryText();

private:
  Pad_String _text;
};

class HTML_EntryTag : public HTML_Entry 
{
public:
          int         Finish_key(HTML *html, int &index, HTML_Key *key);
  virtual Pad_List *  Get_keys (void);
          Pad_String *Get_pathname(char *name, HTML *html);
  virtual void        Layout(HTML *html);
  virtual void        Parse (HTML *html);
          HTML_EntryTag *Parse_meta (HTML *);
          void        Set_tag(HTML_EntryTag *);
  virtual HtmlEntryTagType Set_tag_type(void);

  virtual ~HTML_EntryTag();
  HTML_EntryTag();
  HTML_EntryTag(HTML_EntryTag *);
  HTML_EntryTag(HTML *html, int &index);

protected:
    Pad_String _tag;		// Really only necessary for debugging
    Pad_List * _keyList;		// List of HTML_Key
};

class HTML_EntryTagStart : public HTML_EntryTag 
{
public:
          void          Layout(HTML *html);
          void          Parse (HTML *html);
          HtmlEntryType Type (void);

  virtual ~HTML_EntryTagStart();
  HTML_EntryTagStart();
  HTML_EntryTagStart(HTML_EntryTag *);

protected:
  HtmlEntryTagType _startTag;
};

class HTML_EntryTagEnd : public HTML_EntryTag 
{
public:
          int           Get_tag(void);
          void          Layout(HTML *html);
          void          Parse(HTML *html);
          HtmlEntryType Type(void);

  virtual ~HTML_EntryTagEnd();
  HTML_EntryTagEnd();

private:
    HtmlEntryTagType _endTag;
};

class HTML_TclItem
{
  public:
    Pad_Object *obj;
    float       origSize;
    float       scaledSize;
};

class HTML_EntryTagTcl : public HTML_EntryTag 
{
public:
          void     Add (Pad_Object *obj);
          void     Layout(HTML *html);
  virtual Pad_Bool Remove_obj (Pad_Object *obj);

  virtual ~HTML_EntryTagTcl();
  HTML_EntryTagTcl();
  HTML_EntryTagTcl(HTML_EntryTag *);

private:
    Pad_String *_script;
    Pad_List *  _objects;		// List of HTML_TclItem's
};

class HTML_EntryTagAnchor : public HTML_EntryTag 
{
public:
          void Layout(HTML *html);
          void Parse (HTML *html);

    virtual ~HTML_EntryTagAnchor();
    HTML_EntryTagAnchor();
    HTML_EntryTagAnchor(HTML_EntryTag *);

private:
    Pad_String *_href;
    Pad_String *_name;
};

class HTML_EntryTagImage : public HTML_EntryTag 
{
public:
            void Layout(HTML *html);
            void Parse(HTML *html);
    virtual void Remove_display_item(void);
            void Set_image(Pad_ImageData *image);

    virtual ~HTML_EntryTagImage();
    HTML_EntryTagImage(HTML_EntryTag *, HTML *html);

    HTML *         html;
    Pad_String *   imgUrl;
    unsigned char *data;
    int            index;
    int            space;
    int            width, height;
    Pad_UrlItem *  urlItem;
    Pad_String     altImageUrl;

private:
    HTML_Display * _displayitem;
    Pad_ImageData *_image;
    Pad_ImageData *_altImage;
    Pad_Bool       _found;
    Pad_Bool       _ismap;
    int            _borderWidth;
};

///////////////////////////////////////////////////////////////////////
//
//   Display-related classes
//
///////////////////////////////////////////////////////////////////////

class HTML_Display 
{
public:
    virtual HTML_Entry *Get_entry(void);
            Pad_BBox *Get_bbox(void);
            float Get_x (void) const;
            float Get_y (void) const;
            float Get_size (void) const;
    virtual void  Render (HTML *html);
    virtual void  Set_entry(HTML_Entry *entry);
    virtual void  Set_image(Pad_ImageData *image);
            void  Set_x(float x);
            void  Set_y(float y);
    virtual void  Update_line_position(float dy);

    virtual ~HTML_Display ();
    HTML_Display ();
protected:
    float       _scale;
    float       _x;
    float       _y;
    HTML_Entry *_entry;
    Pad_BBox    _bbox;      // Used just for clipping
};

class HTML_DisplayPadObject : public HTML_Display 
{
public:
    void Render (HTML *html);
    void Update_line_position(float dy);

    virtual ~HTML_DisplayPadObject ();
    HTML_DisplayPadObject ();
    HTML_DisplayPadObject (HTML *html, Pad_Object *o);
private:
    Pad_Object *_obj;
};

class HTML_DisplayText : public HTML_Display 
{
public:
            short Calc_line_width ();
            void  Render (HTML *html);
    virtual void  Update_line_position(float dy);

    virtual ~HTML_DisplayText ();
    HTML_DisplayText (HTML *html, Pad_String *text, Pad_Bool preformatted);
private:
    Pad_List *_stringList;         // List of HTML_DisplayString
    short     _lineWidth;
};

class HTML_DisplayString : public Pad_String 
{
public:
    float Getx(void) const;
    float Gety(void) const;
    void  Sety (float);
    void  Setx (float);

    virtual ~HTML_DisplayString();
    HTML_DisplayString();
    HTML_DisplayString(Pad_String *);
private:
    float _x;
    float _y;
};

class HTML_DisplayImage : public HTML_Display 
{
public:
            void Render (HTML *html);
    virtual void Set_image(Pad_ImageData *image);

    virtual ~HTML_DisplayImage();
    HTML_DisplayImage();
    HTML_DisplayImage(HTML_Entry *entry, HTML *html, Pad_ImageData *image, Pad_ImageData *altImage);
private:
    float          _height;
    float          _width;
    Pad_ImageData *_image;
    Pad_ImageData *_altImage;
};

class HTML_DisplayAnchor : public HTML_Display 
{
public:
            void Remove_anchor(HTML_Anchor *anchor);
            void Render (HTML *html);
    virtual void Set_image(Pad_ImageData *image);
    virtual void Update_line_position(float dy);

    virtual ~HTML_DisplayAnchor();
    HTML_DisplayAnchor();
    HTML_DisplayAnchor(HTML_Entry *entry, HTML *html, Pad_ImageData *image, Pad_ImageData *altImage,
		      int width, int height, Pad_Bool ismap, int borderWidth);
    HTML_DisplayAnchor(HTML *html, Pad_String *text);
    HTML_DisplayAnchor(HTML_Anchor *anc);
private:
    float     _height;
    float     _width;
    Pad_List *_anchors;		// List of HTML_Anchor's
				// This is a list of components of this piece
				// of the anchor. It consists of a single item
				// for image anchors, and one item per line of
				// text.  It does not contain the different
				// pieces that may comprise an anchor.
};

class HTML_DisplayRectangle : public HTML_Display 
{
public:
    void Render (HTML *html);
    void Set_height (float x);
    void Set_width (float y);
    void Set_color(char *penColor);
    
    virtual ~HTML_DisplayRectangle ();
    HTML_DisplayRectangle (HTML_Entry *entry, HTML *html, float w, float h);
    HTML_DisplayRectangle (HTML_Entry *entry, HTML *html, float w, float h, float xpos, float ypos);
    
private:
    HTML *     _html;		// Necessary to free color
    Pad_ColorRef _penColor;
    float      _height;
    float      _width;
};

typedef struct {
    char *name;			// Name of item.
    int type;			// Type of item.
} HTML_TagOrKey;

///////////////////////////////////////////////////////////////////////
//
//   HTML class
//
///////////////////////////////////////////////////////////////////////

class HTML : public Pad_Group
{
  
 public:
    int           inlineImages;	    // Number of inline images requested, but not yet received.
    unsigned char htmlFlags;        // OR'd combination of flags
    Pad_Bool      refineNeeded;     // True if HTML needs refinement
    float         width;	    // Width of HTML page in characters
    float         lineWidth;	    // Width of line border
    float         xmin, ymin, xmax, ymax;
    Pad_List      anchorList;	    // List of type HTML_Anchor
    Pad_List      displayList;	    // List of type HTML_Display
    Pad_List      entryList;	    // List of type HTML_Entry
    Pad_Callback *  updateScript;
    Pad_Callback *  doneScript;
    Pad_Callback *  errorScript;
    Pad_String    url;		    // URL name
    unsigned char *source;	    // Original HTML source
    int           sourceIndex;
    int           sourceSpace;
    Pad_String    docType;	    // Stated Document type
    Pad_String    docLastModified;  // Stated last modification date
    int           docLength;	    // Stated length
    Pad_UrlItem * urlItem;	    // Pointer to urlItem while being retrieved
    Pad_BorderRef border;	    // Border
    Pad_ColorRef  fillColor;        // Fill color
    Pad_ColorRef  textColor;	    // Color of text 
    Pad_ColorRef  linkColor;	    // Color of unvisited link
    Pad_ColorRef  alinkColor;	    // Color of active link
    Pad_ColorRef  vlinkColor;	    // Color of visited link
    Pad_ColorRef  ulinkColor;	    // Color of link not yet loaded
    Pad_ColorRef  smallColor;	    // Color of "greeked" page when too small
  
    virtual Pad_Bool     Add(Pad_Object *obj, Pad_Bool transform);     // Move object into group
            void         Add_anchor(HTML_Anchor *anchor);
            void         Add_display_item(HTML_Display *di);
            void         Add_text(Pad_View &pad_view, char *buf);
            void         Add_token(Pad_View &pad_view);
    virtual Pad_Bool     Check_render_clipped(void);
    virtual void         Compute_bounding_box(void);
    virtual Pad_Bool     Contains_text(char *);
    virtual Pad_Bool     Continue_refinement(void);
            void         Create_display_list(void);
    virtual Pad_Bool     Delete(Pad_Object *obj); // redefine Pad_Group::Delete
            void         Delete_display_list_item(HTML_EntryTag *entry);
            void         Draw_image(Pad_ImageData *image, float x1, float y1, float x2, float y2);
            void         End_tag(Pad_View &pad_view);
            Pad_List *   Flow_string(Pad_String *text, Pad_Bool preformatted);
    virtual void         Get_fillname(Pad_String &fillname);
    virtual void         Get_bordername(Pad_String &bordername);
    virtual float        Get_borderwidth(void);
            char *       Get_done_script(void);
            char *       Get_error_script(void);
            Pad_FontRef *Get_font_ref(void);
            Pad_List *   Get_htmlanchors(void);
            char *       Get_last_modified(void);
            int          Get_length(void);
    virtual Pad_Bool     Get_name(Pad_String &string);
    virtual Pad_Bool     Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified);
            unsigned char *Get_source(void);
    virtual Pad_Bool     Get_text(Pad_String &string);
            char *       Get_type(void);
            char *       Get_update_script(void);
            char *       Get_url(void);
    virtual float        Get_width(void);
            void         Layout (void);
            HTML_Entry * Next_token(int &index, Pad_Bool &eof);
            HTML_Entry * Parse_tag(int &index, Pad_Bool &eof);
            HTML_Entry * Parse_text(int &index, Pad_Bool &eof);
            void         Parse_token(Pad_View &pad_view);
    virtual Pad_Bool     Pick_group(Pad_Event *event);  // True if group object should be picked
							// Move object from group onto surface
    virtual Pad_Bool     Remove(Pad_Object *obj, Pad_Bool transform);
    virtual Pad_Bool     Render(void);
    virtual void         Set_border(char *colorname);
    virtual void         Set_border_default(void);
    virtual void         Set_borderwidth(float borderwidth);
    virtual void         Set_borderwidth_default(void);
            void         Set_done_script(char *script);
            void         Set_done_script_default(void);
            void         Set_error_script(char *script);
            void         Set_error_script_default(void);
    virtual Pad_Bool     Set_fill(char *colorname);
    virtual Pad_Bool     Set_fill(int red, int green, int blue);
    virtual void         Set_fill_default(void);
    virtual void         Set_font(char *fullname);
    virtual void         Set_font_default(void);
    virtual Pad_Bool     Set_option(Pad_Option *option, char *value); 
    virtual Pad_Bool     Set_option_default(Pad_Option *option);
            void         Set_update_script(char *script);
            void         Set_update_script_default(void);
    virtual Pad_Bool     Set_url(char *urlname);
            void         Set_url_default(void);
    virtual Pad_Bool     Set_width(float width);
    virtual void         Set_width_default(void);
            void         Skip_white_space(int &index);
            void         Start_tag(Pad_View &pad_view);
            float        Start_new_line(int request);
            void         Update_line_spacing (void) const;
    
    HTML(Pad *pad);
    virtual ~HTML();
    PAD_TYPE("html");
};

///////////////////////////////////////////////////////////////////////
//
//   Anchor-related class
//
///////////////////////////////////////////////////////////////////////

#define ANCHOR_IMAGEDATA_SET    (1<<0)  // Set if -image  has been configured
#define ANCHOR_FONT_SET         (1<<1)  // Set if -font  has been configured

class HTML_Anchor : public Pad_Object
{
  public:
    unsigned char anchorFlags;
    Pad_String    href;
    Pad_String    name;
    HTML *        html;
    HTML_Anchor * next;
    HTML_DisplayAnchor *displayAnchor;
    Pad_ColorRef   penColor;
    unsigned char state;

            HTML *   Get_html(void);
    virtual void Get_font(Pad_String &fullname);
    virtual Pad_ImageData *Get_imagedata(void);
    virtual Pad_Bool Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified);
            int      Get_state(void);
            char *   Get_url(void);
    virtual Pad_Bool Get_map(void);
            char *   Get_anchor_name(void);
    virtual Pad_Bool Render(void);
    virtual void     Set_font(char *fullname);
    virtual void     Set_font_default(void);
    virtual void     Set_imagedata(Pad_ImageData *image);
    virtual void     Set_imagedata_default(void);
    virtual Pad_Bool Set_option(Pad_Option *option, char *value); 
    virtual Pad_Bool Set_option_default(Pad_Option *option);
            Pad_Bool Set_state(int flag);
            void     Set_state_default(void);
    virtual Pad_Bool Set_url(char *urlname);
    virtual Pad_Bool Set_url_default(void);
    virtual void     Update_line_position(float dy);
    virtual Pad_Bool Write_obj_pre(char *padname, Pad_String &str);
    
    virtual ~HTML_Anchor();
    HTML_Anchor(Pad *pad, HTML *html);
    PAD_TYPE("htmlanchor");
  protected:
    Pad_FontRef    _font;

};

class HTML_AnchorImage : public HTML_Anchor
{
  public:
    virtual void     Compute_bounding_box(void);
    virtual Pad_Bool Continue_refinement(void);         // True if object wants more refinement
            int      Get_height(void);
            int      Get_imageborder_width(void);
            Pad_ImageData *Get_imagedata(void);
            float    Get_width(void);
            Pad_Bool Get_map(void);
            Pad_Bool Render(void);
            void     Set_imageborder_width(int border);
            void     Set_imagedata(Pad_ImageData *image);
            void     Set_imagedata_default(void);
            void     Set_map(Pad_Bool ismap);

    virtual ~HTML_AnchorImage();
    HTML_AnchorImage(Pad *pad, HTML *html, Pad_ImageData *image, Pad_ImageData *altImage, int width, int height);
  private:
    Pad_ImageData *_image;
    Pad_ImageData *_altImage;
    int            _height;
    int            _width;
    Pad_Bool       _ismap;
    int            _borderWidth;
};

class HTML_AnchorText : public HTML_Anchor
{
  public:
    int level;			// Level anchor was rendered at

    virtual void     Compute_bounding_box(void);
    virtual Pad_Bool Contains_text(char *);
    virtual Pad_Bool Continue_refinement(void);         // True if object wants more refinement
            Pad_Bool Render(void);
    virtual Pad_Bool Get_text(Pad_String &string);

    virtual ~HTML_AnchorText();
    HTML_AnchorText(Pad *pad, HTML *html, HTML_DisplayString *str);
  private:
    Pad_String _text;
    Pad_Bool _refineNeeded;
};

#endif
