/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "pad-tcl.h"
#include "tkkpl.h"	// but ignore the "tk" in the name
#include "renderer.h"
#include "win.h"
#include "global.h"

#include <string.h>

////////////////////////////////////////////////////////
//
// KPL Object in a Pad (i.e. procedural render)
//
////////////////////////////////////////////////////////

#define KPL_DEFAULT_BBOX            NULL
#define KPL_DEFAULT_RENDERSCRIPT    NULL

Pad_Kpl::~Pad_Kpl() 
{
    Generate_delete();

    /// N.B. Currently we never delete KPL pcode
    delete kpl;
}

Pad_Bool refinement_needed;
Pad_Object *kpl_render_object = NULL;

Pad_Kpl::Pad_Kpl(Pad *pad) :
Pad_Object(pad)
{
    _type = PAD_KPL;
    kplFlags = PAD_NO_MASK;
    kpl = new Kpl;

    interp = 0;
    renderPcode = 0;
    bbPcode = 0;
}

//
// Gets called with string args when object is created.  This routine must 
// process any arguments up to, but not including the options.  (argc, argv) 
// point to parameters on the creation line just after 'create <type>'.
// This returns the number of args processed.
//
int
Pad_Kpl::Create_obj_args(int, char **)
{
    interp = pad->view->win->interp;

    return(0);
}

//
// Set the object's <option> to the value specified by <sval>.
// Note that each option has a specific setter and getter method
// if the correct argument type is available.
//
Pad_Bool
Pad_Kpl::Set_option(Pad_Option *option, char *sval)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BBOX:
	rc = Set_bboxscript(sval);
	break;
      case PAD_RENDERSCRIPT:
	if (strlen(sval) == 0) {
	    Set_string_renderscript(NULL);
	} else {
	    Set_string_renderscript(sval);
	}
	break;
      default:
	rc = Pad_Object::Set_option(option, sval);
	break;
    }

    return(rc);
}

//
// Set the object's <option> to the default value.
//
Pad_Bool
Pad_Kpl::Set_option_default(Pad_Option *option)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BBOX:
	Set_bboxscript_default();
	break;
      case PAD_RENDERSCRIPT:
	Set_string_renderscript_default();
	break;
      default:
	rc = Pad_Object::Set_option_default(option);
	break;
    }

    return(rc);
}

//
// Return the value of the object's <option> as a string
// in <result>.
//
Pad_Bool
Pad_Kpl::Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BBOX:
	modified = (kplFlags & KPL_BBOX_SET) ? TRUE : FALSE;
	result = Get_bboxscript();
	break;
      case PAD_RENDERSCRIPT:
	modified = (optionFlags & PAD_RENDERSCRIPT_SET) ? TRUE : FALSE;
	result = "";
	break;
      default:
	rc = Pad_Object::Get_option(option, result, modified);
	break;
    }

    return(rc);
}

//
// Setters and getters for renderscript
//
Pad_Bool
Pad_Kpl::Set_string_renderscript(char *new_renderscript)
{
    Pad_Bool rc = TRUE;

    if (kpl_script(new_renderscript) == TCL_ERROR) {
	Pad_errorString = "Error: Trouble compiling KPL render script";
	rc = FALSE;
    }

    return(rc);
}

void 
Pad_Kpl::Set_string_renderscript_default(void)
{
    Set_string_renderscript(KPL_DEFAULT_RENDERSCRIPT);
}

Pad_Callback *
Pad_Kpl::Get_renderscript(void)
{
    return NULL;
}

//
// Setters and getters for bbox
//
Pad_Bool
Pad_Kpl::Set_bboxscript(char *new_bbScript)
{
    Pad_Bool rc = TRUE;

    kplFlags |= KPL_BBOX_SET;
    if (bb_script(new_bbScript) == TCL_ERROR) {
	Pad_errorString = "Error: Trouble compiling KPL bbox script";
	rc = FALSE;
    } else {
	Update();
    }

    return(rc);
}

void
Pad_Kpl::Set_bboxscript_default(void)
{
    Set_bboxscript(KPL_DEFAULT_BBOX);
    kplFlags &= ~KPL_BBOX_SET;
}

char *
Pad_Kpl::Get_bboxscript(void)
{
    return("");		// Can't retrieve KPL bbox script
}

int Pad_Kpl::kpl_script (char *script)
{
    Pcode *pcode = kpl->compile(script);
    if (!pcode) return TCL_ERROR;
    renderPcode = pcode;
    return TCL_OK;
}

int Pad_Kpl::bb_script (char *script)
{
    Pcode *pcode = kpl->compile(script);
    if (!pcode) return TCL_ERROR;
    bbPcode = pcode;
    return TCL_OK;
}

void Pad_Kpl::Compute_bounding_box(void) 
{
    if (bbPcode) {
        kpl->execute(bbPcode);
        int len1=0,len2=0;
        float a[16],b[16];
        kpl->pop(&len2,b);
        kpl->pop(&len1,a);
        if (len1 >= 2 && len2 >= 2) {
	    // Note the studied indifference to the order of
	    // arguments (and even which corners you use):
	    float ax = a[0];
	    float ay = a[1];
	    float bx = b[0];
	    float by = b[1];
	    if (!(optionFlags & PAD_WIDTH_SET)) {
		Set_bbox_xmin(MIN(ax,bx));
                Set_bbox_xmax(MAX(ax,bx));
	    }
	    if (!(optionFlags & PAD_HEIGHT_SET)) {
		Set_bbox_ymin(MIN(ay,by));
                Set_bbox_ymax(MAX(ay,by));
	    }
	}
    }
    Pad_Object::Compute_bounding_box();
}

//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
Pad_Kpl::Render(void)
{
    Pad_Bool rc = TRUE;

    if (renderPcode) 
      {
	  kpl_render_object = this;
	  refinement_needed = FALSE;
	  kpl->execute(renderPcode);
	  kpl_render_object = NULL;

	  rc = !(Pad_prc->win->view->pad->level && Pad_renderer->Is_interrupted(Pad_prc->win));
      }

    return(rc);
}

//
// Render the Kpl object by calling its render script
// (if there is one).
//
Pad_Bool Pad_Kpl::Continue_refinement(void)
{
    return refinement_needed;
}

//
// Kpl objects always get clipped
//
Pad_Bool Pad_Kpl::Check_render_clipped()
{
    return(TRUE);
}
