/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "pad-sound.h"
#include "global.h"
#include "pad-string.h"
#include "hashtab.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef HAS_SOUND

int Pad_SynthSound::SAMPLE_NUMBER = 4096;
int Pad_SynthSound::RATE = 8000;
int Pad_SynthSound::CHANNELS = 1;
Pad_SoundFormat Pad_SynthSound::STYLE = LinearSigned16LSB;
// Pad_SoundFormat Pad_SynthSound::STYLE = LinearSigned8;
int Pad_SynthSound::SAMPLE_SIZE = ((Pad_SynthSound::STYLE == LinearSigned8) ? 1 : 2);
int Pad_SynthSound::BUFFER_SIZE = Pad_SynthSound::SAMPLE_NUMBER * Pad_SynthSound::CHANNELS * Pad_SynthSound::SAMPLE_SIZE;
int Pad_SynthSound::LOW_WATER = Pad_SynthSound::BUFFER_SIZE / 2;

int Pad_Sound::CONFIG_UNSET = -10;

//
// These are used by the Pad++ sound system for configuring sounds
//
Pad_SoundOption Pad_optionSoundDirection;
Pad_SoundOption Pad_optionSoundGain;
Pad_SoundOption Pad_optionSoundMaterial;
Pad_SoundOption Pad_optionSoundMaxWork;
Pad_SoundOption Pad_optionSoundMinWork;
Pad_SoundOption Pad_optionSoundRatio;
Pad_SoundOption Pad_optionSoundSize;
Pad_SoundOption Pad_optionSoundSpeed;
Pad_SoundOption Pad_optionSoundHardness;
Pad_SoundOption Pad_optionSoundImpact;
Pad_SoundOption Pad_optionSoundLoop;
Pad_SoundOption Pad_optionSoundLength;
Pad_SoundOption Pad_optionSoundRoughness;
Pad_SoundOption Pad_optionSoundScrape;
Pad_SoundOption Pad_optionSoundVolume;

////////////////////////////////////////////////////////////////
//
// Pad_SoundConfig utility class
//
//
////////////////////////////////////////////////////////////////
Pad_SoundConfig::Pad_SoundConfig()
{
        // all of the following static variables have been configured
        // before this constructor in Pad_NCDSound::Init(), called
        // from Pad_MasterInit()
    direction = &Pad_optionSoundDirection;
    gain      = &Pad_optionSoundGain;
    material  = &Pad_optionSoundMaterial;
    maxWork   = &Pad_optionSoundMaxWork;
    minWork   = &Pad_optionSoundMinWork;
    ratio     = &Pad_optionSoundRatio;
    size      = &Pad_optionSoundSize;
    speed     = &Pad_optionSoundSpeed;
    hardness  = &Pad_optionSoundHardness;
    impact    = &Pad_optionSoundImpact;
    length    = &Pad_optionSoundLength;
    loop      = &Pad_optionSoundLoop;
    roughness = &Pad_optionSoundRoughness;
    scrape    = &Pad_optionSoundScrape;
    volume    = &Pad_optionSoundVolume;
}

Pad_SoundConfig::~Pad_SoundConfig(){}

Pad_SoundConfig *Pad_SoundConfig::_instance = NULL;

Pad_SoundConfig *
Pad_SoundConfig::Get_instance(void) {
    if (! _instance) {
        _instance = new Pad_SoundConfig ();
    }
    return _instance;
}

void
Pad_SoundConfig::Reset()
{
    direction->data = Pad_Sound::CONFIG_UNSET;
    gain->data      = Pad_Sound::CONFIG_UNSET;
    material->data  = Pad_Sound::CONFIG_UNSET;
    maxWork->data   = Pad_Sound::CONFIG_UNSET;
    minWork->data   = Pad_Sound::CONFIG_UNSET;
    ratio->data     = Pad_Sound::CONFIG_UNSET;
    size->data      = Pad_Sound::CONFIG_UNSET;
    speed->data     = Pad_Sound::CONFIG_UNSET;
    hardness->data  = Pad_Sound::CONFIG_UNSET;
    impact->data    = Pad_Sound::CONFIG_UNSET;
    length->data    = Pad_Sound::CONFIG_UNSET;
    loop->data      = Pad_Sound::CONFIG_UNSET;
    roughness->data = Pad_Sound::CONFIG_UNSET;
    scrape->data    = Pad_Sound::CONFIG_UNSET;
    volume->data    = Pad_Sound::CONFIG_UNSET;
}

////////////////////////////////////////////////////////////////
//
// Pad_SoundOption utility class
//
//
////////////////////////////////////////////////////////////////
Pad_SoundOption::Pad_SoundOption()
{
    type = SoundConfigNone;
}

Pad_SoundOption::~Pad_SoundOption(){}

void
Pad_SoundOption::Bounds(int low, int high)
{
    if (data < low) {
        data = low;
    } else if (data > high) {
        data = high;
    }
}

////////////////////////////////////////////////////////////////
//
// Pad_PlayState Class
//
//
////////////////////////////////////////////////////////////////
Pad_PlayState::Pad_PlayState() {}
Pad_PlayState::~Pad_PlayState()
{
  if (_data) {
    delete [] _data;
  }
}

Pad_PlayState::Pad_PlayState(Pad_Sound *source) {
    _Init_state();
    _source = source;
}

void
Pad_PlayState::_Init_state(void)
{
    _stopped = FALSE;
    _loop    = FALSE;
    _source  = NULL;
    _data    = NULL;
}

void
Pad_PlayState::_Init(void)
{
    Pad_SoundConfig *config;
    
    Pad_optionSoundDirection.type = SoundConfigDirection;
    Pad_optionSoundGain.type      = SoundConfigGain;
    Pad_optionSoundMaterial.type  = SoundConfigMaterial;
    Pad_optionSoundMaxWork.type   = SoundConfigMaxWork;
    Pad_optionSoundMinWork.type   = SoundConfigMinWork;
    Pad_optionSoundRatio.type     = SoundConfigRatio;
    Pad_optionSoundSize.type      = SoundConfigSize;
    Pad_optionSoundSpeed.type     = SoundConfigSpeed;
    Pad_optionSoundHardness.type  = SoundConfigHardness;
    Pad_optionSoundImpact.type    = SoundConfigImpact;
    Pad_optionSoundLength.type    = SoundConfigLength;
    Pad_optionSoundLoop.type      = SoundConfigLoop;
    Pad_optionSoundRoughness.type = SoundConfigRoughness;
    Pad_optionSoundScrape.type    = SoundConfigScrape;
    Pad_optionSoundVolume.type    = SoundConfigVolume;

        // the constructor for Pad_SoundConfig assumes that the previous
        // static variables have been configure
    config = Pad_SoundConfig::Get_instance();
}

void
Pad_PlayState::Update_state(void){}

int
Pad_PlayState::Refill(int)
{
    Pad_errorString = "Pad_PlayState::Refill(int) should never be called";
    return -1;
}

void
Pad_PlayState::Set_stop(Pad_Bool stop) {
    _stopped = stop;
    return;
}

Pad_Bool
Pad_PlayState::Stopped(void) {
    return _stopped;
}

void
Pad_PlayState::Set_loop(Pad_Bool loop) {
    _loop = loop;
    return;
}

Pad_Bool
Pad_PlayState::Loop(void) {
    return _loop;
}


////////////////////////////////////////////////////////////////
//
// Pad_SynthState subclass of Pad_PlayState
//
//
////////////////////////////////////////////////////////////////
Pad_SynthState::Pad_SynthState()
{
    _Init_state();
}

Pad_SynthState::Pad_SynthState(Pad_Sound *source) 
        : Pad_PlayState (source)
{
    _Init_state();
}

Pad_SynthState::~Pad_SynthState() {}

void
Pad_SynthState::_Init_state()
{
    _amp = 1.0;
    _inSample = 0;
    _data = new char[Pad_SynthSound::BUFFER_SIZE];
}

void
Pad_SynthState::Update_state(void)
{
    _amp = Pad_Sound::Db_to_amp(_source->_gain);
    cerr << "Gain: " << _source->_gain << " Amp: " << _amp << endl;
}

int
Pad_SynthState::Refill(int)
{
    Pad_errorString = "Pad_SynthState::Refill(int) should never be called";
    return -1;
}


////////////////////////////////////////////////////////////////
//
// Pad_ObjectState subclass of Pad_PlayState
//
//
////////////////////////////////////////////////////////////////
Pad_ObjectState::Pad_ObjectState()
{
    _Init_state();
}

Pad_ObjectState::Pad_ObjectState(Pad_Sound *source)
        : Pad_SynthState(source)
{
    _Init_state();
}

Pad_ObjectState::~Pad_ObjectState(){}

void
Pad_ObjectState::_Init_state()
{
    _duration = 0;		// no limit
    _threshold = .0001;	// min. avg. of the signal
    _threshold *= _threshold;	// min. avg. of the power of the signal
    _scale = 1.0;		// normalization constant
    _phase = 0;
    for (int i = 0; i < MAXFILTERS; i++) {
        _xnm1 [i] = 0;
        _xnm2 [i] = 0;
        _ynm1 [i] = 0;
        _ynm2 [i] = 0;
    }
	/* no need to initialize out since it is overwritten */
    for (float* p = _power; p < _power + WINSIZE; ) {
        *p++ = 0.0;
    }
    _sum = 0;
}

void
Pad_ObjectState::Update_state()
{
    Pad_SynthState::Update_state();
	// apply a +30dB scaling factor -- this was determined 'experimentally'
    _scale = _amp * 1000.0;
}

float
Pad_ObjectState::_Gen_input (int)
{
    return 0.0;
}

int
Pad_ObjectState::Refill(int numBytes)
{
    float insamp, osamp [MAXFILTERS], outsamp;
    int j;
    int samples;
    Pad_ObjectSound *object = (Pad_ObjectSound *) _source;
    int nsamp;
    char *_write;

    nsamp = numBytes / Pad_SynthSound::SAMPLE_SIZE;
    _write = _data;
    for (samples = 0; samples < nsamp; samples++) {
            /* get input sample */
            /* devide by number of filters to ensure constant energy through the bank */
        insamp = _Gen_input (_phase);
        if (object->_numFilters > 1)
            insamp /= object->_numFilters;
		
            /* pass it through the filter bank */
        outsamp = 0.0;
        for (j = 0; j < object->_numFilters; j++) {
            osamp[j] = ((object->_filters[j][0] * insamp) +
                        (object->_filters[j][1] * _ynm1[j]) -
                        (object->_filters[j][2] * _ynm2[j]));
            _xnm2[j] = _xnm1[j];
            _xnm1[j] = insamp;
            _ynm2[j] = _ynm1[j];
            _ynm1[j] = osamp[j];
            outsamp +=  osamp[j];
            osamp[j] = 0.0;
        }
		
            // scale is the product of the amp resulting from the gain
            // and a normalization constant defined by the subclass
        outsamp *= _scale;
		
//        _data[samples] = (char)((outsamp * 122.5) + 0.5);
        OUTPUT_SAMPLE(outsamp)
                
        _phase++;

        if (_phase % (Pad_SynthSound::BUFFER_SIZE/8) == 0) {
            cerr << "outsamp = "  << outsamp
                 << ", sample = " << _data[samples]
                 << ", gain = "   << object->_gain
                 << ", scale = "  << _scale
                 << ", amp = "    << _amp
                 << endl;
        }

        outsamp *= outsamp;
        if (outsamp > _sum) {
            _sum = outsamp;
        }
        if (_phase % WINSIZE == 0) {
            _sum = 0;
        }
    }
    return TRUE;
}

////////////////////////////////////////////////////////////////
//
// Pad_ScrapeState subclass of Pad_ObjectState
//
//
////////////////////////////////////////////////////////////////
Pad_ScrapeState::Pad_ScrapeState(){}
Pad_ScrapeState::~Pad_ScrapeState(){}

Pad_ScrapeState::Pad_ScrapeState(Pad_Sound *source)
        : Pad_ObjectState(source)
{
    _Init_state();
}

void
Pad_ScrapeState::_Init_state()
{
    _roughness = 10;
    _speed     = 10;
    
    _threshold = -1;            // don't stop when silent
}

void
Pad_ScrapeState::Update_state()
{
    Pad_ObjectState::Update_state();
    
    _inSample = 0.0;
}

float
Pad_ScrapeState::_Gen_input (int phase)
{
    if (_speed == 0) {
        return 0.0;	// nothing when steady
    }
	
    if (phase % (101 - _speed) == 0) {
        _inSample = _roughness * drand48 ();
    }
    
    return _inSample;
}

////////////////////////////////////////////////////////////////
//
// Pad_ImpacteState subclass of Pad_ObjectState
//
//
////////////////////////////////////////////////////////////////
Pad_ImpactState::Pad_ImpactState()
{
    _Init_state();
}

Pad_ImpactState::Pad_ImpactState(Pad_Sound *source)
        : Pad_ObjectState(source){}

Pad_ImpactState::~Pad_ImpactState(){}

void
Pad_ImpactState::_Init_state()
{
    _hardness = 50;
    
	// apply a +30dB scaling factor -- this was determined 'experimentally'
    _scale = _amp * 1000.0;
    	// we limit impact sound duration to 10 msecs 
    _duration = 10;
}

void
Pad_ImpactState::Update_state()
{
    Pad_ObjectState::Update_state();
    
	// 0 < hardness < 100 => 10.1 msecs > length > .1 msecs 
    _impulseLength = (1 + (100 - _hardness)) * Pad_SynthSound::RATE / 10000;
}

float
Pad_ImpactState::_Gen_input (int phase)
{
    if (phase > _impulseLength) {
        return 0.0;
    }
    
    return SINE ((float(phase * TABLELENGTH)) / float(_impulseLength));
}

////////////////////////////////////////////////////////////////
//
// Pad_MachineState subclass of Pad_PlayState
//
//
////////////////////////////////////////////////////////////////
Pad_MachineState::Pad_MachineState()
{
    _Init_state();
}

Pad_MachineState::~Pad_MachineState(){}

Pad_MachineState::Pad_MachineState(Pad_Sound *source) 
        : Pad_SynthState(source)
{
    Pad_MachineSound *src = (Pad_MachineSound*) source;
    
    _Init_state();
    
    _maxWork   = src->_maxWork;
    _speed     = src->_speed;
    _modAmpInc = src->_tableInc;
}

void
Pad_MachineState::_Init_state(void)
{
    _modAmpPhase  = 0;
    _modFreqPhase = 0;
    _carFreqPhase = 0;
	
	// safe default value. Initialized in UpdateState
    _modAmpInc = 0;
}

void
Pad_MachineState::Update_state(void)
{
    Pad_MachineSound *machine =(Pad_MachineSound*) _source;

    Pad_SynthState::Update_state();

    	// .1 < modAmpFreq 
    float modAmpFreq = _speed / 10.0;
    _modAmpInc = modAmpFreq * machine->_tableInc;

}

int
Pad_MachineState::Refill(int numBytes)
{
    float modAmp;		// modulator gain
    float modOut;		// modulator output
    float carFreqInc;           // carrier frequency increment
    float carOut;		// output sample
    int   samples;
    char *_write;
    int nsamp;

    nsamp = numBytes / Pad_SynthSound::SAMPLE_SIZE;
    Pad_MachineSound *src = (Pad_MachineSound *)_source;

    _write = _data;
    for (samples = 0; samples < nsamp; samples++) {
            // the modulator amplitude varies between minWork and maxWork
            // at a frequency that depends on the speed of the machine
        _modAmpPhase += _modAmpInc;
        modAmp = (src->_minWork + _maxWork) / 2.0 +
            (_maxWork - src->_minWork) * SINE (_modAmpPhase) / 2.0;
		
            // the modulator signal is a sine wave whose frequency depends
            // on the ratio of the machine. The modulator's amplitude is
            // modulated by the amplitude modulator computed above.
        _modFreqPhase += src->_modFreqInc;
        modOut = modAmp * src->_modFreq * SINE (_modFreqPhase);
		
            // the output signal is the result of the frequency
            // modulation of the carrier (whose frequency depends on
            // the size of the machine) by the modulator.
        carFreqInc = (src->_carFreq + modOut) * src->_tableInc;
        _carFreqPhase += carFreqInc;
        carOut = _amp * SINE (_carFreqPhase);
		
//        _data[samples] = (char)((carOut * 122.5) + 0.5);
        OUTPUT_SAMPLE(carOut)
        if (samples % (Pad_SynthSound::BUFFER_SIZE/8) == 0) {
            cerr << "carOut = " << carOut
                 << ", sample = " << _data[samples] << endl;
        }
    }
    return TRUE; 
}

////////////////////////////////////////////////////////////////
//
// Pad_LiftState subclass of Pad_PlayState
//
//
////////////////////////////////////////////////////////////////
Pad_LiftState::Pad_LiftState(){}

Pad_LiftState::Pad_LiftState(Pad_Sound *source)
        : Pad_SynthState(source)
{
    _Init_state();
}

Pad_LiftState::~Pad_LiftState(){}

void
Pad_LiftState::_Init_state(void)
{
    for( int i=0; i<=9; i++) {
        _freq[i]=0;
    }
    
    _changed = 0;
    _incFreq = 0;
}

void
Pad_LiftState::Update_state(void)
{
    Pad_SynthState::Update_state();
}

int
Pad_LiftState::Refill(int numBytes)
{
    Pad_LiftSound* src =  (Pad_LiftSound*) _source;
	
    float carOut;               // output sample
    int samples;
    int nsamp;
    char *_write;

    nsamp = numBytes / Pad_SynthSound::SAMPLE_SIZE;
    _write = _data;
    for (samples = 0; samples < nsamp; samples++) {
        src->_inc++;

        if( src->_inc > int(src->_speed)) {
            src->_inc = 0;
            
            if( int(src->_direction)) {
                src->_floor += int(src->_direction);
                if( src->_floor > src->_max1stPartial-1) {	        
                        // up
                    src->_floor = 0;
                    for(int i = 9; i >= 1; i--)
                        _freq[i] = _freq[i-1];
                    _freq[0] = 0;
                    _changed++;
                    _changed %= src->_nbPartial;
//                    fprintf(stderr,"_changed:%i\n",_changed);
                } else {
	        		//down
                    if( src->_floor < 0 ) {
                        src->_floor = src->_max1stPartial-1;
                        for(int i = 0; i <= 8; i++)
                            _freq[i] = _freq[i+1];
                        _freq[9] = 0;
                        _changed = (_changed -1)%src->_nbPartial;
                        if (_changed<0) _changed = src->_nbPartial;
                    }
                }
                src->_baseFreq = int(src->_size) + src->_floor;	          
            }
        }
	      

        carOut = 0;
        float ampPartial;
        float fr = src->_baseFreq; 
	      	        
        for(int i=0;i<=src->_nbPartial-1;i++) {
                // transformation
            ampPartial = fr;

                // enveloppe
	        
            if( ampPartial < src->_xPointMax ) {
                ampPartial = src->_aUp1*ampPartial+src->_bUp1;
            } else {
                ampPartial = src->_aUp2*ampPartial+src->_bUp2;
            }
	        
            _freq[i]+=fr*src->_tableInc;
            _freq[i]= (int)_freq[i];
            carOut += ampPartial*SINE(_freq[i]);
            fr *= src->_rapPartial;
        }
//        _data[samples] = (char)((carOut * 122.5) + 0.5);
        OUTPUT_SAMPLE(carOut)
        if (samples % (Pad_SynthSound::BUFFER_SIZE/8) == 0) {
            cerr << "carOut = " << carOut
                 << ", sample = " << _data[samples] << endl;
        }

        carOut *= 0.2;
    }
    return TRUE;
}

////////////////////////////////////////////////////////////////
//
// Pad_FileState subclass of Pad_PlayState
//
//
////////////////////////////////////////////////////////////////
Pad_FileState::Pad_FileState(){}
Pad_FileState::Pad_FileState(Pad_Sound *source)
        : Pad_PlayState(source){}

Pad_FileState::~Pad_FileState(){}

////////////////////////////////////////////////////////////////
//
// Pad_AudioFormat Class
//   This class implements an abstraction around the different
//   input sound file types. It is currently just a wrapper
//   around the code in libst.a, but in the future it would
//   be nice to have it all in Pad++.
//
////////////////////////////////////////////////////////////////

Pad_AudioFormat::Pad_AudioFormat(){
    _length    = 0;
    _filename  = NULL;
}

Pad_AudioFormat::Pad_AudioFormat(const char *filename){
    _filename  = new Pad_String(filename);
}

Pad_AudioFormat::~Pad_AudioFormat(){
    if (_filename){
        delete [] _filename;
        _filename = NULL;
    }
    if (_data){
        delete [] _data;
        _data = NULL;
    }
}

const Pad_String *
Pad_AudioFormat::Get_Filename(void) const {
    return _filename;
}

long
Pad_AudioFormat::Get_length(void) const{
    return _length;
}

long
Pad_AudioFormat::Get_rate(void) const{
    Pad_errorString = "Pad_AudioFormat::Get_rate(void) should not be called";
    return -1;
}

long
Pad_AudioFormat::Get_channels(void) const{
    Pad_errorString = "Pad_AudioFormat::Get_channels(void) should not be called";
    return -1;
}

long
Pad_AudioFormat::Get_bits_per_sample(void) const{
    Pad_errorString = "Pad_AudioFormat::Bits_per_sample(void) should not be called";
    return -1;
}

void *
Pad_AudioFormat::Get_private(void) const{
    Pad_errorString = "Pad_AudioFormat::Get_private(void) should not be called";
    return NULL;
}

Pad_SoundFormat
Pad_AudioFormat::Get_style(void) const{
    Pad_errorString = "Pad_AudioFormat::Get_style(void) should not be called";
    return NoFormat;
}

//
// Get function for the data field
//
char *
Pad_AudioFormat::Get_data(void) const {
    return _data;
}

////////////////////////////////////////////////////////////////
//
// Pad_Sound Class
//
//
////////////////////////////////////////////////////////////////

Pad_Sound::Pad_Sound(){
//    SoundDevice *dev = SoundDevice::Get_instance();
  
    _states = NULL;
    _gain   = 0;
    _format  = NULL;
}

Pad_Sound::~Pad_Sound(){
    if (_format){
        delete _format;
        _format = NULL;
    }
    if (_states) {
        delete _states;
    }
}

void
Pad_Sound::Update_state (void)
{
    if (_states) {
        Pad_PlayState *state;
        Pad_Iterator oi;
    
        DOLIST (oi,*_states,Pad_PlayState,state) {
            state->Update_state();
        }
    }
}

Pad_Bool
Pad_Sound::Get_configuration (Pad_SoundOption &conf) const
{
    if (conf.type == SoundConfigGain) {
        conf.data = _gain;
    } else {
        return FALSE;
    }
    return TRUE;
}

Pad_Bool
Pad_Sound::Stop(void)
{
    Pad_Iterator li;
    Pad_PlayState *state;

    if (!_states){
        return FALSE;
    }
    
    DOLIST (li, *_states, Pad_PlayState, state) {
        state->Set_stop(TRUE);
    }
    return TRUE;
}

Pad_Bool
Pad_Sound::Play (Pad_SoundConfig*)
{
    Pad_errorString = "Pad_Sound::Play(Pad_SoundConfig*) should never be called";
    return FALSE;
}

Pad_Bool
Pad_Sound::Play (Pad_SoundConfig *config, Pad_PlayState *state)
{
    SoundDevice *dev = (SoundDevice*) SoundDevice::Get_instance();
    
    if (!_states) {
        _states = new Pad_List();
    }
    _states->Push(state);
    state->Update_state();
    return dev->Play(config,state);
}

Pad_Bool
Pad_Sound::Remove(Pad_Sound *sound)
{
    if (_states) {
        if (_states->Remove(sound)) {
            delete sound;
            return TRUE;
        }
    }
    return FALSE;
}

Pad_Bool
Pad_Sound::Configure (Pad_SoundConfig *config)
{
    Pad_SoundOption *option;

    if (config->gain->data != Pad_Sound::CONFIG_UNSET) {
      option = config->gain;
      option->Bounds(-60,30);
      _gain = option->data;
    }
    Update_state();
    return TRUE;
}

Pad_Bool
Pad_Sound::Configure (Pad_List &configList)
{
    Pad_SoundOption *option;
    Pad_Iterator oi;
    
    DOLIST (oi,configList,Pad_SoundOption,option) {
        switch (option->type) {
            case SoundConfigGain :
                option->Bounds(-60,30);
                _gain = option->data;
                break;
            default :
                break;
        }
    }
    return TRUE;
}

Pad_Bool
Pad_Sound::Get_configuration (Pad_List &conf) const
{
    Pad_optionSoundGain.data = _gain;
    conf.Push_last(&Pad_optionSoundGain);

    return TRUE;
}

Pad_Bool
Pad_Sound::Done(Pad_PlayState *state)
{
  return _states->Remove (state);
}


//
// Set function for the internal sound format 
//
Pad_Bool
Pad_Sound::Set_format(Pad_AudioFormat *format) {
    _format = format;
    return TRUE;
}

//
// Get function for the internal sound format 
//
Pad_AudioFormat *
Pad_Sound::Get_format() const {
    return _format;
}

//
// Get function for the sound data
//
char *
Pad_Sound::Get_data() const {
    return _format->Get_data();
}

//
// Get function for the number of 32 samples in the sound
//
int
Pad_Sound::Get_length() const {
    return _format->Get_length();
}

//
// Get function for the number of channels (mono or stereo)
//
int
Pad_Sound::Get_channels() const {
    return _format->Get_channels();
}

//
// Get function for the recording rate of the sound
//
int
Pad_Sound::Get_rate() const {
    return _format->Get_rate();
}

//
// Get function for the sample size (either 8 or 16 bits per sample)
//
int
Pad_Sound::Get_bits_per_sample() const {
    return _format->Get_bits_per_sample();
}

//
// Default Get function for filename
//
char *
Pad_Sound::Get_name() const {
    return NULL;
}

void
Pad_Sound::Init(void)
{
    Pad_SoundConfig *config;
    
    Pad_optionSoundDirection.type = SoundConfigDirection;
    Pad_optionSoundGain.type      = SoundConfigGain;
    Pad_optionSoundMaterial.type  = SoundConfigMaterial;
    Pad_optionSoundMaxWork.type   = SoundConfigMaxWork;
    Pad_optionSoundMinWork.type   = SoundConfigMinWork;
    Pad_optionSoundRatio.type     = SoundConfigRatio;
    Pad_optionSoundSize.type      = SoundConfigSize;
    Pad_optionSoundSpeed.type     = SoundConfigSpeed;
    Pad_optionSoundHardness.type  = SoundConfigHardness;
    Pad_optionSoundImpact.type    = SoundConfigImpact;
    Pad_optionSoundLength.type    = SoundConfigLength;
    Pad_optionSoundLoop.type      = SoundConfigLoop;
    Pad_optionSoundRoughness.type = SoundConfigRoughness;
    Pad_optionSoundScrape.type    = SoundConfigScrape;
    Pad_optionSoundVolume.type    = SoundConfigVolume;

        // the constructor for Pad_SoundConfig assumes that the previous
        // static variables have been configure
    config = Pad_SoundConfig::Get_instance();
}

Pad_Bool
Pad_Sound::Data(void)
{
    return FALSE;
}

//
// The operator<<() functions are for printing out debugging information
//    about the currently loaded sounds
//
ostream&
operator<<(ostream &os, Pad_Sound &sound){
    os << *sound._format;
    return os;
}

ostream&
operator<<(ostream &os, Pad_AudioFormat &format){
    os << *format._filename;
    return os;
}

////////////////////////////////////////////////////////////////
//
// Pad_SynthSound subclass of Pad_Sound
//
//
////////////////////////////////////////////////////////////////
Pad_SynthSound::Pad_SynthSound (){}
Pad_SynthSound::~Pad_SynthSound (){}

int
Pad_SynthSound::Get_rate(void) const
{ 
    return Pad_SynthSound::RATE;
}

int
Pad_SynthSound::Get_channels(void) const
{
    return Pad_SynthSound::CHANNELS;
}

Pad_SoundFormat
Pad_SynthSound::Get_style(void) const 
{
    return Pad_SynthSound::STYLE;
}

Pad_Bool
Pad_SynthSound::Data(void)
{
    return TRUE;
}

// Pad_Bool
// Pad_SynthSound::Get_configuration (Pad_SoundOption &) const
// {
//     return FALSE;
// }

// Pad_Bool
// Pad_SynthSound::Get_configuration (Pad_List &) const
// {
//     return FALSE;
// }

// Pad_Bool
// Pad_SynthSound::Configure (Pad_List &)
// {
//     return FALSE;
// }

// void
// Pad_SynthSound::Update_state(void){}



////////////////////////////////////////////////////////////////
//
// Pad_LiftSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////
float Pad_LiftSound::_rapPartial = 1.2;
int   Pad_LiftSound::_nbPartial  = 6;
float Pad_LiftSound::_pointMax   = 0.5;

Pad_LiftSound::Pad_LiftSound ()
{
    _size      = 440;
    _speed     = 0;
    _direction = 0;
    
	// sound is constructed with nbPartial components: 
	// f rap*f rap^2*f ...  rap^nbpartial*f = ... 
	// hence maxSize = (playFreq / 2)/...
    _maxSize =int((Pad_SynthSound::RATE) / (2*pow(_rapPartial,_nbPartial)));

        // these fields are initialized in UpdateState
        // we just give them reasonable default values just in case...
    _tableInc = TABLELENGTH / float(Pad_SynthSound::RATE);
    _inc = 0;
    _floor = 0;
}
Pad_LiftSound::~Pad_LiftSound (){}

Pad_Bool
Pad_LiftSound::Get_configuration (Pad_List &conf) const
{
    Pad_Sound::Get_configuration (conf);

    Pad_optionSoundDirection.data = _direction;
    conf.Push_last(&Pad_optionSoundDirection);

    Pad_optionSoundSpeed.data = _speed;
    conf.Push_last(&Pad_optionSoundSpeed);

    Pad_optionSoundSize.data = _size;
    conf.Push_last(&Pad_optionSoundSize);

    return TRUE;
}

Pad_Bool
Pad_LiftSound::Get_configuration (Pad_SoundOption &conf) const
{
    if (Pad_Sound::Get_configuration (conf)) {
        return TRUE;
    }
    switch (conf.type) {
        case SoundConfigSpeed :
            conf.data = _speed;
            break;
        case SoundConfigSize :
            conf.data = _size;
            break;
        case SoundConfigDirection :
            conf.data = _direction;
            break;
        default :
            return FALSE;
    }
    return TRUE;
}

Pad_Bool
Pad_LiftSound::Play (Pad_SoundConfig *config)
{
    Pad_LiftState *state = new Pad_LiftState(this);
    return Pad_Sound::Play(config,state);
}

Pad_Bool
Pad_LiftSound::Configure (Pad_SoundConfig *config)
{
    Pad_SoundOption *option;
    Pad_Sound::Configure(config);
    if (config->size->data != Pad_Sound::CONFIG_UNSET) {
      option = config->size;
      option->Bounds(1,_maxSize);
      _size = option->data;
    }
    if (config->speed->data != Pad_Sound::CONFIG_UNSET) {
      option = config->speed;
      option->Bounds(1,10000);
      _speed = option->data;
    }
    if (config->direction->data != Pad_Sound::CONFIG_UNSET) {
      option = config->direction;
      option->Bounds(-1,1);
      _direction = option->data;
    }
    Update_state();
    return TRUE;
}

Pad_Bool
Pad_LiftSound::Configure (Pad_List &configList)
{
    Pad_SoundOption *option;
    Pad_Iterator oi;
    
    Pad_Sound::Configure(configList);
    DOLIST (oi,configList,Pad_SoundOption,option) {
        switch (option->type) {
            case SoundConfigSpeed :
                option->Bounds(1,10000);
                _speed = option->data;
                break;
            case SoundConfigSize :
                option->Bounds(1,_maxSize);
                _size = option->data;
                break;
            case SoundConfigDirection :
                option->Bounds(-1,1);
                _direction = option->data;
                break;
            default :
                break;
        }
    }
    Update_state();
    return TRUE;
}

void
Pad_LiftSound::Update_state (void)
{
    Pad_Sound::Update_state();

    float rnb = pow(_rapPartial,_nbPartial);

    _xPointMax = _size*(1+(rnb-1)*_pointMax);

    _aUp1 = 1/(_xPointMax-_size);
    _bUp1 = -_size*_aUp1;
         
    _aUp2 = -1/(rnb*_size-_xPointMax);
    _bUp2 = -_xPointMax*_aUp2+1;

    _max1stPartial = int(_size)*(_rapPartial-1);
        
    _baseFreq = int(_size)+_floor;
}
    
////////////////////////////////////////////////////////////////
//
// Pad_ObjectSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////
Pad_ObjectSound::Pad_ObjectSound ()
{
    _material = 30;
    _size     = 50;
    _speed    = 50;
    
    _srate = Pad_SynthSound::RATE;
    _tableInc = TABLELENGTH / float(_srate);
}
Pad_ObjectSound::~Pad_ObjectSound (){}

Pad_Bool
Pad_ObjectSound::Get_configuration (Pad_List &conf) const 
{
    Pad_Sound::Get_configuration (conf);

    Pad_optionSoundMaterial.data = _material;
    conf.Push_last(&Pad_optionSoundMaterial);

    Pad_optionSoundSpeed.data = _speed;
    conf.Push_last(&Pad_optionSoundSpeed);

    Pad_optionSoundSize.data = _size;
    conf.Push_last(&Pad_optionSoundSize);

    return TRUE;
}

Pad_Bool
Pad_ObjectSound::Get_configuration (Pad_SoundOption &conf) const
{
    if (Pad_Sound::Get_configuration (conf)) {
        return TRUE;
    }
    switch (conf.type) {
        case SoundConfigSpeed :
            conf.data = _speed;
            break;
        case SoundConfigSize :
            conf.data = _size;
            break;
        case SoundConfigMaterial :
            conf.data = _material;
            break;
        default :
            return FALSE;
    }
    return TRUE;
}

Pad_Bool
Pad_ObjectSound::Configure (Pad_SoundConfig *config)
{
    Pad_SoundOption *option;
    Pad_Sound::Configure(config);
    if (config->size->data != Pad_Sound::CONFIG_UNSET) {
      option = config->size;
      option->Bounds(1,100);
      _size = option->data;
    }
    if (config->material->data != Pad_Sound::CONFIG_UNSET) {
      option = config->material;
      option->Bounds(1,100);
      _material = option->data;
    }
    Update_state();
    return TRUE;
}

Pad_Bool
Pad_ObjectSound::Configure (Pad_List &configList)
{
    Pad_SoundOption *option;
    Pad_Iterator oi;
    
    Pad_Sound::Configure(configList);
    DOLIST (oi,configList,Pad_SoundOption,option) {
        switch (option->type) {
            case SoundConfigSize :
                option->Bounds(1,100);
                _size = option->data;
                break;
            case SoundConfigMaterial :
                option->Bounds(1,100);
                _material = option->data;
                break;
            default :
                break;
        }
    }
    Update_state();
    return TRUE;
}

    
Pad_Bool
Pad_ObjectSound::Play (Pad_SoundConfig *config)
{
    Pad_ObjectState *state;
    
    if (config->scrape->data == 1) {
        if (config->impact->data == 1) {
            Pad_errorString = "only one of impact or scrape can be used at a time to play an object source";
            return FALSE;
        }
        state = new Pad_ScrapeState(this);
    } else if (config->impact->data == 1) {
        state = new Pad_ImpactState(this);
    } else {
        Pad_errorString = "must use impact or scrape in order to play an object source";
        return FALSE;
    }
    return Pad_Sound::Play(config,state);         
}

void
Pad_ObjectSound::Update_state (void)
{
    Pad_Sound::Update_state();
    
	/* update state */
    float bwidth [MAXFILTERS];
    float freq [MAXFILTERS];
    float bbw;
    float bwslope;
	
	/* first set freq and bwidth to log scales */
	/* 0 < size < 100 => 40 < freq[0] < 3000 */
    freq[0] = 40. + ((_size/100.) * (_size/100.) * 2960.);
	
	/* basic bandwidth */
	/* 0 < material < 100 => .005 < bbw < .1 */
    float mat = (100 - _material)/300.;
    bbw = .005 + mat * mat;
	
	/* slope for bandwidth */
	/* 0 < _material < 100 => -.01 < bwslope < .99 */
    bwslope = (_material - 1)/100.;
	
	/* bandwidth of the fundamental */
    bwidth[0] = freq[0] * bbw * bwslope;

	/* now set other freq's, bwidths as function of fund*/
	/* freq corresponds to bar resonances, */
	/* bwidth slopes up depending on initial bw */
    int i;
    for (i = 1; i < MAXFILTERS; i++) {
        int n = (i+1)*2 + 1;
        n *= n;
        freq[i] = freq[0] * n / 9.0;
		
        bwidth[i] = (freq[i] - freq[0]) * bbw * bwslope;
		
            /* limit to 1/3 of sample freq (Nyquist is 1/2) */
        if (freq[i] > _srate/3.) {
            break;
        }
    }
    _numFilters = i;

        /* right.  First set up filter coefficients for numflt filters...*/
    float twopidivsrate = (2.0 * M_PI) / _srate;
    for (i = 0; i < _numFilters; i++) {
        _filters[i][2] = exp(bwidth[i] * - twopidivsrate);
        _filters[i][1] = (4. * _filters[i][2]) *
            cos(freq[i] * twopidivsrate) / (_filters[i][2] + 1);
        _filters[i][0] = (1. - _filters[i][2]) *
            sqrt(1. - ((_filters[i][1] * _filters[i][1]) /
                       (_filters[i][2] * 4.)));
    }
}

////////////////////////////////////////////////////////////////
//
// Pad_MachineSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////
Pad_MachineSound::Pad_MachineSound ()
{
    _minWork = 3;
    _maxWork = 5;
    _size = 90;
    _ratio = 3;
    _speed = 50;
    
    _tableInc = TABLELENGTH / float(Pad_SynthSound::RATE);
    
	// these fields are initialized in Update_state
	// we just give them reasonable default values just in case...
    _carFreq = 50.0;
    _modFreq = _carFreq;
    _modFreqInc = _modFreq * _tableInc;
}
Pad_MachineSound::~Pad_MachineSound (){}

Pad_Bool
Pad_MachineSound::Get_configuration (Pad_List &conf) const 
{
    Pad_Sound::Get_configuration (conf);
    
    Pad_optionSoundRatio.data = _ratio;
    conf.Push_last(&Pad_optionSoundRatio);

    Pad_optionSoundMinWork.data = _minWork;
    conf.Push_last(&Pad_optionSoundMinWork);

    Pad_optionSoundMaxWork.data = _maxWork;
    conf.Push_last(&Pad_optionSoundMaxWork);

    Pad_optionSoundSpeed.data = _speed;
    conf.Push_last(&Pad_optionSoundSpeed);

    Pad_optionSoundSize.data = _size;
    conf.Push_last(&Pad_optionSoundSize);

    return TRUE;
}

Pad_Bool
Pad_MachineSound::Get_configuration (Pad_SoundOption &conf) const
{
    if (Pad_Sound::Get_configuration (conf)) {
        return TRUE;
    }
    switch (conf.type) {
        case SoundConfigSpeed :
            conf.data = _speed;
            break;
        case SoundConfigSize :
            conf.data = _size;
            break;
        case SoundConfigRatio :
            conf.data = _ratio;
            break;
        case SoundConfigMinWork :
            conf.data = _minWork;
            break;
        case SoundConfigMaxWork :
            conf.data = _maxWork;
            break;
        default :
            return FALSE;
    }
    return TRUE;
}

Pad_Bool
Pad_MachineSound::Play (Pad_SoundConfig *config)
{
    Configure(config);
    Pad_MachineState *state = new Pad_MachineState(this);
    return Pad_Sound::Play(config,state);    
}

Pad_Bool
Pad_MachineSound::Configure (Pad_SoundConfig *config)
{
    Pad_SoundOption *option;
    Pad_Sound::Configure(config);
    if (config->speed->data != Pad_Sound::CONFIG_UNSET) {
      option = config->speed;
      option->Bounds(1,10000);
      _speed = option->data;
    }
    if (config->size->data != Pad_Sound::CONFIG_UNSET) {
      option = config->size;
      option->Bounds(1,100);
      _size = option->data;
    }
    if (config->ratio->data != Pad_Sound::CONFIG_UNSET) {
      option = config->ratio;
      option->Bounds(1,100);
      _ratio = option->data;
    }
    if (config->maxWork->data != Pad_Sound::CONFIG_UNSET) {
      option = config->maxWork;
      option->Bounds(_minWork,10);
      _maxWork = option->data;
    }
    if (config->minWork->data != Pad_Sound::CONFIG_UNSET) {
      option = config->minWork;
      option->Bounds(0,10);
      _minWork = option->data;
    }
    Update_state();
    return TRUE;
}

Pad_Bool
Pad_MachineSound::Configure (Pad_List &configList)
{
    Pad_SoundOption *option;
    Pad_Iterator oi;
    
    Pad_Sound::Configure(configList);
    DOLIST (oi,configList,Pad_SoundOption,option) {
        switch (option->type) {
            case SoundConfigSpeed :
                option->Bounds(1,10000);
                _speed = option->data;
                break;
            case SoundConfigSize :
                option->Bounds(1,100);
                _size = option->data;
                break;
            case SoundConfigRatio :
                option->Bounds(1,100);
                _ratio = option->data;
                break;
            case SoundConfigMaxWork :
                option->Bounds(_minWork,10);
                _maxWork = option->data;
                break;
            case SoundConfigMinWork :
                option->Bounds(0,10);
                _minWork = option->data;
                break;
            default :
                break;
        }
    }
    Update_state();
    return TRUE;
}

void
Pad_MachineSound::Update_state(void)
{
    Pad_Sound::Update_state();
    
        // must check that _maxWork > _minWork
    if (_maxWork < _minWork) {
        _maxWork = _minWork;
    }
    
	/* 50 < carFreq < 1040 */
    _carFreq = (40 + (101 - _size)*10);
	
	/* carFreq / 10 < modFreq < carFreq * 10 */
    _modFreq = _carFreq * _ratio / 10.0;
    _modFreqInc = _modFreq * _tableInc;
}

////////////////////////////////////////////////////////////////
//
// Pad_FileSound subclass of Pad_Sound
//
//
////////////////////////////////////////////////////////////////
Pad_FileSound::Pad_FileSound (void){}
Pad_FileSound::~Pad_FileSound (void){}

Pad_Bool
Pad_FileSound::Play (Pad_SoundConfig *config)
{
    Pad_FileState *state = new Pad_FileState(this);
    return Pad_Sound::Play(config,state);
}

char *
Pad_FileSound::Get_name() const {
    return _name.Get();
}

//
// This function loads the sound file specified by filename into memory.
//   It returns the TRUE or FALSE. If it returns false the error is 
//   condition is stored in Pad_errorString.
//
Pad_Bool
Pad_FileSound::Load(char *filename){

    SoundDevice *dev = SoundDevice::Get_instance();
        // must set _name before calling SoundDevice::Load()
    _name = filename;
    return dev->Load(*this);
}

Pad_Bool
Pad_FileSound::Get_configuration (Pad_SoundOption &conf) const
{
    if (Pad_Sound::Get_configuration (conf)) {
        return TRUE;
    }
    return FALSE;
}

Pad_Bool
Pad_FileSound::Get_configuration (Pad_List &conf) const
{
    if (Pad_Sound::Get_configuration (conf)) {
        return TRUE;
    }
    return FALSE;
}

Pad_Bool
Pad_FileSound::Configure (Pad_List &configList)
{
    Pad_Sound::Configure(configList);
    return FALSE;
}

void
Pad_FileSound::Update_state(void)
{
    Pad_Sound::Update_state();
}

////////////////////////////////////////
//
// SoundDevice class
//
////////////////////////////////////////
SoundDevice::SoundDevice()
{
    _sounds = NULL;
}

SoundDevice::~SoundDevice()
{
    if (_sounds) {
        delete _sounds;
    }
}

Pad_List *
SoundDevice::Get_sounds(void) const
{
    return _sounds;
}

//---- table to convert dB gain to amp factors
#define MIN_DB	-60
#define MAX_DB	30


float Pad_Sound::DbToAmpTable [] = {
        0.000001,       /* -60 dB */
        0.000001,       /* -59 dB */
        0.000002,       /* -58 dB */
        0.000002,       /* -57 dB */
        0.000003,       /* -56 dB */
        0.000003,       /* -55 dB */
        0.000004,       /* -54 dB */
        0.000005,       /* -53 dB */
        0.000006,       /* -52 dB */
        0.000008,       /* -51 dB */
        0.000010,       /* -50 dB */
        0.000013,       /* -49 dB */
        0.000016,       /* -48 dB */
        0.000020,       /* -47 dB */
        0.000025,       /* -46 dB */
        0.000032,       /* -45 dB */
        0.000040,       /* -44 dB */
        0.000050,       /* -43 dB */
        0.000063,       /* -42 dB */
        0.000079,       /* -41 dB */
        0.000100,       /* -40 dB */
        0.000126,       /* -39 dB */
        0.000158,       /* -38 dB */
        0.000200,       /* -37 dB */
        0.000251,       /* -36 dB */
        0.000316,       /* -35 dB */
        0.000398,       /* -34 dB */
        0.000501,       /* -33 dB */
        0.000631,       /* -32 dB */
        0.000794,       /* -31 dB */
        0.001000,       /* -30 dB */
        0.001259,       /* -29 dB */
        0.001585,       /* -28 dB */
        0.001995,       /* -27 dB */
        0.002512,       /* -26 dB */
        0.003162,       /* -25 dB */
        0.003981,       /* -24 dB */
        0.005012,       /* -23 dB */
        0.006310,       /* -22 dB */
        0.007943,       /* -21 dB */
        0.010000,       /* -20 dB */
        0.012589,       /* -19 dB */
        0.015849,       /* -18 dB */
        0.019953,       /* -17 dB */
        0.025119,       /* -16 dB */
        0.031623,       /* -15 dB */
        0.039811,       /* -14 dB */
        0.050119,       /* -13 dB */
        0.063096,       /* -12 dB */
        0.079433,       /* -11 dB */
        0.100000,       /* -10 dB */
        0.125893,       /* -9 dB */
        0.158489,       /* -8 dB */
        0.199526,       /* -7 dB */
        0.251189,       /* -6 dB */
        0.316228,       /* -5 dB */
        0.398107,       /* -4 dB */
        0.501187,       /* -3 dB */
        0.630957,       /* -2 dB */
        0.794328,       /* -1 dB */
        1.000000,       /* 0 dB */
        1.258925,       /* 1 dB */
        1.584893,       /* 2 dB */
        1.995262,       /* 3 dB */
        2.511886,       /* 4 dB */
        3.162278,       /* 5 dB */
        3.981072,       /* 6 dB */
        5.011872,       /* 7 dB */
        6.309573,       /* 8 dB */
        7.943282,       /* 9 dB */
        10.000000,      /* 10 dB */
        12.589254,      /* 11 dB */
        15.848932,      /* 12 dB */
        19.952623,      /* 13 dB */
        25.118864,      /* 14 dB */
        31.622777,      /* 15 dB */
        39.810717,      /* 16 dB */
        50.118723,      /* 17 dB */
        63.095734,      /* 18 dB */
        79.432823,      /* 19 dB */
        100.000000,     /* 20 dB */
        125.892541,     /* 21 dB */
        158.489319,     /* 22 dB */
        199.526231,     /* 23 dB */
        251.188643,     /* 24 dB */
        316.227766,     /* 25 dB */
        398.107171,     /* 26 dB */
        501.187234,     /* 27 dB */
        630.957344,     /* 28 dB */
        794.328235,     /* 29 dB */
        1000.000000     /* 30 dB */
};

float
Pad_Sound::Db_to_amp (int db)
{
	if (db == 0)
		return 1.0;
	
	if (db < 0)
		return Pad_Sound::DbToAmpTable [0];
	if (db > MAX_DB)
		return Pad_Sound::DbToAmpTable [MAX_DB];
	
	return Pad_Sound::DbToAmpTable [db - MIN_DB];
}

/*
 * Copyright 1993 by Digital Equipment Corporation, Maynard, Massachusetts.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and its 
 * documentation for any purpose is hereby granted without fee, provided that 
 * the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation, and that the name of Digital not be used in advertising or 
 * publicity pertaining to distribution of the software without specific, 
 * written prior permission.  Digital makes no representations about the 
 * suitability of this software for any purpose.  It is provided "as is" 
 * without express or implied warranty.
 * 
 * DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING 
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL 
 * DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY 
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN 
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF 
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
float
Pad_Sound::AF_sine_float[1024] = {
 0.000000, 0.006136, 0.012272, 0.018407,
 0.024541, 0.030675, 0.036807, 0.042938,
 0.049068, 0.055195, 0.061321, 0.067444,
 0.073565, 0.079682, 0.085797, 0.091909,
 0.098017, 0.104122, 0.110222, 0.116319,
 0.122411, 0.128498, 0.134581, 0.140658,
 0.146730, 0.152797, 0.158858, 0.164913,
 0.170962, 0.177004, 0.183040, 0.189069,
 0.195090, 0.201105, 0.207111, 0.213110,
 0.219101, 0.225084, 0.231058, 0.237024,
 0.242980, 0.248928, 0.254866, 0.260794,
 0.266713, 0.272621, 0.278520, 0.284408,
 0.290285, 0.296151, 0.302006, 0.307850,
 0.313682, 0.319502, 0.325310, 0.331106,
 0.336890, 0.342661, 0.348419, 0.354164,
 0.359895, 0.365613, 0.371317, 0.377007,
 0.382683, 0.388345, 0.393992, 0.399624,
 0.405241, 0.410843, 0.416430, 0.422000,
 0.427555, 0.433094, 0.438616, 0.444122,
 0.449611, 0.455084, 0.460539, 0.465977,
 0.471397, 0.476799, 0.482184, 0.487550,
 0.492898, 0.498228, 0.503538, 0.508830,
 0.514103, 0.519356, 0.524590, 0.529804,
 0.534998, 0.540171, 0.545325, 0.550458,
 0.555570, 0.560662, 0.565732, 0.570781,
 0.575808, 0.580814, 0.585798, 0.590760,
 0.595699, 0.600616, 0.605511, 0.610383,
 0.615232, 0.620057, 0.624860, 0.629638,
 0.634393, 0.639124, 0.643832, 0.648514,
 0.653173, 0.657807, 0.662416, 0.667000,
 0.671559, 0.676093, 0.680601, 0.685084,
 0.689541, 0.693971, 0.698376, 0.702755,
 0.707107, 0.711432, 0.715731, 0.720003,
 0.724247, 0.728464, 0.732654, 0.736817,
 0.740951, 0.745058, 0.749136, 0.753187,
 0.757209, 0.761202, 0.765167, 0.769103,
 0.773010, 0.776888, 0.780737, 0.784557,
 0.788346, 0.792107, 0.795837, 0.799537,
 0.803208, 0.806848, 0.810457, 0.814036,
 0.817585, 0.821102, 0.824589, 0.828045,
 0.831470, 0.834863, 0.838225, 0.841555,
 0.844854, 0.848120, 0.851355, 0.854558,
 0.857729, 0.860867, 0.863973, 0.867046,
 0.870087, 0.873095, 0.876070, 0.879012,
 0.881921, 0.884797, 0.887640, 0.890449,
 0.893224, 0.895966, 0.898674, 0.901349,
 0.903989, 0.906596, 0.909168, 0.911706,
 0.914210, 0.916679, 0.919114, 0.921514,
 0.923880, 0.926210, 0.928506, 0.930767,
 0.932993, 0.935184, 0.937339, 0.939459,
 0.941544, 0.943593, 0.945607, 0.947586,
 0.949528, 0.951435, 0.953306, 0.955141,
 0.956940, 0.958703, 0.960431, 0.962121,
 0.963776, 0.965394, 0.966976, 0.968522,
 0.970031, 0.971504, 0.972940, 0.974339,
 0.975702, 0.977028, 0.978317, 0.979570,
 0.980785, 0.981964, 0.983105, 0.984210,
 0.985278, 0.986308, 0.987301, 0.988258,
 0.989177, 0.990058, 0.990903, 0.991710,
 0.992480, 0.993212, 0.993907, 0.994565,
 0.995185, 0.995767, 0.996313, 0.996820,
 0.997290, 0.997723, 0.998118, 0.998476,
 0.998795, 0.999078, 0.999322, 0.999529,
 0.999699, 0.999831, 0.999925, 0.999981,
 1.000000, 0.999981, 0.999925, 0.999831,
 0.999699, 0.999529, 0.999322, 0.999078,
 0.998795, 0.998476, 0.998118, 0.997723,
 0.997290, 0.996820, 0.996313, 0.995767,
 0.995185, 0.994565, 0.993907, 0.993212,
 0.992480, 0.991710, 0.990903, 0.990058,
 0.989177, 0.988258, 0.987301, 0.986308,
 0.985278, 0.984210, 0.983105, 0.981964,
 0.980785, 0.979570, 0.978317, 0.977028,
 0.975702, 0.974339, 0.972940, 0.971504,
 0.970031, 0.968522, 0.966976, 0.965394,
 0.963776, 0.962121, 0.960431, 0.958703,
 0.956940, 0.955141, 0.953306, 0.951435,
 0.949528, 0.947586, 0.945607, 0.943593,
 0.941544, 0.939459, 0.937339, 0.935184,
 0.932993, 0.930767, 0.928506, 0.926210,
 0.923880, 0.921514, 0.919114, 0.916679,
 0.914210, 0.911706, 0.909168, 0.906596,
 0.903989, 0.901349, 0.898674, 0.895966,
 0.893224, 0.890449, 0.887640, 0.884797,
 0.881921, 0.879012, 0.876070, 0.873095,
 0.870087, 0.867046, 0.863973, 0.860867,
 0.857729, 0.854558, 0.851355, 0.848120,
 0.844854, 0.841555, 0.838225, 0.834863,
 0.831470, 0.828045, 0.824589, 0.821102,
 0.817585, 0.814036, 0.810457, 0.806848,
 0.803208, 0.799537, 0.795837, 0.792107,
 0.788346, 0.784557, 0.780737, 0.776888,
 0.773010, 0.769103, 0.765167, 0.761202,
 0.757209, 0.753187, 0.749136, 0.745058,
 0.740951, 0.736817, 0.732654, 0.728464,
 0.724247, 0.720003, 0.715731, 0.711432,
 0.707107, 0.702755, 0.698376, 0.693971,
 0.689541, 0.685084, 0.680601, 0.676093,
 0.671559, 0.667000, 0.662416, 0.657807,
 0.653173, 0.648514, 0.643832, 0.639124,
 0.634393, 0.629638, 0.624860, 0.620057,
 0.615232, 0.610383, 0.605511, 0.600616,
 0.595699, 0.590760, 0.585798, 0.580814,
 0.575808, 0.570781, 0.565732, 0.560662,
 0.555570, 0.550458, 0.545325, 0.540171,
 0.534998, 0.529804, 0.524590, 0.519356,
 0.514103, 0.508830, 0.503538, 0.498228,
 0.492898, 0.487550, 0.482184, 0.476799,
 0.471397, 0.465977, 0.460539, 0.455084,
 0.449611, 0.444122, 0.438616, 0.433094,
 0.427555, 0.422000, 0.416430, 0.410843,
 0.405241, 0.399624, 0.393992, 0.388345,
 0.382683, 0.377007, 0.371317, 0.365613,
 0.359895, 0.354164, 0.348419, 0.342661,
 0.336890, 0.331106, 0.325310, 0.319502,
 0.313682, 0.307850, 0.302006, 0.296151,
 0.290285, 0.284408, 0.278520, 0.272621,
 0.266713, 0.260794, 0.254866, 0.248928,
 0.242980, 0.237024, 0.231058, 0.225084,
 0.219101, 0.213110, 0.207111, 0.201105,
 0.195090, 0.189069, 0.183040, 0.177004,
 0.170962, 0.164913, 0.158858, 0.152797,
 0.146730, 0.140658, 0.134581, 0.128498,
 0.122411, 0.116319, 0.110222, 0.104122,
 0.098017, 0.091909, 0.085797, 0.079682,
 0.073565, 0.067444, 0.061321, 0.055195,
 0.049068, 0.042938, 0.036807, 0.030675,
 0.024541, 0.018407, 0.012272, 0.006136,
 0.000000, -0.006136, -0.012272, -0.018407,
 -0.024541, -0.030675, -0.036807, -0.042938,
 -0.049068, -0.055195, -0.061321, -0.067444,
 -0.073565, -0.079682, -0.085797, -0.091909,
 -0.098017, -0.104122, -0.110222, -0.116319,
 -0.122411, -0.128498, -0.134581, -0.140658,
 -0.146730, -0.152797, -0.158858, -0.164913,
 -0.170962, -0.177004, -0.183040, -0.189069,
 -0.195090, -0.201105, -0.207111, -0.213110,
 -0.219101, -0.225084, -0.231058, -0.237024,
 -0.242980, -0.248928, -0.254866, -0.260794,
 -0.266713, -0.272621, -0.278520, -0.284408,
 -0.290285, -0.296151, -0.302006, -0.307850,
 -0.313682, -0.319502, -0.325310, -0.331106,
 -0.336890, -0.342661, -0.348419, -0.354164,
 -0.359895, -0.365613, -0.371317, -0.377007,
 -0.382683, -0.388345, -0.393992, -0.399624,
 -0.405241, -0.410843, -0.416430, -0.422000,
 -0.427555, -0.433094, -0.438616, -0.444122,
 -0.449611, -0.455084, -0.460539, -0.465977,
 -0.471397, -0.476799, -0.482184, -0.487550,
 -0.492898, -0.498228, -0.503538, -0.508830,
 -0.514103, -0.519356, -0.524590, -0.529804,
 -0.534998, -0.540171, -0.545325, -0.550458,
 -0.555570, -0.560662, -0.565732, -0.570781,
 -0.575808, -0.580814, -0.585798, -0.590760,
 -0.595699, -0.600616, -0.605511, -0.610383,
 -0.615232, -0.620057, -0.624860, -0.629638,
 -0.634393, -0.639124, -0.643832, -0.648514,
 -0.653173, -0.657807, -0.662416, -0.667000,
 -0.671559, -0.676093, -0.680601, -0.685084,
 -0.689541, -0.693971, -0.698376, -0.702755,
 -0.707107, -0.711432, -0.715731, -0.720003,
 -0.724247, -0.728464, -0.732654, -0.736817,
 -0.740951, -0.745058, -0.749136, -0.753187,
 -0.757209, -0.761202, -0.765167, -0.769103,
 -0.773010, -0.776888, -0.780737, -0.784557,
 -0.788346, -0.792107, -0.795837, -0.799537,
 -0.803208, -0.806848, -0.810457, -0.814036,
 -0.817585, -0.821102, -0.824589, -0.828045,
 -0.831470, -0.834863, -0.838225, -0.841555,
 -0.844854, -0.848120, -0.851355, -0.854558,
 -0.857729, -0.860867, -0.863973, -0.867046,
 -0.870087, -0.873095, -0.876070, -0.879012,
 -0.881921, -0.884797, -0.887640, -0.890449,
 -0.893224, -0.895966, -0.898674, -0.901349,
 -0.903989, -0.906596, -0.909168, -0.911706,
 -0.914210, -0.916679, -0.919114, -0.921514,
 -0.923880, -0.926210, -0.928506, -0.930767,
 -0.932993, -0.935184, -0.937339, -0.939459,
 -0.941544, -0.943593, -0.945607, -0.947586,
 -0.949528, -0.951435, -0.953306, -0.955141,
 -0.956940, -0.958703, -0.960431, -0.962121,
 -0.963776, -0.965394, -0.966976, -0.968522,
 -0.970031, -0.971504, -0.972940, -0.974339,
 -0.975702, -0.977028, -0.978317, -0.979570,
 -0.980785, -0.981964, -0.983105, -0.984210,
 -0.985278, -0.986308, -0.987301, -0.988258,
 -0.989177, -0.990058, -0.990903, -0.991710,
 -0.992480, -0.993212, -0.993907, -0.994565,
 -0.995185, -0.995767, -0.996313, -0.996820,
 -0.997290, -0.997723, -0.998118, -0.998476,
 -0.998795, -0.999078, -0.999322, -0.999529,
 -0.999699, -0.999831, -0.999925, -0.999981,
 -1.000000, -0.999981, -0.999925, -0.999831,
 -0.999699, -0.999529, -0.999322, -0.999078,
 -0.998795, -0.998476, -0.998118, -0.997723,
 -0.997290, -0.996820, -0.996313, -0.995767,
 -0.995185, -0.994565, -0.993907, -0.993212,
 -0.992480, -0.991710, -0.990903, -0.990058,
 -0.989177, -0.988258, -0.987301, -0.986308,
 -0.985278, -0.984210, -0.983105, -0.981964,
 -0.980785, -0.979570, -0.978317, -0.977028,
 -0.975702, -0.974339, -0.972940, -0.971504,
 -0.970031, -0.968522, -0.966976, -0.965394,
 -0.963776, -0.962121, -0.960431, -0.958703,
 -0.956940, -0.955141, -0.953306, -0.951435,
 -0.949528, -0.947586, -0.945607, -0.943593,
 -0.941544, -0.939459, -0.937339, -0.935184,
 -0.932993, -0.930767, -0.928506, -0.926210,
 -0.923880, -0.921514, -0.919114, -0.916679,
 -0.914210, -0.911706, -0.909168, -0.906596,
 -0.903989, -0.901349, -0.898674, -0.895966,
 -0.893224, -0.890449, -0.887640, -0.884797,
 -0.881921, -0.879012, -0.876070, -0.873095,
 -0.870087, -0.867046, -0.863973, -0.860867,
 -0.857729, -0.854558, -0.851355, -0.848120,
 -0.844854, -0.841555, -0.838225, -0.834863,
 -0.831470, -0.828045, -0.824589, -0.821102,
 -0.817585, -0.814036, -0.810457, -0.806848,
 -0.803208, -0.799537, -0.795837, -0.792107,
 -0.788346, -0.784557, -0.780737, -0.776888,
 -0.773010, -0.769103, -0.765167, -0.761202,
 -0.757209, -0.753187, -0.749136, -0.745058,
 -0.740951, -0.736817, -0.732654, -0.728464,
 -0.724247, -0.720003, -0.715731, -0.711432,
 -0.707107, -0.702755, -0.698376, -0.693971,
 -0.689541, -0.685084, -0.680601, -0.676093,
 -0.671559, -0.667000, -0.662416, -0.657807,
 -0.653173, -0.648514, -0.643832, -0.639124,
 -0.634393, -0.629638, -0.624860, -0.620057,
 -0.615232, -0.610383, -0.605511, -0.600616,
 -0.595699, -0.590760, -0.585798, -0.580814,
 -0.575808, -0.570781, -0.565732, -0.560662,
 -0.555570, -0.550458, -0.545325, -0.540171,
 -0.534998, -0.529804, -0.524590, -0.519356,
 -0.514103, -0.508830, -0.503538, -0.498228,
 -0.492898, -0.487550, -0.482184, -0.476799,
 -0.471397, -0.465977, -0.460539, -0.455084,
 -0.449611, -0.444122, -0.438616, -0.433094,
 -0.427555, -0.422000, -0.416430, -0.410843,
 -0.405241, -0.399624, -0.393992, -0.388345,
 -0.382683, -0.377007, -0.371317, -0.365613,
 -0.359895, -0.354164, -0.348419, -0.342661,
 -0.336890, -0.331106, -0.325310, -0.319502,
 -0.313682, -0.307850, -0.302006, -0.296151,
 -0.290285, -0.284408, -0.278520, -0.272621,
 -0.266713, -0.260794, -0.254866, -0.248928,
 -0.242980, -0.237024, -0.231058, -0.225084,
 -0.219101, -0.213110, -0.207111, -0.201105,
 -0.195090, -0.189069, -0.183040, -0.177004,
 -0.170962, -0.164913, -0.158858, -0.152797,
 -0.146730, -0.140658, -0.134581, -0.128498,
 -0.122411, -0.116319, -0.110222, -0.104122,
 -0.098017, -0.091909, -0.085797, -0.079682,
 -0.073565, -0.067444, -0.061321, -0.055195,
 -0.049068, -0.042938, -0.036807, -0.030675,
 -0.024541, -0.018407, -0.012272, -0.006136
};

#else /* HAS_SOUND */

//
// Sound is not enable so these functions all return errors
//

Pad_Sound::Pad_Sound(){
    _format = NULL;
    cerr << "Sound not enabled" << endl;
    Pad_errorString = "Sound not enabled";
}

Pad_Sound::~Pad_Sound(){
    if (_format){
        delete _format;
        _format = NULL;
    }
}

//
// Get function for the internal sound format 
//
Pad_AudioFormat *
Pad_Sound::Get_format() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return NULL;
}

//
// Get function for the sound data
//
char *
Pad_Sound::Get_data() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return NULL;
}

//
// Get function for the number of 32 samples in the sound
//
int
Pad_Sound::Get_length() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return -1;
}

//
// Get function for the number of channels (mono or stereo)
//
int
Pad_Sound::Get_channels() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return -1;
}

//
// Get function for the recording rate of the sound
//
int
Pad_Sound::Get_rate() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return -1;
}

//
// Get function for the sample size (either 8 or 16 bits per sample)
//
int
Pad_Sound::Get_bits_per_sample() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return -1;
}

//
// Get function for the filename the sound was loaded from
//
char *
Pad_Sound::Get_name() const {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return NULL;
}

//
// Play a sound
//
Pad_Bool
Pad_Sound::Play(Pad_SoundConfig*) {
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Play (Pad_SoundConfig *, Pad_PlayState *)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Stop (void)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Get_configuration (Pad_SoundOption &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Get_configuration (Pad_List &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Configure (Pad_List &)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Configure (Pad_SoundConfig*)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

void
Pad_Sound::Update_state(void)
{
    cerr << "Sound not enabled" << endl;
}

Pad_Bool
Pad_Sound::Data(void)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_Sound::Done(Pad_PlayState*)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

////////////////////////////////////////////////////////////////
//
// Pad_SynthSound subclass of Pad_Sound
//
//
////////////////////////////////////////////////////////////////
Pad_SynthSound::Pad_SynthSound (){}
Pad_SynthSound::~Pad_SynthSound (){}

Pad_Bool
Pad_SynthSound::Play (Pad_SoundConfig *)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_SynthSound::Get_configuration (Pad_SoundOption &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_SynthSound::Get_configuration (Pad_List &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_SynthSound::Configure (Pad_List &)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

void
Pad_SynthSound::Update_state(void)
{
    cerr << "Sound not enabled" << endl;
}

int
Pad_SynthSound::Get_rate(void) const
{ 
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

int
Pad_SynthSound::Get_channels(void) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_SoundFormat
Pad_SynthSound::Get_style(void) const 
{
    cerr << "Sound not enabled" << endl;
    return NoFormat;
}

Pad_Bool
Pad_SynthSound::Data(void)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

////////////////////////////////////////////////////////////////
//
// Pad_LiftSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////

Pad_LiftSound::Pad_LiftSound (){}
Pad_LiftSound::~Pad_LiftSound (){}

Pad_Bool
Pad_LiftSound::Get_configuration (Pad_List &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_LiftSound::Get_configuration (Pad_SoundOption &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_LiftSound::Play (Pad_SoundConfig *)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_LiftSound::Configure (Pad_List &)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_LiftSound::Configure (Pad_SoundConfig*)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

void
Pad_LiftSound::Update_state (void)
{
    cerr << "Sound not enabled" << endl;
}
    
////////////////////////////////////////////////////////////////
//
// Pad_ObjectSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////
Pad_ObjectSound::Pad_ObjectSound (){}
Pad_ObjectSound::~Pad_ObjectSound (){}

Pad_Bool
Pad_ObjectSound::Get_configuration (Pad_List &) const 
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_ObjectSound::Get_configuration (Pad_SoundOption &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_ObjectSound::Configure (Pad_List &)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_ObjectSound::Configure (Pad_SoundConfig*)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

    
Pad_Bool
Pad_ObjectSound::Play (Pad_SoundConfig *)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

void
Pad_ObjectSound::Update_state (void)
{
    cerr << "Sound not enabled" << endl;
}

////////////////////////////////////////////////////////////////
//
// Pad_MachineSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////
Pad_MachineSound::Pad_MachineSound (){}
Pad_MachineSound::~Pad_MachineSound (){}

Pad_Bool
Pad_MachineSound::Get_configuration (Pad_List &) const 
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_MachineSound::Get_configuration (Pad_SoundOption &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_MachineSound::Play (Pad_SoundConfig *)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_MachineSound::Configure (Pad_List &)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_MachineSound::Configure (Pad_SoundConfig*)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

void
Pad_MachineSound::Update_state(void)
{
    cerr << "Sound not enabled" << endl;
}

////////////////////////////////////////////////////////////////
//
// Pad_FileSound subclass of Pad_Sound
//
//
////////////////////////////////////////////////////////////////
Pad_FileSound::Pad_FileSound (void){}
Pad_FileSound::~Pad_FileSound (void){}

void
Pad_FileSound::Update_state (void)
{
    cerr << "Sound not enabled" << endl;
}
    
Pad_Bool
Pad_FileSound::Get_configuration (Pad_SoundOption &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_FileSound::Get_configuration (Pad_List &) const
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_FileSound::Configure (Pad_List &)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

Pad_Bool
Pad_FileSound::Play (Pad_SoundConfig *)
{
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

char *
Pad_FileSound::Get_name (void) const
{
    cerr << "Sound not enabled" << endl;
    return NULL;
}

//
// This function loads the sound file specified by filename into memory.
//   It returns the TRUE or FALSE. If it returns false the error is 
//   condition is stored in Pad_errorString.
//
Pad_Bool
Pad_FileSound::Load(char *){
    cerr << "Sound not enabled" << endl;
    return FALSE;
}

//
// The operator<<() functions are for printing out debugging information
//    about the currently loaded sounds
//
ostream&
operator<<(ostream &os, Pad_Sound &){
    cerr << "Sound not enabled" << endl;
    return os;
}

ostream&
operator<<(ostream &os, Pad_AudioFormat &){
    cerr << "Sound not enabled" << endl;
    return os;
}

////////////////////////////////////////
//
// SoundDevice class
//
////////////////////////////////////////
SoundDevice::SoundDevice()
{
    _sounds = NULL;
    cerr << "Sound not enabled" << endl;
}

SoundDevice::~SoundDevice()
{
    if (_sounds) {
        delete _sounds;
    }
}

Pad_List *
SoundDevice::Get_sounds(void) const
{
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return NULL;
}

SoundDevice *
SoundDevice::Get_instance(void)
{
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
    return NULL;
}

#endif /* HAS_SOUND */
