#ifndef PAD_SOUND_H
#define PAD_SOUND_H
/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/
#include <stdio.h>
#include "defs.h"
#include "pad-string.h"
#include "api.h"

//
// The following are for the Pad_SynthSound subclasses
//

#define MAXFILTERS 20
#define WINSIZE    500
#define TABLELENGTH 1024
#define SINE(index)	Pad_Sound::AF_sine_float [((int) (index)) % TABLELENGTH]

enum Pad_SoundFormat {
    NoFormat,
    LinearSigned8,
    LinearSigned16LSB,
    LinearSigned16MSB,
};

class Pad_HashTable;
class Pad_List;
class ostream;

////////////////////////////////////////////////////////////////
//
// Pad_PlayState Class
//
//
////////////////////////////////////////////////////////////////

class Pad_PlayState {
public:
    Pad_PlayState();
    Pad_PlayState(Pad_Sound *source);
    virtual ~Pad_PlayState();
    virtual int Refill(int numBytes);
    virtual void Update_state(void);
    Pad_Sound *_source;
    Pad_Bool Loop(void);
    void     Set_loop(Pad_Bool loop);
    Pad_Bool Stopped(void);
    void     Set_stop(Pad_Bool stop);
protected:
    Pad_Bool   _loop;
    Pad_Bool   _stopped;
    char      *_data;
    
    virtual void _Init_state(void);
            void _Init(void);
};

class Pad_FileState : public Pad_PlayState {
public:
    Pad_FileState();
    Pad_FileState(Pad_Sound *source);
    virtual ~Pad_FileState();
protected:
};

class Pad_SynthState : public Pad_PlayState {
public:
    Pad_SynthState();
    Pad_SynthState(Pad_Sound *source);
    virtual ~Pad_SynthState();
    virtual int Refill(int numBytes);
    virtual void  Update_state(void);
protected:
    float _amp;
    float _inSample;
    
            float _DbToAmp(int db);    
    virtual void  _Init_state(void);
};

class Pad_ObjectState : public Pad_SynthState {
public:
    Pad_ObjectState();
    Pad_ObjectState(Pad_Sound *source);
    virtual ~Pad_ObjectState();
    int Refill(int numBytes);
    virtual void  Update_state(void);
protected:
    int   _phase;
    int	  _duration;	// limit the time to play in secs. (no limit if <= 0)
    float _threshold;	// min. avg. of the square of the power of the signal
    float _scale;		// normalization constant * amp
    float _xnm1 [MAXFILTERS];	// previous imputs
    float _xnm2 [MAXFILTERS];	// previous previous inputs
    float _ynm1 [MAXFILTERS];	// previous outputs
    float _ynm2 [MAXFILTERS];	// previous previous outputs
    float _power [WINSIZE];	// powers of previous samples
    float _sum;			// power of the last WINSIZE samples

    virtual float _Gen_input(int phase);
    virtual void  _Init_state(void);
};

class Pad_ScrapeState : public Pad_ObjectState {
public:
    Pad_ScrapeState();
    Pad_ScrapeState(Pad_Sound *source);
    virtual ~Pad_ScrapeState();
    virtual void  Update_state(void);
protected:
	/* sound-specific attributes */
    int _roughness;
    int _speed;
	
	/* state */
    float _inSample;

            float _Gen_input (int);
    virtual void  _Init_state(void);
};

class Pad_ImpactState : public Pad_ObjectState {
public:
    Pad_ImpactState();
    Pad_ImpactState(Pad_Sound *source);
    virtual ~Pad_ImpactState();
    virtual void  Update_state(void);
protected:
	// sound-specific attributes 
    int _hardness;
	
	// state 
    int _impulseLength;
	
            float _Gen_input (int);
    virtual void  _Init_state(void);
};

class Pad_LiftState : public Pad_SynthState {
public:
    Pad_LiftState();
    Pad_LiftState(Pad_Sound *source);
    virtual ~Pad_LiftState();
    int Refill(int numBytes);
    virtual void Update_state(void);
protected:
    float _freq[10]; 
    int   _incFreq;
    int   _changed;
    
    virtual void _Init_state(void);
};

class Pad_MachineState : public Pad_SynthState {
public:
    Pad_MachineState();
    Pad_MachineState(Pad_Sound *source);
    virtual ~Pad_MachineState();
    int Refill(int numBytes);
    virtual void Update_state(void);
protected:
    	/* sound-specific attributes */
    int _maxWork;
    int _speed;
	/* state */
    float _modAmpPhase;	// amplitude modulator phase
    float _modFreqPhase;	// frequency modulator phase
    float _carFreqPhase;	// carrier phase
    float _modAmpInc;	// amplitude modulator increment

    virtual void _Init_state(void);
};

////////////////////////////////////////////////////////////////
//
// Utility Classes
//
//
////////////////////////////////////////////////////////////////

class Pad_SoundOption {
public:
    Pad_SoundOption();
    ~Pad_SoundOption();
    void Bounds(int low, int high);

    int                 data;
    Pad_SoundOptionType type;
};

class Pad_SoundConfig {
public:
    ~Pad_SoundConfig();
    static Pad_SoundConfig *Get_instance(void);
    void Reset();
    
        // general sound configuration options
    Pad_SoundOption *direction;
    Pad_SoundOption *gain;
    Pad_SoundOption *material;
    Pad_SoundOption *minWork;
    Pad_SoundOption *maxWork;
    Pad_SoundOption *ratio;
    Pad_SoundOption *size;
    Pad_SoundOption *speed;
        // play specific options
    Pad_SoundOption *hardness;
    Pad_SoundOption *impact;
    Pad_SoundOption *length;
    Pad_SoundOption *loop;
    Pad_SoundOption *roughness;
    Pad_SoundOption *scrape;
    Pad_SoundOption *volume;
private:
    Pad_SoundConfig();
    static Pad_SoundConfig *_instance;
};
    
////////////////////////////////////////////////////////////////
//
// Pad_AudioFormat Class
//
//
////////////////////////////////////////////////////////////////

class Pad_AudioFormat{
    friend ostream& operator<<(ostream&, Pad_AudioFormat&);
    friend class Pad_Sound;
public:
    Pad_AudioFormat();
    Pad_AudioFormat(const char*);
    virtual ~Pad_AudioFormat();
  
        ////////////////////////////////////////
        //
        // Utility functions
        //
        ////////////////////////////////////////

    virtual Pad_Bool        Open_format(char*)=0;
    virtual long            Read(void)=0;
    virtual long            Get_length(void) const;
    const   Pad_String *    Get_Filename (void) const;
    virtual Pad_SoundFormat Get_style(void) const;
    virtual long            Get_rate(void) const;
    virtual long            Get_channels(void) const;
    virtual void *          Get_private(void) const;
    virtual char *          Get_data(void) const;
    virtual long            Get_bits_per_sample(void) const;
protected: 
    Pad_String *_filename;
    long        _length;          // number of samples
    char *      _data;            // data samples
    void *      _private;         // private data field
};

////////////////////////////////////////////////////////////////
//
// SoundDevice Class
//
//
////////////////////////////////////////////////////////////////

class SoundDevice{
public:
    virtual ~SoundDevice();
    static SoundDevice *Get_instance();
    
    virtual int       Play(Pad_SoundConfig *, Pad_PlayState *)=0;
    virtual int       Stop(void)=0;
    virtual Pad_Bool  Load(Pad_Sound &sound)=0;
    virtual int       Set_volume(int &master)=0;
    virtual int       Set_volume(int &left, int &right)=0;
    virtual int       Get_volume(int &left, int &right) const=0;
            Pad_List *Get_sounds(void) const;
//    virtual Pad_AudioFormat *New_format(void)=0;
protected:
    SoundDevice();
    static SoundDevice *_instance;
    Pad_List *_sounds;
};

////////////////////////////////////////////////////////////////
//
// Pad_Sound Class
//
//
////////////////////////////////////////////////////////////////

class Pad_Sound{
    friend class Pad_SoundHandle;
    friend class Pad_SynthState;
    friend ostream& operator<<(ostream&, Pad_Sound&);
public:
    static float Db_to_amp (int db);
    static void  Init(void);
    static float DbToAmpTable[];
    static float AF_sine_float[];
    static int   CONFIG_UNSET;
    
    Pad_Sound();
    virtual ~Pad_Sound();
        // Get and Set functions
    Pad_AudioFormat *Get_format(void) const;
    Pad_Bool         Set_format(Pad_AudioFormat *);
    int              Get_length(void) const;
    int              Get_bits_per_sample(void) const;
    virtual char    *Get_data(void) const;
    virtual int      Get_rate(void) const;
    virtual int      Get_channels(void) const;
    virtual char    *Get_name(void) const;

        // configuration functions
    virtual Pad_Bool Configure(Pad_List &options);
    virtual Pad_Bool Configure(Pad_SoundConfig *config);
    virtual Pad_Bool Get_configuration(Pad_List&) const;
    virtual Pad_Bool Get_configuration(Pad_SoundOption&) const;

        // sound API functions
    virtual Pad_Bool Play(Pad_SoundConfig*);
    virtual Pad_Bool Play(Pad_SoundConfig*,Pad_PlayState*);
    virtual Pad_Bool Stop(void);
    virtual Pad_Bool Data(void);
    virtual Pad_Bool Done(Pad_PlayState*);
            Pad_Bool Remove(Pad_Sound *sound);
    virtual void Update_state (void);
protected:
    Pad_AudioFormat *_format;
    int              _gain;
    Pad_List        *_states;
};

////////////////////////////////////////////////////////////////
//
// Pad_FileSound subclass of Pad_Sound
//
//
////////////////////////////////////////////////////////////////

class Pad_FileSound : public Pad_Sound {
public:
    Pad_FileSound();
    ~Pad_FileSound();
    Pad_Bool Configure(Pad_List &options);
    Pad_Bool Get_configuration(Pad_List&) const;
    Pad_Bool Get_configuration(Pad_SoundOption&) const;
    char *   Get_name(void) const;
    Pad_Bool Play(Pad_SoundConfig*);
    Pad_Bool Load(char*);    
    void     Update_state (void);
protected:
    Pad_String _name;
};

////////////////////////////////////////////////////////////////
//
// Pad_SynthSound subclass of Pad_Sound
//
//
////////////////////////////////////////////////////////////////

class Pad_SynthSound : public Pad_Sound {
public:
    Pad_SynthSound();
    virtual ~Pad_SynthSound();
    static int SAMPLE_NUMBER;
    static int RATE;    
    static int CHANNELS;
    static int SAMPLE_SIZE;    
    static int BUFFER_SIZE;    
    static int LOW_WATER;
    static Pad_SoundFormat STYLE;
    
    virtual Pad_Bool Configure(Pad_SoundConfig *config)=0;
    virtual Pad_Bool        Configure(Pad_List &options)=0;
    virtual Pad_Bool        Get_configuration(Pad_List&) const=0;
    virtual Pad_Bool        Get_configuration(Pad_SoundOption&) const=0;
    virtual Pad_SoundFormat Get_style(void) const;
    virtual int             Get_rate(void) const;
    virtual int             Get_channels(void) const;
                            
    virtual Pad_Bool        Play(Pad_SoundConfig*)=0;
    virtual Pad_Bool        Data(void);
    virtual void Update_state (void)=0;
protected:
};

////////////////////////////////////////////////////////////////
//
// Pad_ObjectSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////

class Pad_ObjectSound : public Pad_SynthSound {
    friend class Pad_ObjectState;
public:
    Pad_ObjectSound();
    virtual ~Pad_ObjectSound();
    virtual Pad_Bool Configure(Pad_List &options);
    virtual Pad_Bool Configure(Pad_SoundConfig *config);
    virtual Pad_Bool Get_configuration(Pad_List&) const;
    virtual Pad_Bool Get_configuration(Pad_SoundOption&) const;
    virtual Pad_Bool Play(Pad_SoundConfig*);
    void  Update_state(void);
protected:
        // User configurable attributes
    int _material;
    int _speed;
    int _size;

        // internal parameters
    int	  _srate;
    float _tableInc;
    int	  _numFilters;
    float _filters [MAXFILTERS] [3];

};

////////////////////////////////////////////////////////////////
//
// Pad_MachineSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////

class Pad_MachineSound : public Pad_SynthSound {
    friend class Pad_MachineState;
public:
    Pad_MachineSound();
    virtual ~Pad_MachineSound();
    
    virtual Pad_Bool Configure(Pad_List &options);
    virtual Pad_Bool Configure(Pad_SoundConfig *config);
    virtual Pad_Bool Get_configuration(Pad_List&) const;
    virtual Pad_Bool Get_configuration(Pad_SoundOption&) const;
    virtual Pad_Bool Play(Pad_SoundConfig*);
    void  Update_state(void);
protected:
        // User configurable attributes
    int _minWork;
    int _maxWork;
    int _size;
    int _ratio;
    int _speed;

        // internal parameters
    float _tableInc;	// increment to go through sine table once/second
    float _carFreq;	// carrier frequency
    float _modFreq;	// modulator frequency
    float _modFreqInc;	// increment for modulator frequency
    
};

////////////////////////////////////////////////////////////////
//
// Pad_LiftSound subclass of Pad_SynthSound
//
//
////////////////////////////////////////////////////////////////

class Pad_LiftSound : public Pad_SynthSound {
    friend class Pad_LiftState;
public:
    Pad_LiftSound();
    virtual ~Pad_LiftSound();
    virtual Pad_Bool Configure(Pad_List &options);
    virtual Pad_Bool Configure(Pad_SoundConfig *config);
    virtual Pad_Bool Get_configuration(Pad_List&) const;
    virtual Pad_Bool Get_configuration(Pad_SoundOption&) const;
    virtual Pad_Bool Play(Pad_SoundConfig*);
    virtual void  Update_state(void);
protected:
        // User configurable attributes
    int _size;
    int _direction;
    int _speed;

        // internal parameters
    float        _tableInc; // increment to go through sine table once/second
    float        _lnfr;	    // base frequency
    float        _lnSize;
    float        _baseFreq;
    float        _floor;
    int          _inc;
    float        _aUp1, _bUp1, _aUp2, _bUp2;
    float        _max1stPartial;
    float        _xPointMax;
    int          _maxSize;
    static float _rapPartial;
    static int   _nbPartial;
    static float _pointMax;

};

typedef short sample16;		// 16 bit word
typedef char sample8;		// 8  bit word

// MAG8  is 2^7 -1 (8 bits linear)
// MAG16 is 2^15 -1 (16 bits linear)
#define MAG8  122.0
#define MAG16 32767.0

#define OUTPUT_SAMPLE(sample) \
if (sample > 1.0) { \
	sample = 1.0; \
} else if (sample < -1.0) { \
	sample = -1.0; \
} \
switch (Pad_SynthSound::STYLE) { \
case LinearSigned16LSB: \
    * ((sample16 *) _write) = (sample16) (sample * MAG16); \
    _write += 2; \
    break; \
case LinearSigned8: \
    * ((sample8 *) _write) = (sample8) (sample * MAG8); \
    _write++; \
    break; \
default : \
    cerr << "sound format not supported by OUTPUT_SAMPLE" << endl; \
    return -1; \
} 


#endif /* PAD_SOUND_H */
