/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "pad-string.h"
#include "tkwin.h"
#include "view.h"
#include "pad-tcl.h"
#include "pad-tcl-sound.h"
#include "text.h"
#include "point.h"
#include "list.h"
#include "pad.h"
#include "html.h"
#include "image.h"
#include "callback.h"
#include "global.h"
#include "api.h"

#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef PAD_UNIX
#  include <unistd.h>
#endif

/*
   This file supplies the basic Tcl interface to the Pad widget.  Pad is
   designed to look and act much like an Tk canvas widget.  While many of
   the commands of the canvas exist on the Pad with the same or similar 
   syntax, there are also many other commands designed specifically for the
   interactivity of the Pad.
*/

static int _command(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

//
// This is the implementation of the pad widget command for Tcl.
// It has many options which are used to control, configure, and query the Pad.
// The basic options follow, which each have several sub-options.  The sub-options
// are all parsed by subroutines.
//   * control
//   * configure
//
int Pad_Command(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) 
{
    char c;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    int result = TCL_OK;

    if (argc < 2) 
      {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " ?options?\"", NULL);
	return TCL_ERROR;
      }

    Tcl_Preserve((ClientData)tkwin);

    c = argv[1][0];
    
    if ((c == 'c') && (!strncmp(argv[1], "conf", 4)))           // Configure
      {
	  result = Pad_Configure(clientData, interp, argc-2, argv+2);
      }
    else		                                        // Everything else
      {
	  result = _command(clientData, interp, argc-1, argv+1);
      }

    Tcl_Release((ClientData)tkwin);
    return result;
}

////////////////////////////////////////////////////////
//
// All routines except configure
//
////////////////////////////////////////////////////////

static int
_command(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) 
{
    char c;
    int result = -1;		// Invalid result, it must be set
    int len;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_HashTableIterator hi;
    Pad_Language *language;
    Pad_Bool first;

    if (argc == 0)
      {
	  TCL_ERR(interp, "wrong # args");         //  args are (option ?value?)
      }

    // clear the global error string
    Pad_errorString = "";

    len = strlen(argv[0]);
    c = argv[0][0];
    switch (c) 
      {
	case 'a':
	  if (!strncmp(argv[0], "addgroupmember", MAX(4, len)))
	    {
		result = Padtcl_addgroupmember(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "addmodifier", MAX(4, len)))
	    {
		result = Padtcl_addmodifier(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "addoption", MAX(4, len)))
	    {
		result = Padtcl_addoption(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "addtag", MAX(5, len)))
	    {
		result = Padtcl_addtag(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "addtype", MAX(5, len)))
	    {
		result = Padtcl_addtype(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "allocborder", MAX(6, len)))
	    {
		result = Padtcl_allocborder(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "alloccolor", MAX(6, len)))
	    {
		result = Padtcl_alloccolor(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "allocimage", MAX(6, len)))
	    {
		result = Padtcl_allocimage(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "animate", MAX(2, len)))
	    {
		result = Padtcl_animate(clientData, interp, argc, argv);
	    }
	  break;
	case 'b':
	  if (!strncmp(argv[0], "bbox", MAX(2, len)))
	    {
		result = Padtcl_bbox(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "bind", MAX(5, len)))
	    {
		result = Padtcl_bind(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "bindtags", MAX(5, len)))
	    {
		result = Padtcl_bindtags(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "border", MAX(2, len)))
	    {
		result = Padtcl_border(clientData, interp, argc, argv);
	    }
	  break;
	case 'c':
	  if (!strncmp(argv[0], "cache", MAX(2, len)))
	    {
		result = Padtcl_cache(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "center", MAX(7, len)))
	    {
		result = Padtcl_center(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "centerbbox", MAX(7, len)))
	    {
		result = Padtcl_center_bbox(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "clock", MAX(2, len)))
	    {
		result = Padtcl_clock(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "color", MAX(3, len)))
	    {
		result = Padtcl_color(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "coords", MAX(3, len)))
	    {
		result = Padtcl_coords(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "create", MAX(2, len)))
	    {
		result = Padtcl_create(clientData, interp, argc, argv);
	    }
	  break;
	case 'd':
	  if (!strncmp(argv[0], "damage", MAX(2, len)))
	    {
		result = Padtcl_damage(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "delete", MAX(7, len)))
	    {
		result = Padtcl_delete(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "deletemodifier", MAX(7, len)))
	    {
		result = Padtcl_deletemodifier(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "dtag", MAX(2, len)) ||
		   !strncmp(argv[0], "deletetag", MAX(7, len)))
	    {
		result = Padtcl_dtag(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "drawborder", MAX(5, len)))
	    {
		result = Padtcl_draw_border(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "drawline", MAX(5, len)))
	    {
		result = Padtcl_draw_line(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "drawimage", MAX(5, len)))
	    {
		result = Padtcl_draw_image(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "drawpolygon", MAX(5, len)))
	    {
		result = Padtcl_draw_polygon(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "drawtext", MAX(5, len)))
	    {
		result = Padtcl_draw_text(clientData, interp, argc, argv);
	    }
	  break;
	case 'f':
          if (!strncmp(argv[0], "find", MAX(3, len)))
            {
                result = Padtcl_find(clientData, interp, argc, argv);
            }
          else if (!strncmp(argv[0], "focus", MAX(3, len)))
	    {
		result = Padtcl_focus(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "font", MAX(3, len)))
	    {
		result = Padtcl_font(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "freeborder", MAX(5, len)))
	    {
		result = Padtcl_freeborder(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "freecolor", MAX(5, len)))
	    {
		result = Padtcl_freecolor(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "freeimage", MAX(5, len)))
	    {
		result = Padtcl_freeimage(clientData, interp, argc, argv);
	    }
	  break;
	case 'g':
	  if (!strncmp(argv[0], "getdate", MAX(4, len)))
	    {
		result = Padtcl_getdate(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getgroup", MAX(4, len)))
	    {
		result = Padtcl_getgroup(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getlevel", MAX(4, len)))
	    {
		result = Padtcl_getlevel(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getmag", MAX(5, len)))
	    {
		result = Padtcl_getmag(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getmodifier", MAX(5, len)))
	    {
		result = Padtcl_getmodifier(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getpads", MAX(5, len)))
	    {
		result = Padtcl_getpads(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getportals", MAX(5, len)))
	    {
		result = Padtcl_getportals(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getsize", MAX(4, len)))
	    {
		result = Padtcl_getsize(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "gettags", MAX(5, len)))
	    {
		result = Padtcl_gettags(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "gettextbbox", MAX(5, len)))
	    {
		result = Padtcl_gettextbbox(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getview", MAX(4, len)))
	    {
		result = Padtcl_getview(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "getzoom", MAX(4, len)))
	    {
		result = Padtcl_getzoom(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "grab", MAX(4, len)))
	    {
		result = Padtcl_grab(clientData, interp, argc, argv);
            }
	  else if (!strncmp(argv[0], "grid", MAX(4, len)))
	    {
		result = Padtcl_grid(clientData, interp, argc, argv);
            }
	  break;
	case 'h':
	  if (!strncmp(argv[0], "hastag", MAX(2, len)))
	    {
		result = Padtcl_hastag(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "html", MAX(2, len)))
	    {
		result = Padtcl_html(clientData, interp, argc, argv);
	    }
	  break;
	case 'i':
	  if (!strncmp(argv[0], "image", MAX(2, len)))
	    {
		result = Padtcl_image(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "info", MAX(2, len)))
	    {
		result = Padtcl_info(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "isvisible", MAX(2, len)))
	    {
		result = Padtcl_isvisible(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "itemconfigure", MAX(2, len)) ||
		   !strncmp(argv[0], "ic", MAX(2, len)))
	    {
		result = Padtcl_itemconfigure(clientData, interp, argc, argv);
	    }
	  break;
	case 'l':
	  if (!strncmp(argv[0], "layer", MAX(4, len)))
	    {
		result = Padtcl_layer(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "layout", MAX(4, len)))
	    {
		result = Padtcl_layout(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "line2spline", MAX(2, len)))
	    {
		result = Padtcl_line2spline(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "lower", MAX(2, len)))
	    {
		result = Padtcl_lower(clientData, interp, argc, argv);
	    }
	  break;
	case 'm':
	  if (!strncmp(argv[0], "modifier", MAX(3, len)))
	    {
		result = Padtcl_modifier(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "moveto", MAX(3, len)))
	    {
		result = Padtcl_move_to(clientData, interp, argc, argv);
	    }
	  break;
	case 'n':
	  if (!strncmp(argv[0], "noise", MAX(2, len)))
	    {
		result = Padtcl_noise(clientData, interp, argc, argv);
	    }
	  break;
	case 'p':
	  if (!strncmp(argv[0], "padxy", MAX(3, len)))
	    {
		result = Padtcl_padxy(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "pick", MAX(2, len)))
	    {
		result = Padtcl_pick(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "popcoordframe", MAX(2, len)))
	    {
		result = Padtcl_pop_coord_frame(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "printinfo", MAX(6, len)))
	    {
		tkwin->view->Print_debugging_info();
		result = TCL_OK;
	    }
	  else if (!strncmp(argv[0], "printtree", MAX(6, len)))
	    {
		tkwin->view->pad->Print_tree(*tkwin->view);
		result = TCL_OK;
	    }
	  else if (!strncmp(argv[0], "pushcoordframe", MAX(2, len)))
	    {
		result = Padtcl_push_coord_frame(clientData, interp, argc, argv);
	    }
	  break;
	case 'r':
	  if (!strncmp(argv[0], "raise", MAX(3, len)))
	    {
		result = Padtcl_raise(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "random", MAX(3, len)))
	    {
		result = Padtcl_random(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "read", MAX(3, len)))
	    {
		result = Padtcl_read(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "removegroupmember", MAX(3, len)))
	    {
		result = Padtcl_removegroupmember(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "render", MAX(6, len)))
	    {
		result = Padtcl_render(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "renderitem", MAX(6, len)))
	    {
		result = Padtcl_render_item(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "resetcoordframe", MAX(3, len)))
	    {
		result = Padtcl_reset_coord_frame(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "rotate", MAX(2, len)))
	    {
		result = Padtcl_rotate(clientData, interp, argc, argv);
	    }
	  break;
	case 's':
	  if (!strncmp(argv[0], "scale", MAX(2, len)))
	    {
		result = Padtcl_scale(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setcapstyle", MAX(5, len)))
	    {
		result = Padtcl_setcapstyle(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setcolor", MAX(5, len)))
	    {
		result = Padtcl_setcolor(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setfont", MAX(8, len)))
	    {
		result = Padtcl_setfont(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setfontheight", MAX(8, len)))
	    {
		result = Padtcl_setfontheight(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setid", MAX(4, len)))
	    {
		result = Padtcl_setid(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setjoinstyle", MAX(4, len)))
	    {
		result = Padtcl_setjoinstyle(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setlanguage", MAX(4, len)))
	    {
		result = Padtcl_setlanguage(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setlinewidth", MAX(4, len)))
	    {
		result = Padtcl_setlinewidth(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "setmodifier", MAX(4, len)))
	    {
		result = Padtcl_setmodifier(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "settoplevel", MAX(4, len)))
	    {
		result = Padtcl_settoplevel(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "sound", MAX(2, len)))
	    {
              result = Padtcl_sound(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "slide", MAX(2, len)))
	    {
		result = Padtcl_slide(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "spline2line", MAX(2, len)))
	    {
		result = Padtcl_spline2line(clientData, interp, argc, argv);
	    }
	  break;
	case 't':
	  if (!strncmp(argv[0], "text", MAX(2, len)))
	    {
		result = Padtcl_text(clientData, interp, argc, argv);
	    }
          else if (!strncmp(argv[0], "tree", MAX(4, len)))
            {
                result = Padtcl_tree(clientData, interp, argc, argv);
            }
	  else if (!strncmp(argv[0], "type", MAX(2, len)))
	    {
		result = Padtcl_type(clientData, interp, argc, argv);
	    }
	  break;
	case 'u':
	  if (!strncmp(argv[0], "update", MAX(2, len)))
	    {
		result = Padtcl_update(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "urlfetch", MAX(9, len)))
	    {
		result = Padtcl_urlfetch(clientData, interp, argc, argv);
	    }
	  break;
	case 'w':
	  if (!strncmp(argv[0], "warp", MAX(2, len)))
	    {
		result = Padtcl_warp(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "windowshape", MAX(2, len)))
	    {
		result = Padtcl_windowshape(clientData, interp, argc, argv);
	    }
	  else if (!strncmp(argv[0], "write", MAX(2, len)))
	    {
		result = Padtcl_write(clientData, interp, argc, argv);
	    }
	  break;
	case 'z':
	  if (!strncmp(argv[0], "zoom", MAX(1, len)))
	    {
		result = Padtcl_zoom(clientData, interp, argc, argv);
	    }
	  break;
      }
    if (result == -1) {
	DOTABLE(hi, Pad_languages, Pad_Language, language) {
	    if (language->commandProc && (language->cmdName == argv[0])) {
		result = language->commandProc(clientData, interp, argc, argv);
	    }
	}

	if (result == -1) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "unknown command: ", argv[0], "\n",
     "Try: addgroupmember, addoption, addtag, addtype, animate, bbox, bind, bindtags, border, cache,\n",
     "     center, centerbbox, clock, color, configure, coords, create, damage, delete, deletetag,\n",
     "     find, focus, font, getdate, getgroup, getpads, getsize, gettags, getview, getzoom, grab, grid,\n",
     "     hastag, html, image, info, isvisible, itemconfigure, layer, layout, line2spline, lower, modifier,\n",
     "     moveto, noise, padxy, pick, popcoordframe, printtree, pushcoordframe, raise, random, read,\n",
     "     removegroupmember, render, renderitem, resetcoordframe, rotate, scale, setid, setlanguage,\n",
     "     settoplevel, slide, sound, spline2line, text, tree, type, update, urlfetch, warp, windowshape,\n",
     "     write, zoom,\n",
     NULL);

	    first = TRUE;
	    DOTABLE(hi, Pad_languages, Pad_Language, language) {
		if (language->commandProc != NULL) {
		    if (first) {
			first = FALSE;
			Tcl_AppendResult(interp, "     ", NULL);
		    } else {
			Tcl_AppendResult(interp, ", ", NULL);
		    }
		    Tcl_AppendResult(interp, language->cmdName.Get(), NULL);
		}
	    }
	    
	    return(TCL_ERROR);
	}
    }
    
    return(result);
}


////////////////////////////////////////////////////////
//
// Configure routines
//
////////////////////////////////////////////////////////

int Pad_Configure(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) 
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    int result = TCL_OK;
    
    if (argc == 0) {
	result = Tk_ConfigureInfo(interp, tkwin->Get_tkwindow(), Pad_configSpecs,
				  (char *)tkwin, (char *)NULL, 0);
    } else if (argc == 1) {
	result = Tk_ConfigureInfo(interp, tkwin->Get_tkwindow(), Pad_configSpecs,
				  (char *)tkwin, argv[0], 0);
    } else {
	result = Pad_Configure_widget(clientData, interp, argc, argv);
    }
    
    return(result);
}

//
// Case insensitive version of strstr
//

char *casestrstr(char *s1, char *s2) 
{
    char *p1, *p2;

    if (*s2 == 0) {
	return((char *)0);
    }
    while (*s1) {
	p1 = s1;
	p2 = s2;
	do {
	    if (*p2 == 0) {
		return(s1);
	    }
	    if (tolower(*p1) != tolower(*p2)) {
		break;
	    }
	    p1++;
	    p2++;
	} while (1);
	s1++;
    }

    return((char *)0);
}

int 
Pad_GetSearchType(Tcl_Interp *interp, int argc, char *string, int &searchType)
{
    if (!strcmp(string, "all")) {
	if (argc == 1) {
	    searchType = PAD_SEARCH_ALL;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"all\"");
	}
    } else if (!strcmp(string, "above")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_ABOVE;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"above tagorid\"");
	}
    } else if (!strcmp(string, "below")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_BELOW;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"below tagorid\"");
	}
    } else if (!strcmp(string, "closest")) {
	if ((argc >= 3) && (argc <= 5)) {
	    searchType = PAD_SEARCH_CLOSEST;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"closest x y ?halo ?start??\"");
	}
    } else if (!strcmp(string, "enclosed")) {
	if (argc == 5) {
	    searchType = PAD_SEARCH_ENCLOSED;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"enclosed x1 y1 x2 y2\"");
	}
    } else if (!strcmp(string, "overlapping")) {
	if (argc == 5) {
	    searchType = PAD_SEARCH_OVERLAPPING;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"overlapping x1 y1 x2 y2\"");
	}
    } else if (!strcmp(string, "withlayer")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHLAYER;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withlayer layer\"");
	}
    } else if (!strcmp(string, "withinfo")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHINFO;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withinfo infoString\"");
	}
    } else if (!strcmp(string, "withsticky")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHSTICKY;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withsticky stickyType\"");
	}
    } else if (!strcmp(string, "withtag")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHTAG;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withtag tagOrId\"");
	}
    } else if (!strcmp(string, "withtext")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHTEXT;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withtext text\"");
	}
    } else if (!strcmp(string, "withname")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHNAME;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withname name\"");
	}
    } else if (!strcmp(string, "withtype")) {
	if (argc == 2) {
	    searchType = PAD_SEARCH_WITHTYPE;
	} else {
	    TCL_ERR(interp, "wrong # args: searchCommand should be \"withtype type\"");
	}
    } else {
	Tcl_AppendResult(interp, "unknown searchCommand: ", string, "\n",
			 "try all, withinfo, withlayer, withname, withsticky, withtag, withtext, withtype, or enclosed, overlapping, ",
			 NULL);
	return(TCL_ERROR);
    }
    
    return(TCL_OK);
}



////////////////////////////////////////////////////////
//
// TCL Object in a Pad (i.e. procedural render)
//
////////////////////////////////////////////////////////

#define TCL_DEFAULT_BBOX     NULL

Pad_Tcl::~Pad_Tcl() 
{
    Generate_delete();
}

Pad_Tcl::Pad_Tcl(Pad *pad) :
Pad_Object(pad)
{
    _type = PAD_TCL;
    tclFlags = PAD_NO_MASK;
    interp = NULL;
    win = NULL;
}

//
// Gets called with string args when object is created.  This routine must 
// process any arguments up to, but not including the options.  (argc, argv) 
// point to parameters on the creation line just after 'create <type>'.
// This returns the number of args processed.
//
int
Pad_Tcl::Create_obj_args(int, char **)
{
    interp = pad->view->win->interp;
    win = pad->view->win;

    return(0);
}

//
// Set the object's <option> to the value specified by <sval>.
// Note that each option has a specific setter and getter method
// if the correct argument type is available.
//
Pad_Bool
Pad_Tcl::Set_option(Pad_Option *option, char *sval)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BBOX:
	Set_bboxscript(sval);
	break;
      default:
	rc = Pad_Object::Set_option(option, sval);
	break;
    }

    return(rc);
}

//
// Set the object's <option> to the default value.
//
Pad_Bool
Pad_Tcl::Set_option_default(Pad_Option *option)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BBOX:
	Set_bboxscript_default();
	break;
      default:
	rc = Pad_Object::Set_option_default(option);
	break;
    }

    return(rc);
}

//
// Return the value of the object's <option> as a string
// in <result>.
//
Pad_Bool
Pad_Tcl::Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BBOX:
	modified = (tclFlags & TCL_BBOX_SET) ? TRUE : FALSE;
	result = Get_bboxscript();
	break;
      default:
	rc = Pad_Object::Get_option(option, result, modified);
	break;
    }

    return(rc);
}

//
// Setters and getters for bbox
//
void
Pad_Tcl::Set_bboxscript(char *new_bbScript)
{
    tclFlags |= TCL_BBOX_SET;
    bbScript = new_bbScript;
    Update();
}

void
Pad_Tcl::Set_bboxscript_default(void)
{
    Set_bboxscript(TCL_DEFAULT_BBOX);
    tclFlags &= ~TCL_BBOX_SET;
}

char *
Pad_Tcl::Get_bboxscript(void)
{
    if (bbScript.Length() > 0) {
	return bbScript.Get();
    } else {
	return "";
    }
}

void Pad_Tcl::Compute_bounding_box(void) 
{
    int result;
    int rc;
    int listargc;
    char **listargv;
    double value[4];
    Pad_String *save_result;
	  
    if ((bbScript.Length() > 0) && interp)
      {
	  save_result = new Pad_String(interp->result);
	  result = Tcl_GlobalEval(interp, bbScript.Get());
	  if (result == TCL_ERROR) {
	      Tcl_AddErrorInfo(interp, "\n    (Pad -bb)");
	      Pad_Background_error(interp);
	  }
	  
	  rc = Tcl_SplitList(interp, interp->result, &listargc, &listargv);
	  
	  if (listargc != 4) 
	    {
		cerr << "Warning: Tcl bounding box must have exactly 4 elements" << endl;
	    }
	  else 
	    {
		value[XMIN] = ATOXF(win, listargv[0]);
		value[YMIN] = ATOYF(win, listargv[1]);
		value[XMAX] = ATOXF(win, listargv[2]);
		value[YMAX] = ATOYF(win, listargv[3]);
		
		if (!(optionFlags & PAD_WIDTH_SET)) {
                    Set_bbox_xmin(value[XMIN]);
                    Set_bbox_xmax(value[XMAX]);
		}
		if (!(optionFlags & PAD_HEIGHT_SET)) {
                    Set_bbox_ymin(value[YMIN]);
                    Set_bbox_ymax(value[YMAX]);
		}
	    }

	  if (rc == TCL_OK)
	    ckfree((char *) listargv);

	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp, save_result->Get(), NULL);
	  delete save_result;
      }

    Pad_Object::Compute_bounding_box();
}

//
// Render the Tcl object by calling its render script
// (if there is one).
//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
Pad_Tcl::Render(void)
{
    Pad_View *view;
    Pad_Callback *renderScript = Get_renderscript();

    view = (Pad_View *)Pad_prc->views.Last();
    if (renderScript) {
	renderScript->Eval();
    }

    return(TRUE);
}

//
// Tcl objects always get clipped
//
Pad_Bool Pad_Tcl::Check_render_clipped()
{
    return(TRUE);
}
