/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef PAD_TCL_H
#define PAD_TCL_H 1

#include "defs.h"
#include "object.h"
#include "view.h"


#define PAD_FIND_AND              0
#define PAD_FIND_ERROR            1
#define PAD_FIND_OR               2
#define PAD_FIND_REGEXP           3
#define PAD_FIND_EXACT            4
#define PAD_FIND_GLOB             5
#define PAD_SEARCH_ALL            6
#define PAD_SEARCH_ABOVE          7
#define PAD_SEARCH_BELOW          8
#define PAD_SEARCH_CLOSEST        9
#define PAD_SEARCH_ENCLOSED       10
#define PAD_SEARCH_OVERLAPPING    11
#define PAD_SEARCH_WITHTAG        12
#define PAD_SEARCH_WITHTEXT       13
#define PAD_SEARCH_WITHNAME       14
#define PAD_SEARCH_WITHTYPE       15
#define PAD_SEARCH_WITHLAYER      16
#define PAD_SEARCH_WITHINFO       17
#define PAD_SEARCH_WITHSTICKY     18


//
// Macros to support generating a Tcl result or error.
//
#define TCL_ERR(interp, string) {Tcl_ResetResult(interp); \
			         Tcl_AppendResult(interp, string, (char *)NULL); \
				 return(TCL_ERROR);}
#define TCL_RESULT(interp, string) {Tcl_ResetResult(interp); \
				    Tcl_AppendResult(interp, string, (char *)NULL);}
				 
//
// Support routines
//
int Padtcl_addtag(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_addgroupmember(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_addmodifier(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_addoption(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_addtype(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_allocborder(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_alloccolor(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_allocimage(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_animate(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_bbox(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_bind(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_bindtags(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_border(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_cache(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_center(ClientData, Tcl_Interp *interp, int argc, char **argv);
int Padtcl_center_bbox(ClientData, Tcl_Interp *interp, int argc, char **argv);
int Padtcl_clock(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_color(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_coords(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_create(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_damage(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_delete(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_deletemodifier(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_dtag(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_draw_border(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_draw_line(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_draw_image(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_draw_polygon(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_draw_text(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_find(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_focus(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_font(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_freeborder(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_freecolor(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_freeimage(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getdate(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getgroup(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getlevel(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getmag(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getmodifier(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getpads(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getportals(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getsize(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_gettags(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_gettextbbox(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getview(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_getzoom(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_grab(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_grid(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_hastag(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_html(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_image(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_info(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_isvisible(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_itemconfigure(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_layer(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_layout(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_line2spline(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_lower(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_modifier(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_move_to(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_noise(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_padxy(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_pick(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_push_coord_frame(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_pop_coord_frame(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_raise(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_random(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_read(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_render(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_render_item(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_reset_coord_frame(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_removegroupmember(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_rotate(ClientData, Tcl_Interp *, int argc, char **argv); 
int Padtcl_scale(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setcapstyle(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setcolor(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setfont(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setfontheight(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setid(ClientData, Tcl_Interp *interp, int argc, char **argv);
int Padtcl_setjoinstyle(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setlanguage(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_settoplevel(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setlinewidth(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_setmodifier(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_slide(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_spline2line(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_text(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_tree(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_type(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_update(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_urlfetch(ClientData, Tcl_Interp *interp, int argc, char **argv);
int Padtcl_warp(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_windowshape(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_write(ClientData, Tcl_Interp *, int argc, char **argv);
int Padtcl_zoom(ClientData, Tcl_Interp *, int argc, char **argv);
Pad_Bool   Read_script(istream *fp, Pad_StreamData &sd, Tcl_Interp *interp=NULL);
int  Pad_GetSearchType(Tcl_Interp *interp, int argc, char *string, int &searchType);

class HTML;
class HTML_EntryTagImage;
class Pad_String;
class Pcode;
class Kpl;
class Pad_TkWin;

void Pad_AppendOption(Tcl_Interp *interp, Pad_Object *obj, Pad_Option *option, 
		      Pad_String &result, Pad_Bool nondefault_only, Pad_Bool braces);
void Pad_AppendTclElement(Tcl_Interp *interp, char *element, Pad_String &result);
int  Pad_ItemConfigure(Pad_TkWin *tkwin, Tcl_Interp *interp, Pad_Object *obj, 
			      int argc, char **argv);
char *casestrstr(char *s1, char *s2);

#define TCL_BBOX_SET    (1<<0)  // Set if -bbox has been configured

class Pad_Tcl : public Pad_Object 
{
  public:
    unsigned char tclFlags;	// OR'd combination of flags
    Pad_String    bbScript;
    Tcl_Interp *  interp;
    Pad_Win *     win;
    
    virtual void     Compute_bounding_box(void);
    virtual Pad_Bool Check_render_clipped(void);
    virtual int      Create_obj_args(int argc, char **argv);   // Called when an item is created with string args
    virtual char *   Get_bboxscript(void);
    virtual Pad_Bool Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified);
    virtual Pad_Bool Render(void);
            void     Set_bboxscript(char *bbScript);
            void     Set_bboxscript_default(void);
    virtual Pad_Bool Set_option(Pad_Option *option, char *value); 
    virtual Pad_Bool Set_option_default(Pad_Option *option);

    virtual ~Pad_Tcl();
    Pad_Tcl(Pad *pad);

    PAD_TYPE("tcl");
};

#define KPL_BBOX_SET    (1<<0)  // Set if -bbox has been configured

class Pad_Kpl : public Pad_Object
{
  public:
    unsigned char kplFlags;	// OR'd combination of flags
    Kpl *         kpl;		// KPL interpreter
    Tcl_Interp *  interp;	// Tcl interpreter
    Pcode *       renderPcode;	// compiled render-thyself script
    Pcode *       bbPcode;	// compiled bounding-box script

    int kpl_script (char *script);	// compile into renderPcode
    int bb_script (char *script);	// compile into bbPcode
    
    virtual Pad_Bool Check_render_clipped(void);
    virtual void     Compute_bounding_box(void);
    virtual Pad_Bool Continue_refinement(void);
    virtual int      Create_obj_args(int argc, char **argv);   // Called when an item is created with string args
    virtual char *   Get_bboxscript(void);
    virtual Pad_Bool Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified);
    virtual Pad_Callback * Get_renderscript(void);
    virtual Pad_Bool Render(void);
            Pad_Bool Set_bboxscript(char *bbScript);
            void     Set_bboxscript_default(void);
    virtual Pad_Bool Set_option(Pad_Option *option, char *value); 
    virtual Pad_Bool Set_option_default(Pad_Option *option);
    virtual Pad_Bool Set_string_renderscript(char *new_renderscript);
    virtual void     Set_string_renderscript_default(void);

    virtual ~Pad_Kpl();
    Pad_Kpl(Pad *pad);

    PAD_TYPE("kpl");
};

#endif
