/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

// ************************************************************************
// Pad_StreamItem class implementation
//
// Implements operations to manipulate Pad_StreamItem objects
// ************************************************************************

#include <string.h> // uses memcpy()
#include <iostream.h>

#include "streamdata.h"


// ************************************************************************
// constructors, etc.
// ************************************************************************

Pad_StreamItem::Pad_StreamItem()
{
  type = _no_type;
  dataLength = infoLength = 0;
  info = NULL;
}

Pad_StreamItem::Pad_StreamItem(const Pad_StreamItem& m)
{
  type = m.type;
  dataLength = m.dataLength;
  infoLength = m.infoLength;
  if (infoLength) {
    info = new unsigned char[infoLength];
    memcpy(info, m.info, infoLength);
  }
  else
    info=NULL;
}

Pad_StreamItem& Pad_StreamItem::operator= (const Pad_StreamItem& m)
{
  if (this == &m) return *this;
  if (info) delete [] info;
  type = m.type;
  dataLength = m.dataLength;
  infoLength = m.infoLength;
  if (infoLength) {
    info = new unsigned char[infoLength];
    memcpy(info, m.info, infoLength);
  }
  else
    info = NULL;
  return *this;
}

Pad_StreamItem::~Pad_StreamItem()
{
  if (info) delete [] info;
}


int Pad_StreamItem::marshal(Pad_StreamData& m) const
{
  return marshal_full(m);
}

int Pad_StreamItem::reconstruct(Pad_StreamData& m)
{
  int err = reconstruct_full(m);
  return err;
}

// ************************************************************************
// _full methods
// ************************************************************************

int Pad_StreamItem::marshal_full(Pad_StreamData& m) const
{
  Pad_StreamItem* copy = new Pad_StreamItem(*this);
  m.marshal(copy);
  return noErr;
}

int Pad_StreamItem::reconstruct_full(Pad_StreamData& m)
{
  Pad_StreamItem *item = NULL;
  int err = m.reconstruct(item);
  *this = *item;
  return err;
}

// ************************************************************************
// iostream operators
// ************************************************************************

//
// Write the item.  This is the format:
// If item is a single datum then:
//    <bytes>;           // contents depend on item type and encoding used
// If item is a vector then:
//    short dataLength;  // number of datums in the vector
//    <bytes>;           // contents depend on item type and encoding used
//

ostream&
Pad_StreamItem::Write(ostream &os, Pad_StreamData *sd)
{
  // Item data length if an array
  if ((dataLength > 1) || (type == (_char | _vector))) {
    unsigned char *outbuf=NULL;
    unsigned int outbuf_len=0;
    int length=dataLength;
    sd->drep->EncodeArray(&length, 1, outbuf, outbuf_len);
    os.write(outbuf, outbuf_len);
    if (sd->drep->Type() == PAD_CHAR_FORMAT) os << ' ';
    delete[] outbuf;
    outbuf_len = 0;
  }
  
  // Item data
  os.write((char*)info, infoLength);
  if (sd->drep->Type() == PAD_CHAR_FORMAT) os << endl;

  return os;
}

//
// Read dataLength datums from input.  Assume type and dataLength
// have been set.
//
istream&
Pad_StreamItem::Read(istream &is, Pad_StreamData *sd)
{
  int n;

  if ((n = sd->drep->ReadData(is, type, dataLength, infoLength, info)) > 0) {
    sd->bytesRead += n;
  }

  return is;
}

//
// Compute the total number of bytes when writting the item.
//
int Pad_StreamItem::OutputLength(Pad_DataRep *drep)
{
  int sum=0;

  // Item data length if an array
  if ((dataLength > 1) || (type == (_char | _vector))) {
    unsigned char *outbuf=NULL;
    unsigned int outbuf_len=0;
    int length=dataLength;
    drep->EncodeArray(&length, 1, outbuf, outbuf_len);
    sum = outbuf_len;
    if (drep->Type() == PAD_CHAR_FORMAT) sum++; //for the space
    delete[] outbuf;
    outbuf_len = 0;
  }
  
  // Item data
  sum += infoLength;
  if (drep->Type() == PAD_CHAR_FORMAT) sum++; // for the newline
  
  return sum;
}
