/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

// Nothing in this file has anything to do with Tk, just Tcl.

#include "defs.h"
#include "pad-tcl.h"
#include "object.h"
#include "imagedata.h"
#include "view.h"
#include "tkkpl.h"
#include "renderer.h"
#include "text.h"
#include "global.h"
#include "win.h"
#include "../kpl/src/kpl.h"

#include <string.h>
#include <stdlib.h>

// Nasty unchecked string length limit (per "symbol"---it's not THAT bad)
#define STRLEN 100000	/* probably best to sync with STRLEN in kpl dir */

extern "C" {
extern void install_kpl_g(void);
extern void install_kpl_g_raw(void);
extern void install_kpl_graphics(void);
extern void install_kpl_lib(void);
extern void install_kpl_matrix(void);
extern void install_kpl_physics(void);
extern void install_kpl_pump(void);
extern void install_kpl_sumvec(void);
extern void install_kpl_tcl(void);
extern void install_kpl_vec(void);
extern void install_kpl_w(void);
extern int kpl_get(int id, float *data);
extern void kpl_set(int id, float *data, int len);
}

static char input_file[256] = "";

char *kpl_input_file()
{
	return input_file;
}

extern "C" void set_input_file(char *str);
void set_input_file(char *str)
{
	strcpy(input_file, str);
}


// All Kpl interpreters that you create are currently merely pointers to
// the one true Kpl interpreter.  That's okay, everybody just shares the
// one namespace.

int Kpl::envinit = 0;

extern Pad_Bool refinement_needed;
extern Pad_Object *kpl_render_object;

static void kpl_refine(void)
{
    refinement_needed = TRUE;
}

static void kpl_interrupted(void)
{
    float a[16];
    a[0] = !(Pad_renderer->Is_interrupted(Pad_prc->win));
    kpl_push(a, 1);
}

static void kpl_drawline(void)
{
    float a[16];
    int len;

    len = kpl_pop(a);
    if (len > 2 && !(len & 1)) {
	Pad_renderer->Draw_line(len >> 1, (Pad_Point *)a, FALSE);
    } else {
	fprintf(stderr, "kpl_drawline called with bad length vector: %d\n", len);
    }
}

static void kpl_drawpolygon(void)
{
    float a[16];
    int len;

    len = kpl_pop(a);
    if (len > 2 && !(len & 1)) {
	Pad_renderer->Draw_polygon(len >> 1, (Pad_Point *)a);
    } else {
	fprintf(stderr, "kpl_drawline called with bad length vector: %d\n", len);
    }
}

static void kpl_setabslinewidth(void)
{
    float a[16];
    kpl_pop(a);
    Pad_renderer->Set_abs_line_width(a[0]);
}

static void kpl_setlinewidth(void)
{
    float scale;
    float a[16];
    kpl_pop(a);

    scale = Pad_renderer->Get_mag();
    if (kpl_render_object) {
	scale *= kpl_render_object->transform.Get_scale();
    }
    Pad_renderer->Set_abs_line_width(scale * a[0]);
}

static void kpl_setcolor(void)
{
    float a[16];
    int id;
    Pad_Color *padColor;

    id = kpl_pop(a);
    if (id < 0) {
	char *name = kpl_token_name(-id);
	padColor = Pad_LookupColorByName(name);
	if (padColor)
	  Pad_renderer->Set_color(padColor);
    }
}

static void kpl_setfont(void)
{
    float a[16];
    int id;
    Pad_Font *font;

    id = kpl_pop(a);
    if (id < 0) {
	char *name = kpl_token_name(-id);
	font = Pad_LookupFontByName(name);
	if (font)
	  Pad_renderer->Set_font(font);
    }
}

static void kpl_setfontheight(void)
{
    float a[16];

    kpl_pop(a);
    Pad_renderer->Set_font_height(a[0]);
}

static void kpl_drawtext(void)
{
    float a[16], b[16];
    int len;
    int level;
    char *name;
    char buf[20];

    len = kpl_pop(a);
    len = kpl_pop(b);
    if (len < 0) {
	name = kpl_token_name(-len);
    } else {
	sprintf(buf, "%g", b[0]);
	name = buf;
    }
				// level should depend on
				// current magnification, and
				// size of text
    level = 2;
    Pad_renderer->Draw_string(name, level, a[0], a[1], 0.0);
}

static void kpl_drawborder(void)
{
    float a[16], p1[16], p2[16], width[16], relief[16];
    int relief_len, width_len, p1_len, p2_len;
    int id;
    Pad_Border *border;
    Pad_Point min, max;
    Pad_String reliefName;

    relief_len = kpl_pop(relief);
    id = kpl_pop(a);
    width_len = kpl_pop(width);
    p2_len = kpl_pop(p2);
    p1_len = kpl_pop(p1);

    if ((id < 0) && (p1_len == 2) && (p2_len == 2) && (width_len == 1)) {
	char *name = kpl_token_name(-id);

				// Check for valid relief name
	if (!Pad_Relief_type_to_name((int)relief[0], reliefName)) {
	    return;
	}
	border = Pad_LookupBorderByName(name);
	if (border) {
	    min.Set(p1[0], p1[1]);
	    max.Set(p2[0], p2[1]);
	    Pad_renderer->Set_border(border);
	    Pad_renderer->Draw_filled_3d_rectangle(min, max, width[0], (int)relief[0]);
	}
    }
}

//
// Draw an image
// Usage:  <imagetoken> drawimage
//
static void kpl_drawimage(void)
{
    float pos[16];
    float bbox[4];
    float image[16];
    int image_len;
    char *tokenName;
    Pad_ImageData *imageData;

    if (!Pad_prc) {
	return;
    }

    kpl_pop(pos);
    image_len = kpl_pop(image);

    if (image_len < 0) {
	tokenName = kpl_token_name(-image_len);
	imageData = Pad_prc->win->renderer->Get_image(tokenName);
	if (imageData) {
	    bbox[XMIN] = pos[0];
	    bbox[YMIN] = pos[1];
	    bbox[XMAX] = pos[0] + imageData->width;
	    bbox[YMAX] = pos[1] + imageData->height;
	    Pad_renderer->Draw_image(imageData, bbox[XMIN], bbox[YMIN], bbox[XMAX], bbox[YMAX], FALSE);
	}
    }
}

static void kpl_getsize(void)
{
    float a[16];
    if (kpl_render_object) {
	a[0] = kpl_render_object->pixelDim;
    } else {
	a[0] = 0.0;
    }
    kpl_push(a, 1);
}

static void kpl_getlevel(void)
{
    float a[16];
    a[0] = Pad_prc->win->view->pad->level;
    kpl_push(a, 1);
}

static void kpl_renderitem(void)
{
    int tagorid_len;
    float tagorid[16];
    char *argv[2];
    

    tagorid_len = kpl_pop(tagorid);
    if (tagorid_len < 0) {
	argv[0] = "renderitem";
	argv[1] = kpl_token_name(-tagorid_len);
	Padtcl_render_item(Pad_prc->win, Pad_prc->win->interp, 2, argv);
    }
}


//
// Path support library
//
static float current_x, current_y;
static float start_x, start_y;

static void g_new_path()
{
    Pad_renderer->Begin_polygon();
}

static void g_close_path()
{
    Pad_renderer->V2f(start_x, start_y);
}

static void g_move_to(float x, float y)
{
    Pad_renderer->Begin_polygon();
    Pad_renderer->V2f(x, y);    
    start_x = current_x = x;
    start_y = current_y = y;
}

static void g_line_to(float x, float y)
{
    Pad_renderer->V2f(x, y);    
    current_x = x; 
    current_y = y;
}

static void g_curve_to(float x1, float y1, float x2, float y2, float x3, float y3)
{
    Pad_Point p0, p1, p2, p3;

    p0.Set(current_x, current_y);
    p1.Set(x1, y1);
    p2.Set(x2, y2);
    p3.Set(x3, y3);

    Pad_renderer->B8f(p0, p1, p2, p3);

    current_x = x3;
    current_y = y3;
}

static void g_stroke_path()
{
    Pad_renderer->End_line();
}

static void g_fill_path()
{
    Pad_renderer->End_polygon();
}

// KPL Wrappers

static void kpl_wrap_newpath()
{
    g_new_path();
}

static void kpl_wrap_closepath(void)
{
    g_close_path();
}

static void kpl_wrap_stroke(void)
{
    g_stroke_path();
}

static void kpl_wrap_fill(void)
{
    g_fill_path();
}

static void kpl_wrap_moveto(void)
{    
    float a[16];
    kpl_pop(a);
    g_move_to(a[0], a[1]);
}

static void kpl_wrap_lineto(void)
{
    float a[16];
    kpl_pop(a);
    g_line_to(a[0], a[1]);
}

static void kpl_wrap_curveto(void)
{
    float a[16];
    kpl_pop(a);
    g_curve_to(a[0], a[1], a[2], a[3], a[4], a[5]);
}

void install_bezier_paths(void)
{
    kpl_install("newpath",   (Proc)kpl_wrap_newpath);
    kpl_install("closepath", (Proc)kpl_wrap_closepath);
    kpl_install("stroke",    (Proc)kpl_wrap_stroke);
    kpl_install("fill",      (Proc)kpl_wrap_fill);
    kpl_install("moveto",    (Proc)kpl_wrap_moveto);
    kpl_install("lineto",    (Proc)kpl_wrap_lineto);
    kpl_install("curveto",   (Proc)kpl_wrap_curveto);
}

void install_kpl_newpad(void)
{
				// Install Pad bezier stuff
    void install_bezier_paths(void);
    install_bezier_paths();

				// Install Pad commands
    kpl_install("refine",          (Proc)kpl_refine);
    kpl_install("interrupted",     (Proc)kpl_interrupted);
    kpl_install("setfont",         (Proc)kpl_setfont);
    kpl_install("setfontheight",   (Proc)kpl_setfontheight);
    kpl_install("setcolor",        (Proc)kpl_setcolor);
    kpl_install("setlinewidth",    (Proc)kpl_setlinewidth);
    kpl_install("setabslinewidth", (Proc)kpl_setabslinewidth);
    kpl_install("drawline",        (Proc)kpl_drawline);
    kpl_install("drawpolygon",     (Proc)kpl_drawpolygon);
    kpl_install("drawtext",        (Proc)kpl_drawtext);
    kpl_install("drawborder",      (Proc)kpl_drawborder);
    kpl_install("drawimage",       (Proc)kpl_drawimage);
    kpl_install("getlevel",        (Proc)kpl_getlevel);
    kpl_install("getsize",         (Proc)kpl_getsize);
    kpl_install("renderitem",      (Proc)kpl_renderitem);
}

Kpl::Kpl() 
{
    if (!envinit) {
				// Install KPL standards
	install_kpl_lib();
	install_kpl_matrix();
	install_kpl_newpad();
	install_kpl_physics();
	install_kpl_pump();
	install_kpl_sumvec();
	install_kpl_tcl();
	install_kpl_vec();
	envinit = 1;
    }
    codehead = NULL;
}

//
// KPL does not free pcode!
//

Kpl::~Kpl() {
    while (codehead) {
	//  kpl_free_pcode(codehead->pcode);
        //  This call should traverse Ken's first_pcode..last_pcode
	codefrag *frag = codehead;
	codehead = codehead->next;
	delete frag;
    }
    // Also need to do something about the KPL symtab and other stuff
    if (envinit) {
	// uninstall whatever needs it
	// (and if this is a complete reversal, unset envinit)
    }
}

Pcode *Kpl::compile (char *script) {
    Pcode *p = kpl_parse(script);
    if (!p) return NULL;
    codehead = new codefrag (p, codehead);
    return p;
}

void Kpl::execute (Pcode *p) {
    kpl_execPcode(p);
}

int Kpl::exec (char *script) {
    if (script) {
	Pcode *p = compile(script);
	if (!p) return TCL_ERROR;
	execute(p);
    } else {
	Pcode p,q;
	Symbol s;
	Value v;
	q.next = &p;
	p.symbol = &s;
	p.next = NULL;
	p.branch = NULL;
	s.name = "`";
	s.type = '`';
	s.value = &v;
	v.len = -s.type;
	execute(&q);
    }
    return TCL_OK;
}

char *Kpl::eval (char *script) {
    if (exec(script) == TCL_ERROR) return NULL;
    int n = 1000;
    char *r = (char *) ckalloc(n);
    int i = 0;
    r[0] = '\0';
    while (kpl_stack_level() > 0) {
	float data[16];
	int len = kpl_pop(data);
	char s[STRLEN+1];
	kpl_sprintvalue (s, data, len);
	int slen = strlen(s);
	while (i + slen >= n-2) {
	    n *= 2;
	    r = (char *) realloc(r,n);
	}
	strcpy (&r[i], s);
	i += slen;
	r[i++] = ' ';
	r[i] = '\0';
    }
    if (i > 0) r[i-1] = '\0';
    return r;
}

int Kpl::push (char *value) {
    float data[16];
    int len = sscan (value, data);
    return push (len, data);
}

int Kpl::push (int len, float *data) {
    kpl_push (data, len);
    return TCL_OK;
}

char *Kpl::Pop() {
    if (kpl_stack_level() <= 0) return NULL;
    float data[16];
    int len = kpl_pop(data);
    return sprint (len, data);
}

int Kpl::pop (int *len, float *data) {
    if (kpl_stack_level() <= 0) return TCL_ERROR;
    *len = kpl_pop(data);
    return TCL_OK;
}

char *Kpl::get (char *name) {
    int id = kpl_declare_token (name, P_VAR); // only declares if necessary, actually
    return get(id);
}

char *Kpl::get (int id) {
    float data[16];
    int len;
    len = kpl_get (id, data);
    if (len == 0) return NULL;
    return sprint (len, data);
}

int Kpl::get (int id, int *len, float *data) {
    *len = kpl_get (id, data);
    if (*len == 0) return TCL_ERROR; // with variables clobbered for efficiency
    return TCL_OK;
}

int Kpl::set (char *name, char *value) {
    int id = kpl_declare_token (name, P_VAR);
    return set (id, value);
}

int Kpl::set (int id, char *value) {
    float data[16];
    int len = sscan (value, data);
    return set (id, len, data);
}

int Kpl::set (int id, int len, float *data) {
    if (id < 0 || id >= MAXSYMBOL) return TCL_ERROR;
    kpl_set (id, data, len);
    return TCL_OK;
}

char *Kpl::sprint (int len, float *data) {
    char s[STRLEN+1];
    kpl_sprintvalue (s, data, len);
    int n = strlen(s);
    char *r = (char *)malloc(n+1);
    strcpy (r, s);
    return r;
}

int Kpl::sscan (char *value, float *data) {
    return kpl_sscanvalue (value, data);
}



//
// Called upon system startup to create kpl command
//
extern "C" int Kpl_Init (Tcl_Interp *interp) 
{
    Kpl *kpl = new Kpl;
    Tcl_CreateCommand (interp, "kpl", KplCmd, (ClientData)kpl, DestroyKpl);
    kpl->push (1, (float *)(&interp)); // interp trapped in a KPL vector
    kpl->exec ("tcluse");
    kpl->exec ("g_no_autoinit");       // Tells KPL not to initialize GL or X

    return TCL_OK;
}

//
// Implementation of the kpl command
//
int KplCmd (ClientData clientData, Tcl_Interp *interp, int argc, char **argv) 
{
    Kpl *kpl = (Kpl *)clientData;

    if (argc < 2) {
	Tcl_AppendResult (interp, "wrong # args: should be \"",
			  argv[0], " <subcommand> ?parameters?\"", NULL);
	return TCL_ERROR;
    }

    char *kmd = argv[1];	//  kpl subcommand:
    
    if (!strcmp(kmd, "run")) {	//  kpl run
	if (argc != 3) {
	    Tcl_AppendResult (interp, "wrong # args: should be \"",
			      argv[0], " ", argv[1], " <address>\"", NULL);
	    return TCL_ERROR;
	}
	long address = atol (argv[2]);
	kpl->run((Pcode *)address);
	
    } else if (!strcmp(kmd, "compile")) { //  kpl compile
	if (argc < 3) {
	    Tcl_AppendResult (interp, "wrong # args: should be \"",
			      argv[0], " ", argv[1], " <KPL tokens ...>\"", NULL);
	    return TCL_ERROR;
	}
	char *script = Tcl_Concat (argc-2, argv+2);
	Pcode *pcode = kpl->compile(script);
	free(script);
	if (!pcode) {
	    interp->result = "trouble in KPL compile";
	    return TCL_ERROR;
	}
	sprintf (interp->result, "%ld", (long)pcode);
    } else if (!strcmp(kmd,"exec")) { //  kpl exec ...
	char *script = (argc > 2) ? Tcl_Concat (argc-2, argv+2) : (char *)NULL;
	int rc = kpl->exec(script);
	if (script) ckfree((char *) script);
	if (rc == TCL_ERROR) {
	    interp->result = "trouble in KPL exec";
	    return TCL_ERROR;
	}
    } else if (!strcmp(kmd,"eval")) { //  kpl eval ...
	char *script = (argc > 2) ? Tcl_Concat (argc-2, argv+2) : (char *)NULL;
	char *result = kpl->eval(script);
	if (script) ckfree((char *) script);
	if (!result) {
	    TCL_ERR(interp, "trouble in KPL eval");
	}
	interp->result = result;
	interp->freeProc = (Tcl_FreeProc *)free;
    } else if (!strcmp(kmd,"push")) { //  kpl push <value>
	if (argc != 3) {
	    Tcl_AppendResult (interp, "wrong # args: should be \"",
			      argv[0], " ", argv[1], " <value>\"", NULL);
	    return TCL_ERROR;
	}
	kpl->push(argv[2]);
    } else if (!strcmp(kmd,"pop" )) { //  kpl pop
	if (argc != 2) {
	    Tcl_AppendResult (interp, "wrong # args: should be \"",
			      argv[0], " ", argv[1], "\"", NULL);
	    return TCL_ERROR;
	}
	char *result = kpl->Pop();
	if (!result) {
	    interp->result = "trouble in KPL pop";
	    return TCL_ERROR;
	}
	interp->result = result;
	interp->freeProc = (Tcl_FreeProc *)free;
    } else if (!strcmp(kmd,"get" )) { //  kpl get <name>
	if (argc != 3) {
	    Tcl_AppendResult (interp, "wrong # args: should be \"",
			      argv[0], " ", argv[1], " <name>\"", NULL);
	    return TCL_ERROR;
	}
	char *result = kpl->get(argv[2]);
	if (!result) {
	    interp->result = "trouble in KPL get";
	    return TCL_ERROR;
	}
	interp->result = result;
	interp->freeProc = (Tcl_FreeProc *)free;
    } else if (!strcmp(kmd,"set" )) { //  kpl set <name> <value>
	if (argc != 4) {
	    Tcl_AppendResult (interp, "wrong # args: should be \"",
			      argv[0], " ", argv[1], " <name> <value>\"", NULL);
	    return TCL_ERROR;
	}
	int rc = kpl->set (argv[2], argv[3]);
	if (rc == TCL_ERROR) {
	    interp->result = "trouble in KPL set";
	    return TCL_ERROR;
	}
    } else {
	Tcl_AppendResult (interp, "unknown kpl subcommand \"", kmd,
			  "\" - must be exec, eval, push, pop, get, or set", NULL);
	return TCL_ERROR;
    }
    
    return TCL_OK;
}


void DestroyKpl (ClientData clientData) {
    Kpl *kpl = (Kpl *)clientData;
    delete kpl;
}

