/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef TKKPL_H
#define TKKPL_H 1

// These definitions pertain to the KPL Tcl-command, "kpl",
// as distinct from the KPL Pad-object.

#include "defs.h"
#include "pad-tcl.h"

class Kpl {
  class codefrag {  // A trivial private class
        // This linked list should eventually give way to an associative table
        // that memoizes seen scripts to avoid proliferation of pcode fragments.
    friend class Kpl;
    Pcode *pcode;       // Pointer to pcode fragment
    codefrag *next;
    codefrag (Pcode *p, codefrag *n) : pcode(p), next(n) {}
  };

  static int envinit;   // For stuff common to all KPL interps
  codefrag *codehead;   // Head of a list of KPL pcode fragments

public:

  Kpl();
  ~Kpl();

  Pcode *compile (char *script);   // Get an actual code pointer
  void execute (Pcode *p);	   // The most direct route

  // In all methods that yield "char *", the space was allocated with
  // 'malloc' and should be freed with 'free' when you are done with it;
  // note also that such methods can yield NULL in error cases:
  // (may change to 'ckalloc' and 'ckfree' for Tcl)

  // Omitting script avoids a parse, whereupon these do a KPL "`" (pop & eval):
  int exec (char *script =NULL);	// KPL data stack left alone afterwards
  char *eval (char *script =NULL);	// Entire stack popped and yielded

  int push (char *value);		// push 1 value on the KPL data stack
  int push (int len, float *data =NULL);	// len<0 => id=-len & no data
  char *Pop();				// pop 1 value from the KPL data stack
  int pop (int *len, float *data =NULL);	// fetches len & possibly data

  // The "name" parameters in these methods are considered already quoted:
  char *get (char *name);		// retrieve KPL data value by name
  char *get (int id);			// and by symbol id
  int get (int id, int *len, float *data =NULL);  // gets len & possibly data
  int set (char *name, char *value);	// store KPL data value against name
  int set (int id, char *value);	// and against symbol id
  int set (int id, int len, float *data =NULL);  // and sans conversion

  // Versions of kpl_sprintvalue, kpl_sscanvalue:
  char *sprint (int len, float *data =NULL);	// String for you or Tcl to eat
  int sscan (char *value, float *data =NULL);	// Yields "len", may store data

  void run (Pcode *p) {execute(p);}	// A synonym
};

//
// Function prototypes
//
extern "C" int  Kpl_Init(Tcl_Interp *interp);
int  KplCmd (ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
void DestroyKpl (ClientData clientData);

#endif  // TKKPL_H
