/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

// ************************************************************************
// XDR C++ interface implementation
//
// Implements C++ interface for XDR encoding and decoding utilities.
// ************************************************************************

#include "defs.h"
#include <string.h> // uses strdup()
extern "C" {
#include "xdrutil.h"
}
#include "xdrintf.h"

int
Pad_XDRDataRep::Type()
{
  return PAD_XDR_FORMAT;
}

//
// Method for reading bytes into an array.  This is used by
// Pad_StreamData when reading data before decoding it.
//
int 
Pad_XDRDataRep::ReadData(istream &is, unsigned int type, 
			 unsigned int dataLength, unsigned int &infoLength, 
			 unsigned char *&info)
{
  if (type != (_char | _vector)) {
    return Pad_DataRep::ReadData(is, type, dataLength, infoLength, info);
  } else {
    infoLength = extend(dataLength) + 4;
    info = new unsigned char[infoLength];
    is.read(info, infoLength);
    int bytesRead = infoLength;
    if (is.good())
      return bytesRead;
    else
      return itemErr;
  }
}


// ************************************************************************
// Encoding routines
// ************************************************************************

int Pad_XDRDataRep::EncodeString(const char* input, unsigned char*& output,
  unsigned int& length)
{
  length=XDRextend(strlen(input)+1)+XDRBlockSize;
  output=new unsigned char[length];
  int err=XDREncodeStringC(input, output, length);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const short* input, unsigned int inputLength,
  unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(short));
  output=new unsigned char[outputLength];
  int err=XDREncodeShortArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const int* input, unsigned int inputLength,
  unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(int));
  output=new unsigned char[outputLength];
  int err=XDREncodeIntArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const long* input, unsigned int inputLength,
  unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(long));
  output=new unsigned char[outputLength];
  int err=XDREncodeLongArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const unsigned char* input,
  unsigned int inputLength, unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(unsigned char));
  output=new unsigned char[outputLength];
  int err=XDREncodeUCharArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const unsigned short* input,
  unsigned int inputLength, unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(unsigned short));
  output=new unsigned char[outputLength];
  int err=XDREncodeUShortArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const unsigned int* input,
  unsigned int inputLength, unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(unsigned int));
  output=new unsigned char[outputLength];
  int err=XDREncodeUIntArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const unsigned long* input,
  unsigned int inputLength, unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(unsigned long));
  output=new unsigned char[outputLength];
  int err=XDREncodeULongArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const float* input, unsigned int inputLength,
  unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(float));
  output=new unsigned char[outputLength];
  int err=XDREncodeFloatArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}

int Pad_XDRDataRep::EncodeArray(const double* input, unsigned int inputLength,
  unsigned char*& output, unsigned int& outputLength)
{
  outputLength=inputLength*XDRextend(sizeof(double));
  output=new unsigned char[outputLength];
  int err=XDREncodeDoubleArray(input, inputLength, output, outputLength);
  if (err) delete [] output;
  return err;
}


// ************************************************************************
// Decoding routines
// ************************************************************************

int Pad_XDRDataRep::DecodeString(const unsigned char* input, unsigned int inputLength,
  char* output, unsigned int outputLength)
{
  return XDRDecodeStringC(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  short* output, unsigned int outputLength)
{
  return XDRDecodeShortArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  int* output, unsigned int outputLength)
{
  return XDRDecodeIntArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  long* output, unsigned int outputLength)
{
  return XDRDecodeLongArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  return XDRDecodeUCharArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  unsigned short* output, unsigned int outputLength)
{
  return XDRDecodeUShortArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  unsigned int* output, unsigned int outputLength)
{
  return XDRDecodeUIntArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  unsigned long* output, unsigned int outputLength)
{
  return XDRDecodeULongArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
  float* output, unsigned int outputLength)
{
  return XDRDecodeFloatArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::DecodeArray(const unsigned char* input,
  unsigned int inputLength, double* output, unsigned int outputLength)
{
  return XDRDecodeDoubleArray(input, inputLength, output, outputLength);
}


// ************************************************************************
// Encoding routines for flattening
// ************************************************************************

int Pad_XDRDataRep::flatEncodeString(const char* input, unsigned char* output,
  unsigned int length)
{
  return XDREncodeStringC(input, output, length);
}

int Pad_XDRDataRep::flatEncodeArray(const short* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  return XDREncodeShortArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const int* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  return XDREncodeIntArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const long* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  return XDREncodeLongArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const unsigned char* input,
  unsigned int inputLength, unsigned char* output, unsigned int outputLength)
{
  return XDREncodeUCharArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const unsigned short* input,
  unsigned int inputLength, unsigned char* output, unsigned int outputLength)
{
  return XDREncodeUShortArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const unsigned int* input,
  unsigned int inputLength, unsigned char* output, unsigned int outputLength)
{
  return XDREncodeUIntArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const unsigned long* input,
  unsigned int inputLength, unsigned char* output, unsigned int outputLength)
{
  return XDREncodeULongArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const float* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  return XDREncodeFloatArray(input, inputLength, output, outputLength);
}

int Pad_XDRDataRep::flatEncodeArray(const double* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  return XDREncodeDoubleArray(input, inputLength, output, outputLength);
}




