/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

// ************************************************************************
// XDR C++ interface declaration
//
// Declares C++ interface for XDR encoding and decoding utilities.
// ************************************************************************


#ifndef _xdrintf_h
#define _xdrintf_h

#include "streamdata.h"

// Extending length to 4-byte boundary for XDR encoding

#define XDRBlockSize 4
#define XDRextend(length) (4*(((length)+3)/4))


class Pad_XDRDataRep : public Pad_DataRep {
public:
  Pad_XDRDataRep();

  // Encoding routines
  int EncodeString(const char* input, unsigned char*& output,
		   unsigned int& length);
  int EncodeArray(const short* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const int* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const long* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const unsigned char* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const unsigned short* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const unsigned int* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const unsigned long* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const float* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  int EncodeArray(const double* input, unsigned int inputLength,
		  unsigned char*& output, unsigned int& outputLength);
  
  
  // Decoding routines
  int DecodeString(const unsigned char* input, unsigned int inputLength,
		   char* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  short* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  int* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  long* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  unsigned char* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		 unsigned short* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  unsigned int* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  unsigned long* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  float* output, unsigned int outputLength);
  int DecodeArray(const unsigned char* input, unsigned int inputLength,
		  double* output, unsigned int outputLength);
  
  // Encoding routines for flattening

  int flatEncodeString(const char* input, unsigned char* output,
		       unsigned int length);
  int flatEncodeArray(const short* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const int* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const long* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const unsigned char* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const unsigned short* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const unsigned int* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const unsigned long* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const float* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int flatEncodeArray(const double* input, unsigned int inputLength,
		      unsigned char* output, unsigned int outputLength);
  int extend(int length);
  virtual int ReadData(istream &is, unsigned int type, unsigned int dataLength,
                      unsigned int &infoLength, unsigned char *&info);
  virtual int Type();
};

inline Pad_XDRDataRep::Pad_XDRDataRep() {BlockSize=XDRBlockSize;}
inline int Pad_XDRDataRep::extend(int length) {return XDRextend(length);}

#endif

