/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

/* ************************************************************************
// XDR utilities implementation
//
// Implements utilities for XDR encoding and decoding.
//
// Written in C because of xdr.h incompatibility with C++).
// ************************************************************************/
#define NO_BOOLEAN_ENUM

#ifdef PAD_UNIX
#include <rpc/rpc.h> /*#include "xdr.h"*/ /* for xdr encoding */
#endif

#ifdef PAD_WIN
#include "../win/xdr.h"
#endif

#include <string.h> /* uses strlen() */

#include "defs.h"
extern "C" {
#include "xdrutil.h"
}


/* Extending length to 4-byte boundary for XDR encoding */

#define XDRBlockSize 4
#define XDRextend(length) (4*(((length)+3)/4))


#define noErr 0;
#define xdrErr -1024;


/* ************************************************************************
// Encoding routines
// ************************************************************************/

int XDREncodeStringC(const char* input, unsigned char* output,
  unsigned int length)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, length, XDR_ENCODE);
  succ=xdr_string(&xdrs, (char**)&input, length-XDRBlockSize);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeShortArray(const short* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(short),
    (xdrproc_t)&xdr_short);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeIntArray(const int* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(int),
    (xdrproc_t)&xdr_int);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeLongArray(const long* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(long),
    (xdrproc_t)&xdr_long);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeUCharArray(const unsigned char* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(unsigned char),
    (xdrproc_t)&xdr_u_char);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeUShortArray(const unsigned short* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(unsigned short),
    (xdrproc_t)&xdr_u_short);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeUIntArray(const unsigned int* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(unsigned int),
    (xdrproc_t)&xdr_u_int);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeULongArray(const unsigned long* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(unsigned long),
    (xdrproc_t)&xdr_u_long);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeFloatArray(const float* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(float),
    (xdrproc_t)&xdr_float);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDREncodeDoubleArray(const double* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)output, outputLength, XDR_ENCODE);
  succ=xdr_vector(&xdrs, (char*)input, inputLength, sizeof(double),
    (xdrproc_t)&xdr_double);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}


/* ************************************************************************
// Decoding routines
// ************************************************************************/

int XDRDecodeStringC(const unsigned char* input, unsigned int inputLength,
  char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_string(&xdrs, (char**)&output, outputLength);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeShortArray(const unsigned char* input, unsigned int inputLength,
  short* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(short),
    (xdrproc_t)&xdr_short);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeIntArray(const unsigned char* input, unsigned int inputLength,
  int* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(int),
    (xdrproc_t)&xdr_int);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeLongArray(const unsigned char* input, unsigned int inputLength,
  long* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(long),
    (xdrproc_t)&xdr_long);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeUCharArray(const unsigned char* input, unsigned int inputLength,
  unsigned char* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength,
    sizeof(unsigned char), (xdrproc_t)&xdr_u_char);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeUShortArray(const unsigned char* input, unsigned int inputLength,
  unsigned short* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength,
    sizeof(unsigned short), (xdrproc_t)&xdr_u_short);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeUIntArray(const unsigned char* input, unsigned int inputLength,
  unsigned int* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(unsigned int),
    (xdrproc_t)&xdr_u_int);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeULongArray(const unsigned char* input, unsigned int inputLength,
  unsigned long* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(unsigned long),
    (xdrproc_t)&xdr_u_long);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeFloatArray(const unsigned char* input, unsigned int inputLength,
  float* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(float),
    (xdrproc_t)&xdr_float);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}

int XDRDecodeDoubleArray(const unsigned char* input, unsigned int inputLength,
  double* output, unsigned int outputLength)
{
  int succ;
  XDR xdrs;
  xdrmem_create(&xdrs, (char*)input, inputLength, XDR_DECODE);
  succ=xdr_vector(&xdrs, (char*)output, outputLength, sizeof(double),
    (xdrproc_t)&xdr_double);
  xdr_destroy(&xdrs);
  if (!succ) return xdrErr;
  return noErr;
}
