#ifndef LINUX_SOUND_H
#define LINUX_SOUND_H
/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/
#include <stdio.h>
#include <tcl.h>
#include "../generic/pad-sound.h"
#include "../generic/defs.h"

class LinuxSoundDevice;
struct soundstream;
class Pad_String;

struct Volume_struct {
    int right;
    int left;
};

struct Play_struct {
    long *buf;
    Pad_AudioFormat *format;
    long nsamp;
    Pad_PlayState *state;
};

class LinuxSoundDevice : public SoundDevice{
public:
    static void _Handle_set_volume (ClientData clientdata);
    static void _Handle_play(ClientData cliendata);

    LinuxSoundDevice();
    LinuxSoundDevice(char*);
    virtual ~LinuxSoundDevice();
  
    virtual int Play(Pad_SoundConfig*,Pad_PlayState*);
    virtual int Stop(void);
    virtual int Set_volume(int &master);
    virtual int Set_volume(int &left, int &right);
    virtual int Get_volume(int &left, int &right) const;
    virtual Pad_Bool Load(Pad_Sound &sound);
protected:
    int _dspfd;
    Pad_String *_deviceName;
    int         _bufferSize;
    int         _bufferCount;
    char *      _audioBuffer;
    
    Pad_Bool _Set_bufsize(void);
    Pad_Bool _Set_channels(int channels);
    Pad_Bool _Set_rate(int rate);
    Pad_Bool _Set_samplesize(int samplesize);
    void _Set_play_timer(long *buf, long nsamp, Pad_AudioFormat *format, Pad_PlayState*);
    void _Set_play_timer(Pad_PlayState*);
    void _Set_play_timer(Play_struct *data);
    int  Is_Ready(void) const;
    int Create_volume_handler(Volume_struct *vol) const;
    void _Get_stereo_volumes (int &, int &, int) const;
    void _Set_stereo_volumes (int &vol);
    void _Set_stereo_volumes (const int left, const int right, int &vol);
    int  _Init_play(Pad_AudioFormat *format);
    void _Stop_play(void);
    void _Reset(Pad_Bool hard);
    int  _Flush(void);
    void _Write_short(const soundstream * ft,unsigned int ui);
    Pad_Bool  _Play(Pad_AudioFormat *format, long *buf,long nsamp,Pad_PlayState*);
    void _Put(int c);
};

////////////////////////////////////////////////////////////////
//
// LinuxAudioFormat Class
//
//
////////////////////////////////////////////////////////////////

class LinuxAudioFormat : public Pad_AudioFormat{
public:
    LinuxAudioFormat();
    LinuxAudioFormat(char*);
    LinuxAudioFormat (soundstream * format);
    virtual ~LinuxAudioFormat();
  
        ////////////////////////////////////////
        //
        // Utility functions
        //
        ////////////////////////////////////////

    static LinuxAudioFormat *Get_sound_file_format(char *filename);

    virtual long        Read(void);
    const   soundstream *        Get_format(void) const;
    virtual Pad_SoundFormat Get_style(void) const;
    virtual long        Get_rate(void) const;
    virtual long        Get_channels(void) const;
    virtual long        Get_length(void) const;
    virtual long        Get_bits_per_sample(void) const;
    virtual void *      Get_private(void) const;
protected: 
    virtual void _Stop_read();
    virtual int  _Read(soundstream *, char*, long)=0;

    FILE *      _fp;
    soundstream *_soundstream;            // internal data rep from SoX libst.a
};

class LinuxAUAudioFormat : public LinuxAudioFormat{
public:
    LinuxAUAudioFormat();
    ~LinuxAUAudioFormat();
    LinuxAUAudioFormat (char *);
    LinuxAUAudioFormat (soundstream * format);
  
    virtual Pad_Bool        Open_format(char*);

private:
    virtual int  _Read(soundstream *, char*, long);
};

class LinuxAIFFAudioFormat : public LinuxAudioFormat{
public:
    LinuxAIFFAudioFormat();
    ~LinuxAIFFAudioFormat();
    LinuxAIFFAudioFormat (char *);
    LinuxAIFFAudioFormat (soundstream * format);
  
    virtual Pad_Bool        Open_format(char*);

private:
    virtual int  _Read(soundstream *, char*, long);
};

class LinuxWAVAudioFormat : public LinuxAudioFormat{
public:
    LinuxWAVAudioFormat();
    ~LinuxWAVAudioFormat();
    LinuxWAVAudioFormat (char *);
    LinuxWAVAudioFormat (soundstream * format);

    virtual Pad_Bool        Open_format(char*);

private:
    virtual int  _Read(soundstream *, char*, long);
};

#endif /* LINUX_SOUND_H */
