#ifdef HAS_NCD_SOUND 
/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "ncd-sound.h"
#include "../generic/pad-sound.h"
#include "../generic/api.h"
#include "../generic/list.h"
#include "sound-tools.h"
#include "sound-lib.h"
#include <tcl.h>
#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "../generic/global.h"
#include "soundcard.h"
#include <stdarg.h>

#include <audio/audiolib.h>
#include <audio/soundlib.h>

#include <sys/time.h>
#include <unistd.h>

#undef DEBUG_SOUND
//#define DEBUG_SOUND 1

static int currentVolume;


////////////////////////////////////////////////////////////////
//
// NCDAudioFormat subclass of Pad_AudioFormat
//
//
////////////////////////////////////////////////////////////////
NCDAudioFormat::NCDAudioFormat()
{
    _bucket = 0;
    _flow   = 0;
}

NCDAudioFormat::~NCDAudioFormat(){}

AuBucketID
NCDAudioFormat::Get_bucket(void) const
{
    return _bucket;
}

Pad_Bool
NCDAudioFormat::Set_bucket(AuBucketID bucket)
{
    _bucket = bucket;
    return TRUE;
}


AuFlowID
NCDAudioFormat::Get_flow(void)
{
    return _flow;
}

void
NCDAudioFormat::Set_flow(AuFlowID flow)
{
    _flow = flow;
}

long
NCDAudioFormat::Read(void)
{
    Pad_errorString = "NCDAudioFormat::Read(void) should not be called";
    return -1;
}


//
// The SoundDevice class is a singleton class - only one instance
//    can exist at any time. To access that instance a Get_instance()
//    fucntion is used. Each Subclass of SoundDevice implements its
//    own version of SoundDevice::Get_instance(), and only one version
//    is compiled in at load time.
//
SoundDevice *SoundDevice::_instance = NULL;

SoundDevice *
SoundDevice::Get_instance(void) {
    if (! _instance) {
        _instance = new NCDSoundDevice ();
    }
    return _instance;
}

void
AudioTimer(ClientData clientdata)
{
    AuServer *aud = (AuServer*) clientdata;

    AuHandleEvents(aud);
//    cerr << "handle events" << endl;
    Tcl_CreateTimerHandler(50,AudioTimer,aud);
}


//
// Constructors for NCDSoundDevice should never be called except
//    from Get_instance()
//
NCDSoundDevice::NCDSoundDevice(){
    char **msg;
    
    _audio = AuOpenServer(NULL, // name will be looked up from AUDIOSERVER environment variable
                          0,    // length of authproto
                          NULL, // authproto will be looked up
                          0,    // length of authdata
                          NULL, // authdata will be looked up
                          msg); // return msg from server on failure
    
    currentVolume = 90;
    
    if (_audio == NULL) {
        cerr << "SoundDevice: error: " << *msg << endl;
        return;
    }
    for (int i = 0; i < AuServerNumDevices(_audio); i++) {
	if ((AuDeviceKind(AuServerDevice(_audio, i))
	     == AuComponentKindPhysicalOutput)
	    && AuDeviceNumTracks(AuServerDevice(_audio, i)) == 1) {

	    _device = AuDeviceIdentifier(AuServerDevice(_audio, i));
	    break;
	}
    }
    Tcl_CreateTimerHandler(50,AudioTimer,_audio);
}

//
// Free the sound data
//
void
NCDSoundDevice::Free_sound(Pad_Sound &sound){
        // not all sounds use buckets
    NCDAudioFormat *format = (NCDAudioFormat *) sound.Get_format();
    
    if (format->Get_bucket()) {
        AuDestroyBucket(_audio, format->Get_bucket(), NULL);
    }
}


//
// The destructor. As of yet no data needs to be special deleted
//
NCDSoundDevice::~NCDSoundDevice(){
    AuCloseServer (_audio);
}


static void
doneCB(AuServer *,
       AuEventHandlerRec *,
       AuEvent *ev,
       Pad_Sound *sound)
{
    AuElementNotifyEvent *event = (AuElementNotifyEvent *) ev;

    switch (ev->type) {
        case AuEventTypeElementNotify :
            if (ev->type == AuEventTypeElementNotify &&
                event->kind == AuElementNotifyKindState &&
                event->cur_state == AuStateStop)
                {
                    cerr << "Sound finished: " << sound->Get_name() << endl;
                }
            break;
        default :
            break;
    }
}

AuBucketID
NCDSoundDevice::Bucket_from_data(Pad_Sound &sound)
{
    Sound format;
    AuPointer data;
    AuBucketID bucket;
    AuStatus status;
    NCDAudioFormat *ncd = (NCDAudioFormat*) sound.Get_format();
    
    format = SoundCreate(SoundFileFormatNone,
                         ncd->Get_style(),
                         sound.Get_channels(),
                         sound.Get_rate(),
                         Pad_SynthSound::SAMPLE_NUMBER,
                         "");
    data = sound.Get_data();
    bucket = AuSoundCreateBucketFromData(_audio,
                                         format,
                                         data,
                                         AuAccessAllMasks,
                                         NULL,
                                         &status);

    cerr << "Bucket: " << bucket << endl;
    
    SoundDestroy(format);

    return bucket;
}


static AuBool
EventCallback(AuServer  *aud,  AuEvent  *ev,  AuEventHandlerRec *rec)
{
    Pad_PlayState *state = (Pad_PlayState*) rec->data;
    Pad_Sound *sound = state->_source;
    NCDAudioFormat *ncd = (NCDAudioFormat*) sound->Get_format();
    
    switch (ev->type) {
        case AuEventTypeElementNotify :
        {
            AuElementNotifyEvent *event = (AuElementNotifyEvent *) ev;
            switch  (event->kind) {
                case AuElementNotifyKindLowWater :
                    if (!sound->Stopped()) {
                        state->Refill(event->num_bytes);
                        AuWriteElement(aud,
                                       ncd->Get_flow(),
                                       0,
                                       event->num_bytes,
                                       sound->Get_data(),
                                       AuFalse,
                                       NULL);
                    }
                    cerr << "Low water event: " << event->num_bytes << endl;
                    break;
                case AuElementNotifyKindState :
                    switch (event->cur_state) {
                        case AuStatePause :
                            if (event->reason != AuReasonUser) {
                                if (!sound->Stopped()) {
                                    cerr << "Pause state: " << event->num_bytes << endl;
                                    if (!sound->Loop()) {
                                        sound->Set_stop(TRUE);
                                    }
                                    
                                    state->Refill(event->num_bytes);
                                    AuWriteElement(aud,
                                                   ncd->Get_flow(),
                                                   0,
                                                   event->num_bytes,
                                                   sound->Get_data(),
                                                   AuFalse,
                                                   NULL);
                                } else {
                                // the sound is no longer playing so
                                // we remove it from the list of
                                // sounds in the source
                                    sound->Done(state);
                                }
                            }
                            break;
                        case AuStateStop:
                            sound->Set_stop(TRUE);
                            break;
                    }
            }
        }
    
//         case AuEventTypeGrabNotify :
//             cerr << "Grab event" << endl;
//             break;
//         case AuEventTypeMonitorNotify :
//             cerr << "Monitor event" << endl;
//             break;
//         case AuEventTypeBucketNotify :
//             cerr << "Bucket event" << endl;
//             break;
//         case AuEventTypeDeviceNotify :
//             cerr << "Device event" << endl;
//             break;
        default :
            break;
    }
    
    return AuTrue;
}

int
Time_diff()
{
    struct timeval        tv;
    struct timezone       tz;
    static struct timeval prev_tv;
    int dsec, dusec;

    if (!gettimeofday(&tv,&tz)) {
        dsec = tv.tv_sec - prev_tv.tv_sec;
        dusec = tv.tv_usec - prev_tv.tv_usec;
        dusec += dsec * 1000000;
        dusec = int(dusec/1000);
     
        prev_tv.tv_sec = tv.tv_sec;
        prev_tv.tv_usec = tv.tv_usec;
    }
    return dusec;
}

//
// Play() is the interface function used to play a sound on the
//   SoundDevice. Calling play with a volume, temporarily resets
//   the master volume while the sound is being played.
//
int
NCDSoundDevice::Play(Pad_SoundConfig *config, Pad_PlayState *state)
{
    AuEventHandlerRec *rec;
    AuFixedPoint       vol;
    AuBucketID         bucket;
    AuFlowID           flowID;
//    int                *volume_mult;
//    int                 monitor;
    AuStatus           status;
    NCDAudioFormat    *ncd;
    int                loop_num = 1;
    AuEventHandlerRec *handler;
    AuElement          elements[3];
    int                volume;
    Pad_Sound         *sound = state->_source;
    ncd  = (NCDAudioFormat*) sound->Get_format();
    
    if (config->volume->data == Pad_Sound::CONFIG_UNSET) {
            // We were called from Play(Pad_Sound&, Pad_Bool)
            // so set the volume to the current volume
        volume = currentVolume;
    } else {
        volume = config->volume->data;
    }

    if (sound->Data()) {
            // this case is presumably here in case a sound is already
            // playing in a flow, and Play() has been called again
        if (ncd->Get_flow() != 0) {
            sound->Set_stop(FALSE);
        } else {
            if (!(flowID =  AuCreateFlow(_audio, NULL))) {
                cerr << "Couldn't create flow" << endl;
                return FALSE;
            }
            ncd->Set_flow(flowID);
            AuMakeElementImportClient(&elements[0],
                                      sound->Get_rate(),
                                      ncd->Get_style(),
                                      sound->Get_channels(),
                                      AuTrue,
                                      Pad_SynthSound::SAMPLE_NUMBER,
                                      Pad_SynthSound::LOW_WATER,
                                      0,
                                      NULL);
            AuMakeElementExportDevice(&elements[1],
                                      0,
                                      _device,
                                      sound->Get_rate(),
                                      AuUnlimitedSamples,
                                      0,
                                      NULL);
            AuSetElements(_audio, flowID, AuTrue, 2, elements, NULL);
            if (config->loop->data == 1) {
//                sound->Set_stop(TRUE);
                sound->Set_loop(TRUE);
            }
//             sound->Refill(BUFFER_SIZE);
//             AuWriteElement(_audio,
//                            flowID,
//                            0,
//                            BUFFER_SIZE/2,
//                            sound->Get_private(),
//                            AuFalse,
//                            &status);
        }
    } else {
        bucket = ncd->Get_bucket();
        if (bucket == 0) {
                // Haven't loaded the sound into the server yet ...
            bucket = AuSoundCreateBucketFromFile(_audio,
                                                 sound->Get_name(),
                                                 AuAccessAllMasks,
                                                 NULL,
                                                 NULL);
            ncd->Set_bucket(bucket);
        }
    
        vol = AuFixedPointFromFraction(volume, 100);
//    AuMonitorRate = 10;
        
        rec = AuSoundPlayFromBucket(_audio,
                                    bucket,
                                    _device,
                                    vol,
                                    (void (*)(AuServer  *,  AuEventHandlerRec  *, AuEvent *, AuPointer)) doneCB, // done callback
                                    (AuPointer) sound,  // callback data
                                    loop_num,    // loop count
                                    &flowID, // flowID
                                    NULL,                                
                                    NULL,
                                    NULL);

//        cerr << "FlowID: " << flowID << endl;
    }

    
    if (!(handler = AuRegisterEventHandler(_audio,
                                           AuEventHandlerIDMask,
                                           0,
                                           flowID,
                                           EventCallback,
                                           (AuPointer) sound))) {
        cerr << "Couldn't register event handler" << endl;
    }
    AuStartFlow(_audio,flowID,&status);
        // even though the documentation says otherwise, this call
        // is absolutely necessary...
    AuFlush(_audio);
    return TRUE;
}

int
NCDSoundDevice::Stop()
{
    Pad_errorString = "NCDSoundDevice::Stop() should never be called";
    return FALSE;
    
}

//
// Load() imports a file into the Sound Device.
//
Pad_Bool
NCDSoundDevice::Load(Pad_Sound &sound){
    NCDAudioFormat *ncd = (NCDAudioFormat*) sound.Get_format();
    AuBucketID bucket;
    AuStatus status = -5;
    Pad_Bool rc;
    
    bucket = AuSoundCreateBucketFromFile(_audio,
                                         sound.Get_name(),
                                         AuAccessAllMasks,
                                         NULL,
                                         &status);

    if (bucket == AuNone) {
        rc = FALSE;
        switch (status) {
            case AuBadLength :
                Pad_errorString = "Bucket error: Bad Length";
                break;
            case AuBadAlloc :
                Pad_errorString = "Bucket error: Bad Alloc";
                break;
            case AuBadValue :
                Pad_errorString = "Bucket error: Bad Value";
                break;
            case -5 :
                Pad_errorString = "bad file name: ";
                Pad_errorString += sound.Get_name();
                break;
            default :
                break;
        }
    }
    if (rc == FALSE) {
        return rc;
    }
    
    cerr << "Bucket = " << bucket << endl;
    return ncd->Set_bucket(bucket);    
}
 
//
// Set_volume() sets either the both channels to the same volume, or sets
//    them independantly. If a sound is being played a Tcl timer is created
//    to set the volume when the device is idle.
//
// This procedure sets Pad_resultString on success and Pad_errorString
//   on failure.
//
int
NCDSoundDevice::Set_volume (int &vol){
    currentVolume = vol;
    return TRUE;
}

int
NCDSoundDevice::Set_volume (int &right, int &left){
    currentVolume = (right + left) / 2;
    return TRUE;
}

//
// Get_volume() returns the hardware volume. returns -1 on error.
//
int
NCDSoundDevice::Get_volume (int &left, int &right) const{
    left = currentVolume;
    right = currentVolume;
    return TRUE;
}

#endif /* HAS_NCD_SOUND */
 
