/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "../generic/global.h"
#include "../generic/pad-sound.h"

#ifdef HAS_SGI_SOUND
#include "sgi-sound.h"

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>

SoundDevice *SoundDevice::_instance = NULL;

SoundDevice *
SoundDevice::Get_instance(void) {
    if (! _instance) {
        _instance = new SgiSoundDevice ();
    }
    return _instance;
}

Pad_AudioFormat*
SgiSoundDevice::New_format()
{
  return new SgiAudioFormat();
}

Pad_Bool
SgiSoundDevice::Load(Pad_Sound &sound){
    char *filename = sound.Get_name();
    
    Pad_AudioFormat *format = new SgiAudioFormat();

    if(!format->Open_format(filename)) {
        return FALSE;
    }
    /*
        // shouldn't need this case...
    if (format->Get_filename() == "") {
        Pad_errorString = "Couldn't load sound file: ";
        Pad_errorString += filename;
        Pad_errorString += "\n";
        return FALSE;
    }
    */
    format->Read();
    sound.Set_format(format);
    return TRUE;
}

int
SgiSoundDevice::Play(Pad_SoundConfig *, Pad_PlayState *state)
{
    Pad_Sound *sound = state->_source;
    

      // Set up the audio port
  if(! _Init_play(sound->Get_format())){
    Pad_errorString = "";
    return FALSE;
  }

      // Now really play the sound
  _Play(sound->Get_format(), sound->Get_data(), sound->Get_length());
  
      // Clean up
  _Stop_play();
  return TRUE;
}

SgiSoundDevice::SgiSoundDevice() {
  _audioConfig = ALnewconfig();
  _deviceName  = new Pad_String ("audioport");
  
      // don't try to close the port until it's been configured...
  _audioPort   = NULL;
  _audioBuffer = NULL;
  _bufferCount = 0;
}

SgiSoundDevice::~SgiSoundDevice(){
  ALfreeconfig (_audioConfig);
  ALcloseport (_audioPort);
  if (_deviceName) {
    delete _deviceName;
  }
  if (_audioBuffer) {
    delete _audioBuffer;
  }
}

int
SgiSoundDevice::Stop(void){
  ALcloseport (_audioPort);
  return TRUE;
}
 
void
SgiSoundDevice::_Flush(void){
  ALwritesamps (_audioPort, _audioBuffer, _bufferCount);
}

int
SgiSoundDevice::Set_volume(int &master){
  long pvbuf[4];
  
  if (master < 0 || master > 100){
    Pad_errorString = "Bad Value: accepted range is 0-100.";
    return FALSE;
  }

      // the Sgi volume setting goes from 0 to 255 but the Pad++
      // sound interface only allows from 0 to 100 so we must scale
  master = (float(master) / 100) * 255;
  
  pvbuf[0] = AL_LEFT_SPEAKER_GAIN;
  pvbuf[1] = master;
  pvbuf[2] = AL_RIGHT_SPEAKER_GAIN;
  pvbuf[3] = master;

  if (ALsetparams (AL_DEFAULT_DEVICE, pvbuf, (long)4) == 0){
    return TRUE;
  }
  Pad_errorString = "Error setting volume in SgiSoundDevice::Set_volume(int)";
  return FALSE;
}

int
SgiSoundDevice::Set_volume(int &left,
                           int &right){
  long pvbuf[4];
  
  if (left < 0 || left > 100 || right < 0 || right > 100){
    Pad_errorString = "Bad Value: accepted range is 0-100.";
    return FALSE;
  }

      // the Sgi volume setting goes from 0 to 255 but the Pad++
      // sound interface only allows from 0 to 100 so we must scale
  left = (float(left) / 100) * 255;
  right = (float(right) / 100) * 255;
  
  pvbuf[0] = AL_LEFT_SPEAKER_GAIN;
  pvbuf[1] = left;
  pvbuf[2] = AL_RIGHT_SPEAKER_GAIN;
  pvbuf[3] = right;

  if (ALsetparams (AL_DEFAULT_DEVICE, pvbuf, (long)4) == 0){
    return TRUE;
  }
  Pad_errorString = "Error setting volume in SgiSoundDevice::Set_volume(int,int)";
  
  return FALSE;
}

//
// This procedure sets a timer that will reset the volume in 100ms
//
int
SgiSoundDevice::Reset_volume (int &left, int &right)
{
    Set_volume(left, right);
    return(TRUE);
}

int
SgiSoundDevice::Get_volume(int &left,
                           int &right) const{
  long pvbuf[4];
  pvbuf[0] = AL_LEFT_SPEAKER_GAIN;
  pvbuf[2] = AL_RIGHT_SPEAKER_GAIN;

  if (ALsetparams (AL_DEFAULT_DEVICE, pvbuf, (long)4) == 0){
    left  = pvbuf[1];
    right = pvbuf[3];
    return TRUE;
  }
    Pad_errorString = "Error getting volume in SgiSoundDevice::Get_volume(int&,int&)";
  return FALSE;
}

int
SgiSoundDevice::_Get_output_rate(){
  long buf[4];

  buf[0] = AL_OUTPUT_RATE;
  buf[2] = AL_DIGITAL_INPUT_RATE;
  ALgetparams(AL_DEFAULT_DEVICE, buf, 4);

/* output rate is in Hz -- return it */
  if (buf[1] > 0) 
    return (buf[1]);

/* output rate is logical -- track down what it means */
  else 
  {
    if	    (buf[1] == AL_RATE_AES_1) 
    {
          /* we are clocked off of digital input. find
           * real input rate, if system supports this ability.
           * Otherwise, return AL_RATE_UNDEFINED */
      if (ALgetdefault(AL_DEFAULT_DEVICE,AL_DIGITAL_INPUT_RATE) >= 0) 
        return (buf[3]);
    }
    return (AL_RATE_UNDEFINED);
  }
}

static int	samps_per_frame;	/* frame size */
static long	bits_per_samp;		/* sample resolution */
static int	frames_per_sec;		/* sample rate */
static int	bytes_per_samp;		/* sample width */
static double	secs_per_frame;		/* time to play one audio frame */
static int	frames_per_buf;		/* frames / sample buffer */
static int	samps_per_buf;		/* samples / sample buffer */
static int	bytes_per_buf;		/* bytes / sample buffer */
static double	secs_per_buf;		/* time to play one sample buffer */
static double	file_rate;		/* audio file sample rate */

void
SgiSoundDevice::_Play(Pad_AudioFormat *format, char *, long){
    int num_bufs;
    int leftover_bytes;
    int leftover_samps;
    int leftover_frames;
    int samp_count;
    int frame_count;
    double sec_count;
    int i;
    int samples_read;
    int total_frames;   
    int total_samps;
    int total_samp_bytes;
    char *sampbuf;
    
    sec_count = 0.0;
    
    /*
     * figure out how many reads we have to do
     */
    total_frames    =  format->Get_length();

    if (total_frames > 0) {
      total_samps      =  total_frames * samps_per_frame;
      total_samp_bytes =  total_samps * bytes_per_samp;
      num_bufs         = total_samp_bytes / bytes_per_buf;
      leftover_bytes   = total_samp_bytes % bytes_per_buf;
      leftover_samps   = leftover_bytes / bytes_per_samp;
      leftover_frames  = leftover_samps / samps_per_frame;
    }
    
    /*
     * note that there may be some pad bytes following the valid samples -
     * for example, the sample data area may be padded so that the valid 
     * samples begin on a block boundary and the sample area ends on a block
     * boundary (where blocksize is specified by the user)
     */
    samp_count  = 0;
    frame_count = 0; 
    sec_count   = 0.0;
    sampbuf = (char *) format->Get_data();
    
    for (i=0; num_bufs<0 || i<num_bufs; i++)
    {
        samp_count  += samps_per_buf;
        frame_count += frames_per_buf;
        sec_count   += secs_per_buf;
        
        samples_read = frames_per_buf * samps_per_frame;
        ALwritesamps(_audioPort, sampbuf, samples_read);
        sampbuf += samples_read;
    }
    /*
     * play the leftovers
     */
    ALwritesamps(_audioPort, sampbuf, leftover_samps);
}


int
SgiSoundDevice::_Init_play(Pad_AudioFormat *format) {
  int rude;
  long pvbuf[4];
  long audio_rate;
  long output_rate;
  long samp_type;
  long samp_wordsize;
  int err;
  
      // Reset the audio device, cancelling any sound that may be playing
  if (_audioPort != NULL){
    ALcloseport (_audioPort);
  }
  
  
    /*
     * need to determine whether audio is in use. if not, then we
     * can just go ahead and be "rude."
     */
    pvbuf[0] = AL_OUTPUT_COUNT;
    pvbuf[2] = AL_MONITOR_CTL;
    if (ALgetparams(AL_DEFAULT_DEVICE, pvbuf, 4) < 0) {
	if (oserror() == AL_BAD_DEVICE_ACCESS) {
          Pad_errorString = "Can't init hardware: could not access audio hardware.\n";
          return FALSE;
	}
    }

    
    if ((pvbuf[1] == 0) && (pvbuf[3] == AL_MONITOR_OFF)) {
        rude = 1;
    }

  samps_per_frame   = format->Get_channels();
  file_rate         = format->Get_rate();

  samp_type         = format->Get_style();
  bits_per_samp     = format->Get_bits_per_sample();
  
     //decide on output rate for the audio hardware
  audio_rate = frames_per_sec = file_rate;
  
      // determine current output rate
  output_rate = _Get_output_rate();

     // if the rates are the same, all is well. if not, then we need to proceed
     // in a either a "rude" or "polite" manner.
  if (output_rate != audio_rate) {
        /*
         *  we set the rate to our value in Hz.  ALsetparams will set
	 *  the hardware to the closest rate to audio_rate which the
	 *  hardware supports.
         */
    output_rate = audio_rate;
    pvbuf[0] = AL_OUTPUT_RATE;
    pvbuf[1] = output_rate;
    ALsetparams(AL_DEFAULT_DEVICE, pvbuf, 2);
  }

    /*
     * decide what size blocks of samples we should read from the
     * input file and pass to ALwritesamps
     */
    if (bits_per_samp <= 8)
    {
        bytes_per_samp = 1;
        samp_wordsize  = AL_SAMPLE_8;
    }
    else if (bits_per_samp <= 16)
    {
        bytes_per_samp = 2;
        samp_wordsize  = AL_SAMPLE_16;
    }
    else if (bits_per_samp <= 24)
    {
        bytes_per_samp = 4;
        samp_wordsize  = AL_SAMPLE_24;
    }

  
    /*
     * make the buffer large enough to hold 1/2 sec of audio frames
     * we add one to frames_per_sec before we divide in order to
     * correctly handle the 11025 case
     */
    secs_per_frame = 1.0 / ((double)frames_per_sec);
    frames_per_buf = (frames_per_sec+1)/2;
    
    samps_per_buf = frames_per_buf * samps_per_frame;
    bytes_per_buf = samps_per_buf * bytes_per_samp;
    secs_per_buf  = secs_per_frame * frames_per_buf;

    int bufferSize;
    bufferSize = bytes_per_buf;
  _audioBuffer = (char *)new unsigned char[bufferSize];
  
    /*
     * configure and open audio port
     */

  ALsetchannels(_audioConfig, samps_per_frame);
  ALsetwidth   (_audioConfig, samp_wordsize);
  ALsetsampfmt (_audioConfig, samp_type);

    /*
     * make the ring buffer large enough to hold 1 sec of audio samples
     */
  ALsetqueuesize(_audioConfig, samps_per_buf*2);

  _audioPort = ALopenport (_deviceName->Get(), "w", _audioConfig);
  if (_audioPort == 0) {
    err = oserror();
    if (err == AL_BAD_NO_PORTS) {
      Pad_errorString = "Can't init hardware: no audio ports available.\n";
    }
    else if (err == AL_BAD_OUT_OF_MEM) {
      Pad_errorString = "Can't init hardware: not enough memory to open audio port.\n";
    }
    return FALSE;
  }
  
  return TRUE;
}


void
SgiSoundDevice::_Stop_play(void){
  _Flush();
      // Here's were we set a timer to stop the audio device
}

////////////////////////////////////////////////////////////////
//
// SgiAudioFormat Class
//
////////////////////////////////////////////////////////////////

SgiAudioFormat::SgiAudioFormat(){
}

SgiAudioFormat::SgiAudioFormat(AFfilehandle fh, const char *filename){
  _filename = new Pad_String(filename);
  _bitsPerSample = 0;
  _sampType = 0;
  _rate = 0;
  _channels = 0;
  _fileHandle = fh;
}

SgiAudioFormat::~SgiAudioFormat(){
}

Pad_Bool
SgiAudioFormat::Open_format(char *filename){
  int fd;
  AFfilehandle fh;
  
  if (fd = open(filename, O_RDONLY, 0)){
        // if we successfully opened it 
    if (fh = AFopenfd (fd, "r", 0)){
	_fileHandle = fh;
	return TRUE;
    }
  }
  return FALSE;
}

long
SgiAudioFormat::Read(void){
  int frames_read;
  
  _channels   = AFgetchannels(_fileHandle, AF_DEFAULT_TRACK);
  _rate       = AFgetrate(_fileHandle, AF_DEFAULT_TRACK);
  AFgetsampfmt(_fileHandle, AF_DEFAULT_TRACK, &_sampType, &_bitsPerSample);

				// Weird bug fix.  AFgetsampfmt is supposed
				// to return a nice constant, but doesn't seem
				// to in practice, so we are hard-coding here
				// which is the right value for AIFF files.
  _sampType = AL_SAMPFMT_TWOSCOMP;

      // figure out how long the sound data is
  _length = AFgetframecnt(_fileHandle, AF_DEFAULT_TRACK);
  _data   = new char[_length * _channels * (_bitsPerSample / 8)];

      // read the data into the internal buffer
  AFseekframe(_fileHandle, AF_DEFAULT_TRACK, 0);
  frames_read = AFreadframes(_fileHandle, AF_DEFAULT_TRACK, 
                             _data, _length);
  return frames_read;
}

//
// Get function for the sound data sample stlye. All samples
//   are stored internally as two's complement signed bytes.
//
Pad_SoundFormat
SgiAudioFormat::Get_style() const {
  return (Pad_SoundFormat)_sampType;
}

//
// Get function for the sound data sample frequency
//
long 
SgiAudioFormat::Get_rate() const {
  return _rate;
}

//
// Get function for the sound data width (8/16 bit) 
//
long 
SgiAudioFormat::Get_bits_per_sample() const {
  return _bitsPerSample;
}

//
// Get function for the number of channels in the sound, the constants
//   values are 1 (AL_MONO), or 2 (AL_STEREO).
//
long 
SgiAudioFormat::Get_channels() const {
  return _channels;
}

long
SgiAudioFormat::Get_length(void) const{
  return _length;
}

char *
SgiAudioFormat::Get_data(void) const{
  return _data;
}

void *
SgiAudioFormat::Get_private(void) const{
  return NULL;
}

#endif /* HAS_SGI_SOUND */
