#ifndef SOUND_TOOLS_H
#define SOUND_TOOLS_H

/*
 * July 5, 1991
 * Copyright 1991 Lance Norskog And Sundry Contributors
 * This source code is freely redistributable and may be used for
 * any purpose.  This copyright notice must be maintained. 
 * Lance Norskog And Sundry Contributors are not responsible for 
 * the consequences of using this software.
 */

/*
 * Sound Tools sources header file.
 */

/*
 * Handler structure for each format.
 */
struct soundstream;

enum Filetype {
  FILETYPE_UNDEF,
  FILETYPE_RAW,
  FILETYPE_AU,
  FILETYPE_AIFF,
  FILETYPE_WAV
};

class SoundFormat {
public:
  SoundFormat(){}
  virtual ~SoundFormat(){}
  virtual void startread(soundstream *)=0;			
  virtual int  read(soundstream *, long *, long)=0;			
  virtual void stopread(soundstream *)=0;		
private:
};

class AUSoundFormat : public SoundFormat{
public:
  AUSoundFormat(){}
  virtual ~AUSoundFormat(){}
  virtual void startread(soundstream *);			
  virtual int  read(soundstream *, long *, long);			
  virtual void stopread(soundstream *);		
private:
};

class WAVSoundFormat : public SoundFormat{
public:
  WAVSoundFormat(){}
  virtual ~WAVSoundFormat(){}
  virtual void startread(soundstream *);			
  virtual int  read(soundstream *, long *, long);			
  virtual void stopread(soundstream *);		
private:
  static char *_format_str(unsigned wFormatTag);
};

class AIFFSoundFormat : public SoundFormat{
public:
  AIFFSoundFormat(){}
  virtual ~AIFFSoundFormat(){}
  virtual void startread(soundstream *);			
  virtual int  read(soundstream *, long *, long);
  virtual void stopread(soundstream *);		
private:
  double read_ieee_extended(soundstream *);
  double ConvertFromIeeeExtended(unsigned char *);
};

/* Signal parameters */

struct  signalinfo {
	long		rate;		/* sampling rate */
	int		size;		/* word length of data */
	int		style;		/* format of sample numbers */
	int		channels;	/* number of sound channels */
};

/* Loop parameters */

struct  loopinfo {
	int		start;		/* first sample */
	int		length;		/* length */
	int		count;		/* number of repeats, 0=forever */
	int		type;		/* 0=no, 1=forward, 2=forward/back */
};

/* Instrument parameters */

/* vague attempt at generic information for sampler-specific info */

struct  instrinfo {
	char 		MIDInote;	/* for unity pitch playback */
	char		MIDIlow, MIDIhi;/* MIDI pitch-bend range */
	char		loopmode;	/* semantics of loop data */
	char		nloops;		/* number of active loops */
	unsigned char	smpte[4];	/* SMPTE offset (hour:min:sec:frame) */
					/* this is a film audio thing */
};


#define MIDI_UNITY 60		/* MIDI note number to play sample at unity */

/* Loop modes */
#define LOOP_NONE          0	
#define LOOP_8             1	/* 8 loops: don't know ?? */
#define LOOP_SUSTAIN_DECAY 2	/* AIFF style: one sustain & one decay loop */

/* Pipe parameters */

#include <stdio.h>
struct	pipeinfo {
	FILE	*pout;			/* Output file */
	FILE	*pin;			/* Input file */
};

/*
 *  Format information for input and output files.
 */

#define	PRIVSIZE	100

#define NLOOPS		8

struct soundstream {
	struct	signalinfo info;	/* signal specifications */
	struct  instrinfo instr;	/* instrument specification */
	struct  loopinfo loops[NLOOPS];	/* Looping specification */
	char	swap;			/* do byte- or word-swap */
	char	seekable;		/* can seek on this file */
	char	*filename;		/* file name */
	Filetype filetype;		/* type of file */
	char *       comment;		/* comment string */
	FILE *       fp;                /* File stream pointer */
	SoundFormat *h;			/* format struct for this file */
	double       priv[PRIVSIZE/8];	/* format's private data area */
};

/* flags field */
#define FILE_STEREO	1	/* does file format support stereo? */
#define FILE_LOOPS	2	/* does file format support loops? */
#define FILE_INSTR	4	/* does file format support instrument specificications? */

/* Size field */
#define	BYTE	1
#define	WORD	2
#define	LONG	4
#define	FLOAT	5
#define DOUBLE	6
#define IEEE	7		/* IEEE 80-bit floats.  Is it necessary? */

/* Style field */
#define UNSIGNED	1	/* unsigned linear: Sound Blaster */
#define SIGN2		2	/* signed linear 2's comp: Mac */
#define	ULAW		3	/* U-law signed logs: US telephony, SPARC */
#define ALAW		4	/* A-law signed logs: non-US telephony */

extern char *sizes[], *styles[];

#define	P1(x) x
#define	P2(x,y) x, y
#define	P3(x,y,z) x, y, z
#define	P4(x,y,z,w) x, y, z, w

/* raw file reading and writing utility */
int rawread (soundstream *, long *, long);
void rawwrite(soundstream *, long *, long);

/* Utilities to read and write shorts and longs little-endian and big-endian */
unsigned short rlshort(P1(soundstream * ft));/* short little-end */
unsigned short rbshort(P1(soundstream * ft)); /* short big-end    */
void wlshort(P2(soundstream * ft,
                          unsigned short us)); /* short little-end */
void wbshort(P2(soundstream * ft,
                          unsigned short us)); /* short big-end    */
unsigned long  rllong(P1(soundstream * ft)); /* long little-end  */
unsigned long  rblong(P1(soundstream * ft)); /* long big-end     */
void wllong(P2(soundstream * ft,
                         unsigned long ul)); /* long little-end  */
void  wblong(P2(soundstream * ft,
                         unsigned long ul));	/* long big-end     */

/* Read and write words and longs in "machine format".  Swap if indicated.  */
unsigned short rshort(P1(soundstream * ft));			
void wshort(P2(soundstream * ft, unsigned short us));
unsigned long  rlong(P1(soundstream * ft));		
void  wlong(P2(soundstream * ft, unsigned long ul));
float          rfloat(P1(soundstream * ft));
void           wfloat(P2(soundstream * ft, double f));
double         rdouble(P1(soundstream * ft));
void           wdouble(P2(soundstream * ft, double d));

/* Utilities to byte-swap values */
unsigned short swapw(P1(unsigned short us));		/* Swap short */
unsigned long  swapl(P1(unsigned long ul));		/* Swap long */
float  	       swapf(P1(float f));			/* Swap float */
double 	       swapd(P1(double d));			/* Swap double */

#ifdef ARM
extern double sfloor(P1(double x));   /* Hack our way around the flawed */
extern double sceil(P1(double x));    /* UnixLib floor ceil functions */
#endif


extern void report(P2(char *, ...));
extern void warn(P2(char *, ...));
extern void fail(P2(char *, ...));

typedef	unsigned int u_i;
typedef	unsigned long u_l;
typedef	unsigned short u_s;

extern float volume;	/* expansion coefficient */
extern int dovolume;

extern float amplitude;	/* Largest sample so far */

extern int writing;	/* are we writing to a file? */

extern int verbose;	/* be noisy on stderr */
extern int summary;	/* just print summary of information */

extern char *myname;

extern int soxpreview;	/* Preview mode: be fast and ugly */

#define	MAXRATE	50L * 1024			/* maximum sample rate */

/* Some wacky processors don't have arithmetic down shift, so do divs */
/* Most compilers will turn this into a shift if they can, don't worry */
#define RIGHT(datum, bits)	((datum) / (1L << bits)) 
#define LEFT(datum, bits)	((datum) << bits) 

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

#define READBINARY	"r"
#define WRITEBINARY	"w"

#include <errno.h>

#define REMOVE unlink



#endif /* SOUND_TOOLSH */
