/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "pad-string.h"
#include "tkwin.h"
#include "view.h"
#include "portal.h"
#include "pad-tcl.h"
#include "text.h"
#include "point.h"
#include "list.h"
#include "pad.h"
#include "html.h"
#include "image.h"
#include "group.h"
#include "bind.h"
#include "object.h"
#include "callback.h"
#include "flist.h"
#include "ilist.h"
#include "curves.h"
#include "tree.h"
#include "tree-layout.h"
#include "hashtab.h"
#include "global.h"
#include "imagedata.h"
#include "misc.h"
#include "fontdata.h"
#include "api.h"
#include "anim-path.h"
#include "anim-channel.h"
#include "anim.h"
#include "animatable.h"
#include "cache.h"

#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <time.h>          // For the getdate command.  Need to generalize this for windows.

#define UCHAR(c) ((unsigned char) (c))

#ifdef PAD_UNIX
#  include <unistd.h>
#endif

extern "C" double noise1(double);



//
// Padtcl_damage:
//
//  pathName damage ?tagOrId?
//
//     Damage the specified object (causing it to be redrawn),
//     Or the entire surface if no tagOrId is specified
//
int
Padtcl_damage(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Iterator oi;
    Pad_List objs;
    Pad_Object *obj;
    Pad_Bool prev_damage_enabled;

    if ((argc != 1) && (argc != 2))
      {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName damage ?tagOrId?\"");
	  return(TCL_ERROR);
      }

    if (argc == 1) {
	tkwin->view->Damage();
    } else {
	tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
	DOLIST(oi, objs, Pad_Object, obj)
	  {
				// Normally, damage is disabled within renders
				// Force damage to occur within renders
	      prev_damage_enabled = obj->pad->view->win->damageEnabled;
	      obj->pad->view->win->damageEnabled = TRUE;
	      obj->Damage();
	      obj->pad->view->win->damageEnabled = prev_damage_enabled;
	  }
    }

    return(TCL_OK);
}

/*
 Padtcl_read

  pathName read filename

    Read the specified pad file
*/
int
Padtcl_read(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int rc;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName read filename\"");
    }

    if (tkwin->view->pad->Read(argv[1])) {
        rc = TCL_OK;
    } else {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
        rc = TCL_ERROR;
    }

    return(rc);
}

/*
 Padtcl_write:

  pathName write [-format type] [-relative] filename ?tagOrId ...?

     Write into the specified filename the tcl commands necessary to create
     the specified objects.  If no objects are specified, all objects will
     be written out including a command to restore the current view
     If <filename> is an empty string, then this doesn't
     write into any file, but instead returns the result as a string.  Otherwise,
     this returns an empty string
*/

int
Padtcl_write(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int format, i;
    int rc;
    char *filename;
    Pad_Bool found, relative;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Object *obj;
    Pad_List objs, writeObjs;
    Pad_Iterator oi;
    Pad_String str, result;
    

    if (argc < 2) {
	  Tcl_AppendResult(interp, "wrong # args: should be \"",
			   "pathName write [-format type] [-relative] filename ?tagOrId ...?\"", NULL);
	  return(TCL_ERROR);
    }

				// Check for flags in any order
    relative = FALSE;
    format = PAD_SCRIPT_FORMAT;
    found = TRUE;
    while ((argc >= 2) && found) {
	found = FALSE;
        if (!strcasecmp(argv[1], "-format")) {
	    if (argc >= 3) {
		if (!strcasecmp(argv[2], "text")) {
		    format = PAD_SCRIPT_FORMAT;
		} else if (!strcasecmp(argv[2], "binary-interchange")) {
		    format = PAD_XDR_FORMAT;
		} else if (!strcasecmp(argv[2], "binary-native")) {
		    format = PAD_NATIVE_FORMAT;
		} else if (!strcasecmp(argv[2], "binary-debug")) {
		    format = PAD_CHAR_FORMAT;
		} else {
		    TCL_ERR(interp, "-format must be followed by format tye: text, binary-interchange, or binary-native");
		}
		found = TRUE;
		argc -= 2;
		argv += 2;
	    } else {
		TCL_ERR(interp, "-format must be followed by format tye: text, binary-interchange, or binary-native");
	    }
	} else if (!strcasecmp(argv[1], "-relative")) {
	    relative = TRUE;
	    found = TRUE;
	    argc--;
	    argv++;
	}
    }

    if (argv[1][0] == '\0') {
	filename = NULL;
    } else {
	filename = argv[1];
    }

    for (i=2; i<argc; i++) {
	tkwin->view->pad->Find_with_tagorid(argv[i], objs, TRUE);
	DOLIST(oi, objs, Pad_Object, obj) {
	    writeObjs.Push_last(obj);
	}
    }
    
    if (tkwin->view->pad->Write(filename, writeObjs, result, format, relative)) {
	Tcl_ResetResult(interp);
	if (filename == NULL) {
	    Tcl_AppendResult(interp, result.Get(), NULL);
	}
        rc = TCL_OK;
    } else {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
        rc = TCL_ERROR;
    }

    return(rc);
}

/*
 Padtcl_push_coord_frame

  pathName pushcoordframe tagOrId
     or
  pathName pushcoordframe x1 y1 x2 y2

  Push the specified bounding box (or bbox of tagOrId) onto the coordinate frame stack,
  and use it as the frame for all future coordinates.

  Returns resulting coordinate frame
 */
int
Padtcl_push_coord_frame(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Object *obj;
    Pad_List objs;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    float *frame = NULL;

    if ((argc != 2) && (argc != 5))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pushcoordframe tagOrId\",\nor \"pushcoordframe x1 y1 x2 y2\"");
      }

    frame = NULL;
    if (argc == 2) {
	tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
	obj = (Pad_Object *)objs.First();
	if (obj) {
	    frame = new float[4];
	    obj->Get_global_bbox(frame);
	}
    }
    if (argc == 5) {
	frame = new float[4];
	frame[XMIN] = ATOXF(tkwin, argv[1]);
	frame[YMIN] = ATOYF(tkwin, argv[2]);
	frame[XMAX] = ATOXF(tkwin, argv[3]);
	frame[YMAX] = ATOYF(tkwin, argv[4]);
    }

    if (frame) {
	if ((frame[XMIN] == frame[XMAX]) ||
	    (frame[YMIN] == frame[YMAX])) {
	    delete [] frame;
	    TCL_ERR(interp, "Coordinate frame must have non-zero width and height");
	} else {
	    Pad_coordFrames.Push(frame);

	    Tcl_ResetResult(interp);
	    sprintf(Pad_result, "%f %f %f %f",
		    XTOUNITS(tkwin, frame[XMIN]),
		    YTOUNITS(tkwin, frame[YMIN]),
		    XTOUNITS(tkwin, frame[XMAX]),
		    YTOUNITS(tkwin, frame[YMAX]));
	    TCL_RESULT(interp, Pad_result);
	}
    } else {
	Tcl_ResetResult(interp);
    }

    return(TCL_OK);
}

/*
 Padtcl_pop_coord_frame

  pathName popcoordframe

  Pop the top coordinate frame off the stack.
  If the stack is empty, return an error.
  Otherwise, return the frame taken off the top of the stack.
 */
int
Padtcl_pop_coord_frame(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{
    float *frame;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName popcoordframe\"");
      }

    if (Pad_coordFrames.Is_empty()) {
	TCL_ERR(interp, "Coordinate stack empty");
    }

    frame = (float *)Pad_coordFrames.Pop();

    Tcl_ResetResult(interp);
    sprintf(Pad_result, "%f %f %f %f",
	    XTOUNITS(tkwin, frame[XMIN]),
	    YTOUNITS(tkwin, frame[YMIN]),
	    XTOUNITS(tkwin, frame[XMAX]),
	    YTOUNITS(tkwin, frame[YMAX]));
    TCL_RESULT(interp, Pad_result);

    delete [] frame;

    return(TCL_OK);
}

/*
 Padtcl_reset_coord_frame

  pathName resetcoordframe

  Pop all frames off the coordinate frame stack.

  Return empty string.
 */
int
Padtcl_reset_coord_frame(ClientData , Tcl_Interp *interp, int argc, char **)
{
    float *frame;

    if (argc != 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName resetcoordframe\"");
      }

    while (!Pad_coordFrames.Is_empty()) {
	frame = (float *)Pad_coordFrames.Pop();
	delete [] frame;
    }

    return(TCL_OK);
}

/*
 Padtcl_text

  pathName text tagOrId <command>
  Where <command> is one of:
    bbox    index1 ?index2?
    compare index1 op index2
    delete  index1 ?index2?
    get     index1 ?index2?
    index   index ?char?
    insert  index chars
    set     option value

  Handles text subcommand
 */
int
Padtcl_text(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int rc = TCL_OK;
    int len;
    Pad_List objs;
    Pad_Object *obj;
    Pad_Text *text = NULL;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc < 3)
	{
	    TCL_ERR(interp, "wrong # args: should be \"pathName text tagOrId option\"");
	}

    len = strlen(argv[2]);
    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj) {
	text = obj->Get_text_obj();
    }
    if (!text) {
	Tcl_AppendResult(interp, "Item not a valid text object: ",
			 argv[1], NULL);
	return(TCL_ERROR);
    } else {
	if (!strncmp(argv[2], "bbox", MAX(1, len)))
	    {
		rc = text->Bbox_cmd(interp, argc-2, argv+2);
	    }
	else if (!strncmp(argv[2], "compare", MAX(1, len)))
	    {
		rc = text->Compare(interp, argc-2, argv+2);
	    }
	else if (!strncmp(argv[2], "delete", MAX(1, len)))
	    {
		rc = text->Del(interp, argc-2, argv+2);
	    }
	else if (!strncmp(argv[2], "get", MAX(1, len)))
	    {
		rc = text->Get(interp, argc-2, argv+2);
	    }
	else if (!strncmp(argv[2], "index", MAX(3, len)))
	    {
		rc = text->Index(interp, argc-2, argv+2);
	    }
	else if (!strncmp(argv[2], "insert", MAX(3, len)))
	    {
		rc = text->Insert(interp, argc-2, argv+2);
	    }
	else if (!strncmp(argv[2], "mark", MAX(1, len)))
	    {
		rc = text->Mark(interp, argc-2, argv+2);
	    }
	else {
	    TCL_ERR(interp, "Invalid option, should be bbox, compare, delete, get, index, insert, or mark");
	}
    }

    return(rc);
}

/*
 Padtcl_update:

  pathName update ?-dissolve <speed>? ?withRefinement?

     Force scheduled screen updates to happen immediately.
     If 'withRefinement' is specified, then also force all scheduled
     refinements to happen now as well.

     If -dissolve is specified, then use a dissolve effect when updating.
 */
int
Padtcl_update(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    if ((argc < 1) || (argc > 4))
      {
	  TCL_ERR(interp, "wrong # args: should be \"update ?-dissolve <speed>? ?withRefinement?\"");
      }

    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    int dissolve_speed = 0;

    if (argc >= 3) {
        if (!strcmp(argv[1], "-dissolve")) {
	    dissolve_speed = atoi(argv[2]);    
	    argc -= 2;
	}
    }
    tkwin->view->Update_display(dissolve_speed, argc == 1 ? FALSE : TRUE);
    return(TCL_OK);
}

static Pad_List urls;

class Url {
  public:
    int token;
    Pad_Callback *updateCb;
    Pad_Callback *doneCb;
    Pad_Callback *errorCb;
    Pad_String data;

    ~Url();
    Url();
};

Url::~Url()
{
    if (updateCb) {
	delete updateCb;
	updateCb = NULL;
    }
    if (doneCb) {
	delete doneCb;
	doneCb = NULL;
    }
    if (errorCb) {
	delete errorCb;
	errorCb = NULL;
    }
}

Url::Url()
{
    updateCb = NULL;
    doneCb = NULL;
    errorCb = NULL;
}

//////////////////////////////////////////////////
//  URL stuff
//////////////////////////////////////////////////

//
// Callback for when a URL is updated (i.e., more data arrived)
//
static Pad_Bool
Url_UpdateCB(ClientData clientData, int token, char *data, int len)
{
    int result;
    Url *url;
    Pad_Iterator oi;
    Pad_String cmd;
    Tcl_Interp *interp = (Tcl_Interp *)clientData;

    DOLIST(oi, urls, Url, url) {
	if (url->token == token) {
	    url->data.Set(data, len);
	    if (url->updateCb) {
		cmd = url->updateCb->Get();
		cmd += " url";
		cmd += token;
		result = Pad_GlobalEval(url->updateCb->Language(), cmd.Get());
		if (result == TCL_ERROR) {
		    Tcl_AddErrorInfo(interp, "\n    (Pad URL callback)");
		    Pad_Background_error(interp);
		}
	    }
	    break;
	}
    }

    return(TRUE);
}

//
// Callback for when a URL has completed loading
//
static Pad_Bool
Url_DoneCB(ClientData clientData, int token)
{
    int result;
    Url *url;
    Pad_Iterator oi;
    Pad_String cmd;
    Tcl_Interp *interp = (Tcl_Interp *)clientData;

    DOLIST(oi, urls, Url, url) {
	if (url->token == token) {
	    if (url->doneCb) {
		cmd = url->doneCb->Get();
		cmd += " url";
		cmd += token;
		result = Pad_GlobalEval(url->doneCb->Language(), cmd.Get());
		if (result == TCL_ERROR) {
		    Tcl_AddErrorInfo(interp, "\n    (Pad URL callback)");
		    Pad_Background_error(interp);
		}
	    }
	    break;
	}
    }

    return(TRUE);
}

//
// Callback for when there has been an error with the URL
//
static Pad_Bool
Url_ErrorCB(ClientData clientData, int token, char *data, int len)
{
    int result;
    Url *url;
    Pad_Iterator oi;
    Pad_String cmd;
    Tcl_Interp *interp = (Tcl_Interp *)clientData;

    DOLIST(oi, urls, Url, url) {
	if (url->token == token) {
	    url->data.Set(data, len);
	    if (url->errorCb) {
		cmd = url->errorCb->Get();
		cmd += " url";
		cmd += token;
		result = Pad_GlobalEval(url->errorCb->Language(), cmd.Get());
		if (result == TCL_ERROR) {
		    Tcl_AddErrorInfo(interp, "\n    (Pad URL callback)");
		    Pad_Background_error(interp);
		}
	    }
	    break;
	}
    }

    return(TRUE);
}

//
//    .pad urlfetch <url> ?option value ...?
//    where valid options are:
//       -file <filename>
//       -var <var>
//       -updatescript <script>     # <token> gets added onto the end of <script>
//       -donescript <script>       # <token> gets added onto the end of <script>
//       -errorscript <script>      # <token> gets added onto the end of <script>
//    Returns a token to be used by future calls to urlfetch
//             or
//    .pad urlfetch <token>
//    Returns the most recent data fetched.
//
int
Padtcl_urlfetch(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int token;
    char *urlname;
    char *file;
    char *var;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Url *url;
    Pad_Bool found;
    Pad_Iterator oi;
    Pad_UrlItem *urlItem;

    if (argc < 2)
      {
	  Tcl_AppendResult(interp,
			   "wrong # args: should be \"pathName urlfetch URL ?option value ...?\"\n",
			   "                     or \"pathName urlfetch token\"", NULL);
	  return TCL_ERROR;
      }

    if (!strncmp(argv[1], "url", 3)) {
	token = atoi(&argv[1][3]);
	DOLIST(oi, urls, Url, url) {
	    if (url->token == token) {
		Tcl_AppendResult(interp, url->data.Get(), NULL);
		break;
	    }
	}
    } else {
	url = new Url;
	urlname = argv[1];

				// Check for flags in any order
	file = NULL;
	var = NULL;
	found = TRUE;
	while ((argc >= 4) && found) {
	    found = FALSE;
	    if (!strcmp(argv[2], "-updatescript")) {
		url->updateCb = new Pad_Callback(argv[3], Pad_callbackLanguage);
		found = TRUE;
	    }
	    else if (!strcmp(argv[2], "-donescript")) {
		url->doneCb = new Pad_Callback(argv[3], Pad_callbackLanguage);
		found = TRUE;
	    }
	    else if (!strcmp(argv[2], "-errorscript")) {
		url->errorCb = new Pad_Callback(argv[3], Pad_callbackLanguage);
		found = TRUE;
	    }
	    else if (!strcmp(argv[2], "-file")) {
		file = argv[3];
		found = TRUE;
	    }
	    else if (!strcmp(argv[2], "-var")) {
		var = argv[3];
		found = TRUE;
	    }
	    if (found) {
		argc -= 2;
		argv += 2;
	    }
	}
	
	urlItem = tkwin->view->pad->Url_fetch(urlname, file, var, tkwin->interp, TRUE, 
						   Url_UpdateCB, Url_DoneCB, Url_ErrorCB);
	if (urlItem) {
	    url->token = urlItem->token;
	    urls.Push_last(url);
	} else {
	    delete url;
	    TCL_ERR(interp, "Error fetching URL");
	}
	
	sprintf(Pad_result, "url%d", url->token);
	TCL_RESULT(interp, Pad_result);
    }

    return(TCL_OK);
}

/*
 Padtcl_center:

  pathName center ?-twostep? tagOrId ?animationTime ?x y ?z ?portalID ...????

     Change the view so as to center the specified object so it's bounding box
     fills three quarters of the screen. Or if x, y, and z are specified, then
     tune to taste (percentages - default to .5, .5, .75).
     If '-twostep' is specified, and it is appropriate, break up animation into two
     motions, with a midpoint at the point halfway between the start and endpoints,
     zoomed out such that both points are visible.
     If a list of portalID's is specified, change the view within the last one specified.
 */
int
Padtcl_center(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    int id;
    float x, y, z;
    int animationTime;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Object *obj;
    Pad_List objs;
    Pad_View *view;
    Pad *pad;
    Pad_Bool twostep, rc;

    if (argc < 2)
      {
	  TCL_ERR(interp,
		  "wrong # args: should be \"pathName center ?-twostep? tagOrId ?animationTime ?x y ?z ?portalID ...????\"");
      }

    animationTime = 0;
    x = 0.5;
    y = 0.5;
    z = 0.75;
    rc = 0;
    twostep = 0;

    if (!strcmp(argv[1], "-twostep")) {
	twostep = 1;
	argc--;
	argv++;
    }

    if (argc >= 3)
      {
	  animationTime = atoi(argv[2]);
	  if (argc >= 5)
	    {
		x = atof(argv[3]);
		y = atof(argv[4]);
		if (argc >= 6)
		  {
		      z = atof(argv[5]);
		  }
	    }
      }

				// Extract last portal from list
    view = tkwin->view;
    pad = view->pad;
    for (i=6; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    id = atoi(argv[i]);
	    obj = pad->Get_object_from_id(id);
	    if (obj &&
		(obj->Type() == PAD_PORTAL)) {
		view = (Pad_View *)obj;
		pad = ((Pad_Portal *)obj)->lookon->view->pad;
	    } else {
		Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				 view->win->Get_name(), NULL);
		return(TCL_ERROR);
	    }
	}
    }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    view->Center(objs, animationTime, twostep, x, y, z);

    return(TCL_OK);
}

/*
 Padtcl_center_bbox:

  pathName centerbbox ?-twostep? x1 y1 x2 y2 ?animationTime ?x y ?z ?portalID ...????

     Change the view so as to center the specified bounding box so that it
     fills three quarters of the screen. Or if x, y, and z are specified, then
     tune to taste (percentages - default to .5, .5, .75).
     If '-twostep' is specified, and it is appropriate, break up animation into two
     motions, with a midpoint at the point halfway between the start and endpoints,
     zoomed out such that both points are visible.
     If a list of portalID's is specified, change the view within the last one specified.
 */
int
Padtcl_center_bbox(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, id;
    float x, y, z;
    float x1, y1, x2, y2;
    int animationTime;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Bool twostep, rc;
    Pad *pad;
    Pad_View *view;
    Pad_Object *obj;

    if (argc < 5)
      {
	  TCL_ERR(interp,
		  "wrong # args: should be \"pathName centerbbox ?-twostep? x1 y1 x2 y2 ?animationTime ?x y ?z ?portalID ...????\"");
      }

    animationTime = 0;
    x = 0.5;
    y = 0.5;
    z = 0.75;
    rc = 0;
    twostep = 0;

    if (!strcmp(argv[1], "-twostep")) {
	twostep = 1;
	argc--;
	argv++;
    }

    x1 = ATOXF(tkwin, argv[1]);
    y1 = ATOYF(tkwin, argv[2]);
    x2 = ATOXF(tkwin, argv[3]);
    y2 = ATOYF(tkwin, argv[4]);
    if (argc >= 6) {
	animationTime = atoi(argv[5]);
	if (argc >= 7)
	  {
	      x = atof(argv[6]);
	      y = atof(argv[7]);
	      if (argc >= 9)
		{
		    z = atof(argv[8]);
		}
	  }
    }

				// Extract last portal from list
    view = tkwin->view;
    pad = view->pad;
    for (i=9; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    id = atoi(argv[i]);
	    obj = pad->Get_object_from_id(id);
	    if (obj &&
		(obj->Type() == PAD_PORTAL)) {
		view = (Pad_View *)obj;
		pad = ((Pad_Portal *)obj)->lookon->view->pad;
	    } else {
		Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				 view->win->Get_name(), NULL);
		return(TCL_ERROR);
	    }
	}
    }

    view->Center(x1, y1, x2, y2, animationTime, twostep, x, y, z);

    return(TCL_OK);
}

//
// Padtcl_move_to
//
//   pathName moveto ?-twostep? xview yview zoom ?animationTime ?portalID portalID ...??
//
//     Move to new position.  If zoom is given as "", then don't change the zoom.
//     If '-twostep' is specified, and it is appropriate, break up animation into two
//     motions, with a midpoint at the point halfway between the start and endpoints,
//     zoomed out such that both points are visible.
//     If animationTime is specified, move with animation of 'animationTime' milliseconds.
//     If a list of portalID's is specified, change the view within the last one specified.
//     Return resultant xview, yview, zoom.
//
int
Padtcl_move_to(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, id;
    int animationTime;
    Pad_Bool twostep = FALSE;
    Pad_Bool x_flag, y_flag, zoom_flag;
    float xview, yview, zoom;
    Pad *pad;
    Pad_View *view;
    Pad_Object *obj;

    if (argc < 4)
      {
	  TCL_ERR(interp,
		  "wrong # args: should be \"pathName moveto ?-twostep? xview yview zoom ?animationTime ?portalID ...??\"");
      }

    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (!strcmp(argv[1], "-twostep")) {
	twostep = 1;
	argc--;
	argv++;
    }

    if (!strcmp(argv[1], "")) {
	x_flag = FALSE;
	zoom = 1.0;
    } else {
	xview = ATOXF(tkwin, argv[1]);
	x_flag = TRUE;
    }
    if (!strcmp(argv[2], "")) {
	y_flag = FALSE;
	zoom = 1.0;
    } else {
	yview = ATOYF(tkwin, argv[2]);
	y_flag = TRUE;
    }
    if (!strcmp(argv[3], "")) {
	zoom_flag = FALSE;
	zoom = 1.0;
    } else {
	zoom = atof(argv[3]);
	zoom_flag = TRUE;
    }
    animationTime = 0;

    if (zoom <= 0.0)
      {
	  Tcl_AppendResult(interp, "moveto Error: invalid zoom = \"", argv[3], "\"", NULL);
	  return(TCL_ERROR);
      }

    if (argc > 4)
      {
	  animationTime = atoi(argv[4]);
      }

    view = tkwin->view;
    pad = view->pad;
    for (i=5; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    id = atoi(argv[i]);
	    obj = pad->Get_object_from_id(id);
	    if (obj &&
		(obj->Type() == PAD_PORTAL)) {
		view = (Pad_View *)obj;
		pad = ((Pad_Portal *)obj)->lookon->view->pad;
	    } else {
		Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				 view->win->Get_name(), NULL);
		return(TCL_ERROR);
	    }
	}
    }

				// If args not specified, then don't change them.
    if (x_flag == FALSE) {
	xview = view->xview;
    }
    if (y_flag == FALSE) {
	yview = view->yview;
    }
    if (zoom_flag == FALSE) {
	zoom = view->zoom;
    }

    view->Animate_to(xview, yview, zoom, animationTime, twostep);

    Tcl_ResetResult(interp);
    sprintf(Pad_result, "%g %g %g",
	    XTOUNITS(view->win, view->xview),
	    YTOUNITS(view->win, view->yview),
	    view->zoom);
    TCL_RESULT(interp, Pad_result);
    return(TCL_OK);
}

//
// Padtcl_getview
//
//   pathName getview ?portalID portalID ...?
//
//     Return current position (xoffset, yoffset, zoom).
//     If a list of portalID's is specified, return the view within the last one specified.
//
int
Padtcl_getview(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, id;
    Pad_Object *obj;
    Pad *pad;
    Pad_View *view;
    float zoom = 1.0;

    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    view = tkwin->view;
    pad = view->pad;
    zoom = view->zoom;
    for (i=1; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    id = atoi(argv[i]);
	    obj = pad->Get_object_from_id(id);
	    if (obj &&
		(obj->Type() == PAD_PORTAL)) {
		view = (Pad_View *)obj;
		zoom *= view->zoom;
		pad = ((Pad_Portal *)obj)->lookon->view->pad;
	    } else {
		Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				 view->win->Get_name(), NULL);
		return(TCL_ERROR);
	    }
	}
    }

    Tcl_ResetResult(interp);
    sprintf(Pad_result, "%g %g %g",
	    XTOUNITS(view->win, view->xview),
	    YTOUNITS(view->win, view->yview),
	    zoom);
    TCL_RESULT(interp, Pad_result);
    return(TCL_OK);
}

//
// Padtcl_getzoom
//
//   pathName getzoom ?portalID portalID ...?
//
//     Return current zoom.
//     If a list of portalID's is specified, return the zoom within the last one specified.
//
int
Padtcl_getzoom(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, id;
    Pad_Object *obj;
    Pad *pad;
    Pad_View *view;
    float zoom = 1.0;

    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    view = tkwin->view;
    pad = view->pad;
    zoom = view->zoom;
    for (i=1; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    id = atoi(argv[i]);
	    obj = pad->Get_object_from_id(id);
	    if (obj &&
		(obj->Type() == PAD_PORTAL)) {
		view = (Pad_View *)obj;
		zoom *= view->zoom;
		pad = ((Pad_Portal *)obj)->lookon->view->pad;
	    } else {
		Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				 view->win->Get_name(), NULL);
		return(TCL_ERROR);
	    }
	}
    }

    Tcl_ResetResult(interp);
    sprintf(Pad_result, "%g", zoom);
    TCL_RESULT(interp, Pad_result);
    return(TCL_OK);
}

//
// Zoom around specified pad position by multiplicitive factor.
// If an optional list of portals is specified, then change the view within the
// last portal.  The entire list is necessary in case the last portal is sitting
// on a different surface then this function is called with.
//
int
Padtcl_zoom(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    float zmult;
    float x, y;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    int i, id;
    int animationTime;
    Pad *pad;
    Pad_View *view;
    Pad_Object *obj;

    if (argc < 4)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName zoom zoomFactor padXloc padYloc ?animationTime ?portalID ...??\"");
      }

    zmult = atof(argv[1]);
    if (zmult < 0) {
	TCL_ERR(interp, "zoomFactor must be > 0.0");
    }

    x = ATOXF(tkwin, argv[2]);
    y = ATOYF(tkwin, argv[3]);
    view = tkwin->view;
    pad = view->pad;

    if (argc > 4) {
	animationTime = atoi(argv[4]);
	argc--;
	argv++;
    } else {
	animationTime = 0;
    }

    for (i=4; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    id = atoi(argv[i]);
	    obj = pad->Get_object_from_id(id);
	    if (obj &&
		(obj->Type() == PAD_PORTAL)) {
		view = (Pad_View *)obj;
		pad = ((Pad_Portal *)obj)->lookon->view->pad;
	    } else {
		Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				 view->win->Get_name(), NULL);
		return(TCL_ERROR);
	    }
	}
    }
    view->Zoom_around(x, y, zmult, animationTime);

    return(TCL_OK);
}

//
// Set Pad_Focus to specified object if specified.
// If a list of portals is specified, then the object specified
// sits on the surface the last portals looks onto.
// Return current focus.
//
int
Padtcl_focus(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, id;
    Pad_Object *obj;
    Pad_List objs;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_View *view;
    Pad *pad;

    if (argc < 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName focus ?tagOrId ?PortalID ...??\"");
      }

    if (argc >= 2)
      {
	  if (argv[1][0] == 0) {
	      tkwin->view->Update_focus(NULL);
	  } else {

	      view = tkwin->view;
	      pad = view->pad;

	      for (i=2; i<argc; i++) {
		  if (argv[i][0] != '\0') {
		      id = atoi(argv[i]);
		      obj = pad->Get_object_from_id(id);
		      if (obj &&
			  (obj->Type() == PAD_PORTAL)) {
			  view = (Pad_View *)obj;
			  pad = ((Pad_Portal *)obj)->lookon->view->pad;
		      } else {
			  Tcl_AppendResult(interp, "Error: portalID ", argv[i],
					   " is not a portal on ",
					   view->win->Get_name(), NULL);
			  return(TCL_ERROR);
		      }
		  }
	      }

	      pad->Find_with_tagorid(argv[1], objs, TRUE);
	      obj = (Pad_Object *)objs.First();
	      if (obj) {
		  tkwin->view->Update_focus(obj);
	      }
	  }
      }

    Tcl_ResetResult(interp);
    if (Pad_focus)
      {
	  if (Pad_focus->pad->view->win != tkwin) {
				// Focus on other surface
	      sprintf(Pad_result, "%d %s", Pad_focus->id, Pad_focus->pad->view->win->Get_name());
	  } else {
	      sprintf(Pad_result, "%d", Pad_focus->id);
	  }
	  TCL_RESULT(interp, Pad_result);
      }

    return(TCL_OK);
}

//
// Set the language that all callback scripts created in the future
// should be evaluated in.
//
int
Padtcl_setlanguage(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool rc;

    if ((argc != 1) && (argc != 2))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setlanguage ?language?\"");
      }

    if (argc == 2)
      {
	  if (!strcmp(argv[1], "automatic")) {
	      Pad_callbackLanguage = NULL;
	  } else {
	      rc = Pad_GetLanguage(argv[1], &Pad_callbackLanguage);
	      if (!rc) {
		  Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
		  Tcl_AppendResult(interp, ", or automatic", NULL);
		  return(TCL_ERROR);
	      }
	  }
      }
    if (Pad_callbackLanguage) {
	Tcl_AppendResult(interp, Pad_callbackLanguage->name.Get(), NULL);
    } else {
	Tcl_AppendResult(interp, "automatic", NULL);
    }

    return(TCL_OK);
}

//
// Set the top-level interpreter language.
//
int
Padtcl_settoplevel(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool rc;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName settoplevel language\"");
      }

    rc = Pad_GetLanguage(argv[1], &Pad_topLevelLanguage);
    if (!rc) {
	Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
	return(TCL_ERROR);
    }

    return(TCL_OK);
}

//
// Process configure options for image command
// (argc, argv) contains just options and values.
//
static int
Image_config(Pad_ImageData *imageData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_String optionValue;

				// Return all option-value pairs
    if (argc == 0) {
	optionValue.Printf("-dimensions {%d %d}", imageData->width, imageData->height);
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue = "-name ";
	if (imageData->name) {
	    optionValue += imageData->name;
	} else {
	    optionValue += "{}";
	}
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-rgb %d", imageData->rgb);
	Tcl_AppendElement(interp, optionValue.Get());
    } else {
	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-dimensions", MAX(2, len))) {
		if (argc == 1) {
		    sprintf(Pad_result, "%d %d", imageData->width, imageData->height);
		    TCL_RESULT(interp, Pad_result);
		} else {
		    TCL_ERR(interp, "-dimensions is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-name", MAX(2, len))) {
		if (argc == 1) {
		    if (imageData->name) {
			Tcl_AppendResult(interp, imageData->name, NULL);
		    }
		} else {
		    TCL_ERR(interp, "-dimensions is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-rgb", MAX(2, len))) {
		if (argc >= 2) {
		    if (!strcmp(argv[1], "0")) {
			imageData->Free_rgb();
		    } else {
			if (Pad_rgb && !imageData->rgb) {
			    TCL_ERR(interp, "Can't restore rgb data once it has been freed");
			}
		    }
		    argc--;
		    argv++;
		}
		sprintf(Pad_result, "%d", imageData->rgb);
		Tcl_AppendResult(interp, Pad_result, NULL);
	    } else {
		TCL_ERR(interp, "Invalid option, Try: -dimensions, -name, or -rgb");
	    }
	    argc--;
	    argv++;
	}
    }

    return(TCL_OK);
}

//
// Padtcl_image
//
//   pathName image alloc imagefile ?option value ...?
//   pathName image free imagetoken
//   pathName image configure ...
//   pathName image names
//
//     Allocate an image for future use, or free one.
//     Given an image filename, this allocates that image.
//     The returned image token can then be used for new
//     image objects, and for the draw command.
//     'names' returns all the current image tokens
//
//     configure imagetoken -dimensions   => Returns image dimensions
//     configure imagetoken -name  => Returns image name
//     configure imagetoken -rgb   => Controls whether rgb data should be stored
//
int
Padtcl_image(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_ImageData *imageData;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_String name;
    Pad_Bool rc;
    Pad_HashTableIterator hi;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName image subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "alloc", MAX(1, len))) {
	if (argc >= 3) {
	    rc = Pad_Expand_pathname(interp, argv[2], name);
	    if (!rc) {
		TCL_ERR(interp, Pad_errorString.Get());
	    }
	    imageData = tkwin->renderer->Alloc_image(name.Get());
	    if (!imageData) {
		Tcl_AppendResult(interp, "Error creating image: ", argv[2], NULL);
		return(TCL_ERROR);
	    }
	    if (argc > 3) {
		rc = Image_config(imageData, interp, argc - 3, argv + 3);
		if (rc == TCL_ERROR) {
		    return(TCL_ERROR);
		}
	    }
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, imageData->Get_token(), NULL);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName image alloc filename ?option value ...?\"");
	}
    } else if (!strncmp(argv[1], "free", MAX(1, len))) {
	if (argc == 3) {
	    rc = tkwin->renderer->Free_image(argv[2]);
	    if (!rc) {
		TCL_ERR(interp, Pad_errorString.Get());
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName image free imagetoken\"");
	}
    } else if (!strncmp(argv[1], "configure", MAX(1, len))) {
	if (argc >= 3) {
	    imageData = tkwin->renderer->Get_image(argv[2]);
	    if (imageData) {
		rc = Image_config(imageData, interp, argc - 3, argv + 3);
		if (rc == TCL_ERROR) {
		    return(TCL_ERROR);
		}
	    } else {
		Tcl_AppendResult(interp, "invalid image token: ", argv[2], NULL);
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName image configure imagetoken ?option ?value? ...?\"");
	}
    } else if (!strncmp(argv[1], "names", MAX(1, len))) {
	if (argc == 2) {
	    DOTABLE(hi, Pad_imageNameTable, Pad_ImageData, imageData) {
		Tcl_AppendElement(interp, imageData->Get_token());
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName image names\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: alloc, configure, free, or names");
    }

    return(TCL_OK);
}


////////////////////////////////////////////////
// Animation Path functions that follow
////////////////////////////////////////////////
static int _Animate_create_path(ClientData,      Tcl_Interp *interp, int argc, char **argv);
static int _Animate_config_path(ClientData,      Tcl_Interp *interp,  Pad_String pathToken, int argc, char **argv);
static int _Animate_delete_path(ClientData,      Tcl_Interp*, Pad_String pathToken);
static int _Animate_get_path_data(ClientData,     Tcl_Interp *interp, int, Pad_String pathToken, Pad_String &result);
static int _Animate_get_path_interp_val(ClientData, Tcl_Interp *interp, int, char **argv);

static int
_Animate_create_path(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_String pathToken;
    Pad_String surfaceName;
    Pad_SurfaceHandle surface;
    Pad_TkWin  *tkwin;

    // Attach surface to pad    
    tkwin = (Pad_TkWin*)clientData;
    surfaceName = tkwin->Get_name();
    surface.Attach(surfaceName.Get());

    // Create token to get back ptr to path. The token is regestered
    // in the Pad_AnimPath::tokenTable in Pad_SurfaceHandle::Create_anim_path()
    // and the tokenIndex is advanced there.
    sprintf(Pad_result, "path%d", Pad_AnimPath::tokenIndex);
    pathToken = Pad_result;
    // Use surface to create path and pass it a Pad_AnimPathHandle
    if (surface.Create_anim_path()) {
	if (argc >= 2) {// handle option value pairs if supplied upon creation
	    _Animate_config_path(clientData, interp, pathToken, argc, argv);
	}
    } else {
	sprintf(Pad_result, "path allocation failed");
    }
	
    Tcl_AppendElement(interp, Pad_result);
 
    return(TCL_OK);
}

//
// Process configure options for animation path
// (argc, argv) contains just options and values.
//
static int
_Animate_config_path(ClientData clientData, Tcl_Interp *interp, Pad_String pathToken, int argc, char **argv)
{
    int len;
    float value;
    Pad_AnimPathHandle pathHandle;

    Pad_String reportString;
    Pad_AnimPath::PRE_POST post;
    Pad_AnimPath::PRE_POST pre;

    // These variables are for allocating the data
    int    i;
    int    listrc;                 // check to see if split was successful
    int    listargc;               // number of datum
    char **listargv;               // individual datum
    int    sublistrc;       
    int    sublistargc;
    char **sublistargv;
    int    numData=0;  
    Pad_Bool first = TRUE;         // allocate memory only once
    Pad_AnimPath::anim_data* data; // To store data
    int order;                     // Path order
    int _ord;                      // iterator 
    // End allocating data variables

    // Return all option-value pairs
    if (!pathToken.Get()) {// check that token was passed
	    TCL_ERR(interp, "wrong # args: should be \"pathName animate configure pathToken|chanToken|animToken ...?\"");
    } else if (!(Pad_AnimPathHandle::Path_exists(pathToken))) { // check that token exists
	TCL_ERR(interp, "Invalid path token");
    } else if (argc == 0 && Pad_AnimPathHandle::Path_exists(pathToken)) {

	// Make a handle to the Pad_AnimPath associated with pathToken
	pathHandle.Attach((Pad_AnimPath*)Pad_AnimPath::tokenTable.Get(pathToken.Get()));


	reportString += "{";
	if ((pathHandle.Get_time_path_flag())) {
	    reportString += "-timepath ";
	} else {
	    reportString += "-path ";
	}
	_Animate_get_path_data(clientData, interp, argc, pathToken, reportString); // print out path
	reportString += "} ";

	reportString += "{";
	reportString += "-begintime";
	reportString += " ";
	reportString += pathHandle.Get_begin_time();
	reportString += "} ";

	reportString += "{";
	reportString += "-endtime";
	reportString += " ";
	reportString += pathHandle.Get_end_time();
	reportString += "} ";

	reportString += "{";
	reportString += "-intime";
	reportString += " ";
	reportString += pathHandle.Get_in_time();
	reportString += "} ";

	reportString += "{";
	reportString += "-outtime";
	reportString += " ";
	reportString += pathHandle.Get_out_time();
	reportString += "} ";

	reportString += "{";
	reportString += "-order";
	reportString += " ";
	reportString += pathHandle.Get_order();
	reportString += "} ";

	reportString += "{";
	reportString += "-siso";
	reportString += " ";
	if (pathHandle.Get_siso()) {
	    reportString += "true";
	} else {
	    reportString += "false";
	}
	reportString += "} ";

	reportString += "{";
	reportString += "-pre";
	reportString += " ";
	pre = pathHandle.Get_pre();
	if (pre == Pad_AnimPath::CONSTANT) {
	    reportString += "constant";
	} else if (pre == Pad_AnimPath::CYCLE) {
	    reportString += "cycle";
	} else if (pre == Pad_AnimPath::OSCILLATE) {
	    reportString += "oscillate";
	} else {
	    reportString += "error: no animation path corresponding to pathToken";
	}
	reportString += "} ";

	reportString += "{";
	reportString += "-post";
	reportString += " ";
	post = pathHandle.Get_post();
	if (post == Pad_AnimPath::CONSTANT) {
	    reportString += "constant";
	} else if (post == Pad_AnimPath::CYCLE) {
	    reportString += "cycle";
	} else if (post == Pad_AnimPath::OSCILLATE) {
	    reportString += "oscillate";
	} else {
	    reportString += "error: no animation path corresponding to pathToken";
	}
	reportString += "} ";

    } else {
	// Make handle to the Pad_AnimPath associated with pathToken
	pathHandle.Attach((Pad_AnimPath*)Pad_AnimPath::tokenTable.Get(pathToken.Get()));

	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-path", MAX(2, len))) {

		if (argc==1){
		    if (pathHandle.Get_time_path_flag() && pathHandle.Has_data()) {
			TCL_ERR(interp, "path contains -timepath data");
		    } else {
			_Animate_get_path_data(clientData, interp, argc, pathToken, reportString);// print out path
		    }
		} else { // set data to new data

		    data = new Pad_AnimPath::anim_data;

		    // split the user data list into individual datum
		    listrc = Tcl_SplitList(interp, argv[1], &listargc, &listargv);
		    numData = listargc;
		    if (listrc == TCL_OK) {
			if (listargc < 2) {
			    TCL_ERR(interp, "path needs 2 or more points");
			} else {
			    DOTIMES(i, numData) {
				sublistrc = Tcl_SplitList(interp, listargv[i], &sublistargc, &sublistargv);
				if (sublistrc == TCL_OK) {
				    if (first) {
					first = FALSE;// turn off first flag so allocation occurs only once
					order = sublistargc;
					for (_ord=0; _ord<order; _ord++){// allocate memory to each order of data
					    data->data[_ord] = new Pad_AnimPath::datum [listargc];
					}
					
					if (order>3 || order<1) {// check if order not acceptable value
					    TCL_ERR(interp, "path must have order 1,2, or 3");
					}
					data->order = order;
				    } else {
					if (sublistargc != order) {// check that all data if of same order
					    free(sublistargv);
					    TCL_ERR(interp, "order of element not the same as order of first element");
					}
				    }
				    
				    for (_ord=0; _ord<order; _ord++){ // set values of data from user data
					data->data[_ord][i] = atof(sublistargv[_ord]);
				    }
				    
				    free(sublistargv);
				} else {
				    TCL_ERR(interp, "element in path needs valid list of points");
				}
			    }
			}
			ckfree((char *) listargv);
		    } else {
			TCL_ERR(interp, "path needs valid list of points");
		    }

		    pathHandle.Set_path_data(numData, data, order, FALSE);

		    // free memory allocated to arrays in struct data
		    for (_ord=0; _ord<data->order; _ord++){
			delete [] data->data[_ord];
		    }
		    // free memory for the struct data
		    delete data;
		    reportString = "";
		}
		////////////////////// Begin -timepath setting

	    } else if (!strncmp(argv[0], "-timepath", MAX(6, len))) {

		if (argc==1){
		    if (pathHandle.Get_time_path_flag() && pathHandle.Has_data()) {
			_Animate_get_path_data(clientData, interp, argc, pathToken, reportString);// print out path
		    } else {
			TCL_ERR(interp, "path contains -path data");
		    }
		} else { // set data to new data

		    data = new Pad_AnimPath::anim_data;

		    // split the user data list into individual datum
		    listrc = Tcl_SplitList(interp, argv[1], &listargc, &listargv);
		    numData = listargc;
		    if (listrc == TCL_OK) {
			if (listargc < 2) {
			    TCL_ERR(interp, "path needs 2 or more points");
			} else {
			    DOTIMES(i, numData) {
				sublistrc = Tcl_SplitList(interp, listargv[i], &sublistargc, &sublistargv);
				if (sublistrc == TCL_OK) {
				    if (first) {
					if (sublistargc < 2) {
					    TCL_ERR(interp, "-timepath needs both time and data values");
					}
					first = FALSE;// turn off first flag so allocation occurs only once
					order = sublistargc - 1;
					    // allocate memory for time values
					data->times = new Pad_AnimPath::datum [listargc];
					    // allocate memory to each order of data
					for (_ord=0; _ord<order; _ord++){
					    data->data[_ord] = new Pad_AnimPath::datum [listargc];
					}
					    // check if order not acceptable value
					if (order>3 || order<1) {
					    TCL_ERR(interp, "path must have order 1,2, or 3");
					}
					data->order = order;
				    } else {
					if (sublistargc-1 != order) {// check that all data if of same order
					    free(sublistargv);
					    TCL_ERR(interp, "order of element not the same as order of first element");
					}
				    }
				    data->times[i] = atof(sublistargv[0]);// +1 skips time value
				        // set values of data from user data
				    for (_ord=0; _ord<order; _ord++){ 
					data->data[_ord][i] = atof(sublistargv[_ord+1]);// +1 skips time value
				    }
				    
				    free(sublistargv);
				} else {
				    TCL_ERR(interp, "element in path needs valid list of points");
				}
			    }
			}
			ckfree((char *) listargv);
		    } else {
			TCL_ERR(interp, "path needs valid list of points");
		    }
		        // Set path data values
		    pathHandle.Set_path_data(numData, data, order, TRUE);

		        // free memory allocated to arrays in struct data
		    delete [] data->times;
		    for (_ord=0; _ord<data->order; _ord++){
			delete [] data->data[_ord];
		    }
		        // free memory for the struct data
		    delete data;
		    reportString = "";

		}
		/////////////////// End -timepath setting

	    } else if (!strncmp(argv[0], "-begintime", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-begintime";
		    reportString += " ";
		    reportString += pathHandle.Get_begin_time();
		    reportString += " ";
		    
		} else {
		    value =  atof(argv[1]);
		    pathHandle.Set_begin_time(value);
		    
		    argc--;
		    argv++;
		}

	    } else if (!strncmp(argv[0], "-endtime", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-endtime";
		    reportString += " ";
		    reportString += pathHandle.Get_end_time();
		    reportString += " ";
		} else {
		    value =  atof(argv[1]);
		    pathHandle.Set_end_time(value);
		    
		    argc--;
		    argv++;
		}
	    } else if (!strncmp(argv[0], "-intime", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-intime";
		    reportString += " ";
		    reportString += pathHandle.Get_in_time();
		    reportString += " ";
		    
		} else {
		    value =  atof(argv[1]);
		    pathHandle.Set_in_time(value);
		    
		    argc--;
		    argv++;
		}

	    } else if (!strncmp(argv[0], "-outtime", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-outtime";
		    reportString += " ";
		    reportString += pathHandle.Get_out_time();
		    reportString += " ";
		} else {
		    value =  atof(argv[1]);
		    pathHandle.Set_out_time(value);

		    argc--;
		    argv++;
		}

	    } else if (!strncmp(argv[0], "-order", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-order";
		    reportString += " ";
		    reportString += pathHandle.Get_order();
		    reportString += " ";
		}

	    } else if (!strncmp(argv[0], "-pre", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-pre";
		    reportString += " ";
		    pre = pathHandle.Get_pre();
		    if (pre == Pad_AnimPath::CONSTANT) {
			reportString += "constant";
		    } else if (pre == Pad_AnimPath::CYCLE) {
			reportString += "cycle";
		    } else if (pre == Pad_AnimPath::OSCILLATE) {
			reportString += "oscillate";
		    } else {
			reportString += "error: no animation path corresponding to pathToken";
		    }

		} else {
		    if (!strncmp(argv[1], "constant", MAX(2, len))) {
			pre =  Pad_AnimPath::CONSTANT;
			reportString += "";
		    } else if (!strncmp(argv[1], "cycle", MAX(2, len))) {
			pre = Pad_AnimPath::CYCLE;
			reportString += "";
		    } else if (!strncmp(argv[1], "oscillate", MAX(2, len))) {
			pre = Pad_AnimPath::OSCILLATE;
			reportString += "";
		    } else {
			TCL_ERR(interp, "error, unavailable pre condition: must be; -constant, -cycle, or -oscillate");
		    }
		    pathHandle.Set_pre(pre);
		    
		    argc--;
		    argv++;
		}

	    } else if (!strncmp(argv[0], "-post", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-post";
		    reportString += " ";
		    post = pathHandle.Get_post();
		    if (post == Pad_AnimPath::CONSTANT) {
			reportString += "constant";
		    } else if (post == Pad_AnimPath::CYCLE) {
			reportString += "cycle";
		    } else if (post == Pad_AnimPath::OSCILLATE) {
			reportString += "oscillate";
		    } else {
			reportString += "error: no animation path corresponding to pathToken";
		    }

		} else {
		    if (!strncmp(argv[1], "constant", MAX(2, len))) {
			post =  Pad_AnimPath::CONSTANT;
			reportString += "";
		    } else if (!strncmp(argv[1], "cycle", MAX(2, len))) {
			post = Pad_AnimPath::CYCLE;
			reportString += "";
		    } else if (!strncmp(argv[1], "oscillate", MAX(2, len))) {
			post = Pad_AnimPath::OSCILLATE;
			reportString += "";
		    } else {
			TCL_ERR(interp, "error, unavailable post condition: must be; -constant, -cycle, or -oscillate");
		    }
		    pathHandle.Set_post(post);
		    
		    argc--;
		    argv++;
		} 
	    } else if (!strncmp(argv[0], "-siso", MAX(2, len))) {
		if (argc == 1) {
		    reportString += "-siso";
		    reportString += " ";
 
		    if (pathHandle.Get_siso()) {
			reportString += "true";
		    } else {
			reportString += "false";
		    }
		} else {
		    len = strlen(argv[1]);
		    if (!strncmp(argv[1], "1", MAX(1, len))) {
			pathHandle.Set_siso(TRUE);
		    } else {
			pathHandle.Set_siso(FALSE);
		    }

		}
	    }

	    argc--;
	    argv++;
	}
    }

    Tcl_AppendResult(interp, reportString.Get(), (char*) NULL);
    return(TCL_OK);
}

static int
_Animate_delete_path(ClientData, Tcl_Interp*, Pad_String pathToken)
{
    int rc = TCL_OK; 
    Pad_AnimPathHandle pathHandle;

    // Make a handle to the Pad_AnimPath associated with pathToken
    pathHandle.Attach((Pad_AnimPath*)Pad_AnimPath::tokenTable.Get(pathToken.Get()));
    
    rc = pathHandle.Delete_path();
    
    return(rc);
}

static int
_Animate_get_path_data(ClientData, Tcl_Interp *interp, int, Pad_String pathToken, Pad_String &result)
{
    int i;
    int len;
    int order;

    Pad_AnimPath::anim_data* data;// danger; pointer is pointer to actual data
    Pad_AnimPathHandle       pathHandle;

    // Make a handle to the Pad_AnimPath associated with pathToken
    pathHandle.Attach((Pad_AnimPath*)Pad_AnimPath::tokenTable.Get(pathToken.Get()));

    order = pathHandle.Get_order();
    if (order == 0) {
	TCL_ERR(interp, "Invalid path token");
    }

    len = pathHandle.Get_path_data(data);
    if (!(pathHandle.Get_time_path_flag())) {
	switch (order) {
	case 1:
	    for (i=0; i<len; i++) {
		result += data->data[0][i];
		result += " ";
	    }
	    break;
	case 2:
	    for (i=0; i<len; i++) {
		result += "{";
		result += data->data[0][i];
		result += " ";
		result += data->data[1][i];
		result += "}";
		result += " ";
	    }
	    break;
	case 3:
	    for (i=0; i<len; i++) {
		result += "{";
		result += data->data[0][i];
		result += " ";
		result += data->data[1][i];
		result += " ";
		result += data->data[2][i];
		result += "}";
		result += " ";
	    }
	    break;
	}
    } else {
	switch (order) {
	case 1:
	    for (i=0; i<len; i++) {
		result += "{";
		result += data->times[i];
		result += " ";
		result += data->data[0][i];
		result += "} ";
		result += " ";
	    }
	    break;
	case 2:
	    for (i=0; i<len; i++) {
		result += "{";
		result += data->times[i];
		result += " ";
		result += data->data[0][i];
		result += " ";
		result += data->data[1][i];
		result += "}";
		result += " ";
	    }
	    break;
	case 3:
	    for (i=0; i<len; i++) {
		result += "{";
		result += data->times[i];
		result += " ";
		result += data->data[0][i];
		result += " ";
		result += data->data[1][i];
		result += " ";
		result += data->data[2][i];
		result += "}";
		result += " ";
	    }
	    break;
	}
    }	

    return(TCL_OK);
}

static int
_Animate_get_path_interp_val(ClientData, Tcl_Interp *interp, int, char **argv)
{
    int order;
    Pad_AnimPathHandle   pathHandle;
    Pad_AnimPath::datum3 data3; // array of three floats for returning result
    Pad_String           result;
    float                time;
    Pad_String pathToken; // pad string to hold the token

    pathToken = argv[0];

    // Make a handle to the Pad_AnimPath associated with pathToken
    pathHandle.Attach((Pad_AnimPath*)Pad_AnimPath::tokenTable.Get(pathToken.Get()));

    time = atof(argv[1]);

    order = pathHandle.Get_order();
    //order = Pad_GetAnimPathOrder(argv[0]);
    if (order == 0) {
	TCL_ERR(interp, "Invalid path token or no -path has been set for path");
    }

    pathHandle.Get_interp_value(time, data3);
    switch (order) {
    case 1:
	result += data3[0];

	break;
    case 2:
	result += "{";
	result += data3[0];
	result += " ";
	result += data3[1];
	result += "}";
	result += " ";

	break;
    case 3:
	result += "{";
	result += data3[0];
	result += " ";
	result += data3[1];
	result += " ";
	result += data3[2];
	result += "}";
	result += " ";
	
	break;
    }

    Tcl_AppendResult(interp, result.Get(), NULL);
    return(TCL_OK);
}

//
// Animation Channel Tcl interface declarations
//
static int _Animate_create_channel(ClientData clientData, Tcl_Interp* interp, int argc, char** argv);
static int _Animate_start_channel(ClientData clientData, Tcl_Interp* interp, Pad_String token); 
static int _Animate_config_channel(ClientData clientData, Tcl_Interp*, Pad_String token, int argc, char** argv);
static int _Animate_delete_channel(ClientData, Tcl_Interp*, Pad_String chanToken);
static int _Animate_stop_channel(ClientData, Tcl_Interp*, Pad_String chanToken);

static int
_Animate_create_channel(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_String chanToken;
    Pad_String surfaceName;
    Pad_SurfaceHandle surface;
    Pad_TkWin  *tkwin;

    // Attach surface to pad    
    tkwin = (Pad_TkWin*)clientData;
    surfaceName = tkwin->Get_name();
    surface.Attach(surfaceName.Get());

    // Create token to get back ptr to channel. The token is regestered
    // in the Pad_AnimChannel::tokenTable in Pad_SurfaceHandle::Create_anim_channel()
    // and the tokenIndex is advanced there.
    sprintf(Pad_result, "chan%d", Pad_Animatable::tokenIndex);
    chanToken = Pad_result;
    // Use surface to create channel and pass it a Pad_AnimChannelHandle
    if (surface.Create_anim_channel()) {
	if (argc >= 2) {// handle option value pairs if supplied upon creation
	    _Animate_config_channel(clientData, interp, chanToken, argc, argv);
	}
    } else {
	sprintf(Pad_result, "channel allocation failed");
    }
	
    Tcl_AppendElement(interp, Pad_result);
 
    return(TCL_OK);
}

//
// Process the options for in the channel specified by token
//
static int
_Animate_config_channel(ClientData clientData, Tcl_Interp* interp, Pad_String chanToken, int argc, char** argv)
{
    Pad_String  reportString;
    int         len;

    Pad_TkWin      *tkwin;
    Pad_List        objs;
    Pad_Bool        successful = FALSE;
    Pad_String      pathToken;
    
    Pad_String            tagOrId;
    Pad_String            surfaceName;
    Pad_AnimChannelHandle chanHandle;
    Pad_AnimPathHandle    pathHandle;
    Pad_ObjHandle         objHandle;
    Pad_SurfaceHandle     surface;
			
    tkwin = (Pad_TkWin *)clientData;

    if (!chanToken.Get()) {
	TCL_ERR(interp, "wrong # args: should be \"pathName animate configure chanToken|pathToken|animToken ...?\"");
    } else if (!Pad_AnimChannelHandle::Channel_exists(chanToken)) {
	TCL_ERR(interp, "Invalid channel token");
    } else if (argc == 0) {
	    // Attach a Pad_AnimChannelHandle to the channel associated with chanToken
	chanHandle.Attach((Pad_AnimChannel*)Pad_Animatable::tokenTable.Get(chanToken.Get()));
		
	reportString += "{";
	reportString += "-path";
	reportString += " ";
	if (chanHandle.Has_path()) {
	    reportString += chanHandle.Get_path_token();
	} else {
	    reportString += "{}";
	}
	reportString += "} ";

	reportString += "{";
	reportString += "-object";
	reportString += " ";
	if (chanHandle.Has_object()) {
	    reportString += chanHandle.Get_object_tag_or_id();
	} else {
	    reportString += "{}";
	}
	reportString += "} ";

	reportString += "{";
	reportString += "-option";
	reportString += " ";
	if (chanHandle.Has_option()) {
	    reportString += (int)(chanHandle.Get_option());
	} else {
	    reportString += "{}";
	}
	reportString += "} ";

	reportString += "{";
	reportString += "-timeresolution";
	reportString += " ";
       	reportString += chanHandle.Get_time_resolution();
	reportString += "} ";


	reportString += "{";
	reportString += "-begintime";
	reportString += " ";
	if (chanHandle.Has_path()) {
	    reportString += chanHandle.Get_path_begin_time();
	} else {
	    reportString += "{}";
	}
	reportString += "} ";
	
	reportString += "{";
	reportString += "-endtime";
	reportString += " ";
	if (chanHandle.Has_path()) {
	    reportString += chanHandle.Get_path_end_time();
	} else {
	    reportString += "{}";
	}
	reportString += "} ";
	
	reportString += "{";
	reportString += "-order";
	reportString += " ";
	if (chanHandle.Has_path()) {
	    reportString += chanHandle.Get_path_order();
	} else {
	    reportString += "{}";
	}
	reportString += "} ";

    } else {
	    // Attach a Pad_AnimChannelHandle to the channel associated with chanToken
	chanHandle.Attach((Pad_AnimChannel*)Pad_Animatable::tokenTable.Get(chanToken.Get()));

	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-path", MAX(2, len))) {

		if (argc==1) {
		    reportString += chanHandle.Get_path_token();// print out path token
		} else { 
		    pathToken = argv[1];
		    if (Pad_AnimPathHandle::Path_exists(pathToken)) {
			pathHandle.Attach((Pad_AnimPath*)Pad_AnimPath::tokenTable.Get(pathToken.Get()));
			chanHandle.Set_path(pathHandle); // expect argv[1] to be path token
		    } else {
			reportString += "Invalid path token, Channel has no path";
		    }
		}
		argc--;
		argv++;

	    } else if (!strncmp(argv[0], "-endtime", MAX(2, len))) {

		if (argc==1) {
		    if (chanHandle.Has_path()) {
			reportString += chanHandle.Get_path_end_time(); 
		    } else {
			reportString += "channel has no path";
		    }
		} else { 
		    if (chanHandle.Has_path()) {
			successful = chanHandle.Set_path_end_time(atof(argv[1]));			
			if (successful) {
			    //reportString += Pad_GetAnimEndTime(chanToken); 
			} else {
			    reportString += "set endtime unsuccessful";
			}
		    } else {
			reportString += "channel has no path";
		    }
		}
		argc--;
		argv++;

	    } else if (!strncmp(argv[0], "-option", MAX(2, len))) {

		if (argc==1) {
		    if (chanHandle.Has_option()) {
			reportString += chanHandle.Get_option(); 
		    } else {
			reportString += "channel has no option specified";
		    }
		} else { 
		    surfaceName = tkwin->Get_name();
		    surface.Attach(surfaceName.Get());
		    // expect argv[1] to be a string option such as -place
		    successful = chanHandle.Set_option(surface.Get_option_type(argv[1]));
		    		    
		    if (successful) {
			//reportString += Pad_GetAnimChannelOption(chanToken); 
		    } else {
			reportString += "option set failed";
		    }
		}   
		argc--;
		argv++;

	    } else if (!strncmp(argv[0], "-begintime", MAX(2, len))) {

		if (argc==1) {
		    if (chanHandle.Has_path()) {
			reportString += chanHandle.Get_path_begin_time(); 
		    } else {
			reportString += "channel has no path";
		    }
		} else { 
		    if (chanHandle.Has_path()) {
			successful = chanHandle.Set_path_begin_time(atof(argv[1]));			
			if (successful) {
			    //reportString += Pad_GetAnimBeginTime(chanToken); 
			} else {
			    reportString += "set endtime unsuccessful";
			}
		    } else {
			reportString += "channel has no path";
		    }
		}
		argc--;
		argv++;

	    } else if (!strncmp(argv[0], "-timeresolution", MAX(2, len))) {

		if (argc==1) {
		    reportString += chanHandle.Get_time_resolution(); 

		} else { 
		    chanHandle.Set_time_resolution(atoi(argv[1]));
		}
		argc--;
		argv++;

	    } else if (!strncmp(argv[0], "-order", MAX(2, len))) {

		if (argc==1) {
		    reportString += chanHandle.Get_path_order(); 
		}

	    } else if (!strncmp(argv[0], "-start", MAX(2, len))) {
		    if (chanHandle.Has_path()) {
			successful = chanHandle.Play();			
			if (successful) {
			    //reportString += "channel started";
			} else {
			    reportString += "channel start FAILED";
			}
		    } else {
			reportString += "channel has no path";
		    }

	    } else if (!strncmp(argv[0], "-object", MAX(2, len))) {
		
		if (argc==1) {
 		    reportString += chanHandle.Get_object_tag_or_id(); 

 		} else { 
		    tagOrId = argv[1];
		    surfaceName = tkwin->Get_name();
		    objHandle.Attach(surfaceName.Get(), tagOrId.Get());
		    chanHandle.Set_object(objHandle);

		    //reportString += "-object  ";
 		    //reportString += objHandle.Get_id();
 		}
		argc--;
		argv++;
	    }
	    argc--;
	    argv++;

	}
    }

    Tcl_AppendResult(interp, reportString.Get(), (char*) NULL);
    return(TCL_OK);
}

static int
_Animate_delete_channel(ClientData, Tcl_Interp*, Pad_String chanToken)
{
    int rc = TCL_OK; 
    Pad_AnimChannelHandle chanHandle;

    // Make a handle to the Pad_AnimPath associated with pathToken
    chanHandle.Attach((Pad_AnimChannel*)Pad_Animatable::tokenTable.Get(chanToken.Get()));
    
    rc = chanHandle.Delete_channel();
    
    return(rc);
}

static int
_Animate_stop_channel(ClientData, Tcl_Interp*, Pad_String chanToken)
{
    int rc = TCL_OK; 
    Pad_AnimChannelHandle chanHandle;

    // Make a handle to the Pad_AnimPath associated with pathToken
    chanHandle.Attach((Pad_AnimChannel*)Pad_Animatable::tokenTable.Get(chanToken.Get()));
    
    rc = chanHandle.Set_interrupt(TRUE);
    
    return(rc);
}

static int 
_Animate_start_channel(ClientData, Tcl_Interp* interp, Pad_String token)
{
    Pad_String            report;
    Pad_Bool              successful = FALSE;
    Pad_AnimChannelHandle chanHandle;
	
    if (Pad_AnimChannelHandle::Channel_exists(token)) {
            // Attach a Pad_AnimChannelHandle to the channel associated with chanToken
	chanHandle.Attach((Pad_AnimChannel*)Pad_Animatable::tokenTable.Get(token.Get()));

	successful = chanHandle.Play();			
	if (successful) {
	    //report += "channel started";
	} else {
	    report += "channel start FAILED";
	}
    } else {
	report += "Invalid channel token";
    }

    Tcl_AppendResult(interp, report.Get(), (char*) NULL);
    return(TCL_OK);
}

// Declarations for animations
static int _Animate_create_animation(ClientData clientData, Tcl_Interp* interp, int argc, char** argv);
static int _Animate_start_animation(ClientData clientData,  Tcl_Interp* interp, Pad_String token);
static int _Animate_config_animation(ClientData clientData, Tcl_Interp*,Pad_String token, int argc, char** argv);
static int _Animate_delete_animation(ClientData, Tcl_Interp*, Pad_String animToken);
static int _Animate_stop_animation(ClientData, Tcl_Interp*, Pad_String animToken);

static int
_Animate_create_animation(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_String animToken;
    Pad_String surfaceName;
    Pad_SurfaceHandle surface;
    Pad_TkWin  *tkwin;

    // Attach surface to pad    
    tkwin = (Pad_TkWin*)clientData;
    surfaceName = tkwin->Get_name();
    surface.Attach(surfaceName.Get());

    // Create token to get back ptr to animation. The token is regestered
    // in the Pad_Animation::tokenTable in Pad_SurfaceHandle::Create_animation()
    // and the tokenIndex is advanced there.
    sprintf(Pad_result, "anim%d", Pad_Animation::tokenIndex);
    animToken = Pad_result;
    // Use surface to create animation and pass it a Pad_AnimationHandle
    if (surface.Create_animation()) {
	if (argc >= 2) {// handle option value pairs if supplied upon creation
	    _Animate_config_animation(clientData, interp, animToken, argc, argv);
	}
    } else {
	sprintf(Pad_result, "animation allocation failed");
    }
	
    Tcl_AppendElement(interp, Pad_result);
 
    return(TCL_OK);
}

static int
_Animate_config_animation(ClientData, Tcl_Interp* interp, Pad_String animToken, int argc, char** argv)
{
    Pad_String  reportString;
    int         len;

    Pad_AnimChannelHandle chanHandle;
    Pad_AnimationHandle   animHandle;
    Pad_AnimationHandle   animHandle2;
    Pad_Animatable*       animatable;

    Pad_Bool      successful = FALSE;

    Pad_String members;
    Pad_String membToken;
			
    if (!animToken.Get()) {
	TCL_ERR(interp, "wrong # args: should be \"pathName animate configure pathToken|chanToken|animToken ...?\"");
    } else if (!Pad_AnimationHandle::Animation_exists(animToken)) {
	TCL_ERR(interp, "Invalid animation token");
    } else if (argc == 0) {
	    // Attach a Pad_AnimationHandle to the animation associated with animToken
	animHandle.Attach((Pad_Animation*)Pad_Animation::tokenTable.Get(animToken.Get()));
		
	reportString += "{";
	reportString += "-begintime";
	reportString += " ";
	reportString += animHandle.Get_begin_time();
	reportString += "} ";

	reportString += "{";
	reportString += "-endtime";
	reportString += " ";
	reportString += animHandle.Get_end_time();
	reportString += "} ";

	reportString += "{";
	reportString += "-speedfactor";
	reportString += " ";
	reportString += animHandle.Get_speed_factor();
	reportString += "} ";

	reportString += "{";
	reportString += "-members";
	reportString += " ";
	animHandle.Get_animatable_tokens(members); 
	if (members.Length() != 0) {
	    reportString += members;
	} else {
	    reportString += "{}";
	}
	reportString += "} ";

    } else {
	    // Attach a Pad_AnimationHandle to the animation associated with animToken
	animHandle.Attach((Pad_Animation*)Pad_Animation::tokenTable.Get(animToken.Get()));

	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-endtime", MAX(2, len))) {
		if (argc==1) {
		    reportString += animHandle.Get_end_time(); 
		} else { 
		    successful = animHandle.Set_end_time(atof(argv[1]));			
		    if (successful) {
			//reportString += Pad_GetAnimEndTime(animToken); 
		    } else {
			reportString += "set endtime unsuccessful";
		    }
		}
		argc--;
		argv++;
	    } else if (!strncmp(argv[0], "-begintime", MAX(2, len))) {
		if (argc==1) {
		    reportString += animHandle.Get_begin_time(); 
		} else { 
		    successful = animHandle.Set_begin_time(atof(argv[1]));			
		    		    
		    if (successful) {
			//reportString += Pad_GetAnimBeginTime(animToken); 
		    } else {
			reportString += "set -begintime unsuccessful";
		    }
		}
		argc--;
		argv++;
	    } else if (!strncmp(argv[0], "-speedfactor", MAX(2, len))) {
		if (argc==1) {
		    reportString += animHandle.Get_speed_factor(); 
		} else { 
		    successful = animHandle.Set_speed_factor(atof(argv[1]));			
		    		    
		    if (successful) {
			//reportString += Pad_GetAnimBeginTime(animToken); 
		    } else {
			reportString += "set speedfactor unsuccessful";
		    }
		}
		argc--;
		argv++;
	    } else if (!strncmp(argv[0], "-members", MAX(4, len))) {

		if (argc==1) {
		    animHandle.Get_animatable_tokens(members); 
		    reportString += members;
		} else { 
		    int listIter;
		    int listrc;
		    int listargc;
		    char **listargv;
		    Pad_String animatableToken;
		    listrc = Tcl_SplitList(interp, argv[1], &listargc, &listargv);

		    if (listrc == TCL_OK) {
			for (listIter = 0; listIter<listargc; listIter++) {

			    // get the token for the animatable
			    animatableToken  = listargv[listIter];
			    // get the animatable from a has that animatable hash table
			    animatable = (Pad_Animatable*)Pad_Animatable::tokenTable.Get(animatableToken.Get());

	                    // Check the type of the animatable being added
			    // attach the proper handle to it
			    // add it as a member to the target animation
			    if (animatable->Get_type() == Pad_Animatable::CHAN) {
			        // Attach a Pad_ChannelHandle to the channel associated with animatableToken
				chanHandle.Attach((Pad_AnimChannel*)animatable);
				successful = animHandle.Add_member(chanHandle);			
			    } else if (animatable->Get_type() == Pad_Animatable::ANIM) {
			        // Attach a Pad_ChannelHandle to the channel associated with animatableToken
				animHandle2.Attach((Pad_Animation*)animatable);
				successful = animHandle.Add_member(animHandle2);			
			    }
			}
		    }
		    if (successful) {
			//reportString += Pad_GetAnimEndTime(animToken); 
		    } else {
			//reportString += "add member command unsuccessful";
		    }
		}
		argc--;
		argv++;
	    } else if (!strncmp(argv[0], "-start", MAX(2, len))) {
		successful = animHandle.Play();			
		if (successful) {
		    //reportString += "animation started";
		} else {
		    reportString += "animation start FAILED";
		}
	    }
	    argc--;
	    argv++;
	}
    }
    Tcl_AppendResult(interp, reportString.Get(), (char*) NULL);
    return(TCL_OK);
}

static int
_Animate_delete_animation(ClientData, Tcl_Interp*, Pad_String animToken)
{
    int rc = TCL_OK; 
    Pad_AnimationHandle animHandle;

    // Make a handle to the Pad_AnimPath associated with pathToken
    animHandle.Attach((Pad_Animation*)Pad_Animatable::tokenTable.Get(animToken.Get()));
    
    rc = animHandle.Delete_animation();
    
    return(rc);
}

static int
_Animate_stop_animation(ClientData, Tcl_Interp*, Pad_String animToken)
{
    int rc = TCL_OK; 
    Pad_AnimationHandle animHandle;

    // Make a handle to the Pad_AnimPath associated with pathToken
    animHandle.Attach((Pad_Animation*)Pad_Animatable::tokenTable.Get(animToken.Get()));
    
    rc = animHandle.Set_interrupt(TRUE);
    
    return(rc);
}
  
static int 
_Animate_start_animation(ClientData, Tcl_Interp* interp, Pad_String token)
{
    Pad_String          report;
    Pad_Bool            successful = FALSE;
    Pad_AnimationHandle animHandle;

    if (Pad_AnimationHandle::Animation_exists(token)) {
	    // Attach a Pad_AnimationHandle to the animation associated with animToken
	animHandle.Attach((Pad_Animation*)Pad_Animation::tokenTable.Get(token.Get()));
	successful = animHandle.Play();			
	if (successful) {
	    //report += "animation started";
	} else {
	    report += "animation start FAILED";
	}
    } else {
	report += "Invalid animation token";
    }
    Tcl_AppendResult(interp, report.Get(), (char*) NULL);
    return(TCL_OK);
}


int
Padtcl_animate(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len1;
    int len2;
    int rc = TCL_OK;
    Pad_String token;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName animate subcommand ?arg ...?\"");
    }

    len1 = strlen(argv[1]);
    if (argc >= 3) {
	len2 = strlen(argv[2]);
    }

    if (!strncmp(argv[1], "create", MAX(2, len1))) {
	if (argc >= 3) {

	    if (!strncmp(argv[2], "path", MAX(1, len2))) {
		rc = _Animate_create_path(clientData, interp, argc - 3, argv + 3);
	    } else if (!strncmp(argv[2], "channel", MAX(1, len2))) {
		rc = _Animate_create_channel(clientData, interp, argc - 3, argv + 3);
	    } else if (!strncmp(argv[2], "animation", MAX(1, len2))) {
		rc = _Animate_create_animation(clientData, interp, argc - 3, argv + 3);
	    } else {
		TCL_ERR(interp, "You can only create a \"path\" a \"channel\" or an \"animation\"");
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName animate create (path | channel | animation) ?option value ...?\"");
	}
    } else if (!strncmp(argv[1], "start", MAX(1, len1))) {
	if (argc == 3) {
	    token = argv[2];
	    if (!strncmp(argv[2], "chan", MAX(1, 4))) {
		rc = _Animate_start_channel(clientData, interp, token);
	    } else if (!strncmp(argv[2], "anim", MAX(1, 4))) {
		rc = _Animate_start_animation(clientData, interp, token);
	    } else {
		TCL_ERR(interp, "You can only start a \"path\" or a \"channel\"");
	    }
	} else {
	    TCL_ERR(interp, "An animationToken a channelToken must be provided to the animation \"start\" command");
	}
    } else if (!strncmp(argv[1], "stop", MAX(1, len1))) {
	if (argc == 3) {
	    token = argv[2];
	    if (!strncmp(argv[2], "chan", MAX(1, 4))) {
		rc = _Animate_stop_channel(clientData, interp, token);
	    } else if (!strncmp(argv[2], "anim", MAX(1, 4))) {
		rc = _Animate_stop_animation(clientData, interp, token);
	    } else {
		TCL_ERR(interp, "You can only stop a \"path\" or a \"channel\"");
	    }
	} else {
	    TCL_ERR(interp, "An animationToken a channelToken must be provided to the animation \"stop\" command");
	}

    } else if (!strncmp(argv[1], "delete", MAX(1, len1))) {
	if (argc == 3) {
	    token = argv[2];
	    if (!strncmp(argv[2], "path", MAX(1, 4))) {
		rc = _Animate_delete_path(clientData, interp, token);
	    } else if (!strncmp(argv[2], "chan", MAX(1, 4))) {
		rc = _Animate_delete_channel(clientData, interp, token);
	    } else if (!strncmp(argv[2], "anim", MAX(1, 4))) {
		rc = _Animate_delete_animation(clientData, interp, token);
	    } else {
		TCL_ERR(interp, "You can only start a \"path\" or a \"channel\"");
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName animate deletepath pathtoken | channeltoken | animationtoken \"");
	}

    } else if (!strncmp(argv[1], "getvalue", MAX(4, len1))) {
	if (argc == 4) {
	    rc = _Animate_get_path_interp_val(clientData, interp, argc - 2, argv + 2);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName animate getpathinterpval pathtoken time\"");
	}
    } else if (!strncmp(argv[1], "configure", MAX(2, len1))) {
	
	if (!strncmp(argv[2], "path", MAX(1, 4))) {
	    if (argc >= 3) {
		token = argv[2];
		rc = _Animate_config_path(clientData, interp, token, argc - 3, argv + 3);
	    }
	} else if (!strncmp(argv[2], "chan", MAX(1, 4))) {
	    if (argc >= 3) {
		token = argv[2];
		rc = _Animate_config_channel(clientData, interp, token, argc - 3, argv + 3);
	    }
	} else if (!strncmp(argv[2], "anim", MAX(1, 4))) {
	    if (argc >= 3) {
		token = argv[2];
		rc = _Animate_config_animation(clientData, interp, token, argc - 3, argv + 3);
	    }
	} else {
	    TCL_ERR(interp, "You can only animate configure a \"path\" or a \"channel\" or an \"animation\" ");
	}
  } else if (!strncmp(argv[1], "schedule", MAX(1, len1))) {
	if (argc >= 3) {
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName animate schedule channeltoken ?option value ...?\"");
	}
  } else {
	TCL_ERR(interp, "Invalid subcommand, should be: configure, create, delete, start, getvalue");
    }

    return(rc);
}
// end animation stuff

int
Padtcl_draw_border(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    float x0, y0, x1, y1;
    float width;
    Pad_Border *border;
    int relief;
    int rc;
    Pad_Point min, max;
    Pad_String reliefName;

    if (argc != 8)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName drawborder border type width x1 y1 x2 y2\"");
      }
    if (!Pad_prc) {
	TCL_ERR(interp, "not a render callback: drawborder may only be called during a render callback");
    }

    border = Pad_LookupBorderByName(argv[1]);
    if (border) {
	rc = Pad_Relief_name_to_type(argv[2], relief);
	if (!rc) {
	    Tcl_AppendResult(interp, argv[2], " is an invalid relief", NULL);
	    return(TCL_ERROR);
	}
	
	width = ATOXF(tkwin, argv[3]);
	
	x0 = ATOXF(tkwin, argv[4]);
	y0 = ATOYF(tkwin, argv[5]);
	x1 = ATOXF(tkwin, argv[6]);
	y1 = ATOYF(tkwin, argv[7]);
	min.Set(x0, y0);
	max.Set(x1, y1);

	Pad_renderer->Set_border(border);
	Pad_renderer->Draw_filled_3d_rectangle(min, max, width, relief);
    } else {
	Tcl_AppendResult(interp, "Border not previously allocated by allocborder: ",
			 argv[1], NULL);
	return(TCL_ERROR);
    }

    return(TCL_OK);
}

int
Padtcl_draw_line(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, npts;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc < 5)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName drawline x1 y1 x2 y2 ...\"");
      }

    if (!Pad_prc) {
	TCL_ERR(interp, "not a render callback: drawline may only be called during a render callback");
    }

    argc--;
    argv++;

    if ((argc%2) != 0)
      TCL_ERR(interp, "draw_line needs an even number of points");

    npts = argc / 2;
    Pad_renderer->Begin_line();
    for (i=0; i<npts; i++) {
	Pad_renderer->V2f(ATOXF(tkwin, argv[2*i]), ATOYF(tkwin, argv[2*i + 1]));
    }
    Pad_renderer->End_line();

    return(TCL_OK);
}

int
Padtcl_draw_image(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    float x, y;
    float bbox[4];
    Pad_ImageData *imageData;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 4) {
	TCL_ERR(interp, "wrong # args: should be \"pathName drawimage imagetoken xloc yloc\"");
    }
    if (!Pad_prc) {
	TCL_ERR(interp, "not a render callback: drawimage may only be called during a render callback");
    }

    x = ATOXF(tkwin, argv[2]);
    y = ATOYF(tkwin, argv[3]);
    
    imageData = Pad_renderer->Get_image(argv[1]);
    if (imageData) {
	bbox[XMIN] = x;
	bbox[YMIN] = y;
	bbox[XMAX] = x + imageData->width;
	bbox[YMAX] = y + imageData->height;
	Pad_renderer->Draw_image(imageData, bbox[XMIN], bbox[YMIN], bbox[XMAX], bbox[YMAX], FALSE);
    } else {
	Tcl_AppendResult(interp, "imagetoken not allocated by allocimage: ", argv[1], NULL);
	return(TCL_ERROR);
    }

    return(TCL_OK);
}

int
Padtcl_draw_polygon(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, npts;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc < 5)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName drawpolygon x1 y1 x2 y2 ...\"");
      }

    argc--;
    argv++;

    if ((argc%2) != 0)
      TCL_ERR(interp, "draw_polygon needs an even number of points");

    npts = argc / 2;
    Pad_renderer->Begin_polygon();
    for (i=0; i<npts; i++) {
	Pad_renderer->V2f(ATOXF(tkwin, argv[2*i]), ATOYF(tkwin, argv[2*i + 1]));
    }
    Pad_renderer->End_polygon();

    return(TCL_OK);
}

int
Padtcl_draw_text(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int level;
    float x, y;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 4) {
	TCL_ERR(interp, "wrong # args: should be \"pathName drawtext string xloc yloc\"");
    }

    x = ATOXF(tkwin, argv[2]);
    y = ATOYF(tkwin, argv[3]);

				// Level should depend on size.
    level = 2;

    Pad_renderer->Draw_string(argv[1], level, x, y, 0.0);

    return(TCL_OK);
}

int
Padtcl_setcolor(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setcolor color\"");
      }

    if (Pad_prc) {
	Pad_Color *color = Pad_LookupColorByName(argv[1]);
	if (color) {
	    Pad_renderer->Set_color(color);
	} else {
	    TCL_ERR(interp, "attempt to use setcolor before a color has been allocated");
	}
    } else {
	TCL_ERR(interp, "not a render callback: setcolor may only be called during a render callback");
    }

    return(TCL_OK);
}

int
Padtcl_setfont(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setfont fontname\"");
      }

    Pad_renderer->Set_font(Pad_LookupFontByName(argv[1]));

    return(TCL_OK);
}

int
Padtcl_setcapstyle(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setcapstyle capstyle\"");
      }

    Pad_renderer->Set_cap_style(argv[1]);

    return(TCL_OK);
}

int
Padtcl_setjoinstyle(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setjoinstyle joinstyle\"");
      }

    Pad_renderer->Set_join_style(argv[1]);

    return(TCL_OK);
}

int
Padtcl_setfontheight(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setfontheight fontsize\"");
      }

    Pad_renderer->Set_font_height(ATOYF(tkwin, argv[1]));

    return(TCL_OK);
}

int
Padtcl_gettextbbox(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_BBox bbox;
    float *bbox_data;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName gettextbbox string\"");
      }

    Pad_renderer->Get_string_bbox(argv[1], &bbox);
    bbox_data = bbox.Get();
    sprintf(Pad_result, "%f %f %f %f",
	    XTOUNITS(tkwin, bbox_data[0]),
	    YTOUNITS(tkwin, bbox_data[1]),
	    XTOUNITS(tkwin, bbox_data[2]),
	    YTOUNITS(tkwin, bbox_data[3]));
    TCL_RESULT(interp, Pad_result);

    return(TCL_OK);
}

//
// Padtcl_getgroup
//
//   pathName getgroup tagOrID
//
//     Return the group tagOrId is a member of, or an empty string
//     if it is not a member of a group.
//
int
Padtcl_getgroup(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_List objs;
    Pad_Object *obj;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName getgroup tagOrID\"");
      }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj) {
	if (obj->group) {
	    sprintf(Pad_result, "%d", obj->group->id);
	    TCL_RESULT(interp, Pad_result);
	}
    }

    return(TCL_OK);
}

//
// Padtcl_addgroupmember
//
//   pathName addgroupmember ?-notransform? tagOrID groupTagOrID
//
//     Add the items specified by the first tagOrID to the
//     group specified by the second tagOrID.  By default, items
//     are transformed so they don't change their location when 
//     added to the group - even if the group has a transformation.
//     This is implemented by transforming the item's transformation
//     to be the inverse of the group's transformation.  If
//     the -notransform flag is specified, this inverse transformation
//     is not applied, and the item will move by the group's transformation
//     when added.
//
int
Padtcl_addgroupmember(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Bool transform;
    Pad_List objs;
    Pad_Object *obj;
    Pad_Group *group;
    Pad_Iterator oi;

    if ((argc != 3) && (argc != 4))
      {
	  TCL_ERR(interp, 
	  "wrong # args: should be \"pathName addgroupmember ?-notransform? tagOrID groupTagOrID\"");
      }

    transform = TRUE;
    if (!strcmp(argv[1], "-notransform")) {
	transform = FALSE;
	argc--;
	argv++;
	       
    }

    tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj == NULL) {
	Tcl_AppendResult(interp, argv[2], " does not specify any objects", NULL);
	return(TCL_ERROR);
    }

    if (!obj->Is_group()) {
	Tcl_AppendResult(interp, argv[2], " is not a group", NULL);
	return(TCL_ERROR);
    }
    group = (Pad_Group *)obj;

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    DOLIST(oi, objs, Pad_Object, obj) {
	group->Add(obj, transform);
    }

    return(TCL_OK);
}

//
// Padtcl_removegroupmember
//
//   pathName removegroupmember ?-notransform? tagOrID
//
//     Remove the elements specified by the first tagOrID from the group they
//     are a member of, and return them to the regular pad surface.  By default, 
//     items are transformed so they don't change their location when 
//     removed from the group - even if the group has a transformation.
//     This is implemented by transforming the item's transformation
//     to be the inverse of the group's transformation.  If
//     the -notransform flag is specified, this inverse transformation
//     is not applied, and the item will move by the inverse of the
//     group's transformation when removed.
//
int
Padtcl_removegroupmember(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Bool transform;
    Pad_List objs;
    Pad_Object *obj;
    Pad_Group *group;
    Pad_Iterator oi;

    if ((argc != 2) && (argc != 3))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName removegroupmember ?-notransform? tagOrID\"");
      }

    transform = TRUE;
    if (!strcmp(argv[1], "-notransform")) {
	transform = FALSE;
	argc--;
	argv++;
    }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    DOLIST(oi, objs, Pad_Object, obj) {
	group = obj->group;
	if (group) {
	    group->Remove(obj, transform);
	}
    }

    return(TCL_OK);
}


//
// Padtcl_bindtags
//
//   pathName bindtags tagOrId ?type?
//
//     Change the ordering of event firings on all objects referred to by tagOrId.
//     If type is "general" (the default) events fire most generally first 
//     i.e., all, tags, id.  If type is "specific", then events fire
//     most specific first i.e., id, tags, all.
//
//     If tagOrId is pathName, then it does not change the ordering of any
//     objects, but controls the default ordering of future events.
//
int
Padtcl_bindtags(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_List objs;
    Pad_Object *obj;
    Pad_Iterator oi;

    if ((argc != 2) && (argc != 3))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName bindtags tagOrId ?type?\"");
      }

    if (!strcmp(argv[1], tkwin->Get_name())) {
	if (argc == 3) {
	    if (!strcmp(argv[2], "general")) {
		tkwin->view->pad->idEventFirst = FALSE;
	    } else if (!strcmp(argv[2], "specific")) {
		tkwin->view->pad->idEventFirst = TRUE;
	    } else {
		TCL_ERR(interp, "type must be \"general\" or \"specific\"");
	    }
	}
	if (tkwin->view->pad->idEventFirst) {
	    Tcl_AppendResult(interp, "specific", NULL);
	} else {
	    Tcl_AppendResult(interp, "general", NULL);
	}
    } else {
	tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
	Tcl_ResetResult(interp);
	DOLIST(oi, objs, Pad_Object, obj) {
	    if (argc == 3) {
		if (!obj->Set_event_order(argv[2])) {
		    TCL_ERR(interp, "type must be \"general\" or \"specific\"");
		}
	    }
	}
	obj = (Pad_Object *)objs.First();
	if (obj) {
	    Tcl_AppendResult(interp, obj->Get_event_order(), NULL);
	}
    }

    return(TCL_OK);
}

//
// Padtcl_addtype
//
//   pathName addtype <typename> <createScript>
//
//     Add <typename> to the list of allowed user defined types.
//     When a new object of type <typename> is created, the
//     <createScript> will be executed, and it must return
//     an object id.  When <createScript> is executed, the
//     pad widget the object is being created on will be added
//     on as an extra argument.
//
int
Padtcl_addtype(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool result;
    int rc = TCL_OK;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 3)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName addtype typename createScript\"");
      }

    result = tkwin->view->pad->Add_type(argv[1], argv[2], PAD_USER_TYPE);
    if (result == FALSE) {
	Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
	rc = TCL_ERROR;
    }

    return(rc);
}

//
// Padtcl_addoption
//
//   pathName addoption ?-nowrite? <typename> <optionname> <optionScript> <default>
//
//     Add a new option (named <optionname>) to all objects of
//     type <typename>.  <typename> must either be a built-in type,
//     or a user-defined type previously defined by addtype.
//     When <optionScript> is called, the following arguments
//     will be added on to the end of the script.
//       pathName:  The name of the pad widget the object is on
//       item:      The item id of the object being configured
//       [value]:   Optional value.  If value is specified, then the
//                  option must be set to this value.
//    <optionScript> must return the current (or new) value of the option.
//    <default> specifies the default value of this option.  This is used
//    to determine if the option should be written out (it only will if the
//    value is different than the default).
//    If -nowrite is specified, then this option won't be written out
//
int
Padtcl_addoption(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool result;
    Pad_Bool write;
    int rc = TCL_OK;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if ((argc != 5) && (argc != 6))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName addoption ?-nowrite? typename optionname optionScript default\"");
      }

    write = TRUE;
    if (!strcmp(argv[1], "-nowrite")) {
	write = FALSE;
	argc--;
	argv++;
    }
    result = tkwin->view->pad->Add_option(argv[1], argv[2], argv[3], PAD_USER_OPTION, write, argv[4]);
    if (result == FALSE) {
	Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
	rc = TCL_ERROR;
    }

    return(rc);
}

//
// Padtcl_modifier
//
//   pathName modifier create modifier
//   pathName modifier delete modifier
//   pathName modifier set modifier
//   pathName modifier get
//
//   Controls manipulation of user-defined event modifiers.
//   Modifiers may be created, and deleted.
//   And, the current modifier may be set and get.
//
int
Padtcl_modifier(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Bool rc;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName modifier subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "create", MAX(1, len))) {
	if (argc == 3) {
	    Pad_AddBindingModifier(argv[2]);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName modifier create modifier\"");
	}
    } else if (!strncmp(argv[1], "delete", MAX(1, len))) {
	if (argc == 3) {
	    rc = Pad_DeleteBindingModifier(argv[2]);
	    if (!rc) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "Modifier doesn't exist: ", argv[2], NULL);
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName modifier delete modifier\"");
	}
    } else if (!strncmp(argv[1], "get", MAX(1, len))) {
	if (argc == 2) {
	    Tcl_AppendResult(interp, tkwin->view->pad->Get_modifier(), NULL);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName modifier get\"");
	}
    } else if (!strncmp(argv[1], "set", MAX(1, len))) {
	if (argc == 3) {
	    rc = tkwin->view->pad->Set_modifier(argv[2]);
	    if (!rc) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "Modifier doesn't exist: ", argv[2], NULL);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName modifier set modifier\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: create, delete, get or set");
    }

    return(TCL_OK);
}

//
// Padtcl_addmodifier
//
//   pathName addmodifier modifier
//
//     Add <modifier> to the list of allowed user event modifiers
//     for event bindings.
//
int
Padtcl_addmodifier(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName addmodifier modifier\"");
      }

    Pad_AddBindingModifier(argv[1]);

    return(TCL_OK);
}

//
// Padtcl_deletemodifier
//
//   pathName deletemodifier modifier
//
//     Delete <modifier> from the list of allowed user event modifiers
//     for event bindings.
//
int
Padtcl_deletemodifier(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    int rc;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName deletemodifier modifier\"");
      }

    rc = Pad_DeleteBindingModifier(argv[1]);
    if (!rc) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "modifier doesn't exist: ", argv[1], NULL);
	return(TCL_ERROR);
    }

    return(TCL_OK);
}

//
// Padtcl_getmodifier
//
//   pathName getmodifier
//
//     Return the current active modifier
//
int
Padtcl_getmodifier(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName getmodifier\"");
      }

    Tcl_AppendResult(interp, tkwin->view->pad->Get_modifier(), NULL);

    return(TCL_OK);
}

//
// Padtcl_setmodifier
//
//   pathName setmodifier modifier
//
//     Make modifier be the active modifier for pathName.
//
int
Padtcl_setmodifier(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool rc;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setmodifier modifier\"");
      }

    rc = tkwin->view->pad->Set_modifier(argv[1]);
    if (!rc) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "modifier doesn't exist: ", argv[1], NULL);
    }

    return(TCL_OK);
}

//
// Padtcl_color
//
//   pathName color alloc color
//
//     Allocate a color for future use within a render callback.
//     Given a standard colorname, this allocates that color.
//     The colorname can then be used for the render commands.
//
//   pathName color free color
//
//     Frees a color previously allocated by alloccolor.
//
int
Padtcl_color(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName color subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "alloc", MAX(1, len))) {
	if (argc == 3) {
	    Pad_AllocColorByName(argv[2]);
	    Tcl_AppendResult(interp, argv[2], NULL);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName color alloc color\"");
	}
    } else if (!strncmp(argv[1], "free", MAX(1, len))) {
	if (argc == 3) {
	    Pad_FreeColorByName(argv[2]);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName color free color\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: alloc or free");
    }


    return(TCL_OK);
}

//
// Padtcl_alloccolor
//
//   pathName alloccolor color
//
//     Allocate a color for future use within a render callback.
//     Given a standard colorname, this allocates that color.
//     The colorname can then be used for the setcolor (and kpl set_color)
//     commands.
//
//     Returns color
//
int
Padtcl_alloccolor(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName alloccolor color\"");
      }

    Pad_AllocColorByName(argv[1]);
    Tcl_AppendResult(interp, argv[1], NULL);

    return(TCL_OK);
}

//
// Padtcl_freecolor
//
//   pathName freecolor color
//
//     Frees a color previously allocated by alloccolor.
//
int
Padtcl_freecolor(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName freecolor color\"");
      }

    Pad_FreeColorByName(argv[1]);

    return(TCL_OK);
}

//
// Process configure options for render command
// (argc, argv) contains just options and values.
//
static int
Render_config(Pad_TkWin *tkwin, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_Bool found;
    Pad_String value, optionValue, result;
    Pad_Iterator oi;
    Pad_View *view;
    Pad_Object *obj;

				// Return all option-value pairs
    if (argc == 0) {
	optionValue.Printf("-capstyle %s", Pad_renderer->Get_cap_style_name());
	Tcl_AppendElement(interp, optionValue.Get());

	Pad_renderer->Get_color(value);
	optionValue.Printf("-color %s", value.Get());
	Tcl_AppendElement(interp, optionValue.Get());

	Pad_renderer->Get_font(value);
	optionValue.Printf("-font %s", value.Get());
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-fontheight %f", YTOUNITS(tkwin, Pad_renderer->Get_font_height()));
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-joinstyle %s", Pad_renderer->Get_join_style_name());
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-level %d", Pad_prc->win->view->pad->level);
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-linewidth %f", YTOUNITS(tkwin, Pad_renderer->Get_line_width()));
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-magnification %f", Pad_renderer->Get_mag());
	Tcl_AppendElement(interp, optionValue.Get());

	obj = (Pad_Object*)Pad_prc->objects.First();
	if (obj) {
	  optionValue.Printf("-object %d", obj->id);
	} else {
	  optionValue.Printf("-object {}");
	}
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue = "-portals ";
	result = interp->result;
	Tcl_ResetResult(interp);
	found = FALSE;
	DOLIST(oi, Pad_prc->views, Pad_View, view) {
	    if (view->Type() == PAD_PORTAL) {
		found = TRUE;
		sprintf(Pad_result, "%d", view->id);
		Tcl_AppendElement(interp, Pad_result);
	    }
	}
	if (!found) {
	    Tcl_AppendResult(interp, "{}", NULL);
	}
	optionValue += interp->result;
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, result.Get(), NULL);
	Tcl_AppendElement(interp, optionValue.Get());

    } else {
	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-capstyle", MAX(2, len))) {
		if (argc >= 2) {
		    Pad_renderer->Set_cap_style(argv[1]);
		    argc--;
		    argv++;
		}
		Tcl_AppendResult(interp, Pad_renderer->Get_cap_style_name(), NULL);
	    } else if (!strncmp(argv[0], "-color", MAX(2, len))) {
		if (argc >= 2) {
		    Pad_Color *color = Pad_LookupColorByName(argv[1]);
		    if (color) {
			Pad_renderer->Set_color(color);
		    } else {
			TCL_ERR(interp, "attempt to use color before it has been allocated");
		    }
		    argc--;
		    argv++;
		}
		Pad_renderer->Get_color(value);
		Tcl_AppendResult(interp, value.Get(), NULL);
	    } else if (!strncmp(argv[0], "-font", MAX(2, len))) {
		if (argc >= 2) {
		    Pad_renderer->Set_font(Pad_LookupFontByName(argv[1]));
		    argc--;
		    argv++;
		}
		Pad_renderer->Get_font(value);
		Tcl_AppendResult(interp, value.Get(), NULL);
	    } else if (!strncmp(argv[0], "-fontheight", MAX(2, len))) {
		if (argc >= 2) {
		    Pad_renderer->Set_font_height(ATOYF(tkwin, argv[1]));
		    argc--;
		    argv++;
		}
		sprintf(Pad_result, "%f", YTOUNITS(tkwin, Pad_renderer->Get_font_height()));
		Tcl_AppendResult(interp, Pad_result, NULL);
	    } else if (!strncmp(argv[0], "-joinstyle", MAX(2, len))) {
		if (argc >= 2) {
		    Pad_renderer->Set_join_style(argv[1]);
		    argc--;
		    argv++;
		}
		Tcl_AppendResult(interp, Pad_renderer->Get_join_style_name(), NULL);
	    } else if (!strncmp(argv[0], "-level", MAX(2, len))) {
		if (argc == 1) {
		    sprintf(Pad_result, "%d", Pad_prc->win->view->pad->level);
		    TCL_RESULT(interp, Pad_result);
		} else {
		    TCL_ERR(interp, "-level is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-linewidth", MAX(2, len))) {
		if (argc >= 2) {
		    Pad_renderer->Set_line_width(ATOVALUEF(Pad_prc->win, argv[1]));
		    argc--;
		    argv++;
		}
		sprintf(Pad_result, "%f", YTOUNITS(tkwin, Pad_renderer->Get_line_width()));
		Tcl_AppendResult(interp, Pad_result, NULL);
	    } else if (!strncmp(argv[0], "-magnification", MAX(2, len))) {
		if (argc == 1) {
		    sprintf(Pad_result, "%f", Pad_renderer->Get_mag());
		    TCL_RESULT(interp, Pad_result);
		} else {
		    TCL_ERR(interp, "-magnification is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-object", MAX(2, len))) {
		if (argc == 1) {
		    obj = (Pad_Object*)Pad_prc->objects.First();
		    if (obj) {
		        sprintf(Pad_result, "%d", obj->id);
		    }
		    TCL_RESULT(interp, Pad_result);
		} else {
		    TCL_ERR(interp, "-object is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-portals", MAX(2, len))) {
		if (argc == 1) {
		    DOLIST(oi, Pad_prc->views, Pad_View, view) {
			if (view->Type() == PAD_PORTAL) {
			    sprintf(Pad_result, "%d", view->id);
			    Tcl_AppendElement(interp, Pad_result);
			}
		    }
		} else {
		    TCL_ERR(interp, "-portals is a read-only option");
		}
	    } else {
		Tcl_AppendResult(interp, "Invalid option, Try: -capstyle, -color, -font, -fontheight,\n",
				         "-joinstyle, -level, -linewidth, -magnification, or -portals", NULL);
		return(TCL_ERROR);
	    }
	    argc--;
	    argv++;
	}
    }

    return(TCL_OK);
}

//
// Process draw subcommand for render command.
//
static int
Render_draw(Pad_TkWin *tkwin, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    int level;
    int relief;
    int i, npts;
    float x, y;
    float width;
    float x0, y0, x1, y1;
    Pad_Bool rc;
    Pad_Border *border;
    Pad_Point min, max;
    Pad_ImageData *imageData;

    len = strlen(argv[0]);
    if ((!strncmp(argv[0], "border", MAX(1, len))) || (!strncmp(argv[0], "filledborder", MAX(1, len)))) {
	if (argc == 8) {
	    border = Pad_LookupBorderByName(argv[1]);
	    if (border) {
		rc = Pad_Relief_name_to_type(argv[2], relief);
		if (!rc) {
		    Tcl_AppendResult(interp, argv[2], " is an invalid relief", NULL);
		    return(TCL_ERROR);
		}
			
		width = ATOXF(tkwin, argv[3]);
			
		x0 = ATOXF(tkwin, argv[4]);
		y0 = ATOYF(tkwin, argv[5]);
		x1 = ATOXF(tkwin, argv[6]);
		y1 = ATOYF(tkwin, argv[7]);
		min.Set(x0, y0);
		max.Set(x1, y1);
			
		Pad_renderer->Set_border(border);
		if (argv[0][0] == 'b') {
		    Pad_renderer->Draw_3d_rectangle(min, max, width, relief);
		} else {
		    Pad_renderer->Draw_filled_3d_rectangle(min, max, width, relief);
		}
	    } else {
		Tcl_AppendResult(interp, "Border not previously allocated: ", argv[1], NULL);
		return(TCL_ERROR);
	    }
	} else {
	    if (argv[0][0] == 'b') {
		TCL_ERR(interp, "wrong # args: should be \"pathName render draw border bordertoken relief width x1 y1 x2 y2\"");
	    } else {
		TCL_ERR(interp, "wrong # args: should be \"pathName render draw filledborder bordertoken relief width x1 y1 x2 y2\"");
	    }
	}
    } else if (!strncmp(argv[0], "image", MAX(1, len))) {
	if (argc == 4) {
	    imageData = Pad_renderer->Get_image(argv[1]);
	    if (imageData) {
		x = ATOXF(tkwin, argv[2]);
		y = ATOYF(tkwin, argv[3]);
	    
		x0 = x;
		y0 = y;
		x1 = x + imageData->width;
		y1 = y + imageData->height;
		Pad_renderer->Draw_image(imageData, x0, y0, x1, y1, FALSE);
	    } else {
		Tcl_AppendResult(interp, "imagetoken not allocated: ", argv[1], NULL);
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render draw image imagetoken xloc yloc\"");
	}
    } else if (!strncmp(argv[0], "line", MAX(1, len))) {
	if (argc >= 2) {
	    argc--;
	    argv++;
	    
	    if ((argc%2) != 0)
		TCL_ERR(interp, "\"render draw line\" needs an even number of points");
	    
	    npts = argc / 2;
	    Pad_renderer->Begin_line();
	    for (i=0; i<npts; i++) {
		Pad_renderer->V2f(ATOXF(tkwin, argv[2*i]), ATOYF(tkwin, argv[2*i + 1]));
	    }
	    Pad_renderer->End_line();
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render draw line x1 y1 ...\"");
	}
    } else if (!strncmp(argv[0], "polygon", MAX(1, len))) {
	if (argc >= 2) {
	    argc--;
	    argv++;
	    
	    if ((argc%2) != 0)
		TCL_ERR(interp, "draw_polygon needs an even number of points");
	    
	    npts = argc / 2;
	    Pad_renderer->Begin_polygon();
	    for (i=0; i<npts; i++) {
		Pad_renderer->V2f(ATOXF(tkwin, argv[2*i]), ATOYF(tkwin, argv[2*i + 1]));
	    }
	    Pad_renderer->End_polygon();
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render draw polygon x1 y1 x2 y2 ...\"");
	}
    } else if (!strncmp(argv[0], "text", MAX(1, len))) {
	if (argc == 4) {
	    x = ATOXF(tkwin, argv[2]);
	    y = ATOYF(tkwin, argv[3]);

	    level = 2;		// Level should depend on size.
	    Pad_renderer->Draw_string(argv[1], level, x, y, 0.0);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render draw text string xloc yloc\"");
	}
    }

    return(TCL_OK);
}

//
// Padtcl_render
//
//   pathName render draw ...
//   pathName render configure ...
//
//     These are all the commands that are specific to render callbacks
//     (i.e., in a render script).  It involves configuring the current
//     render state, and rendering graphics.
//
//     scale dz
//     translate dx dy
//     draw border ...
//     draw filledborder ...
//     draw image ...
//     draw line ...
//     draw polygon
//     draw text
//     configure -capstyle
//     configure -joinstyle
//     configure -linestyle
//     configure -color
//     configure -font
//     configure -fontheight
//     configure -level       (read-only)
//     configure -mag         (read-only)
//     configure -portals     (read-only)
//     configure -object      (read-only)
//
int
Padtcl_render(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_Bool rc;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (!Pad_prc) {
	TCL_ERR(interp, "not a render callback: \"render\" may only be called during a render callback");
    }

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName render subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "draw", MAX(1, len))) {
	if (argc >= 3) {
	    rc = Render_draw(tkwin, interp, argc - 2, argv + 2);
	    if (rc == TCL_ERROR) {
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render draw subcommand ?arg ...?\"");
	}
    } else if (!strncmp(argv[1], "configure", MAX(1, len))) {
	if (argc >= 2) {
	    rc = Render_config(tkwin, interp, argc - 2, argv + 2);
	    if (rc == TCL_ERROR) {
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render configure ?option ?value? ...?\"");
	}
    } else if (!strncmp(argv[1], "scale", MAX(1, len))) {
	if (argc == 3) {
	    Pad_renderer->Scale(atof(argv[2]));
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render scale dz\"");
	}
    } else if (!strncmp(argv[1], "translate", MAX(1, len))) {
	if (argc == 4) {
	    Pad_renderer->Translate(ATOXF(tkwin, argv[2]), ATOYF(tkwin, argv[3]));
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName render translate dx dy\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: configure, draw, scale, or translate");
    }

    return(TCL_OK);
}

//
// Padtcl_allocimage
//
//   pathName allocimage imagefile ?-norgb?
//
//     Allocate an image for future use.
//     Given an image filename, this allocates that image.
//     The returned image token can then be used for new
//     image objects, and for the drawimage command.
//     If -norgb is specified, then rgb information is not stored,
//     and image can not be dithered.
//
//     Returns image token
//
int
Padtcl_allocimage(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_ImageData *imageData;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_String name;
    Pad_Bool rc;

    if ((argc != 2) && (argc != 3)) {
	TCL_ERR(interp, "wrong # args: should be \"pathName allocimage imagefile ?-norgb?\"");
    }

    rc = Pad_Expand_pathname(interp, argv[1], name);
    if (!rc) {
	TCL_ERR(interp, Pad_errorString.Get());
    }
    imageData = tkwin->renderer->Alloc_image(name.Get());
    if (!imageData) {
	Tcl_AppendResult(interp, "Error creating image: ", argv[1], NULL);
	return(TCL_ERROR);
    }
    if ((argc == 3) && !strcmp(argv[2], "-norgb")) {
	imageData->Free_rgb();
    }

    Tcl_AppendResult(interp, imageData->Get_token(), NULL);

    return(TCL_OK);
}

//
// Padtcl_freeimage
//
//   pathName freeimage imagetoken
//
//     Frees an image previously allocated by allocimage.
//
int
Padtcl_freeimage(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Bool rc;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName freeimage imagetoken\"");
      }

    rc = tkwin->renderer->Free_image(argv[1]);
    if (!rc) {
	TCL_ERR(interp, Pad_errorString.Get());
    }

    return(TCL_OK);
}

//
// Padtcl_font
//
//   pathName font bbox string fontname fontsize path
//
//     Interact with fonts.
//     
//     bbox - returns the bounding box of a string given a font
//     path - Specifies the font path
//     names - Returns the list of font names
//     loadbitmaps - Loads bitmaps for a given font
//
int
Padtcl_font(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_BBox bbox;
    float *bboxData;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName font subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "bbox", MAX(1, len))) {
	if ((argc == 4) || (argc == 5)) {
	    Pad_renderer->Set_font(Pad_LookupFontByName(argv[3]));
	    if (argc == 5) {
		Pad_renderer->Set_font_height(ATOYF(tkwin, argv[4]));
	    }
	    Pad_renderer->Get_string_bbox(argv[2], &bbox);
	    bboxData = bbox.Get();
	    sprintf(Pad_result, "%f %f %f %f",
		    XTOUNITS(tkwin, bboxData[0]),
		    YTOUNITS(tkwin, bboxData[1]),
		    XTOUNITS(tkwin, bboxData[2]),
		    YTOUNITS(tkwin, bboxData[3]));
	    TCL_RESULT(interp, Pad_result);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName font bbox string fontname ?fontsize?\"");
	}
    } else if (!strncmp(argv[1], "path", MAX(1, len))) {
	// change the font path (doesn't effect fonts already loaded)
	if (argc > 2 && argv[2][0]) {
	  if (argv[2][0] == '+') {
	    Pad_FontData::Append_to_path(argv[2] + 1);
	  } else {
	    Pad_FontData::Set_path(argv[2]);
	  }
	  // now damage the view to repaint the fonts
	  tkwin->view->Damage();
	}
	Pad_String pathName;
	Pad_FontData::Get_path(pathName);
	TCL_RESULT(interp, pathName.Get());	
    } else if (!strncmp(argv[1], "names", MAX(1, len))) {
	// return list of available fonts (unsorted)
	Pad_List fonts;
	Pad_Iterator li;
	Pad_String *name;
	Pad_FontData::Get_available_fonts(fonts);
	DOLIST(li, fonts, Pad_String, name) {
	    Tcl_AppendResult(interp, name->Get(), NULL);
	    Tcl_AppendResult(interp, " ", NULL);
	    delete name;
	}
    } else if (!strncmp(argv[1], "loadbitmaps", MAX(1, len))) {
	// load bitmap versions for fonts
	void Pad_load_x_font_bitmaps(Display *dpy, Pad_Font *font);
	int i;
	for (i = 2; i < argc; i++) {
	    Pad_Font *tmpFont = new Pad_Font(argv[i]);
	    Pad_load_x_font_bitmaps(tkwin->dpy->display, tmpFont);
	    delete tmpFont;

	    // now damage the view to repaint the fonts
	    tkwin->view->Damage();
	}
	// now damage the view to repaint the fonts
	tkwin->view->Damage();

    } else if (!strncmp(argv[1], "maxbitmapsize", MAX(1, len))) {
	// set max size at which bitmaps are used
	void Pad_set_x_font_maxsize(Display *dpy, int size);
	if (argc > 2) {
	    Pad_set_x_font_maxsize(tkwin->dpy->display, atoi(argv[2]));
            // now damage the view to repaint the fonts
            tkwin->view->Damage();
	}
	int Pad_get_x_font_maxsize(Display *);
	sprintf(Pad_result, "%d", Pad_get_x_font_maxsize(tkwin->dpy->display));
	TCL_RESULT(interp, Pad_result);

    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: bbox, loadbitmap, maxbitmapsize, names, or path");
    }

    return(TCL_OK);
}

int
Padtcl_setid(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_List objs;
    Pad_Object *obj;
    int id_num;

    if (argc != 3)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setid tagorid id\"");
      }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    if (!objs.Is_empty()) {
	obj = (Pad_Object *)objs.First();

	id_num = atoi(argv[2]);
	if (id_num > 1) {
	    if (!obj->Set_id(id_num)) {
		TCL_ERR(interp, "id already in use");
	    }
	} else {
	    TCL_ERR(interp, "id must be greater than 1");
	}
    } else {
	TCL_ERR(interp, "tagorid does specify an object");
    }

    Tcl_ResetResult(interp);

    return(TCL_OK);
}

//
// Padtcl_border
//
//   pathName border alloc border
//   pathName border free border
//
//     Allocate a border for future use within a render callback.
//     Given a standard border name, this allocates that border.
//     The border can then be used for render.
//
//     Frees a border previously allocated.
//
int
Padtcl_border(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_Bool rc;
    Pad_Border *border;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName border subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "alloc", MAX(1, len))) {
	if (argc == 3) {
	    border = Pad_AllocBorderByName(argv[2]);
	    if (border) {
		Tcl_AppendResult(interp, argv[2], NULL);
	    } else {
		Tcl_AppendResult(interp, "Invalid border: ",
				 argv[2], NULL);
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName border alloc border\"");
	}
    } else if (!strncmp(argv[1], "free", MAX(1, len))) {
	if (argc == 3) {
	    rc = Pad_FreeBorderByName(argv[2]);
	    if (!rc) {
		Tcl_AppendResult(interp, "Border not previously allocated: ",
				 argv[2], NULL);
		return(TCL_ERROR);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName border free border\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: alloc or free");
    }

    return(TCL_OK);
}

//
// Padtcl_allocborder
//
//   pathName allocborder color
//
//     Allocate a border for future use within a render callback.
//     Given a standard colorname, this allocates a border using that color.
//     The border can then be used for the drawborder (and kpl draw_border)
//     commands.
//
//     Returns bordername
//
int
Padtcl_allocborder(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Border *border;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName allocborder color\"");
      }

    border = Pad_AllocBorderByName(argv[1]);
    if (border) {
	Tcl_AppendResult(interp, argv[1], NULL);
    } else {
	Tcl_AppendResult(interp, "Invalid border: ",
			 argv[1], NULL);
	return(TCL_ERROR);
    }

    return(TCL_OK);
}

//
// Padtcl_freeborder
//
//   pathName freeborder border
//
//     Frees a border previously allocated by allocborder.
//
int
Padtcl_freeborder(ClientData , Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool rc;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName freeborder border\"");
      }

    rc = Pad_FreeBorderByName(argv[1]);

    if (!rc) {
	Tcl_AppendResult(interp, "Border not previously allocated by allocborder: ",
			 argv[1], NULL);
	return(TCL_ERROR);
    }

    return(TCL_OK);
}

//
// Padtcl_render_item
//
//   pathName renderitem ?tagOrId?
//
//     This renders the current (or specified item).  Only callable within a render callback,
//     this gives control on exactly when (and what) gets rendered.
//
int
Padtcl_render_item(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Object *obj, *currentObj;
    Pad_List objs;
    Pad_Iterator oi;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if ((argc != 1) && (argc != 2))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName renderitem ?tagOrId?\"");
      }

    if (Pad_prc) {
	Pad_prc->result = 1;
	currentObj = (Pad_Object *)Pad_prc->objects.First();
	if (!currentObj) {
	    currentObj = tkwin->view->pad;
	}
	if (argc == 1) {
	    Pad_prc->result = currentObj->Render();
	} else {
	    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
	    DOLIST(oi, objs, Pad_Object, obj)
	      {
		  if (obj->Type() != PAD_PAD) {
		      Pad_prc->result = obj->Render();
				// Stop rendering if interrupted
		      if (!Pad_prc->result) {
			  break;
		      }
		  }
	      }
	}
    } else {
	TCL_ERR(interp, "not a render callback: renderitem may only be called during a render callback");
    }

    return(TCL_OK);
}

int
Padtcl_setlinewidth(ClientData , Tcl_Interp *interp, int argc, char **argv)
{
    float width;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName setlinewidth width\"");
      }

    if (Pad_prc) {
	width = ATOVALUEF(Pad_prc->win, argv[1]);
	Pad_renderer->Set_line_width(width);
    } else {
	TCL_ERR(interp, "not a render callback: setlinewidth may only be called during a render callback");
    }

    return(TCL_OK);
}

int
Padtcl_getpads(ClientData, Tcl_Interp *interp, int, char **)
{
    Pad_Bool first;
    Pad_Name padName;
    Pad_List pads;
    Pad_Iterator oi;

    Pad_Win::Get_pads(pads);

    first = TRUE;
    DOLISTI(oi, pads, Pad_Name, padName) {
	if (!first) {
	    Tcl_AppendResult(interp, " ", NULL);
	}
	Tcl_AppendResult(interp, padName, NULL);
	first = FALSE;
    }

    return(TCL_OK);
}

int
Padtcl_getportals(ClientData , Tcl_Interp *interp, int, char **)
{
    Pad_View *view;
    Pad_Iterator pi;

    if (Pad_prc)
      {
	DOLIST(pi, Pad_prc->views, Pad_View, view)
	  {
	      if (view->Type() == PAD_PORTAL) {
		  sprintf(Pad_result, "%d", ((Pad_Portal *)view)->id);
		  Tcl_AppendElement(interp, Pad_result);
	      }
	  }
      }
    else
      {
	TCL_ERR(interp, "not a render callback: getportals may only be called during a render callback");
      }

    return(TCL_OK);
}

int
Padtcl_noise(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    float arg;

    if ((argc != 1) && (argc != 2))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName noise ?arg?\"");
      }

    arg = 0.0;
    if (argc == 2)
      arg = atof(argv[1]);

    sprintf(Pad_result, "%f", noise1(arg));
    TCL_RESULT(interp, Pad_result);

    return(TCL_OK);
}

//
// Padtcl_padxy
//
// pathName padxy ?-sticky? ?-portals? screenx screeny ?-gridspacing value?
//
//   Given screen coordinates, this command
//   returns the Pad coordinates that are displayed at
//   that location.  If gridspacing is specified, then the
//   pad coordinate is rounded to the nearest multiple of
//   gridspacing units.
//
//   If -sticky is specified, then the calculation is done for sticky objects
//   (i.e., independent of the current view).  Can't specify portals when
//   sticky is used.
//
//   If -portals is specified, then the pad coordinate is mapped
//   through the portals.  Otherwise, not.
//

int
Padtcl_padxy(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int mx, my;
    float x, y;
    float gridSpacing = 0;
    Pad_TkWin *tkwin;
    Pad_Event *event;
    Pad_Bool sticky;
    Pad_Bool portals;
    Pad_Bool found;

    if ((argc < 3) || (argc > 7))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName padxy ?-sticky? ?-portals? screenx screeny ?-gridspacing value?\"");
      }

    tkwin = (Pad_TkWin *)clientData;

				// Check for flags in any order
    sticky = FALSE;
    portals = FALSE;

    found = TRUE;
    while ((argc >= 2) && found) {
	found = FALSE;
	if (!strcmp(argv[1], "-sticky")) {
	    sticky = TRUE;
	    found = TRUE;
	}
	if (!strcmp(argv[1], "-portals")) {
	    portals = TRUE;
	    found = TRUE;
	}
	if (found) {
	    argc--;
	    argv++;
	}
    }

    if (sticky && portals) {
	TCL_ERR(interp, "can't specify portals and -sticky simultaneously");
    }

    if (argc >= 4)
      {
	  if (!strcmp(argv[3], "-gridspacing")) {
	      if (argc < 5) {
		  TCL_ERR(interp, "wrong # args: -gridspacing must have value");
	      }
	      gridSpacing = atof(argv[4]);
	      if (gridSpacing <= 0) {
		  TCL_ERR(interp, "gridspacing must be > 0");
	      }
	      argc -= 2;
	      argv += 2;
	  }
      }

    mx = atoi(argv[1]);
    my = atoi(argv[2]);
    event = new Pad_Event(tkwin, mx, my);
    if (portals) {
	tkwin->view->pad->Find_pick(event);
    } else {
	event->Update_coords(mx, my);
    }

    if (sticky) {
	x = XTOUNITS(tkwin, event->stickyPt.x);
	y = YTOUNITS(tkwin, event->stickyPt.y);
    } else {
	x = XTOUNITS(tkwin, event->pt.x);
	y = YTOUNITS(tkwin, event->pt.y);
    }

    if (gridSpacing > 0) {
	x = (int)((x + 0.5 * gridSpacing * SIGN(x)) / gridSpacing) * gridSpacing;
	y = (int)((y + 0.5 * gridSpacing * SIGN(y)) / gridSpacing) * gridSpacing;
    }
    delete event;

    sprintf(Pad_result, "%f %f", x, y);
    TCL_RESULT(interp, Pad_result);
    return(TCL_OK);
}

//
// clock
//
// pathName clock ?clockName ?rest? ?delete??
//
//    Return the name of a new clock.
//    Or if a clock is specified (as clockName), then the elapsed time
//    of this clock in milliseconds is returned.
//    If delete is specified, then delete the named clock and return an empty string.
//
int
Padtcl_clock(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    static int clock_cnt = 0;
    static Tcl_HashTable clock_table;
    static Pad_Bool init = FALSE;
    Tcl_HashEntry *entryPtr;
    int newentry;
    int clock;
    Pad_Time time;
    struct clock_data {
	long sec;
	long usec;
    };
    struct clock_data *clock_data_ptr;

    Pad_TkWin *tkwin;

    if ((argc != 1) && (argc != 2) && (argc != 3))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName clock ?clockName ?reset? ?delete??\"");
      }
    tkwin = (Pad_TkWin *)clientData;

    if (!init) {
	Tcl_InitHashTable(&clock_table, TCL_STRING_KEYS);
	init = TRUE;
    }

    if (argc == 1) {
	sprintf(Pad_result, "clock%d", clock_cnt++);
	TCL_RESULT(interp, Pad_result);

	clock_data_ptr = new clock_data;
	time.Update();
	clock_data_ptr->sec = time.Get_sec();
	clock_data_ptr->usec = time.Get_usec();

	entryPtr = Tcl_CreateHashEntry(&clock_table, Pad_result, &newentry);
	Tcl_SetHashValue(entryPtr, clock_data_ptr);
    } else {
	entryPtr = Tcl_FindHashEntry(&clock_table, argv[1]);
	if (entryPtr) {
	    clock_data_ptr = (clock_data *)Tcl_GetHashValue(entryPtr);
	    if (argc == 2) {
		time.Update();
		clock = (int)((1000000 * (time.Get_sec() - clock_data_ptr->sec) +
			       (time.Get_usec() - clock_data_ptr->usec)) / 1000);
		sprintf(Pad_result, "%d", clock);
		TCL_RESULT(interp, Pad_result);
	    } else {
		if (!strcmp(argv[2], "delete")) {
		    delete clock_data_ptr;
		    Tcl_DeleteHashEntry(entryPtr);
		} else if (!strcmp(argv[2], "reset")) {
		    time.Update();
		    clock_data_ptr->sec = time.Get_sec();
		    clock_data_ptr->usec = time.Get_usec();
		} else {
		    TCL_ERR(interp,
			    "wrong # args: should be \"pathName clock ?clockName ?reset? ?delete??\"");
		}
	    }
	} else {
	    Tcl_AppendResult(interp, "invalid clock name: ", argv[1], NULL);
	    return(TCL_ERROR);
	}
    }

    return(TCL_OK);
}

//
// Process configure options for html command
// (argc, argv) contains just options and values.
//
static int
HTML_config(HTML *html, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_String value, optionValue;

				// Return all option-value pairs
    if (argc == 0) {
	value = html->Get_last_modified();
	if (value == "") {
	    optionValue.Printf("-lastchangedate {}");
	} else {
	    optionValue.Printf("-type %s", value.Get());
	}
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-length %d", html->Get_length());
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-source %s", html->Get_source());
	Tcl_AppendElement(interp, optionValue.Get());

	value = html->Get_type();
	if (value == "") {
	    optionValue.Printf("-type {}");
	} else {
	    optionValue.Printf("-type %s", value.Get());
	}
	Tcl_AppendElement(interp, optionValue.Get());
    } else {
	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-lastchangedate", MAX(3, len))) {
		if (argc == 1) {
		    Tcl_AppendResult(interp, html->Get_last_modified(), NULL);
		} else {
		    TCL_ERR(interp, "-lastchangedate is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-length", MAX(3, len))) {
		if (argc == 1) {
		    sprintf(Pad_result, "%d", html->Get_length());
		    Tcl_AppendResult(interp, Pad_result, NULL);
		} else {
		    TCL_ERR(interp, "-length is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-source", MAX(2, len))) {
		if (argc == 1) {
		    Tcl_AppendResult(interp, html->Get_source(), NULL);
		} else {
		    TCL_ERR(interp, "-source is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-type", MAX(2, len))) {
		if (argc == 1) {
		    Tcl_AppendResult(interp, html->Get_type(), NULL);
		} else {
		    TCL_ERR(interp, "-type is a read-only option");
		}
	    } else {
		TCL_ERR(interp, "Invalid option, Try: -latechangedate, -length, -source, or -type");
	    }
	    argc--;
	    argv++;
	}
    }

    return(TCL_OK);
}

//
// Process configure options for html command
// (argc, argv) contains just options and values.
//
static int
HTMLAnchor_config(HTML_Anchor *htmlAnchor, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    HTML *html;
    Pad_ImageData *imageData;
    Pad_String value, optionValue;

				// Return all option-value pairs
    if (argc == 0) {
	html = htmlAnchor->Get_html();
	optionValue.Printf("-html %d", html->id);
	Tcl_AppendElement(interp, optionValue.Get());

	imageData = htmlAnchor->Get_imagedata();
	if (imageData) {
	    optionValue.Printf("-image %s", imageData->Get_token());
	} else {
	    optionValue = "-image {}";
	}
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-ismap %d", htmlAnchor->Get_map());
	Tcl_AppendElement(interp, optionValue.Get());

	value = htmlAnchor->Get_anchor_name();
	if (value == "") {
	    optionValue.Printf("-name {}");
	} else {
	    optionValue.Printf("-name %s", value.Get());
	}
	Tcl_AppendElement(interp, optionValue.Get());
	
	switch (htmlAnchor->Get_state()) {
	  case HTML_ANCHOR_UNVISITED:
	    value = "unvisited";
	    break;
	  case HTML_ANCHOR_ACTIVE:
	    value = "active";
	    break;
	  case HTML_ANCHOR_VISITED:
	    value = "visited";
	    break;
	  case HTML_ANCHOR_NOTLOADED:
	    value = "notloaded";
	    break;
	}
	optionValue.Printf("-state %s", value.Get());
	Tcl_AppendElement(interp, optionValue.Get());

	optionValue.Printf("-url %s", htmlAnchor->Get_url());
	Tcl_AppendElement(interp, optionValue.Get());
    } else {
	while (argc > 0) {
	    len = strlen(argv[0]);
	    if (!strncmp(argv[0], "-html", MAX(2, len))) {
		if (argc == 1) {
		    html = htmlAnchor->Get_html();
		    value = html->id;
		    Tcl_AppendResult(interp, value.Get(), NULL);
		} else {
		    TCL_ERR(interp, "-html is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-image", MAX(3, len))) {
		if (argc == 1) {
		    imageData = htmlAnchor->Get_imagedata();
		    if (imageData) {
			value = imageData->Get_token();
		    } else {
			value = "";
		    }
		    Tcl_AppendResult(interp, value.Get(), NULL);
		} else {
		    TCL_ERR(interp, "-image is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-ismap", MAX(3, len))) {
		if (argc == 1) {
		    value = htmlAnchor->Get_map();
		    Tcl_AppendResult(interp, value.Get(), NULL);
		} else {
		    TCL_ERR(interp, "-ismap is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-name", MAX(2, len))) {
		if (argc == 1) {
		    Tcl_AppendResult(interp, htmlAnchor->Get_anchor_name(), NULL);
		} else {
		    TCL_ERR(interp, "-name is a read-only option");
		}
	    } else if (!strncmp(argv[0], "-state", MAX(2, len))) {
		if (argc >= 2) {
		    if (!strcmp(argv[1], "unvisited")) {
			htmlAnchor->Set_state(HTML_ANCHOR_UNVISITED);
		    } else if (!strcmp(argv[1], "active")) {
			htmlAnchor->Set_state(HTML_ANCHOR_ACTIVE);
		    } else if (!strcmp(argv[1], "visited")) {
			htmlAnchor->Set_state(HTML_ANCHOR_VISITED);
		    } else if (!strcmp(argv[1], "notloaded")) {
			htmlAnchor->Set_state(HTML_ANCHOR_NOTLOADED);
		    } else {
			Pad_errorString = "Error: Invalid state: ";
			Pad_errorString += argv[1];
			Pad_errorString += ", try unvisited, active, or visited";
			TCL_ERR(interp, Pad_errorString.Get());
		    }
		    argc--;
		    argv++;
		}
		switch (htmlAnchor->Get_state()) {
		  case HTML_ANCHOR_UNVISITED:
		    value = "unvisited";
		    break;
		  case HTML_ANCHOR_ACTIVE:
		    value = "active";
		    break;
		  case HTML_ANCHOR_VISITED:
		    value = "visited";
		    break;
		  case HTML_ANCHOR_NOTLOADED:
		    value = "notloaded";
		    break;
		}
		Tcl_AppendResult(interp, value.Get(), NULL);
	    } else if (!strncmp(argv[0], "-url", MAX(2, len))) {
		if (argc == 1) {
		    value = htmlAnchor->Get_url();
		    Tcl_AppendResult(interp, value.Get(), NULL);
		} else {
		    TCL_ERR(interp, "-url is a read-only option");
		}
	    } else {
		TCL_ERR(interp, "Invalid option, Try: -html, -image, -ismap, -name, -state, or -url");
	    }
	    argc--;
	    argv++;
	}
    }

    return(TCL_OK);
}

//
// Padtcl_html
//
//   pathName html configure ...
//   pathName html anchor configure ...
//
//     Configure an html page, and retrieve information about it.
//
//     configure tagOrId -source          => Returns HTML source
//     configure tagOrId -type            => Returns Mime type
//     configure tagOrId -lastchangedata  => Returns last time/date data was changed (from server)
//     configure tagOrId -length          => Returns length of html source in bytes
//
//     anchor configure tagOrId -name     => Returns anchor name                      (read-only)
//     anchor configure tagOrId -html     => Returns html object anchor belongs to    (read-only)
//     anchor configure tagOrId -image    => Returns image associated with anchor     (read-only)
//     anchor configure tagOrId -ismap    => Returns TRUE if anchor is an image map   (read-only)
//     anchor configure tagOrId -url      => Returns url this anchor will take you to (read-only)
//     anchor configure tagOrId -state    => Returns state of anchor (unvisited, active, visited, or notloaded)
//
int
Padtcl_html(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    HTML *html;
    HTML_Anchor *htmlAnchor;
    Pad_Object *obj;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Bool rc;
    Pad_List objs;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName html subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "anchor", MAX(1, len))) {
	if (argc >= 3) {
	    len = strlen(argv[2]);
	    if (!strncmp(argv[2], "configure", MAX(1, len))) {
		if (argc >= 4) {
		    tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
		    obj = (Pad_Object *)objs.First();
		    if (obj) {
			if ((obj->Type() == PAD_HTML_ANCHOR) || 
			    (obj->Type() == PAD_HTML_ANCHOR_TEXT) || 
			    (obj->Type() == PAD_HTML_ANCHOR_IMAGE)) {
			    htmlAnchor = (HTML_Anchor *)obj;
			    rc = HTMLAnchor_config(htmlAnchor, interp, argc - 4, argv + 4);
			    if (rc == TCL_ERROR) {
				return(TCL_ERROR);
			    }
			} else {
			    TCL_ERR(interp, "tagOrId doesn't specify htmlanchor object");
			}
		    }
		} else {
		    TCL_ERR(interp, "wrong # args: should be \"pathName html anchor configure tagOrId ?option ?value? ...?\"");
		}
	    } else {
		TCL_ERR(interp, "Invalid subcommand, should be: configure");
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName html anchor subcommand ?arg ...?\"");
	}
    } else if (!strncmp(argv[1], "configure", MAX(1, len))) {
	if (argc >= 3) {
	    tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
	    obj = (Pad_Object *)objs.First();
	    if (obj) {
		if (obj->Type() == PAD_HTML) {
		    html = (HTML *)obj;
		    rc = HTML_config(html, interp, argc - 3, argv + 3);
		    if (rc == TCL_ERROR) {
			return(TCL_ERROR);
		    }
		} else {
		    TCL_ERR(interp, "tagOrId doesn't specify html object");
		}
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName html configure tagOrId ?option ?value? ...?\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: anchor or configure");
    }

    return(TCL_OK);
}

//
// pathName info subcommand
//
//   Return basic info as directed by the specified subcommand
//   Valid subcommands are:
//      html getsource tagOrId
//      html gettype tagOrId
//      html getlastchangedate tagOrId
//      html getlength tagOrId
//
//      htmlanchor getname tagOrId
//
//      image getdim token
//      image getname token
//
//      status render
//      status sharedmemory
//
int
Padtcl_info(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Object *obj;
    Pad_List objs;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_String status;

    if (argc < 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName info subcommand ?options ...?\"");
      }

    if (!strcmp(argv[1], "html")) {
	if (argc != 4) {
	    TCL_ERR(interp, "wrong # args: should be \"pathName info html option tagOrId\"");
	}

	HTML *html;

	tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
	obj = (Pad_Object *)objs.First();
	if (obj) {
	    if (!obj->Type() == PAD_HTML) {
		TCL_ERR(interp, "tagOrId doesn't specify html object");
	    }
	    html = (HTML *)obj;
	    if (!strcmp(argv[2], "getsource")) {
		Tcl_AppendResult(interp, html->Get_source(), NULL);
	    } else if (!strcmp(argv[2], "gettype")) {
		Tcl_AppendResult(interp, html->Get_type(), NULL);
	    } else if (!strcmp(argv[2], "getlastchangedate")) {
		Tcl_AppendResult(interp, html->Get_last_modified(), NULL);
	    } else if (!strcmp(argv[2], "getlength")) {
		sprintf(Pad_result, "%d", html->Get_length());
		Tcl_AppendResult(interp, Pad_result, NULL);
	    } else {
		TCL_ERR(interp, "invalid option, try: getsource, gettype, getlastchangedate, or getlength");
	    }
	}
    } else if (!strcmp(argv[1], "htmlanchor")) {
	if (argc != 4) {
	    TCL_ERR(interp, "wrong # args: should be \"pathName info htmlanchor option tagOrId\"");
	}

	int type;
	HTML_Anchor *htmlanchor;

	tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
	obj = (Pad_Object *)objs.First();
	if (obj) {
	    type = obj->Type();
	    if ((type != PAD_HTML_ANCHOR) && 
		(type != PAD_HTML_ANCHOR_IMAGE) && 
		(type != PAD_HTML_ANCHOR_TEXT)) {
		TCL_ERR(interp, "tagOrId doesn't specify htmlanchor object");
	    }
	    htmlanchor = (HTML_Anchor *)obj;
	    if (!strcmp(argv[2], "getname")) {
		Tcl_AppendResult(interp, htmlanchor->Get_anchor_name(), NULL);
	    } else {
		TCL_ERR(interp, "invalid option, try: getname");
	    }
	}
    } else if (!strcmp(argv[1], "image")) {
	if (argc != 4) {
	    TCL_ERR(interp, "wrong # args: should be \"pathName info image option token\"");
	}

	Pad_ImageData *imageData;
	
	imageData = tkwin->renderer->Get_image(argv[3]);
	if (imageData) {
	    if (!strcmp(argv[2], "getdim")) {
		sprintf(Pad_result, "%d %d", imageData->width, imageData->height);
		TCL_RESULT(interp, Pad_result);
	    } else if (!strcmp(argv[2], "getname")) {
		if (imageData->name) {
		    Tcl_AppendResult(interp, imageData->name, NULL);
		}
	    } else {
		TCL_ERR(interp, "invalid option, try: getdim, or getname");
	    }
	} else {
	    Tcl_AppendResult(interp, "invalid image token: ", argv[3], NULL);
	    return(TCL_ERROR);
	}
    } else if (!strcmp(argv[1], "status")) {
	if (argc != 3) {
	    TCL_ERR(interp, "wrong # args: should be \"pathName info status option\"");
	}

	if (!strcmp(argv[2], "render")) {
	    Pad *pad;

	    pad = tkwin->view->pad;
	    status = "O:";
	    status += pad->Sizeof_tree();
	    status += ", R:";
	    status += pad->objectsRendered;
	    status += ", L:";
	    status += pad->level;
	    status += ", T:";
	    status += pad->view->renderTime;
	    Tcl_AppendResult(interp, status.Get(), NULL);
	} else if (!strcmp(argv[2], "sharedmemory")) {
	    status = Pad_sharedMemory;
	    Tcl_AppendResult(interp, status.Get(), NULL);
	} else {
	    TCL_ERR(interp, "invalid option, try: render or sharedmemory");
	}
    } else {
	TCL_ERR(interp, "invalid subcommand, try: html, htmlanchor, image, or status");
    }
    
    return(TCL_OK);
}

//
// Generate a random integer between min and max, inclusively
//
int
Padtcl_random(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    int min, max;
    Pad_String str;

    if (argc != 3)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName random min max\"");
      }

    min = atoi(argv[1]);
    max = atoi(argv[2]);
    if (min >= max) {
	TCL_ERR(interp, "min must be strictly less than max");
    }

    str = Pad_Random(min, max);
    Tcl_AppendResult(interp, str.Get(), NULL);
    return(TCL_OK);
}

int
Padtcl_raise(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool rc;
    Pad_Bool one, first, layerp, found;
    Pad_Object *obj, *above, *previous_obj;
    Pad_List objs, refs;
    Pad_Iterator oi;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if ((argc < 2) || (argc > 4))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName raise ?-layer? ?-one? tagOrId ?aboveThis?\"");
      }

    one = FALSE;
    layerp = FALSE;
    found = TRUE;
    while ((argc >= 2) && found) {
	found = FALSE;
	if (!strcmp(argv[1], "-one")) {
	    one = TRUE;
	    found = TRUE;
	}
	if (!strcmp(argv[1], "-layer")) {
	    layerp = TRUE;
	    found = TRUE;
	}
	if (found) {
	    argc--;
	    argv++;
	}
    }
    if (one && layerp) {
	TCL_ERR(interp, "Can't specify -one and -layer flags together");
    }

				// If raising layers
    if (layerp) {
        rc = TRUE;
        if (argc == 3) {
	    rc = tkwin->view->pad->Raise_layer(argv[1], argv[2]);
	} else if (argc == 2) {
	    rc = tkwin->view->pad->Raise_layer(argv[1]);
	}
	if (!rc) {
	    TCL_ERR(interp, Pad_errorString.Get());
	}
	    
	return(TCL_OK);
    }

				// Else, raising items
    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (!obj) {
	return(TCL_OK);
    }
    if (one) {
	above = obj->next;
    } else {
	above = NULL;
    }

				// Find the highest object in aboveThis
    if (argc == 3) {
	if (one) {
	    TCL_ERR(interp, "Can't specify -one flag and aboveThis");
	}
	tkwin->view->pad->Find_with_tagorid(argv[2], refs, TRUE);
	above = (Pad_Object *)refs.Last();
    }

    first = TRUE;
    previous_obj = obj;		// For compiler warnings only
    DOLIST(oi, objs, Pad_Object, obj)
      {
	  if (first) {
	      rc = obj->Raise(above);
	      if (!rc) {
		  TCL_ERR(interp, Pad_errorString.Get());
	      }
	      first = FALSE;
	  } else {
	      rc = obj->Raise(previous_obj);
	      if (!rc) {
		  TCL_ERR(interp, Pad_errorString.Get());
	      }
	  }
	  previous_obj = obj;
      }

    Tcl_ResetResult(interp);

    return(TCL_OK);
}

int
Padtcl_lower(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Bool rc;
    Pad_Bool one, first, layerp, found;
    Pad_Object *obj, *below, *previous_obj;
    Pad_List objs, refs;
    Pad_Iterator oi;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if ((argc < 2) || (argc > 4))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName lower ?-layer? ?-one? tagOrId ?belowThis?\"");
      }

    one = FALSE;
    layerp = FALSE;
    found = TRUE;
    while ((argc >= 2) && found) {
	found = FALSE;
	if (!strcmp(argv[1], "-one")) {
	    one = TRUE;
	    found = TRUE;
	}
	if (!strcmp(argv[1], "-layer")) {
	    layerp = TRUE;
	    found = TRUE;
	}
	if (found) {
	    argc--;
	    argv++;
	}
    }
    if (one && layerp) {
	TCL_ERR(interp, "Can't specify -one and -layer flags together");
    }

				// If lowering layers
    if (layerp) {
        if (argc == 3) { 
	    rc = tkwin->view->pad->Lower_layer(argv[1], argv[2]);
	} else {
	    rc = tkwin->view->pad->Lower_layer(argv[1]);
	}
	if (!rc) {
	    TCL_ERR(interp, Pad_errorString.Get());
	}
	    
	return(TCL_OK);
    }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (!obj) {
	return(TCL_OK);
    }
    if (one) {
	below = obj->prev;
    } else {
	below = NULL;
    }

				// Find the lowest object in belowThis
    if (argc == 3) {
	if (one) {
	    TCL_ERR(interp, "Can't specify -one flag and belowThis");
	}
	tkwin->view->pad->Find_with_tagorid(argv[2], refs, TRUE);
	below = (Pad_Object *)refs.First();
    }
    first = TRUE;
    previous_obj = obj;		// For compiler warnings only
    DOLIST(oi, objs, Pad_Object, obj)
      {
	  if (first) {
	      rc = obj->Lower(below);
	      if (!rc) {
		  TCL_ERR(interp, Pad_errorString.Get());
	      }
	      first = FALSE;
	  } else {
	      rc = obj->Raise(previous_obj);
	      if (!rc) {
		  TCL_ERR(interp, Pad_errorString.Get());
	      }
	  }
	  previous_obj = obj;
      }

    Tcl_ResetResult(interp);

    return(TCL_OK);
}

//
// Padtcl_layer:
//
// pathName layer subcommand ?arg ...?
//    create
//    delete
//    names
//
int
Padtcl_layer(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad_Layer *layer;
    Pad_Iterator oi;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 "pathName layer subcommand ?arg ...?\"",
			 (char *) NULL);
	return(TCL_ERROR);
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "create", MAX(1, len))) {
	if (argc == 3) {
	    tkwin->view->pad->Create_layer(argv[2]);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layer create layerName\"");
	}
    } else if (!strncmp(argv[1], "delete", MAX(1, len))) {
	if (argc == 3) {
	    layer = tkwin->view->pad->Get_layer_from_name(argv[2]);
	    if (layer) {
		tkwin->view->pad->Delete_layer(layer);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layer delete layerName\"");
	}
    } else if (!strncmp(argv[1], "names", MAX(1, len))) {
	if (argc == 2) {
	    DOLIST(oi, tkwin->view->pad->layers, Pad_Layer, layer) {
		Tcl_AppendElement(interp, layer->name);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layer names\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: create, delete, or names");
    }

    return(TCL_OK);
}

//
// Padtcl_layout:
//
// pathName layout subcommand ?arg ...?
//    align ...
//    distribute ...
//    snap ...
//    position ...
//    size ...
//
int
Padtcl_layout(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    int len;
    int type;
    Pad_List objs;
    Pad *pad;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be ",
			 "\"pathName layout subcommand ?arg ...?\"",
			 (char *) NULL);
	return(TCL_ERROR);
    }

    pad = tkwin->view->pad;
    len = strlen(argv[1]);
    if (!strncmp(argv[1], "align", MAX(1, len))) {
	Pad_Bool anchor, overlapOnly;
	Pad_PList coords;

	if (argc >= 4) {
	    if (!strcmp(argv[2], "-left")) {
		type = PAD_LAYOUT_LEFT;
	    } else if (!strcmp(argv[2], "-right")) {
		type = PAD_LAYOUT_RIGHT;
	    } else if (!strcmp(argv[2], "-top")) {
		type = PAD_LAYOUT_TOP;
	    } else if (!strcmp(argv[2], "-bottom")) {
		type = PAD_LAYOUT_BOTTOM;
	    } else {
		TCL_ERR(interp, "Invalid alignment type.  Must be: -left, -right, -top or -bottom");
	    }
	    if (!strcmp(argv[3], "-anchor")) {
		anchor = TRUE;
		argc--;
		argv++;
	    } else {
		anchor = FALSE;
	    }
	    if (!strncmp(argv[3], "-coords", MAX(6, strlen(argv[3])))) {
		if (argc >= 6) {
		    int listrc;
		    int listargc;
		    char **listargv;
		    Pad_Point point;

		    listrc = Tcl_SplitList(interp, argv[4], &listargc, &listargv);
		    if (listrc == TCL_OK) {
			if ((listargc % 2) != 0) {
			    TCL_ERR(interp, "-coords needs an even number of points");
			} else if (listargc < 2) {
			    TCL_ERR(interp, "-coords needs at least 1 point");
			} else {
			    DOTIMES(i, listargc/2) {
				point.Set(ATOXF(tkwin, listargv[2*i]), ATOYF(tkwin, listargv[2*i+1]));
				coords.Push_last(&point);
			    }
			}
			ckfree((char *) listargv);
		    } else {
			TCL_ERR(interp, "-coords needs valid list of points");
		    }
		    if (!strcmp(argv[5], "-overlaponly")) {
			overlapOnly = TRUE;
			argc--;
			argv++;
		    } else {
			overlapOnly = FALSE;
		    }
		    argc -= 2;
		    argv += 2;
		} else {
		    TCL_ERR(interp, "wrong # args: -coords must be specified with a coordinate list");
		}
	    }
	    for (i=3; i<argc; i++) {
		pad->Find_with_tagorid(argv[i], objs, TRUE, TRUE);
	    }
	    if (coords.Is_empty()) {
		pad->Align(objs, type, anchor);
	    } else {
		pad->Align(objs, type, coords, overlapOnly, anchor);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layout align <type> [-anchor] [-coords {x y ...} [-overlaponly]] tagOrId [tagOrId ...]\"");
	}
    } else if (!strncmp(argv[1], "distribute", MAX(1, len))) {
	if (argc >= 4) {
	    float space;
	    Pad_Bool spacing;
	    Pad_Point point;
	    Pad_PList points;

	    len = strlen(argv[2]);
	    if (!strncmp(argv[2], "-horizontal", MAX(2, len))) {
		type = PAD_LAYOUT_HORIZONTAL;
	    } else if (!strncmp(argv[2], "-vertical", MAX(2, len))) {
		type = PAD_LAYOUT_VERTICAL;
	    } else if (!strncmp(argv[2], "-coords", MAX(2, len))) {
		int listrc;
		int listargc;
		char **listargv;

		type = PAD_LAYOUT_COORDS;
		listrc = Tcl_SplitList(interp, argv[3], &listargc, &listargv);
		if (listrc == TCL_OK) {
		    if ((listargc % 2) != 0) {
			TCL_ERR(interp, "-coords needs an even number of points");
		    } else if (listargc < 4) {
			TCL_ERR(interp, "-coords needs at least 2 points");
		    } else {
			DOTIMES(i, listargc/2) {
			    point.Set(ATOXF(tkwin, listargv[2*i]), ATOYF(tkwin, listargv[2*i+1]));
			    points.Push_last(&point);
			}
		    }
		    ckfree((char *) listargv);
		} else {
		    TCL_ERR(interp, "-coords needs valid list of points");
		}
		argc--;
		argv++;
	    } else {
		TCL_ERR(interp, "Invalid alignment type.  Must be: \"-horizontal\", \"-vertical\", or \"-coords {x1 y1 x2 y2 ...}\"");
	    }
	    if (argc < 4) {
				// Check again, type may have reduced total # of args
		TCL_ERR(interp, "wrong # args: should be \"pathName layout distribute <type> [-space space] tagorId [tagOrId ...]\"");
	    }
	    if (!strcmp(argv[3], "-space")) {
		if (argc >= 5) {
		    spacing = TRUE;
		    if (type == PAD_LAYOUT_HORIZONTAL) {
			space = ATOXF(tkwin, argv[4]);
		    } else if (type == PAD_LAYOUT_VERTICAL) {
			space = ATOYF(tkwin, argv[4]);
		    } else {
			space = ATOVALUEF(tkwin, argv[4]);
		    }
		    argc -= 2;
		    argv += 2;
		} else {
		    TCL_ERR(interp, "wrong # args: -space must be specified with a number");
		}
	    } else {
		spacing = FALSE;
	    }
	    if (argc < 4) {
				// Check again, flags may have reduced total # of args
		TCL_ERR(interp, "wrong # args: should be \"pathName layout distribute <type> [-space space] tagorId [tagOrId ...]\"");
	    }
	    for (i=3; i<argc; i++) {
		pad->Find_with_tagorid(argv[i], objs, TRUE, TRUE);
	    }
	    if (type == PAD_LAYOUT_COORDS) {
		if (spacing) {
		    pad->Distribute(objs, points, space);
		} else {
		    pad->Distribute(objs, points);
		}
	    } else {
		if (spacing) {
		    pad->Distribute(objs, type, space);
		} else {
		    pad->Distribute(objs, type);
		}
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layout distribute <type> [-space space] tagorId [tagOrId ...]\"");
	}
    } else if (!strncmp(argv[1], "position", MAX(1, len))) {
	if (argc >= 9) {
	    Pad_Point targetPt, srcPt;
	    Pad_Object *targetObj;
	    Pad_BBox targetBBox;
	    Pad_Bool objType;
	    int animationTime = 0;

	    if (!strcasecmp(argv[2], "-time")) {
		animationTime = atoi(argv[3]);
		argc -= 2;
		argv += 2;
	    }

	    targetPt.Set(ATOXF(tkwin, argv[2]), ATOYF(tkwin, argv[3]));
	    if (!strcmp(argv[4], "-ref") || !strcmp(argv[4], "-reference")) {
		objType = TRUE;
		pad->Find_with_tagorid(argv[5], objs, TRUE);
		targetObj = (Pad_Object *)objs.First();
		if (targetObj == NULL){
		    TCL_ERR(interp, "invalid reference id");
		}
	    } else if (!strcmp(argv[4], "-bbox")) {
		int listrc;
		int listargc;
		char **listargv;

		objType = FALSE;
		listrc = Tcl_SplitList(interp, argv[5], &listargc, &listargv);
		if (listrc == TCL_OK) {
		    if (listargc != 4) {
			TCL_ERR(interp, "-bbox needs at least 2 points");
		    } else {
			targetBBox.Set(ATOXF(tkwin, listargv[0]), ATOYF(tkwin, listargv[1]),
				       ATOXF(tkwin, listargv[2]), ATOYF(tkwin, listargv[3]));
		    }
		    ckfree((char *) listargv);
		} else {
		    TCL_ERR(interp, "-bbox needs valid list of coords");
		}
	    } else {
		TCL_ERR(interp, "Invalid position type.  Must be: \"-ref tagOrId\", or \"-bbox {x1 y1 x2 y2}\"");
	    }
	    objs.Make_empty();
	    srcPt.Set(ATOXF(tkwin, argv[6]), ATOYF(tkwin, argv[7]));
	    for (i=8; i<argc; i++) {
		pad->Find_with_tagorid(argv[i], objs, TRUE, TRUE);
	    }
 	    if (objType) {
		pad->Layout_position(targetObj, targetPt, objs, srcPt, animationTime);
	    } else {
		pad->Layout_position(targetBBox, targetPt, objs, srcPt, animationTime);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layout position [-time animationTime] x1 y1 <type> x2 y2 tagOrId [tagOrId ...]\"");
	}
    } else if (!strncmp(argv[1], "size", MAX(2, len))) {
	if (argc >= 4) {
	    float scale = 1.0;
	    float newDim = 1.0;
	    Pad_Object *obj = NULL;
	    Pad_List items;

	    if (!strcmp(argv[2], "-width")) {
		type = PAD_LAYOUT_WIDTH;
	    } else if (!strcmp(argv[2], "-height")) {
		type = PAD_LAYOUT_HEIGHT;
	    } else {
		TCL_ERR(interp, "Invalid size type.  Must be: \"-width\", or \"-height\"");
	    }
	    if (!strcmp(argv[3], "-ref") || !strcmp(argv[3], "-reference")) {
		if (argc >= 5) {
		    pad->Find_with_tagorid(argv[4], items, TRUE);
		    obj = (Pad_Object *)items.First();
		    if (obj == NULL){
		      TCL_ERR(interp, "invalid reference id");
		    }
		    argc -= 2;
		    argv += 2;
		} else {
		    TCL_ERR(interp, "wrong # args: -ref must be specified with a tagOrId");
		}
	    }
	    if (!strcmp(argv[3], "-scale")) {
		if (argc >= 5) {
		    if (obj) {
			scale = atof(argv[4]);
		    } else {
			if (type == PAD_LAYOUT_WIDTH) {
			    newDim = ATOXF(tkwin, argv[4]);
			} else {
			    newDim = ATOYF(tkwin, argv[4]);
			}
		    }
		    argc -= 2;
		    argv += 2;
		} else {
		    TCL_ERR(interp, "wrong # args: -scale must be specified with a scale");
		}
	    }

	    for (i=3; i<argc; i++) {
		pad->Find_with_tagorid(argv[i], objs, TRUE, TRUE);
	    }
	    if (obj) {
		pad->Layout_size(type, obj, objs, scale);
	    } else {
		pad->Layout_size(type, objs, newDim);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layout size <type> [-ref tagOrId] [-scale scale] tagOrId [tagOrId ...]\"");
	}
    } else if (!strncmp(argv[1], "snap", MAX(2, len))) {
	if (argc >= 4) {
	    float grid;

	    grid = ATOVALUEF(tkwin, argv[2]);
	    for (i=3; i<argc; i++) {
		pad->Find_with_tagorid(argv[i], objs, TRUE, TRUE);
	    }
	    pad->Snap(grid, objs);
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName layout snap grid tagOrId [tagOrId ...]\"");
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: align, distribute, position, size, or snap");
    }

    return(TCL_OK);
}

//
// Padtcl_bind:
//
// pathName bind tagOrId ?sequence? ?command?
//
int
Padtcl_bind(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int rc;
    ClientData object;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if ((argc < 2) || (argc > 5)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 "pathName bind tagOrId ?sequence? ?command?\"",
			 (char *) NULL);
	return(TCL_ERROR);
    }

	/*
	 * Figure out what object to use for the binding (individual
	 * item vs. tag).
	 */
    object = NULL;
    if (isdigit(argv[1][0])) {
	long id;
	char *end;
	
	id = strtol(argv[1], &end, 0);
	if (*end != 0) {
	    goto bindByTag;
	}
	object = (ClientData)tkwin->view->pad->Get_object_from_id((int)id);
	if (object == 0) {
	    Tcl_AppendResult(interp, "item \"", argv[1],
			     "\" doesn't exist", NULL);
	    return(TCL_ERROR);
	}
    } else {
      bindByTag:
	object = (ClientData) Tk_GetUid(argv[1]);
    }

    if (argc == 4) {
	int append = 0;
	unsigned long mask;
	
	if (argv[3][0] == 0) {
	    rc = Pad_DeleteBinding(interp, tkwin->bindingTable, object, argv[2]);
	    return(rc);
	}
	if (argv[3][0] == '+') {
	    argv[3]++;
	    append = 1;
	}
	mask = Pad_CreateBinding(interp, tkwin->bindingTable,
				object, argv[2], argv[3], append);
	if (mask == 0) {
	    return(TCL_ERROR);
	}
	if (mask & (unsigned) ~(ButtonMotionMask|Button1MotionMask
				|Button2MotionMask|Button3MotionMask|Button4MotionMask
				|Button5MotionMask|ButtonPressMask|ButtonReleaseMask
				|EnterWindowMask|LeaveWindowMask|KeyPressMask
				|KeyReleaseMask|PointerMotionMask
				| Pad_PortalInterceptMask 
				| Pad_CreateMask
				| Pad_ModifyMask
				| Pad_DeleteMask
				| Pad_WriteMask)) {
	    Pad_DeleteBinding(interp, tkwin->bindingTable,
			      object, argv[2]);
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "requested illegal events; ",
			     "only key, button, motion, enter/leave\n",
			     "PortalIntercept, Create, Modify, Delete, or Write "
			     "events may be used", (char *) NULL);
	    return(TCL_ERROR);
	}
    } else if (argc == 3) {
	char *command;
	
	command = Pad_GetBinding(interp, tkwin->bindingTable,
				 object, argv[2]);
	if (command == NULL) {
	    return(TCL_ERROR);
	}
	interp->result = command;
    } else {
	Pad_GetAllBindings(interp, tkwin->bindingTable, object);
    }

    return(TCL_OK);
}

/*
 Padtcl_getlevel:

 pathName getlevel

 Returns the current render level
*/

int
Padtcl_getlevel(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{
    Pad_TkWin *tkwin;

    tkwin = (Pad_TkWin *)clientData;

    if (argc != 1) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 "pathName getlevel\"", NULL);
	return(TCL_ERROR);
    }

    if (Pad_prc) {
	sprintf(Pad_result, "%d", Pad_prc->win->view->pad->level);
	TCL_RESULT(interp, Pad_result);
    } else {
	TCL_ERR(interp, "not a render callback: getlevel may only be called during a render callback");
    }

    return(TCL_OK);
}

/*
 Padtcl_getsize:

 pathName getsize tagOrId ?portalID ..?

 Returns the size (max dimension in current units) of the specified object.
 as seen through the specified portal list.
*/

int
Padtcl_getsize(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    float lod;
    float global_width, global_height;
    Pad_TkWin *tkwin;
    Pad_Object *obj, *view_obj;
    Pad *pad;
    Pad_View *view;
    Pad_List views, objs;

    tkwin = (Pad_TkWin *)clientData;
    pad = tkwin->view->pad;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 "pathName getsize tagOrId ?portalID ...?\"", NULL);
	return(TCL_ERROR);
    }
    
    views.Push(tkwin->view);
				// Extract portals from list
    view = tkwin->view;
    for (i=2; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    pad->Find_with_tagorid(argv[i], objs, TRUE);
	    view_obj = (Pad_Object *)objs.First();
	    if (view_obj) {
		if (view_obj->Type() == PAD_PORTAL) {
		    view = (Pad_View *)view_obj;
		    views.Push_last(view);
		    pad = ((Pad_Portal *)view)->lookon->view->pad;
		} else {
		    Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				     view->win->Get_name(), NULL);
		    return(TCL_ERROR);
		}
	    }
	}
    }
	
    pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj) {
	lod = obj->Compute_dimensions(views, global_width, global_height);
	sprintf(Pad_result, "%f", VALUETOUNITS(tkwin, lod));
	TCL_RESULT(interp, Pad_result);
    }

    return(TCL_OK);
}

/*
 Padtcl_isvisible:

 pathName isvisible tagOrId ?portalID ..?

 Returns true if the first item specified by tagOrId is visible
 in the main view, or within the last portal specified by the portal list.
*/

int
Padtcl_isvisible(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    Pad_Bool visible;
    Pad_TkWin *tkwin;
    Pad_Object *obj, *view_obj;
    Pad *pad;
    Pad_View *view;
    Pad_List objs;

    tkwin = (Pad_TkWin *)clientData;
    pad = tkwin->view->pad;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 "pathName isvisible tagOrId ?portalID ...?\"", NULL);
	return(TCL_ERROR);
    }
    
				// Extract portals from list
    view = tkwin->view;
    for (i=2; i<argc; i++) {
	if (argv[i][0] != '\0') {
	    pad->Find_with_tagorid(argv[i], objs, TRUE);
	    view_obj = (Pad_Object *)objs.First();
	    if (view_obj) {
		if (view_obj->Type() == PAD_PORTAL) {
		    view = (Pad_View *)view_obj;
		    pad = ((Pad_Portal *)view)->lookon->view->pad;
		} else {
		    Tcl_AppendResult(interp, "Error: portalID ", argv[i], " is not a portal on ",
				     view->win->Get_name(), NULL);
		    return(TCL_ERROR);
		}
	    }
	}
    }
	
    pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj) {
	visible = obj->Is_visible(view->viewBBox.Get(), view);
	sprintf(Pad_result, "%d", visible);
	TCL_RESULT(interp, Pad_result);
    }

    return(TCL_OK);
}

/*
 Padtcl_getmag:

 pathName getmag tagOrId

 Returns the current magnification of the specified object.
 Based on the current view and the scale of the object.
*/

int
Padtcl_getmag(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int id;
    Pad_TkWin *tkwin;
    Pad_Object *obj;

    tkwin = (Pad_TkWin *)clientData;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 "pathName getmag tagOrId\"", NULL);
	return(TCL_ERROR);
    }

    if (Pad_prc) {
	id = atoi(argv[1]);
	obj = tkwin->view->pad->Get_object_from_id(id);
	if (obj) {
	    sprintf(Pad_result, "%f", Pad_renderer->Get_mag());
	    TCL_RESULT(interp, Pad_result);
	}
    } else {
	TCL_ERR(interp, "not a render callback: getmag may only be called during a render callback");
    }

    return(TCL_OK);
}

/*
 Padtcl_bbox:

 pathName bbox ?-sticky? tagOrId ?tagOrId tagOrId ...?

 Returns a list with four elements giving an approximate
 bounding box for all the items named by the tagOrId
 arguments.  The list has the form "x1 y1 x2 y2" such
 that the drawn  areas of all the named elements are
 within the region bounded by x1 on the left, x2 on  the
 right, y1 on the bottom, and y2 on the top.
 If no items match any of the tagOrId arguments then an
 empty string is returned.

 If -sticky is specified:
   If the object is not sticky, return the bbox of the object in sticky coordinates,
      that is, the coords of the equivalent sticky object that would appear at
      the same location on the screen.
   If the object is sticky or is the pad surface,
      return the bbox independent of the current view.
 */

int
Padtcl_bbox(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, k;
    Pad_Bool first = TRUE;
    float bb[4], bbox[4];
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator oi;
    Pad_TkWin *tkwin;
    Pad_View *view;
    Pad_Bool sticky;

    tkwin = (Pad_TkWin *)clientData;

    if (argc <= 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName bbox ?-sticky? tagOrId ?tagOrId tagOrId ...?\"");
      }

    if (!strcmp(argv[1], "-sticky")) {
	sticky = TRUE;
	argc--;
	argv++;
    } else {
	sticky = FALSE;
    }

    for (i=1; i<argc; i++)
      {
	  tkwin->view->pad->Find_with_tagorid(argv[i], objs, TRUE);
	  DOLIST(oi, objs, Pad_Object, obj)
	    {
		obj->Get_global_bbox(bbox);
		if (!obj->Get_sticky() && sticky) {
		    view = tkwin->view;
		    bbox[0] = (bbox[0] - view->xview) * view->zoom;
		    bbox[1] = (bbox[1] - view->yview) * view->zoom;
		    bbox[2] = (bbox[2] - view->xview) * view->zoom;
		    bbox[3] = (bbox[3] - view->yview) * view->zoom;
		} 
		if (first)
		  {
		      first = FALSE;
		      for (k=0; k<4; k++) {
			  bb[k] = bbox[k];
		      }
		  } else {
		      if (bbox[XMIN] < bb[XMIN]) bb[XMIN] = bbox[XMIN];
		      if (bbox[XMAX] > bb[XMAX]) bb[XMAX] = bbox[XMAX];
		      if (bbox[YMIN] < bb[YMIN]) bb[YMIN] = bbox[YMIN];
		      if (bbox[YMAX] > bb[YMAX]) bb[YMAX] = bbox[YMAX];
		  }
	    }
      }

    if (first)			// No objects found
      {
	  Tcl_ResetResult(interp);
      }
    else
      {
	  sprintf(Pad_result, "%f %f %f %f",
		  XTOUNITS(tkwin, bb[0]),
		  YTOUNITS(tkwin, bb[1]),
		  XTOUNITS(tkwin, bb[2]),
		  YTOUNITS(tkwin, bb[3]));
	  TCL_RESULT(interp, Pad_result);
      }

    return(TCL_OK);
}

/*
 Padtcl_delete:

 pathName delete tagOrId ?tagOrId tagOrId ...?

 Deletes each item specified by tagOrId's
 */

int
Padtcl_delete(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, j;
    Pad *pad;
    Pad_IList ids;

    pad = ((Pad_TkWin *)clientData)->view->pad;

    if (argc <= 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName delete ?tagOrId tagOrId ...?\"");
      }

    for (i=1; i<argc; i++)
      {
				// Delete by id in case any objects are deleted by
				// the act of deleting others.
	  pad->Find_ids_with_tagorid(argv[i], ids, TRUE);
	  for (j=0; j<ids.Length(); j++) {
	      pad->Delete_obj(ids.Nth(j));
	  }
      }

    Tcl_ResetResult(interp);
    return(TCL_OK);
}


/*
 Padtcl_type:

 pathName type tagOrId

 Returns the type of the specified item.  If tagOrId specifies more
 than one item, then this returns the type of the first item.
 */

int
Padtcl_type(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    Pad_Object *obj;
    Pad_List objs;
    Pad_TkWin *tkwin;
    Pad_String *userType;

    tkwin = (Pad_TkWin *)clientData;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName type tagOrId\"");
      }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj)
      {
	  userType = obj->Get_usertype();
	  if (userType) {
	      sprintf(Pad_result, "%s", userType->Get());
	  } else {
	      sprintf(Pad_result, "%s", obj->Type_name());
	  }
	  TCL_RESULT(interp, Pad_result);
	
	  for (i=0; i<(int)strlen(interp->result); i++)
	    interp->result[i] = tolower(interp->result[i]);
      }

    return(TCL_OK);
}

/*
 Padtcl_find:

 pathName find ?-groupmembers? searchCommand ?arg arg ...?

 This command returns a list consisting of all the items
 that meet the constraints specified by searchCommand
 and arg's. The objects are returned in display list order,
 and if -groupmembers is specified, then group members are
 returned, otherwise, they are not.

 SearchCommand may take any of these forms:

    all
      Selects all the items in the pad.

    above tagorid
      Selects the item just after the items specified by tagorid in the display list

    below tagorid
      Selects the item just before the items specified by tagorid in the display list

    closest x y ?halo ?start??
      Selects the item closest to the point (x, y).  If halo is specified, it finds
      the closest item within halo of the point (x, y).  If start is specified, it
      finds the closest item below start in the display list.  If no such item exists,
      the selection behaves as if start wasn't specified.

    withinfo info
      Selects all the items containing <info> in their
      info itemconfigure option.

    withlayer layer
      Selects all the items on the layer <layer>.

    withname name
      Selects all the items having <name>.

    withtag tagOrId
      Selects all the items given by <tagOrId>.

    withtext text
      Selects all the items containing <text>.

    withtype type
      Selects all the items of type <type>.

    withsticky stickyType
      Selects all sticky items of sticky type <stickyType>.
      If <stickyType> is "all", then all sticky items are returned.

    enclosed x1 y1 x2 y2
      Selects all the items completely enclosed within
      the  rectangular region given by x1, y1, x2, and y2.
      X1 must be no greater then x2 and y1 must be no
      greater than y2.

    overlapping x1 y1 x2 y2
      Selects all the items that overlap or are enclosed
      within the rectangular region given by x1, y1, x2,
      and y2.  X1 must be no greater thenx2 and y1 must
      be no greater than y2.
*/
int
Padtcl_find(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin;
    Pad *pad;
    Pad_Bool groupMembers  = FALSE;
    Pad_Bool found         = FALSE;
    Pad_Bool error         = FALSE;
    int   matchType;                                           // what type of matching will be used 
                                                               // (glob, regexp,exact)
    Pad_List objs;
    Pad_Iterator oi;
    Pad_Object *obj;
    Pad_View  *view = NULL;

    tkwin = (Pad_TkWin *)clientData;
    view = tkwin->view;
    pad = view->pad;

    //_Pad_Print_argv(argc, argv); cerr << "=====" << endl;
    argc--;                                                   // get past "find"
    argv++;
    groupMembers = FALSE;
    matchType = -1;

    if ((argc >= 1) && !strcmp(argv[0], "-groupmembers")) {
        groupMembers = TRUE;
        argc--;
        argv++;
    }
    if ( argc >= 1) {
        do {
            found = FALSE;
            if (!strcmp(argv[0], "-groupmembers")) {
                groupMembers = TRUE;
                found = TRUE;
            } else if (!strcmp(argv[0], "-regexp")) {
	        if (matchType != -1) {
                    error=TRUE;
	        }
                matchType = PAD_FIND_REGEXP;
                found = TRUE;
            } else if (!strcmp(argv[0], "-glob")) {
	        if (matchType != -1) {
                    error=TRUE;
	        }
                matchType = PAD_FIND_GLOB;
                found = TRUE;
            }
	    if (found) {
	        argc--;
	        argv++;
	    }
        } while ((argc >= 1) && found && !error);
    }
    if ((argc < 1) || error) {
	  TCL_ERR(interp,
		  "wrong # args: should be \"pathName find ?-groupmembers? ?-|glob|regexp? [searchCommand ?arg arg ...?] ?[&& | ||] [search Command ?arg arg ...?]? ...\"");
    }

    if (!pad->Find_eval(objs, interp, clientData, groupMembers, matchType, argc, argv)) {
        TCL_ERR(interp, Pad_errorString.Get());
    }

    Tcl_ResetResult(interp);
    DOLIST(oi, objs, Pad_Object, obj) {
	sprintf(Pad_result, "%d", obj->id);
	Tcl_AppendElement(interp, Pad_result);
    }
    return(TCL_OK);
}


/*
 Padtcl_slide:

 //pathName slide tagOrId ?dx dy ?animationTime??
 Slide each of the items given by tagOrId in the pad
 coordinate space by adding dx to the x-coordinate
 of each point associated ith the item and dy to
 the y-coordinate of each point associated with the
 item.  

 This command returns the resulting position.
 */
int
Padtcl_slide(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    double dx, dy;
    int animationTime = 0;
    Pad_List objs;
    Pad_Bool groups;
    Pad_TkWin *tkwin;

    if ((argc != 2) && (argc != 4) && (argc != 5))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName slide tagOrId ?xAmount yAmount ?animationTime??");
      }

    tkwin = (Pad_TkWin *)clientData;

				// If scaling everything, then don't want to modify group members
				// since they will get scaled by their groups
    groups = TRUE;
    if (!strcmp(argv[1], "all")) {
	groups = FALSE;
    }
    tkwin->view->pad->Find_with_tagorid(argv[1], objs, groups);

    if (argc >= 4)
      {
	  dx = ATOXF(tkwin, argv[2]);
	  dy = ATOYF(tkwin, argv[3]);
	  if (argc == 5) {
	      animationTime = atoi(argv[4]);
	  }
	
	  tkwin->view->pad->Slide(objs, dx, dy, TRUE, animationTime);
      }

    if (!objs.Is_empty()) {
	sprintf(Pad_result, "%f %f",
		XTOUNITS(tkwin, ((Pad_Object *)objs.First())->anchorpt.x),
		YTOUNITS(tkwin, ((Pad_Object *)objs.First())->anchorpt.y));
	TCL_RESULT(interp, Pad_result);
    }
    return(TCL_OK);
}

/*
 Padtcl_scale:

 pathName scale tagOrId ?scaleAmount ?centerX centerY? animationTime?

 Scale each of the items given by tagOrId in the pad
 coordinate space by multiplying scaleAmount with the size
 of the item.  This command returns the scale of the first object.
 Scale around the specified point in pad space (or around the object's
 center if centerX and centerY are not specified).
 */

int
Padtcl_scale(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int animationTime;
    float ctrptx, ctrpty;
    double zmult;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator oi;
    Pad_Bool groups;
    Pad_TkWin *tkwin;

    if ((argc < 2) || (argc > 6))
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName scale tagOrId ?scaleAmount ?centerX centerY? animationTime?");
      }

    tkwin = (Pad_TkWin *)clientData;

				// If scaling everything, then don't want to modify group members
				// since they will get scaled by their groups
    animationTime = 0;
    groups = TRUE;
    if (!strcmp(argv[1], "all")) {
	groups = FALSE;
    }
    tkwin->view->pad->Find_with_tagorid(argv[1], objs, groups);

    if (argc >= 3)
      {
	  if (Tcl_GetDouble(interp, argv[2], &zmult) == TCL_ERROR)
	    return(TCL_ERROR);

	  if (zmult <= 0) {
	      TCL_ERR(interp, "scale must be > zero");
	  }
	
	  if (argc == 6) {
	      ctrptx = ATOXF(tkwin, argv[3]);
	      ctrpty = ATOYF(tkwin, argv[4]);
	      animationTime = atoi(argv[5]);
	      tkwin->view->pad->Scale(objs, ctrptx, ctrpty, zmult, TRUE, animationTime);
	  } else {
	      DOLIST(oi, objs, Pad_Object, obj) {
		  if (obj->Type() != PAD_PAD) {
		      if (argc >= 5) {                // Zoom around specified point
		          ctrptx = ATOXF(tkwin, argv[3]);
			  ctrpty = ATOYF(tkwin, argv[4]);
		      } else {
				// Zoom around anchor point
		          ctrptx = obj->anchorpt.x;
			  ctrpty = obj->anchorpt.y;
		      }
		      obj->Scale(ctrptx, ctrpty, zmult, TRUE);
		  }
	      }
	  }
      }

    if (!objs.Is_empty()) {
	sprintf(Pad_result, "%f", ((Pad_Object *)objs.First())->transform.Get_scale());
	TCL_RESULT(interp, Pad_result);
    }
    return(TCL_OK);
}

//
// Padtcl_rotate 
//
//
int
Padtcl_rotate(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_Point ctrpt;
    double theta;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator oi;
    Pad_Bool groups;
    Pad_TkWin *tkwin;
    int len;
    Pad_Bool didrot;

    if ((argc < 3) || (argc > 6))
	{
	    TCL_ERR(interp, 
		    "wrong # args: should be \"pathName rotate tagOrId ?angle(deg) ?centerX centerY??");
	}

    tkwin = (Pad_TkWin *)clientData;

    len = strlen(argv[1]);
    
    // If rotating everything, then don't want to modify group members
    // since they will get rotated by their groups
    groups = TRUE;
    if (!strcmp(argv[1], "all")) {
	groups = FALSE;
    }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, groups);

    if (argc >= 3)
	{
	    if (Tcl_GetDouble(interp, argv[2], &theta) == TCL_ERROR)
		return(TCL_ERROR);
	
	    DOLIST(oi, objs, Pad_Object, obj)


		{
		    if (obj->Type() != PAD_PAD) {
			if (argc==5) {      // Rotate around specified point in group coords
			    ctrpt.x = ATOXF(tkwin, argv[3]);
			    ctrpt.y = ATOYF(tkwin, argv[4]);
			    didrot = obj->Rotate(theta, ctrpt);
			    if (!didrot) {
				TCL_RESULT(interp, Pad_errorString.Get());
				return(TCL_ERROR);
			    }
		
			} 
			if (argc==3) {            // Rotate around object's anchor point
			    didrot = obj->Rotate(theta);
			    if (!didrot) {
				TCL_RESULT(interp, Pad_errorString.Get());
				return(TCL_ERROR);
			    }

			}      
		
		    }
		}

	  
	}
    return(TCL_OK);
}


/*
 Padtcl_pick:

 pathName pick ?-divisible? ?-indivisible? x y

 Return the item id that is picked by the point at X coordinates (x, y).  
 If -divisible is specified, then treat all groups as divisible.  
 If -indivisible is specified, then treat all groups as indivisible.  
 Otherwise, treat groups according to their own specifications.
 */

int
Padtcl_pick(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int x, y;
    int divisible;
    Pad_Object *obj;
    Pad_TkWin *tkwin;
    Pad_Event *event;
    Pad_Bool found;

    if (argc < 3)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName pick ?-divisible? ?-indivisible? x y");
      }

    tkwin = (Pad_TkWin *)clientData;

				// Check for flags in any order
    divisible = AUTOMATIC;
    found = TRUE;
    while ((argc >= 2) && found) {
	found = FALSE;
	if (!strcmp(argv[1], "-divisible")) {
	    divisible = TRUE;
	    found = TRUE;
	}
	if (!strcmp(argv[1], "-indivisible")) {
	    divisible = FALSE;
	    found = TRUE;
	}
	if (found) {
	    argc--;
	    argv++;
	}
    }

    x = atoi(argv[1]);
    y = atoi(argv[2]);
    event = new Pad_Event(tkwin, x, y);
    event->Set_divisible(divisible);

    obj = tkwin->view->pad->Find_pick(event);
    delete event;

    if (obj->id) {
	sprintf(Pad_result, "%d", obj->id);
	TCL_RESULT(interp, Pad_result);
    }
    return(TCL_OK);
}

static Pad_FList SplinePoints;

static void
Pad_AddSplinePoint(float *pts)
{
    int i;

    if (SplinePoints.Is_empty()) {
	SplinePoints.Push_last(pts[0]);
	SplinePoints.Push_last(pts[1]);
    }
    for (i=2; i<8; i++) {
	SplinePoints.Push_last(pts[i]);
    }
}

static void
Pad_AddLinePoint(float *pt)
{
    SplinePoints.Push_last(pt[0]);    
    SplinePoints.Push_last(pt[1]);
}

int
Padtcl_spline2line(ClientData clientData, Tcl_Interp *interp, int argc, char **argv){
    int i;
    float points[8];
    float x, y;
    float error;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc < 8)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName spline2line error x0 y0 ... xn yn");
      }

    argc--;
    argv++;

    error = atof(argv[0]);

    argc--;
    argv++;

    if (argc%2 != 0) {
	TCL_ERR(interp, "line2spline: must have even number of points");
    }

    SplinePoints.Make_empty();

    while (argc >= 8) {
	for (i = 0; i < 8; i += 2) {
	    points[i]   = ATOXF(tkwin, argv[i]);
	    points[i+1] = ATOYF(tkwin, argv[i+1]);
	}

	if (SplinePoints.Is_empty()) {
	    Pad_AddLinePoint(points); // add the first point
	}

	Pad_Bezier(points[0], points[1],
		   points[2], points[3],
		   points[4], points[5],
		   points[6], points[7],
		   error,
		   Pad_AddLinePoint);
	argc -= 6;
	argv += 6;
    }

    if (argc >= 6) {
	for (i = 0; i < 6; i += 2) {
	    points[i]   = ATOXF(tkwin, argv[i]);
	    points[i+1] = ATOYF(tkwin, argv[i+1]);
	}

	if (SplinePoints.Is_empty()) {
	    Pad_AddLinePoint(points); // add the first point
	}

	Pad_Bezier(points[0], points[1],
		   points[2], points[3],
		   points[4], points[5],
		   (float)SplinePoints.Nth(0), 
		   (float)SplinePoints.Nth(1),
		   error,
		   Pad_AddLinePoint);
	
    }

    for (i = 0; i < (SplinePoints.Length()); i+= 2) {
	x = (float)SplinePoints.Nth(i);
	y = (float)SplinePoints.Nth(i+1);
	if (i != 0) {
	    Tcl_AppendResult(interp, " ", NULL);
	}
	sprintf(Pad_result, "%f %f", XTOUNITS(tkwin, x), YTOUNITS(tkwin, y));
	Tcl_AppendResult(interp, Pad_result, NULL);
    }

    return(TCL_OK);
}

int
Padtcl_line2spline(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    float x, y;
    float *points;
    float smoothness;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (argc < 6)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName line2spline error x0 y0 ... xn yn");
      }

    argc--;
    argv++;

    smoothness = atof(argv[0]);

    argc--;
    argv++;

    if (argc%2 != 0) {
	TCL_ERR(interp, "line2spline: must have even number of points");
    }

    points = new float[argc];
    for (i=0; i<argc; i += 2) {
	points[i] =   ATOXF(tkwin, argv[i]);
	points[i+1] = ATOYF(tkwin, argv[i+1]);
    }

    SplinePoints.Make_empty();
    Pad_FitBezier(points, argc / 2, smoothness, Pad_AddSplinePoint);
    delete [] points;

    for (i = 0; i < (SplinePoints.Length()); i+= 2) {
	x = (float)SplinePoints.Nth(i);
	y = (float)SplinePoints.Nth(i+1);
	if (i != 0) {
	    Tcl_AppendResult(interp, " ", NULL);
	}
	sprintf(Pad_result, "%f %f", XTOUNITS(tkwin, x), YTOUNITS(tkwin, y));
	Tcl_AppendResult(interp, Pad_result, NULL);
    }

    return(TCL_OK);
}

/*
 Padtcl_coords:

 pathName coords ?-objectcoords? ?-append? ?-nooutput? tagOrId ?x0 y0 ...?

 Query or modify the coordinates that define an item.  If no
 coordinates are specified, this command returns a list whose elements
 are the coordinates of the item named by tagOrId in global coordinate
 space.

 If -objectcoords is specified, then the coordinates are
 specified and returned in object coordinate space.  This means that the original
 list of specified points are modified by the current object transformation,
 rather then using them as absolutes and resetting the transformation.
 Also, the coordinates  are returned (not including any translationg
 or scale) rather than their actual position on the Pad.

 If -append is specified, then the coordinates are added on to the existing
 coordinates rather than replacing all of them. If -nooutput is specified,
 then this command returns an empty string.
					
 If coordinates are specified, then they replace the current
 coordinates for the named item.  If tagOrId refers to multiple items,
 then the first one in the display list is used.
*/

int
Padtcl_coords(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i, type;
    Pad_Bool objectCoords;
    Pad_Bool append;
    Pad_Bool output;
    Pad_Bool found;
    Pad_Object *obj;
    Pad_TkWin *tkwin;
    int length;
    Pad_PList pts;
    Pad_Point pt;
    Pad_Bool rc;
    Pad_List objs;

    if (argc < 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName coords ?-objectcoords? ?-append? ?-nooutput? tagOrId ?x0 y0 ...?");
      }

    tkwin = (Pad_TkWin *)clientData;

				// Check for flags in any order
    objectCoords = FALSE;
    output = TRUE;
    append = FALSE;

    do {
	found = FALSE;
	if (!strcmp(argv[1], "-objectcoords")) {
	    objectCoords = TRUE;
	    found = TRUE;
	}
	if (!strcmp(argv[1], "-nooutput")) {
	    output = FALSE;
	    found = TRUE;
	}
	if (!strcmp(argv[1], "-append")) {
	    append = TRUE;
	    found = TRUE;
	}
	if (found) {
	    argc--;
	    argv++;
	}
    } while ((argc >= 2) && found);

				// Check # of args again after flags have been parsed.
    if (argc < 2) {
	  TCL_ERR(interp, "wrong # args: should be \"pathName coords ?-objectcoords? ?-append? ?-nooutput? tagOrId ?x0 y0 ...?");
    }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj)
      {
	  type = obj->Type();
	  if ((type != PAD_PORTAL) &&
	      (type != PAD_LINE) &&
	      (type != PAD_SPLINE) &&
	      (type != PAD_RECTANGLE) &&
	      (type != PAD_OVAL) &&
	      (type != PAD_POLYGON)) 
	    {
		TCL_ERR(interp, "Coords command only applies to line, spline, rectangle, polygon, oval, and portal items");
	    }

				// Set new points
	  if (argc > 2) {
	      argc -= 2;
	      argv += 2;

	      if ((argc%2) != 0) {
		  TCL_ERR(interp, "Need an even number of points");
	      }

	      if (type == PAD_SPLINE) {
		  if ((argc%6 != 0) && (argc%6 != 2)) {
		      TCL_ERR(interp, "Need 3n or 3n+1 points");
		  }
	      }

	      length = argc / 2;
	      
	      DOTIMES(i, length) {
		  pt.Set(ATOXF(tkwin, argv[2*i]), ATOYF(tkwin, argv[2*i + 1]));
		  pts.Push_last(&pt);
	      }
	      if (append) {
		  rc = obj->Append_coords(pts, objectCoords);
	      } else {
		  rc = obj->Set_coords(pts, objectCoords);
	      }

	      if (!rc) {
		  TCL_ERR(interp, "Invalid coordinates");
	      }
	  }

				// Return points
	  Tcl_ResetResult(interp);
	  if (output) {
	      obj->Get_coords(pts, objectCoords);
	      DOTIMES(i, pts.Length()) {
		  if (i > 0) {
		      Tcl_AppendResult(interp, " ", NULL);
		  }
		  sprintf(Pad_result, "%f %f", XTOUNITS(tkwin, pts.Nth(i)->x), YTOUNITS(tkwin, pts.Nth(i)->y));
		  Tcl_AppendResult(interp, Pad_result, NULL);
	      }
	  }
      }
    return(TCL_OK);
}
		
/*
 Padtcl_getdate:

 pathName getdate

 Returns the current time in "hh:mm:ss" form
*/

int
Padtcl_getdate(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{
    Pad_TkWin *tkwin;

    tkwin = (Pad_TkWin *)clientData;

    if (argc != 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName getdate\"");
      }

    time_t t;
    char *c;
    t = time(NULL);
    c = asctime(localtime(&t));
    sprintf(tkwin->interp->result, "%s", c);

    return(TCL_OK);
}

/*
 Padtcl_gettags:

 pathName gettags tagOrId

 Return a list whose elements are the tags associated
 with the item given by tagOrId.  If tagOrId refers to
 more than one item, then the tags are returned from the
 first such item in the display list.  If tagOrId
 doesn't refer to any items, or if the item contains no
 tags, then an empty string is returned.
 */

int
Padtcl_gettags(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator ti;
    Tk_Uid tag;

    tkwin = (Pad_TkWin *)clientData;

    if (argc != 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName gettags tagOrId\"");
      }

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj)
      {
	  Tcl_ResetResult(interp);
	  DOLISTI(ti, obj->tags, Tk_Uid, tag) {
	      Tcl_AppendElement(interp, tag);
	  }
      }
    else
      {
	  TCL_RESULT(interp, "");
      }

    return(TCL_OK);
}

/*
 Padtcl_hastag:

 pathName hastag tagToCheck tagOrId

 Return "1" if the item specified by tagOrId contains
 the tagToCheck, or "0" otherwise.  If tagOrId refers to
 more than one item, then the comparison is performed on the
 first such item in the display list.  If tagOrId
 doesn't refer to any items, then "0" is returned.
*/

int
Padtcl_hastag(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator ti;
    Tk_Uid tag, user_tag;
    Pad_Bool found;

    tkwin = (Pad_TkWin *)clientData;

    if (argc != 3)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName hastag tagToCheck tagOrId\"");
      }

    tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
    obj = (Pad_Object *)objs.First();
    if (obj)
      {
	  Tcl_ResetResult(interp);
	  user_tag = Tk_GetUid(argv[1]);
	  found = FALSE;
	  DOLISTI(ti, obj->tags, Tk_Uid, tag) {
	      if (tag == user_tag) {
		  found = TRUE;
		  break;
	      }
	  }
	  if (found) {
	      TCL_RESULT(interp, "1");
	  } else {
	      TCL_RESULT(interp, "0");
	  }
      }
    else 
      {
	  TCL_RESULT(interp, "0");
      }

    return(TCL_OK);
}

/*
 Padtcl_dtag:

 pathName deletetag tagToDelete tagOrId ?tagOrId ...?

 Remove <tagToAdd> from every object specified by <tagOrId>'s
 This command returns an empty string as result.
 */

int
Padtcl_dtag(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    Pad_TkWin *tkwin;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator oi;

    tkwin = (Pad_TkWin *)clientData;

    if (argc < 3)
      {
	  TCL_ERR(interp,
		  "wrong # args: should be \"pathName deletetag tagToDelete tagOrId ?tagOrId ...?\"");
      }

    for (i=2; i<argc; i++)
      {
	  tkwin->view->pad->Find_with_tagorid(argv[i], objs, TRUE);

	  DOLIST(oi, objs, Pad_Object, obj) {
	      obj->Delete_tag(argv[1]);
	  }
      }
    Tcl_ResetResult(interp);

    return(TCL_OK);
}

/*
 Padtcl_addtag:

 pathName addtag tagToAdd tagOrId ?tagOrId ...?

 Add <tagToAdd> to every object specified by <tagOrId>'s
 This command returns an empty string as result.
 */

int
Padtcl_addtag(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    Pad_TkWin *tkwin;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator oi;

    tkwin = (Pad_TkWin *)clientData;

    if (argc < 3)
      {
	  TCL_ERR(interp,
		  "wrong # args: should be \"pathName addtag tagToAdd tagOrId ?tagOrId ...?\"");
      }

    for (i=2; i<argc; i++)
      {
	  tkwin->view->pad->Find_with_tagorid(argv[i], objs, TRUE);

	  DOLIST(oi, objs, Pad_Object, obj) {
	      obj->Add_tag(argv[1]);
	  }
      }
    Tcl_ResetResult(interp);

    return(TCL_OK);
}

/*
 Padtcl_windowshape:

 pathName windowshape [{innter-coords} {outer-coords}]

 Assign a shape to the top-level window holding a Pad
 */

int
Padtcl_windowshape(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
#ifdef PAD_WIN
    TCL_ERR(interp, "windowshape not implemented on Windows");
#endif

#ifdef PAD_UNIX
    Pad_TkWin *tkwin;

    extern void Pad_ComputeShapeCoords(Pad_List &); // See x-win-shape.C
    extern void Pad_Reshape(Pad_Win *, Pad_Bool);      // See win.C

    tkwin = (Pad_TkWin *)clientData;

    if (argc >= 2) {

	// Free existing points 
	Pad_Iterator li;
	Pad_Point *p;
	DOLIST(li, tkwin->shapePoints, Pad_Point, p) {
	    delete p;
	}
	tkwin->shapePoints.Make_empty();
	tkwin->ninner = tkwin->nouter = 0;

	// Assign new points
	int i;
	int listrc;
	int listargc;
	char **listargv;

	// Do inner points
	listrc = Tcl_SplitList(interp, argv[1], &listargc, &listargv);
	if (listargc >= 6) {
	    int npts = listargc & ~1;
	    for (i=0; i < npts; i += 2) {
		Pad_Point *p = new Pad_Point;
		p->x = atof(listargv[i]);
		p->y = atof(listargv[i + 1]);
		tkwin->shapePoints.Push_last((void*)p);
		tkwin->ninner++;
	    }
	}    
	if (listrc == TCL_OK) {
	  ckfree((char *) listargv);
	}

	if (argc >= 3) {
	    // Do outer points
	    listrc = Tcl_SplitList(interp, argv[2], &listargc, &listargv);
	    if (listargc >= 6) {
		int npts = listargc & ~1;
		for (i=0; i < npts; i += 2) {
		    Pad_Point *p = new Pad_Point;
		    p->x = atof(listargv[i]);
		    p->y = atof(listargv[i + 1]);
		    tkwin->shapePoints.Push_last((void*)p);
		    tkwin->nouter++;
		}
	    }
	    if (listrc == TCL_OK) {
	      ckfree((char *) listargv);
	    }    
	}    

	// This makes the coordinates map from 0..1, with Y inverted
	Pad_ComputeShapeCoords(tkwin->shapePoints);

	// This installs the shape on the Pad window
	Pad_Reshape(tkwin, TRUE);
    }

    // Now output a point data
    Pad_Point *p;
    Pad_String tmp;
    int i;

    tmp = "";
    for (i = 0; i < tkwin->ninner; i++) {
	if (i > 0) {
	    tmp += " ";
	}
	p = (Pad_Point*)tkwin->shapePoints.Nth(i);
	tmp += p->x;
	tmp += " ";
	tmp += 1 - p->y;
    }
    Tcl_AppendElement(interp, tmp.Get());

    tmp = "";
    for (i = 0; i < tkwin->nouter; i++) {
	if (i > 0) {
	    tmp += " ";
	}
	p = (Pad_Point*)tkwin->shapePoints.Nth(i + tkwin->ninner);
	tmp += p->x;
	tmp += " ";
	tmp += 1 - p->y;
    }
    Tcl_AppendElement(interp, tmp.Get());

    return(TCL_OK);
#endif
}


//////////////////////////////////////////////////////////////////////////////////////////
//											//
//  Padtcl_tree										//
//  dwhr 20jan96									//
//											//
//  Pad_TreeNode									//
//											//
//////////////////////////////////////////////////////////////////////////////////////////


// helper function for Padtcl_tree
// checks to see that an object is valid, and that it has a tree node
// associated with it
static int 
Is_treeobj (Pad_Object *obj, Tcl_Interp *interp) {
   if (!obj) {
       TCL_RESULT(interp, "invalid tagOrId");
       return(FALSE);
   }
   if (!obj->Get_treenode()) {
       sprintf(Pad_result, "invalid tagOrId - no Pad_TreeNode associated with Id %d", obj->id);
       TCL_RESULT(interp, Pad_result);
       return(FALSE);
   }
   return(TRUE);
}


//
// Padtcl_tree main
//
int 
Padtcl_tree(ClientData clientData, Tcl_Interp *interp, int argc, char **argv) 
{
    Pad_TkWin *  tkwin = (Pad_TkWin *)clientData;
    Pad_List     objs;
    Pad_Object * obj;
    Pad_Object * obj2;
    Pad_TreeNode *node;
    Pad_Bool      stat;
    Pad_Iterator  pi;
    Pad_TreeNode *treenode=NULL;
    
    if (argc < 2) {
       TCL_RESULT(interp, "wrong # args: should be \"pathName tree subcommand ?arg ...?\"");
       return(TCL_ERROR);
    }

				//  If there isn't a treeroot, create one
    if (!tkwin->view->pad->treeroot) {
	tkwin->view->pad->Create_treeroot();
    }

    //
    // add an existing node to an existing node in a parent/child relationship
    //
    if (!strcmp(argv[1], "addnode")) {
	if ((argc != 3) && (argc != 4)) {
	    TCL_RESULT(interp, "wrong # args: should be \"pathName tree addnode childTagOrId parentTagOrId \"");
	    return(TCL_ERROR);
	}

	tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
	obj = (Pad_Object *)objs.Pop();

	if (Is_treeobj(obj, interp)) {
	    treenode = obj->Get_treenode();
	    if (argc == 3) {
		tkwin->view->pad->treeroot->Add_node(obj);
	    } else {
		tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
		obj2 = (Pad_Object *)objs.Pop();
		if (Is_treeobj(obj2, interp)) {
		    if (!obj2->Get_treenode()->Add_node_safe(treenode)) {
			TCL_ERR(interp, Pad_errorString.Get());
		    }
		} else {
		    return(TCL_ERROR);
		}
	    }
	    return(TCL_OK);
	} else {
	    return(TCL_ERROR);
	}

    }

    //
    // animate a pre-computed layout
    // NOTE - must be used after "computelayout" command has been called,
    //        and before ANY movement or zooming of the objects in the hierarchy,
    //        or it will perform an inaccurate animation
    //
    if (!strcmp(argv[1], "animatelayout")) {
       float x, y, z;

       if ((argc != 3) && (argc != 5)) { 
          // .pad tree animatelayout tagOrId -view view
          //      0    1             2       3     4
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree animatelayout tagOrId ?-view view?\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (!Is_treeobj(obj, interp)) {
         return(TCL_ERROR); 
       }
       treenode = obj->Get_treenode();

       if (argc == 3) {
				// No view
	  treenode->Get_layout()->Layout_main();
	  return(TCL_OK);

       } else {
				// View requested
	   if (!strcmp(argv[3], "-view")) {
	       int listrc;
	       int listargc;
	       char **listargv;

	       listrc = Tcl_SplitList(interp, argv[4], &listargc, &listargv);
	       if (listrc == TCL_OK) {
		   if ((listargc != 2) && (listargc != 3)) {
		       TCL_ERR(interp, "-view needs list of 2 or 3 numbers");
		   } else {
		       x = ATOXF(tkwin, listargv[0]);
		       y = ATOYF(tkwin, listargv[1]);
		       if (listargc == 3) {
			   z = atof(listargv[2]);
		       } else {
			   z = treenode->Get_pad()->view->zoom;
		       }
		   }
		   ckfree((char *) listargv);
	       } else {
		   TCL_ERR(interp, "-view needs valid list of 2 or 3 numbers");
	       }
	       
	       treenode->Get_layout()->Set_final_view_state(x, y, z);
	       treenode->Get_layout()->Layout_main();
	       return(TCL_OK);
	       
	   } else {
	       TCL_RESULT(interp, "wrong subcommand for \"animatelayout\" - try \"-view\"");
	       return(TCL_ERROR);
	   }
       }
    }


    //
    // set the animateview flag at a node
    // controls the view animation on an hierachical layout
    //
    if (!strcmp(argv[1], "animateview")) {
       if ((argc != 3) || (argc != 4)) {
          // .pad tree animateview tagOrId 
          //      0    1       2       
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree animateview tagOrId ?value?\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (!Is_treeobj(obj, interp)) return(TCL_ERROR); 
       treenode = obj->Get_treenode();

       if (argc == 3) {
          treenode->Get_layout()->Set_animate_view(TRUE);
	  return(TCL_OK);
       } else {
	 if (!strcmp(argv[3], "0")) {
	   treenode->Get_layout()->Set_animate_view(FALSE);
	   return(TCL_OK);
	 } else { 
	   treenode->Get_layout()->Set_animate_view(TRUE);
	   return(TCL_OK);
         }
       }
    }



    //
    // connects a node to it's parent by drawing the link
    // which resides at a layout.
    //
    if (!strcmp(argv[1], "connect")) {
       if (argc != 3) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree connect tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();
       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   if (treenode->Get_layout()) {
	       treenode->Get_layout()->Connect_node();
	   } else {
	       TCL_RESULT(interp, "ERROR - no layout present at this node\n");
	       return(TCL_ERROR);
	   }
	   return(TCL_OK);
       } else {
	   return(TCL_ERROR);
       }
    }

    //
    // computes all information in a hierachy in preparation for the animatelayout command
    // NOTE: information computed by this function is valid only as long as objects
    //       within the hierarchy remain unmoved and unzoomed
    //
    if (!strcmp(argv[1], "computelayout")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree computelayout tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	  float bb[4];

	  treenode = obj->Get_treenode();
	  obj->Get_global_bbox(bb);
          treenode->Get_layout()->Calc_bboxes();
          treenode->Get_layout()->Set_final_layout_state(bb[0], (bb[1] + 0.5*(bb[3]-bb[1])));
	  return(TCL_OK);
       }
       return(TCL_ERROR);
    }

    // -------------------------------
    if (!strcmp(argv[1], "create")) {
       if (argc != 3) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree create tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (!obj) {
	   TCL_ERR(interp, "invalid tagOrId");
       }
       
       node = new Pad_TreeNode(obj, tkwin->view->pad->treeroot);

       return(TCL_OK);
    }


    // -------------------------------
    if (!strcmp(argv[1], "createroot")) {
       if (argc != 2) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree createroot\"");
	  return(TCL_ERROR);
       }

       node = new Pad_TreeNode(tkwin->view->pad->treeroot, tkwin->view->pad);

       sprintf(Pad_result, "%d", node->Get_obj()->id);
       TCL_RESULT(interp, Pad_result);

       return(TCL_OK);
    }


    //
    // delete the tree information and the objects for a node or a subtree
    //
    if (!strcmp(argv[1], "delete")) {
       if ((argc != 3) && (argc != 4)){
          // .pad tree delete ?-subtree? tagOrId 
          //      0    1     2          3
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree delete ?-subtree? tagOrId\""); 
	  return(TCL_ERROR);
       }

       if (argc == 3) {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();
          if (Is_treeobj(obj, interp)) {
	     treenode = obj->Get_treenode();
             treenode->Delete_node(TRUE);
             return(TCL_OK);
          } else {
	     return(TCL_ERROR);
          }
       } else if (!strcmp(argv[2], "-subtree")) {
          tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();
          if (Is_treeobj(obj, interp)) {
	     treenode = obj->Get_treenode();
             treenode->Delete_subtree(TRUE);
             return(TCL_OK);
          } else {
	     return(TCL_ERROR);
          }
       } else {
	  TCL_RESULT(interp, "wrong subcommand for delete - try \"-subtree\""); 
	  return(TCL_ERROR);
       }
    }




    // -------------------------------
    if (!strcmp(argv[1], "getchildren")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree getchildren tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	  treenode = obj->Get_treenode();
	  Tcl_ResetResult(interp);
	  DOLIST(pi, *(treenode->Get_children()), Pad_TreeNode, node) {
              sprintf(Pad_result, "%d ", node->Get_obj()->id);
              Tcl_AppendResult(interp, Pad_result, NULL);
	  }
          return(TCL_OK);
       } else {
         return(TCL_ERROR);
       }
    }



    // -------------------------------
    if (!strcmp(argv[1], "getfocus")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree getfocus tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   sprintf(Pad_result, "%f", treenode->Get_focus());
	   TCL_RESULT(interp, Pad_result);
	   return(TCL_OK);
       } else {
	   return(TCL_ERROR);
       }
    }
    
    // -------------------------------
    if (!strcmp(argv[1], "getlayoutbbox")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree getlayoutbbox tagOrId\"");
	  return(TCL_ERROR);
       }
       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
         float bb[4];
	 
	 treenode = obj->Get_treenode();
         treenode->Get_layout()->Get_bbox_final(bb);
         Tcl_ResetResult(interp);
         sprintf(Pad_result, "%f %f %f %f ", bb[0], bb[1], bb[2], bb[3]);
         Tcl_AppendResult(interp, Pad_result, NULL);
         return(TCL_OK);
       } else {
         return(TCL_ERROR);
       }
    }


    // -------------------------------
    if (!strcmp(argv[1], "getlink")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree getlink tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   if (treenode->Get_layout()->Get_link()) {
               sprintf(Pad_result, "%d", treenode->Get_layout()->Get_link()->id);
               TCL_RESULT(interp, Pad_result);
	   } else {
               sprintf(Pad_result, "-1"); 
               TCL_RESULT(interp, Pad_result);
	   }
       
	   return(TCL_OK);
       } else {
	   return(TCL_ERROR);
       }
    }

    // -------------------------------
    if (!strcmp(argv[1], "getparent")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree getparent tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	  treenode = obj->Get_treenode();
          if (treenode->Get_parent()) {
	     if (treenode->Get_parent()->Get_obj() != NULL)  {
               sprintf(Pad_result, "%d", treenode->Get_parent()->Get_obj()->id);
               TCL_RESULT(interp, Pad_result);
	     } else {
               sprintf(Pad_result, "%d", -1); 
               TCL_RESULT(interp, Pad_result);
             }
          } 

          return(TCL_OK);
       } else {
         return(TCL_ERROR);
       }
    }


    // -------------------------------
    if (!strcmp(argv[1], "getroot")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree getroot tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	  treenode = obj->Get_treenode();
	  
          if (treenode->Get_root()) {
             sprintf(Pad_result, "%d", treenode->Get_root()->Get_obj()->id);
             TCL_RESULT(interp, Pad_result);
          }

          return(TCL_OK);
       } else {
         return(TCL_ERROR);
       }
    }


    // -------------------------------
    if (!strcmp(argv[1], "isnode")) {
       if ((argc != 3)) { 
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree isnode tagOrId\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (obj) {
	   treenode = obj->Get_treenode();
	   if (treenode) {
	       TCL_RESULT(interp, "1"); 
	   } else {
	       TCL_RESULT(interp, "0"); 
	   }
       } else {
	   TCL_RESULT(interp, "0"); 
       }
       return(TCL_OK);
    }

    // -------------------------------
    // .pad tree layout tagOrId -view view
    //      0    1      2       3     4
    //      1    2      3       4     5
    if (!strcmp(argv[1], "layout")) {
       if ((argc < 3) || (argc == 4) || (argc == 5) || (argc > 7)) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree layout tagOrId ?-view view?\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   float x = 0.0;
	   float y = 0.0;
	   float z = treenode->Get_pad()->view->zoom;
	   
	   if (argc == 3) {
	     treenode->Layout();
	     return(TCL_OK);

           } else if (!strcmp(argv[3], "-view")) {
	       int listrc;
	       int listargc;
	       char **listargv;
	       
	       listrc = Tcl_SplitList(interp, argv[4], &listargc, &listargv);
	       if (listrc == TCL_OK) {
		   if ((listargc != 2) && (listargc != 3)) {
		       TCL_ERR(interp, "-view needs list of 2 or 3 numbers");
		   } else {
		       x = ATOXF(tkwin, listargv[0]);
		       y = ATOYF(tkwin, listargv[1]);
		       if (listargc == 3) {
			   z = atof(listargv[2]);
		       } else {
			   z = treenode->Get_pad()->view->zoom;
		       }
		   }
		   ckfree((char *) listargv);
	       } else {
		   TCL_ERR(interp, "-view needs valid list of 2 or 3 numbers");
	       }

	     float bb[4];
	     obj->Get_global_bbox(bb);
             treenode->Get_layout()->Calc_bboxes();
             treenode->Get_layout()->Set_final_layout_state(bb[0], (bb[1] + 0.5*(bb[3]-bb[1])));
	     treenode->Get_layout()->Set_final_view_state(x, y, z);
	     treenode->Get_layout()->Layout_main();

	     return(TCL_OK);

	   } else {
	     TCL_RESULT(interp, "wrong subcommand for \"layout\" - try \"-view\"");
	     return(TCL_ERROR);
	   }
       } else {
	   Tcl_ResetResult(interp);
	   sprintf(Pad_result, "ERROR - %s is not a member of a padTree", argv[2]); 
	   TCL_RESULT(interp, Pad_result);
	   return(TCL_ERROR);
       }
    }



    // -------------------------------
    // .pad tree lower tagOrId ?tagOrId? 
    //      0    1      2       3     
    if (!strcmp(argv[1], "lower")) {
       if ((argc != 3) && (argc != 4)) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree lower tagOrId ?tagOrId?\"");
	  return(TCL_ERROR);
       }

       if (argc == 3) {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();

          if (Is_treeobj(obj,interp)) {
	     treenode = obj->Get_treenode();
             treenode->Lower();
             return(TCL_OK);
          } else {
             return(TCL_ERROR);
          }
       } else {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj  = (Pad_Object *)objs.Pop();
          tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
          obj2 = (Pad_Object *)objs.Pop();

          if (Is_treeobj(obj,interp) && Is_treeobj(obj2,interp)) {
	     treenode = obj->Get_treenode();
             treenode->Lower(obj2->Get_treenode());
             return(TCL_OK);
          } else {
             return(TCL_ERROR);
          }
       }
    }



    // -------------------------------
    // .pad tree raise tagOrId ?tagOrId? 
    //      0    1      2       3     
    if (!strcmp(argv[1], "raise")) {
       if ((argc != 3) && (argc != 4)) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree raise tagOrId ?tagOrId?\"");
	  return(TCL_ERROR);
       }

       if (argc == 3) {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();

          if (Is_treeobj(obj,interp)) {
	     treenode = obj->Get_treenode();
             treenode->Raise();
             return(TCL_OK);
          } else {
             return(TCL_ERROR);
          }

       } else {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj  = (Pad_Object *)objs.Pop();
          tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
          obj2 = (Pad_Object *)objs.Pop();

          if (Is_treeobj(obj,interp) && Is_treeobj(obj2,interp)) {
	     treenode = obj->Get_treenode();
             treenode->Raise(obj2->Get_treenode());
             return(TCL_OK);
          } else {
             return(TCL_ERROR);
          }
       } 
    }


    //
    // remove the tree information objects associated with a node
    // and promote the children in the hierarchy
    //
    if (!strcmp(argv[1], "removenode")) {
       if (argc != 3) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree removenode ?-subtree? tagOrId\""); 
	  return(TCL_ERROR);
       }

       if (argc == 3) {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();
          if (Is_treeobj(obj, interp)) {
	     treenode = obj->Get_treenode();
             treenode->Delete_node(FALSE);
             return(TCL_OK);
          } else {
	     return(TCL_ERROR);
          }
       } else if (!strcmp(argv[2], "-subtree")) {
          tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();
          if (Is_treeobj(obj, interp)) {
	     treenode = obj->Get_treenode();
             treenode->Delete_subtree(FALSE);
             return(TCL_OK);
          } else {
	     return(TCL_ERROR);
          }
       } else {
	  TCL_RESULT(interp, "wrong subcommand for delete - try \"-subtree\""); 
	  return(TCL_ERROR);
       }
    }


    //
    // reparent a node or a subtree in the hierarchy
    //
    if (!strcmp(argv[1], "reparent")) {
       if ((argc != 4) && (argc != 5)) {
          // .pad tree reparent ?-subtree? tagOrId tagOrId
          //      0    1        2          3       4 
	  Tcl_AppendResult(interp, "wrong # args: should be \"pathName tree reparent ",
			           "?-subtree? nodeTagOrId newparentTagOrId\"", NULL);
	  return(TCL_ERROR);
       }

       if ((argc == 4) && (!strcmp(argv[2], "-subtree"))) {
          Tcl_AppendResult(interp, "wrong # args: should be \"pathName tree reparent ",
			           "?-subtree? nodeTagOrId newparentTagOrId\"", NULL);
	  return(TCL_ERROR);
       } else if ((argc == 5) && (!strcmp(argv[2], "-subtree"))) {
          tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();
          tkwin->view->pad->Find_with_tagorid(argv[4], objs, TRUE);
          obj2 = (Pad_Object *)objs.Pop();

          if ((Is_treeobj(obj, interp)) && (Is_treeobj(obj2, interp))) { 
	     treenode = obj->Get_treenode();
             stat = treenode->Reparent_subtree(obj2->Get_treenode()); 
             return(TCL_OK);
          }
       } else {
          tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
          obj = (Pad_Object *)objs.Pop();
          tkwin->view->pad->Find_with_tagorid(argv[3], objs, TRUE);
          obj2 = (Pad_Object *)objs.Pop();
          if ((Is_treeobj(obj, interp)) && (Is_treeobj(obj2, interp))) { 
	      treenode = obj->Get_treenode();
              stat = treenode->Reparent_subtree(obj2->Get_treenode()); 
              return(TCL_OK);
          }
       } 
    }


    // -------------------------------
    if (!strcmp(argv[1], "setanimatespeed")) {
       if (argc != 4) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree setanimatespeed tagOrId value\"");
	  return(TCL_ERROR);
       }

       Pad_Bool rc;

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   rc = treenode->Get_layout()->Set_animation_speed(atoi(argv[3]));
	   if (!rc) {
	       TCL_RESULT(interp, Pad_errorString.Get());
	       return(TCL_ERROR);
	   }
       }

       return(TCL_OK);
    }


    // -------------------------------
    if (!strcmp(argv[1], "setfocus")) {
	if ((argc != 3) && (argc != 4) && (argc != 5) && (argc != 6)) {
	    TCL_RESULT(interp, "wrong # args: should be \"pathName tree setfocus tagOrId ?value ?levels ?falloff???\"");
	    return(TCL_ERROR);
	}
	
	float f;
	float falloff;
	int levels;

	if (argc >= 4) {
	    f = atof(argv[3]);
	} else {
	    f = 1.0;
	}

	if (argc >= 5) {
	    levels = atoi(argv[4]);
	} else {
	    levels = 0;
	}

	if (argc >= 6) {
	    falloff = atof(argv[5]);
	}

	if ((f < 0.0) || (f > 1.0)) {
	    TCL_RESULT(interp, "\nERROR - focus value must be [0.0,1.0]\n");
	    return(TCL_ERROR);
	}
	
	tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
	obj = (Pad_Object *)objs.Pop();
	if (Is_treeobj(obj, interp)) {
	    treenode = obj->Get_treenode();
	    if (argc >= 6) {
		treenode->Propogate_focus(f, levels, falloff);
	    } else {
		treenode->Propogate_focus(f, levels);
	    }
	    return(TCL_OK);
	} else {
	    return(TCL_ERROR);
	}
    }
    
    // -------------------------------
    if (!strcmp(argv[1], "setfocusmag")) {
       if (argc != 4) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree setfocusmag tagOrId value\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   treenode->Set_focus_mag(atof(argv[3]));
       }

       return(TCL_OK);
    }

    // -------------------------------
    if (!strcmp(argv[1], "setscale")) {
       if (argc != 4) {
	  TCL_RESULT(interp, "wrong # args: should be \"pathName tree setscale tagOrId value\"");
	  return(TCL_ERROR);
       }

       tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
       obj = (Pad_Object *)objs.Pop();

       if (Is_treeobj(obj, interp)) {
	   treenode = obj->Get_treenode();
	   treenode->Set_treenode_scale(atof(argv[3]));
       }

       return(TCL_OK);
    }

    
    // -------------------------------
    // .pad tree setspacing tagOrId x ?y?
    //      0    1             2       3  4
    
    if (!strcmp(argv[1], "setspacing")) {
	if ((argc != 4) && (argc != 5)) {
	    TCL_RESULT(interp, "wrong # args: should be \"pathName tree setspacing tagOrId xvalue ?yvalue?\"");
	    return(TCL_ERROR);
	}
	
	float x, y;

	x = ATOXF(tkwin, argv[3]);
	if (argc == 5) {
	    y = ATOYF(tkwin, argv[4]);
	} else {
	    y = x;
	}
	
	if ((x < 0) || (y < 0)) {
	    TCL_RESULT(interp, "\nERROR - spacing value must be non-negative"); 
	    return(TCL_ERROR);
	}
	
	tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
	obj = (Pad_Object *)objs.Pop();
        if (Is_treeobj(obj, interp)) {
	    treenode = obj->Get_treenode();
	    treenode->Get_layout()->Set_dist(x, y);
	    return(TCL_OK);
        } else {
	    return(TCL_ERROR);
        }
    }

    if (!strcmp(argv[1], "setlinkmode")) {
	if ((argc != 4) && (argc != 3))  {
	    TCL_RESULT(interp, "wrong # args: should be \"pathName tree setlinkmode tagOrId [fixed | scaling]\"");
	    return(TCL_ERROR);
	}

	tkwin->view->pad->Find_with_tagorid(argv[2], objs, TRUE);
	obj = (Pad_Object *)objs.Pop();
        if ( !Is_treeobj(obj, interp)) {
	  return(TCL_ERROR);
	}

	Pad_Bool isfixed;
	
	if (argc == 4) {
	  isfixed = (strcmp(argv[3],"fixed")) ? FALSE : TRUE;

	  treenode = obj->Get_treenode();
	  treenode->Get_layout()->Set_link_mode( isfixed );
	  return(TCL_OK);
	}
	if (argc == 3) {
	  treenode = obj->Get_treenode();
	  isfixed = (treenode->Get_layout()->Get_link_mode() );

	  sprintf(Pad_result, "%d", (isfixed == TRUE) ? 1 : 0 );
	  TCL_RESULT(interp, Pad_result);
	  return(TCL_OK);

	}
    }
    
    // -----------------------------------------------------------------------
    Tcl_AppendResult(interp, 
	"\nInvalid command - ", argv[1], "\n", 
	"Valid options are: \n",
	"   addnode, animatelayout, animateview, connect, computelayout, create,\n",
	"   createroot, delete, getchildren, getfocus, getlayoutbbox, getlink, \n",
	"   getparent, getroot, isnode, layout, lower, raise, removenode, reparent, \n",
	"   setanimatespeed, setfocus, setfocusmag, setlinkmode, setscale, or setspacing.", 
	NULL);
    
    return(TCL_ERROR);
}

//											//
//	End of Tree code								//
//											//
//////////////////////////////////////////////////////////////////////////////////////////


int
Padtcl_cache(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int len;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad *pad = tkwin->view->pad;
    Pad_List objs;
    Pad_Iterator iter;
    Pad_Object *obj;

    if (argc < 2) {
	TCL_ERR(interp, "wrong # args: should be \"pathName cache subcommand ?arg ...?\"");
    }

    len = strlen(argv[1]);
    if (!strncmp(argv[1], "in", MAX(1, len))) {
	if (argc >= 3) {
	    pad->Find_with_tagorid(argv[2], objs, TRUE);
	    DOLIST(iter, objs, Pad_Object, obj) {
		Pad_cacheMgr.CacheIn(obj);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName cache in tagOrId\"");
	}
    } else if (!strncmp(argv[1], "out", MAX(1, len))) {
	if (argc >= 3) {
	    pad->Find_with_tagorid(argv[2], objs, TRUE);
	    DOLIST(iter, objs, Pad_Object, obj) {
		Pad_cacheMgr.CacheOut(obj);
	    }
	} else {
	    TCL_ERR(interp, "wrong # args: should be \"pathName cache out tagOrId\"");
	}
    } else if (!strncmp(argv[1], "configure", MAX(1, len))) {
	Pad_String value, optionValue, result;
	
				// Return all option-value pairs
	if (argc == 2) {
	    value = Pad_cacheMgr.Dir();
	    if (value == "") {
		optionValue.Printf("-dir {}");
	    } else {
		optionValue.Printf("-dir %s", value.Get());
	    }
	    Tcl_AppendElement(interp, optionValue.Get());
	    
	    optionValue.Printf("-size %f", Pad_cacheMgr.Size());
	    Tcl_AppendElement(interp, optionValue.Get());
	    
	    optionValue.Printf("-viewmultiple %f", Pad_cacheMgr.ViewMultiple());
	    Tcl_AppendElement(interp, optionValue.Get());
	    
	    optionValue.Printf("-delay %d", (int)(Pad_cacheMgr.UpdateDelay()/1000));
	    Tcl_AppendElement(interp, optionValue.Get());
	} else {
	    while (argc > 2) {
		len = strlen(argv[2]);
		if (!strncmp(argv[2], "-dir", MAX(3, len))) {
		    if (argc > 3) {
			if (!Pad_cacheMgr.Dir(argv[3])) {
			    Tcl_AppendResult(interp, Pad_errorString.Get(), NULL);
			    return (TCL_ERROR);
			}
			argc--;
			argv++;
		    }
		    optionValue.Printf("%s", Pad_cacheMgr.Dir());
		    Tcl_AppendResult(interp, optionValue.Get(), NULL);
		} else if (!strncmp(argv[2], "-size", MAX(2, len))) {
		    if (argc > 3) {
			Pad_cacheMgr.Size(atof(argv[3]));
			argc--;
			argv++;
		    }
		    optionValue.Printf("%f", Pad_cacheMgr.Size());
		    Tcl_AppendResult(interp, optionValue.Get(), NULL);
		} else if (!strncmp(argv[2], "-viewmultiple", MAX(2, len))) {
		    if (argc > 3) {
			Pad_cacheMgr.ViewMultiple(atof(argv[3]));
			argc--;
			argv++;
		    }
		    optionValue.Printf("%f", Pad_cacheMgr.ViewMultiple());
		    Tcl_AppendResult(interp, optionValue.Get(), NULL);
		} else if (!strncmp(argv[2], "-delay", MAX(2, len))) {
		    if (argc > 3) {
			Pad_cacheMgr.UpdateDelay(atol(argv[3])*1000);
			argc--;
			argv++;
		    }
		    optionValue.Printf("%d", (int)Pad_cacheMgr.UpdateDelay()/1000);
		    Tcl_AppendResult(interp, optionValue.Get(), NULL);
		} else {
		    Tcl_AppendResult(interp, "Invalid option, Try: -dir, -size, -viewmultiple, or -delay", NULL);
		    return(TCL_ERROR);
		}
		argc--;
		argv++;
	    }
	}
    } else {
	TCL_ERR(interp, "Invalid subcommand, should be: configure, in, or out");
    }

    return TCL_OK;

}

int
Padtcl_warp(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pad_TkWin *tkwin;

    tkwin = (Pad_TkWin *)clientData;

    if (argc < 2)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName warp deltax deltay\"");
      }
    tkwin->Warp (atoi (argv[1]), atoi (argv[2]));
    Tcl_ResetResult(interp);

    return(TCL_OK);
}

int
Padtcl_grab(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int x1, y1, x2, y2;
    Pad_Bool dim;
    Pad_Bool found;
    int xdim, ydim;
    Pad_TkWin *tkwin;
    Pad_ImageData *imageData;
    Window winId;
    char *usage = "wrong # args: should be \"pathName grab [-root | -path pathName | -win winId] [-dim {width height}] x y width height\"";

#ifdef PAD_WIN
    TCL_ERR(interp, "grab not implemented on Windows");
#endif

    tkwin = (Pad_TkWin *)clientData;

    if (argc < 5) {
	TCL_ERR(interp, usage);
    }

			// Check for flags in any order
    winId = tkwin->id;  // defaults
    dim = FALSE;
    found = TRUE;
    while ((argc > 5) && found) {
	found = FALSE;
	if (!strcmp(argv[1], "-root")) {
	    winId = None;
	    found = TRUE;
	} else if (!strcmp(argv[1], "-path")) {
	    Tk_Window tkWindow = Tk_NameToWindow(interp, argv[2], tkwin->Get_tkwindow());
	    if (tkWindow) {
		winId = Tk_WindowId(tkWindow);
	    } else {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "Invalid pathName: ", argv[2], NULL);
		return(TCL_ERROR);
	    }
	    argc--;
	    argv++;
	    found = TRUE;
	} else if (!strcmp(argv[1], "-win")) {
	    winId = strtol(argv[2], NULL, 0);
	    argc--;
	    argv++;
	    found = TRUE;
	} else if (!strcmp(argv[1], "-dim")) {
	    int listrc;
	    int listargc;
	    char **listargv;
	    char *dimUsage = "Invalid -dim (dimension) parameter.  Format is {width height}";
	    
	    listrc = Tcl_SplitList(interp, argv[2], &listargc, &listargv);
	    if (listrc == TCL_OK) {
		if (listargc != 2) {
		    ckfree((char *) listargv);
		    TCL_ERR(interp, dimUsage);
		} else {
		    xdim = atoi(listargv[0]);
		    ydim = atoi(listargv[1]);
		    ckfree((char *) listargv);
		}
	    } else {
		TCL_ERR(interp, dimUsage);
	    }
	    argc--;
	    argv++;
	    dim = TRUE;
	    found = TRUE;
	}
	if (found) {
	    argc--;
	    argv++;
	}
    }
    if (argc < 5) {
	TCL_ERR(interp, usage);
    }

    x1 = atoi(argv[1]);
    y1 = atoi(argv[2]);
    x2 = atoi(argv[3]);
    y2 = atoi(argv[4]);

    if (dim) {
	imageData = tkwin->dpy->Grab(winId, xdim, ydim, x1, y1, x2, y2);
    } else {
	imageData = tkwin->dpy->Grab(winId, x1, y1, x2, y2);
    }
    if (imageData) {
	Tcl_AppendResult(interp, imageData->token, NULL);
    } else {
	TCL_ERR(interp, Pad_errorString.Get());
    }

    return(TCL_OK);
}
